(** Bundled by bspack 08/16-09:46 *)
module 
Reason_config
= struct
#1 "reason_config.ml"
(**
 * Copyright (c) 2015-present, Facebook, Inc. All rights reserved.
 *)

let recoverable = ref false

let configure ~r = (
  recoverable := r
)

end
module 
Syntax_util
= struct
#1 "syntax_util.ml"
open Ast_mapper
open Asttypes
open Parsetree
open Longident

(** [is_prefixed prefix i str] checks if prefix is the prefix of str
  * starting from position i
  *)
let is_prefixed prefix str i =
  let len = String.length prefix in
  if i + len > String.length str then false else
  let rec loop j =
    if j >= len then true else
      if String.unsafe_get prefix j <> String.unsafe_get str (i + j) then false else loop (j + 1)
    in
  loop 0

(**
 * pick_while returns a tuple where first element is longest prefix (possibly empty) of the list of elements that satisfy p
 * and second element is the remainder of the list
 *)
let rec pick_while p = function
  | [] -> [], []
  | hd::tl when p hd ->
                  let (satisfied, not_satisfied) = pick_while p tl in
                  hd :: satisfied, not_satisfied
  | l -> ([], l)


let rec replace_string_ old_str new_str i str buffer =
  if i >= String.length str then
    ()
  else
    (* found match *)
    if is_prefixed old_str str i then
      (* split string *)
      let old_str_len = String.length old_str in
      Buffer.add_string buffer new_str;
      replace_string_ old_str new_str (i + old_str_len) str buffer
    else
      let start = String.sub str i 1 in
      Buffer.add_string buffer start;
      replace_string_ old_str new_str (i + 1) str buffer


(** [replace_string old_str new_str str] replaces old_str to new_str in str *)
let replace_string old_str new_str str =
  let buffer = Buffer.create (String.length old_str * 2) in
  replace_string_ old_str new_str 0 str buffer;
  Buffer.contents buffer


module StringMap = Map.Make (String)


(** Generate a suitable extension node for Merlin's consumption,
    for the purposes of reporting a syntax error - only used
    in recovery mode.
 *)
let syntax_error_extension_node loc message =
  let str = Location.mkloc "merlin.syntax-error" loc in
  let payload = PStr [{
    pstr_loc = Location.none;
    pstr_desc = Pstr_eval (
      {
        pexp_loc = Location.none;
        pexp_desc = Pexp_constant (Asttypes.Const_string (message, None));
        pexp_attributes = [];
      },
      []
    );
  }]
 in
 (str, payload)

let reason_to_ml_swapping_alist = [
  "===",  "==";
  "==",  "=";
  (* ===\/ and !==\/ are not representable in OCaml but
   * representable in Reason
   *)
  "\\!==", "!==";
  "\\===", "===";
  "!=", "<>";
  "!==", "!=";
]

let swap_txt map txt =
  if StringMap.mem txt map then
    StringMap.find txt map
  else
    txt

(** identifier_mapper maps all identifiers in an AST with a mapping function f
  *)
let identifier_mapper f =
{ default_mapper with
  expr = begin fun mapper expr ->
    let expr =
      match expr with
        | {pexp_desc=Pexp_ident ({txt} as id);
           pexp_loc;
           pexp_attributes} ->
             let swapped = match txt with
               | Lident s -> Lident (f s)
               | Ldot(longPrefix, s) -> Ldot(longPrefix, f s)
               | Lapply (y,s) -> Lapply (y, s)
             in
             {expr with pexp_desc=Pexp_ident ({id with txt=swapped})}
        | _ -> expr
    in
    default_mapper.expr mapper expr
  end;
  pat = begin fun mapper pat ->
    let pat =
      match pat with
        | {ppat_desc=Ppat_var ({txt} as id);
           ppat_loc;
           ppat_attributes} ->
             {pat with ppat_desc=Ppat_var ({id with txt=(f txt)})}
        | _ -> pat
    in
    default_mapper.pat mapper pat
  end;
}

(** unescape_stars_slashes_mapper unescapes all stars and slases in an AST
  *)
let unescape_stars_slashes_mapper =
  let unescape_stars_slashes str =
    let len = String.length str in
    if len < 2 then
      str
    else
      let ending = String.sub str 1 (len - 1) in
    String.sub str 0 1 ^
      replace_string "\\*" "*"
        (replace_string ("\\/") "/" ending)
  in
  identifier_mapper unescape_stars_slashes

(** escape_stars_slashes_mapper escapes all stars and slases in an AST
  *)
let escape_stars_slashes_mapper =
  let escape_stars_slashes str =
    let len = String.length str in
    if len < 2 then
      str
    else
      let ending = String.sub str 1 (len -1) in
      String.sub str 0 1 ^
        replace_string "*" "\\*"
          (replace_string "/" "\\/" ending)
  in
  identifier_mapper escape_stars_slashes

(**
 * swap_operator_mapper is a mapper that swaps two operators at parse/print time.
 * We need this since we want to transform operator such as "=" in Ocaml to "==" in Reason.
 * In this case, in the parser, everytime we see a token "==" in Reason, we transform it into "=";
 * Similarly, in the printer, everytime we see a token "=", we transform it into "==";
 *)
let swap_operator_mapper map = identifier_mapper (swap_txt map)

let reason_to_ml_swap_map = List.fold_left
                              (fun map (op1, op2) -> (StringMap.add op1 op2 map))
                              StringMap.empty
                              reason_to_ml_swapping_alist

let ml_to_reason_swap_map = List.fold_left
                              (fun map (op1, op2) -> (StringMap.add op2 op1 map))
                              StringMap.empty
                              reason_to_ml_swapping_alist

(* To be used in parser, transform a token into an ast node with different identifier
 *)
let reason_to_ml_swap_operator_mapper = swap_operator_mapper reason_to_ml_swap_map

(* To be used in printer, transform an ast node into a token with different identifier
 *)
let ml_to_reason_swap_operator_mapper = swap_operator_mapper ml_to_reason_swap_map

(* attribute_equals tests an attribute is txt
 *)
let attribute_equals to_compare = function
  | ({txt; _}, _) -> txt = to_compare

(* attribute_exists tests if an attribute exists in a list
 *)
let attribute_exists txt attributes = List.exists (attribute_equals txt) attributes

(* conflicted_attributes tests if both attribute1 and attribute2
 * exist
 *)
let attributes_conflicted attribute1 attribute2 attributes =
  attribute_exists attribute1 attributes &&
  attribute_exists attribute2 attributes

(* normalized_attributes removes attribute from a list of attributes
 *)
let normalized_attributes attribute attributes =
  List.filter (fun x -> not (attribute_equals attribute x)) attributes

(*
 * apply_mapper_chain family applies an ast_mapper_chain to an ast,
 * ordering from left to right.
 *)
let apply_mapper_chain_to_structure =
  List.fold_left (fun s mapper -> mapper.structure mapper s )
let apply_mapper_chain_to_signature =
  List.fold_left (fun s mapper -> mapper.signature mapper s )
let apply_mapper_chain_to_type =
  List.fold_left (fun s mapper -> mapper.typ mapper s )
let apply_mapper_chain_to_expr =
  List.fold_left (fun s mapper -> mapper.expr mapper s )
let apply_mapper_chain_to_pattern =
  List.fold_left (fun s mapper -> mapper.pat mapper s )

let apply_mapper_chain_to_toplevel_phrase toplevel_phrase chain =
  match toplevel_phrase with
  | Ptop_def x -> Ptop_def (apply_mapper_chain_to_structure x chain)
  | x -> x

let apply_mapper_chain_to_use_file use_file chain =
  List.map (fun x -> apply_mapper_chain_to_toplevel_phrase x chain) use_file

(* The following logic defines our own Error object
 * and register it with ocaml so it knows how to print it
 *)

type error = Syntax_error of string

exception Error of Location.t * error

let report_error ppf (Syntax_error err) =
  Format.(fprintf ppf "%s" err)

let () =
  Location.register_error_of_exn
    (function
     | Error (loc, err) ->
        Some (Location.error_of_printer loc report_error err)
     | _ ->
        None
     )

end
module Reason_parser : sig 
#1 "reason_parser.mli"

(* The type of tokens. *)

type token = 
  | WITH
  | WHILE
  | WHEN
  | VIRTUAL
  | VAL
  | UNDERSCORE
  | UIDENT of (string)
  | TYPE
  | TRY
  | TRUE
  | TO
  | TILDE
  | THEN
  | SWITCH
  | STRUCT
  | STRING of (string * string option)
  | STAR
  | SIG
  | SHARP
  | SEMISEMI
  | SEMI
  | RPAREN
  | REC
  | RBRACKET
  | RBRACE
  | QUOTE
  | QUESTION
  | PRIVATE
  | PREFIXOP of (string)
  | PLUSEQ
  | PLUSDOT
  | PLUS
  | PERCENT
  | OR
  | OPTIONAL_NO_DEFAULT
  | OPEN
  | OBJECT
  | NONREC
  | NEW
  | NATIVEINT of (nativeint)
  | MUTABLE
  | MODULE
  | MINUSGREATER
  | MINUSDOT
  | MINUS
  | METHOD
  | MATCH
  | LPAREN
  | LIDENT of (string)
  | LET
  | LESSMINUS
  | LESSGREATER
  | LESSDOTDOTGREATER
  | LESS
  | LBRACKETPERCENTPERCENT
  | LBRACKETPERCENT
  | LBRACKETLESS
  | LBRACKETGREATER
  | LBRACKETBAR
  | LBRACKETATATAT
  | LBRACKETATAT
  | LBRACKETAT
  | LBRACKET
  | LBRACELESS
  | LBRACE
  | LAZY
  | INT64 of (int64)
  | INT32 of (int32)
  | INT of (int)
  | INITIALIZER
  | INHERIT
  | INFIXOP4 of (string)
  | INFIXOP3 of (string)
  | INFIXOP2 of (string)
  | INFIXOP1 of (string)
  | INFIXOP0 of (string)
  | INCLUDE
  | IN
  | IF
  | GREATERRBRACKET
  | GREATERRBRACE
  | GREATER
  | FUNCTOR
  | FUNCTION
  | FUN
  | FOR
  | FLOAT of (string)
  | FALSE
  | EXTERNAL
  | EXPLICITLY_PASSED_OPTIONAL
  | EXCEPTION
  | EQUALGREATER
  | EQUAL
  | EOL
  | EOF
  | END
  | ELSE
  | DOWNTO
  | DOTDOTDOT
  | DOTDOT
  | DOT
  | DONE
  | DO
  | CONSTRAINT
  | COMMENT of (string * Location.t)
  | COMMA
  | COLONGREATER
  | COLONEQUAL
  | COLONCOLON
  | COLON
  | CLASS
  | CHAR of (char)
  | BEGIN
  | BARRBRACKET
  | BARBAR
  | BAR
  | BANG
  | BACKQUOTE
  | ASSERT
  | AS
  | AND
  | AMPERSAND
  | AMPERAMPER

(* This exception is raised by the monolithic API functions. *)

exception Error

(* The monolithic API. *)

val use_file: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.toplevel_phrase list)

val toplevel_phrase: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.toplevel_phrase)

val parse_pattern: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.pattern)

val parse_expression: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.expression)

val parse_core_type: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.core_type)

val interface: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.signature)

val implementation: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.structure)

end = struct
#1 "reason_parser.ml"

module Basics = struct
  
  exception Error
  
  type token = 
    | WITH
    | WHILE
    | WHEN
    | VIRTUAL
    | VAL
    | UNDERSCORE
    | UIDENT of (
# 910 "reason_parser.mly"
       (string)
# 17 "reason_parser.ml"
  )
    | TYPE
    | TRY
    | TRUE
    | TO
    | TILDE
    | THEN
    | SWITCH
    | STRUCT
    | STRING of (
# 902 "reason_parser.mly"
       (string * string option)
# 30 "reason_parser.ml"
  )
    | STAR
    | SIG
    | SHARP
    | SEMISEMI
    | SEMI
    | RPAREN
    | REC
    | RBRACKET
    | RBRACE
    | QUOTE
    | QUESTION
    | PRIVATE
    | PREFIXOP of (
# 887 "reason_parser.mly"
       (string)
# 47 "reason_parser.ml"
  )
    | PLUSEQ
    | PLUSDOT
    | PLUS
    | PERCENT
    | OR
    | OPTIONAL_NO_DEFAULT
    | OPEN
    | OBJECT
    | NONREC
    | NEW
    | NATIVEINT of (
# 876 "reason_parser.mly"
       (nativeint)
# 62 "reason_parser.ml"
  )
    | MUTABLE
    | MODULE
    | MINUSGREATER
    | MINUSDOT
    | MINUS
    | METHOD
    | MATCH
    | LPAREN
    | LIDENT of (
# 863 "reason_parser.mly"
       (string)
# 75 "reason_parser.ml"
  )
    | LET
    | LESSMINUS
    | LESSGREATER
    | LESSDOTDOTGREATER
    | LESS
    | LBRACKETPERCENTPERCENT
    | LBRACKETPERCENT
    | LBRACKETLESS
    | LBRACKETGREATER
    | LBRACKETBAR
    | LBRACKETATATAT
    | LBRACKETATAT
    | LBRACKETAT
    | LBRACKET
    | LBRACELESS
    | LBRACE
    | LAZY
    | INT64 of (
# 847 "reason_parser.mly"
       (int64)
# 97 "reason_parser.ml"
  )
    | INT32 of (
# 846 "reason_parser.mly"
       (int32)
# 102 "reason_parser.ml"
  )
    | INT of (
# 845 "reason_parser.mly"
       (int)
# 107 "reason_parser.ml"
  )
    | INITIALIZER
    | INHERIT
    | INFIXOP4 of (
# 842 "reason_parser.mly"
       (string)
# 114 "reason_parser.ml"
  )
    | INFIXOP3 of (
# 841 "reason_parser.mly"
       (string)
# 119 "reason_parser.ml"
  )
    | INFIXOP2 of (
# 840 "reason_parser.mly"
       (string)
# 124 "reason_parser.ml"
  )
    | INFIXOP1 of (
# 839 "reason_parser.mly"
       (string)
# 129 "reason_parser.ml"
  )
    | INFIXOP0 of (
# 838 "reason_parser.mly"
       (string)
# 134 "reason_parser.ml"
  )
    | INCLUDE
    | IN
    | IF
    | GREATERRBRACKET
    | GREATERRBRACE
    | GREATER
    | FUNCTOR
    | FUNCTION
    | FUN
    | FOR
    | FLOAT of (
# 827 "reason_parser.mly"
       (string)
# 149 "reason_parser.ml"
  )
    | FALSE
    | EXTERNAL
    | EXPLICITLY_PASSED_OPTIONAL
    | EXCEPTION
    | EQUALGREATER
    | EQUAL
    | EOL
    | EOF
    | END
    | ELSE
    | DOWNTO
    | DOTDOTDOT
    | DOTDOT
    | DOT
    | DONE
    | DO
    | CONSTRAINT
    | COMMENT of (
# 917 "reason_parser.mly"
       (string * Location.t)
# 171 "reason_parser.ml"
  )
    | COMMA
    | COLONGREATER
    | COLONEQUAL
    | COLONCOLON
    | COLON
    | CLASS
    | CHAR of (
# 806 "reason_parser.mly"
       (char)
# 182 "reason_parser.ml"
  )
    | BEGIN
    | BARRBRACKET
    | BARBAR
    | BAR
    | BANG
    | BACKQUOTE
    | ASSERT
    | AS
    | AND
    | AMPERSAND
    | AMPERAMPER
  
end

include Basics

let _eRR =
  Basics.Error

type _menhir_env = {
  _menhir_lexer: Lexing.lexbuf -> token;
  _menhir_lexbuf: Lexing.lexbuf;
  _menhir_token: token;
  mutable _menhir_error: bool
}

and _menhir_state = 
  | MenhirState2754
  | MenhirState2751
  | MenhirState2747
  | MenhirState2742
  | MenhirState2739
  | MenhirState2737
  | MenhirState2736
  | MenhirState2733
  | MenhirState2732
  | MenhirState2731
  | MenhirState2730
  | MenhirState2727
  | MenhirState2726
  | MenhirState2725
  | MenhirState2724
  | MenhirState2711
  | MenhirState2704
  | MenhirState2703
  | MenhirState2701
  | MenhirState2700
  | MenhirState2699
  | MenhirState2698
  | MenhirState2695
  | MenhirState2694
  | MenhirState2692
  | MenhirState2691
  | MenhirState2684
  | MenhirState2682
  | MenhirState2681
  | MenhirState2677
  | MenhirState2676
  | MenhirState2675
  | MenhirState2674
  | MenhirState2673
  | MenhirState2671
  | MenhirState2667
  | MenhirState2663
  | MenhirState2660
  | MenhirState2659
  | MenhirState2658
  | MenhirState2657
  | MenhirState2656
  | MenhirState2654
  | MenhirState2651
  | MenhirState2649
  | MenhirState2648
  | MenhirState2646
  | MenhirState2645
  | MenhirState2643
  | MenhirState2641
  | MenhirState2639
  | MenhirState2638
  | MenhirState2636
  | MenhirState2634
  | MenhirState2633
  | MenhirState2631
  | MenhirState2630
  | MenhirState2629
  | MenhirState2626
  | MenhirState2625
  | MenhirState2624
  | MenhirState2623
  | MenhirState2621
  | MenhirState2620
  | MenhirState2614
  | MenhirState2613
  | MenhirState2607
  | MenhirState2606
  | MenhirState2604
  | MenhirState2603
  | MenhirState2602
  | MenhirState2600
  | MenhirState2599
  | MenhirState2598
  | MenhirState2596
  | MenhirState2595
  | MenhirState2594
  | MenhirState2591
  | MenhirState2590
  | MenhirState2589
  | MenhirState2587
  | MenhirState2586
  | MenhirState2585
  | MenhirState2584
  | MenhirState2583
  | MenhirState2582
  | MenhirState2581
  | MenhirState2580
  | MenhirState2578
  | MenhirState2577
  | MenhirState2576
  | MenhirState2573
  | MenhirState2572
  | MenhirState2570
  | MenhirState2568
  | MenhirState2564
  | MenhirState2559
  | MenhirState2557
  | MenhirState2551
  | MenhirState2550
  | MenhirState2549
  | MenhirState2548
  | MenhirState2546
  | MenhirState2545
  | MenhirState2542
  | MenhirState2534
  | MenhirState2533
  | MenhirState2532
  | MenhirState2531
  | MenhirState2530
  | MenhirState2527
  | MenhirState2522
  | MenhirState2521
  | MenhirState2517
  | MenhirState2516
  | MenhirState2515
  | MenhirState2514
  | MenhirState2510
  | MenhirState2509
  | MenhirState2505
  | MenhirState2504
  | MenhirState2501
  | MenhirState2498
  | MenhirState2497
  | MenhirState2495
  | MenhirState2494
  | MenhirState2492
  | MenhirState2491
  | MenhirState2488
  | MenhirState2487
  | MenhirState2485
  | MenhirState2484
  | MenhirState2481
  | MenhirState2480
  | MenhirState2478
  | MenhirState2477
  | MenhirState2475
  | MenhirState2474
  | MenhirState2473
  | MenhirState2472
  | MenhirState2470
  | MenhirState2469
  | MenhirState2467
  | MenhirState2466
  | MenhirState2464
  | MenhirState2460
  | MenhirState2459
  | MenhirState2458
  | MenhirState2457
  | MenhirState2455
  | MenhirState2453
  | MenhirState2451
  | MenhirState2450
  | MenhirState2443
  | MenhirState2436
  | MenhirState2435
  | MenhirState2433
  | MenhirState2428
  | MenhirState2427
  | MenhirState2426
  | MenhirState2422
  | MenhirState2420
  | MenhirState2417
  | MenhirState2416
  | MenhirState2415
  | MenhirState2412
  | MenhirState2411
  | MenhirState2410
  | MenhirState2409
  | MenhirState2405
  | MenhirState2401
  | MenhirState2400
  | MenhirState2399
  | MenhirState2390
  | MenhirState2389
  | MenhirState2387
  | MenhirState2385
  | MenhirState2383
  | MenhirState2381
  | MenhirState2376
  | MenhirState2373
  | MenhirState2371
  | MenhirState2370
  | MenhirState2369
  | MenhirState2368
  | MenhirState2362
  | MenhirState2360
  | MenhirState2358
  | MenhirState2356
  | MenhirState2354
  | MenhirState2352
  | MenhirState2348
  | MenhirState2345
  | MenhirState2344
  | MenhirState2341
  | MenhirState2340
  | MenhirState2337
  | MenhirState2333
  | MenhirState2331
  | MenhirState2320
  | MenhirState2318
  | MenhirState2317
  | MenhirState2314
  | MenhirState2309
  | MenhirState2308
  | MenhirState2306
  | MenhirState2304
  | MenhirState2303
  | MenhirState2301
  | MenhirState2300
  | MenhirState2298
  | MenhirState2297
  | MenhirState2296
  | MenhirState2295
  | MenhirState2290
  | MenhirState2289
  | MenhirState2288
  | MenhirState2287
  | MenhirState2286
  | MenhirState2285
  | MenhirState2283
  | MenhirState2282
  | MenhirState2275
  | MenhirState2272
  | MenhirState2271
  | MenhirState2269
  | MenhirState2268
  | MenhirState2266
  | MenhirState2265
  | MenhirState2264
  | MenhirState2263
  | MenhirState2262
  | MenhirState2253
  | MenhirState2252
  | MenhirState2251
  | MenhirState2249
  | MenhirState2247
  | MenhirState2246
  | MenhirState2242
  | MenhirState2241
  | MenhirState2232
  | MenhirState2229
  | MenhirState2228
  | MenhirState2227
  | MenhirState2226
  | MenhirState2225
  | MenhirState2224
  | MenhirState2223
  | MenhirState2218
  | MenhirState2217
  | MenhirState2216
  | MenhirState2215
  | MenhirState2214
  | MenhirState2213
  | MenhirState2212
  | MenhirState2211
  | MenhirState2208
  | MenhirState2207
  | MenhirState2205
  | MenhirState2204
  | MenhirState2202
  | MenhirState2196
  | MenhirState2192
  | MenhirState2191
  | MenhirState2186
  | MenhirState2185
  | MenhirState2182
  | MenhirState2181
  | MenhirState2177
  | MenhirState2176
  | MenhirState2173
  | MenhirState2171
  | MenhirState2167
  | MenhirState2166
  | MenhirState2163
  | MenhirState2162
  | MenhirState2159
  | MenhirState2155
  | MenhirState2152
  | MenhirState2149
  | MenhirState2147
  | MenhirState2143
  | MenhirState2141
  | MenhirState2138
  | MenhirState2137
  | MenhirState2136
  | MenhirState2134
  | MenhirState2133
  | MenhirState2132
  | MenhirState2131
  | MenhirState2128
  | MenhirState2126
  | MenhirState2123
  | MenhirState2122
  | MenhirState2121
  | MenhirState2120
  | MenhirState2119
  | MenhirState2117
  | MenhirState2116
  | MenhirState2112
  | MenhirState2111
  | MenhirState2110
  | MenhirState2108
  | MenhirState2107
  | MenhirState2103
  | MenhirState2097
  | MenhirState2090
  | MenhirState2087
  | MenhirState2086
  | MenhirState2082
  | MenhirState2081
  | MenhirState2080
  | MenhirState2077
  | MenhirState2076
  | MenhirState2072
  | MenhirState2071
  | MenhirState2070
  | MenhirState2069
  | MenhirState2068
  | MenhirState2062
  | MenhirState2059
  | MenhirState2057
  | MenhirState2056
  | MenhirState2055
  | MenhirState2052
  | MenhirState2051
  | MenhirState2050
  | MenhirState2049
  | MenhirState2048
  | MenhirState2046
  | MenhirState2045
  | MenhirState2043
  | MenhirState2042
  | MenhirState2040
  | MenhirState2038
  | MenhirState2037
  | MenhirState2035
  | MenhirState2033
  | MenhirState2031
  | MenhirState2030
  | MenhirState2027
  | MenhirState2026
  | MenhirState2021
  | MenhirState2019
  | MenhirState2010
  | MenhirState2000
  | MenhirState1999
  | MenhirState1995
  | MenhirState1993
  | MenhirState1992
  | MenhirState1991
  | MenhirState1989
  | MenhirState1988
  | MenhirState1985
  | MenhirState1984
  | MenhirState1983
  | MenhirState1982
  | MenhirState1981
  | MenhirState1979
  | MenhirState1978
  | MenhirState1977
  | MenhirState1976
  | MenhirState1974
  | MenhirState1973
  | MenhirState1972
  | MenhirState1971
  | MenhirState1969
  | MenhirState1968
  | MenhirState1967
  | MenhirState1966
  | MenhirState1965
  | MenhirState1963
  | MenhirState1959
  | MenhirState1956
  | MenhirState1952
  | MenhirState1949
  | MenhirState1946
  | MenhirState1944
  | MenhirState1943
  | MenhirState1942
  | MenhirState1941
  | MenhirState1939
  | MenhirState1938
  | MenhirState1937
  | MenhirState1933
  | MenhirState1929
  | MenhirState1926
  | MenhirState1923
  | MenhirState1918
  | MenhirState1917
  | MenhirState1916
  | MenhirState1915
  | MenhirState1914
  | MenhirState1913
  | MenhirState1911
  | MenhirState1906
  | MenhirState1904
  | MenhirState1903
  | MenhirState1900
  | MenhirState1898
  | MenhirState1896
  | MenhirState1894
  | MenhirState1892
  | MenhirState1890
  | MenhirState1889
  | MenhirState1888
  | MenhirState1887
  | MenhirState1885
  | MenhirState1882
  | MenhirState1881
  | MenhirState1880
  | MenhirState1879
  | MenhirState1876
  | MenhirState1875
  | MenhirState1874
  | MenhirState1871
  | MenhirState1870
  | MenhirState1869
  | MenhirState1866
  | MenhirState1865
  | MenhirState1864
  | MenhirState1861
  | MenhirState1858
  | MenhirState1855
  | MenhirState1853
  | MenhirState1852
  | MenhirState1851
  | MenhirState1850
  | MenhirState1849
  | MenhirState1848
  | MenhirState1847
  | MenhirState1840
  | MenhirState1836
  | MenhirState1828
  | MenhirState1826
  | MenhirState1825
  | MenhirState1821
  | MenhirState1820
  | MenhirState1819
  | MenhirState1818
  | MenhirState1817
  | MenhirState1814
  | MenhirState1813
  | MenhirState1810
  | MenhirState1809
  | MenhirState1808
  | MenhirState1803
  | MenhirState1802
  | MenhirState1801
  | MenhirState1800
  | MenhirState1799
  | MenhirState1795
  | MenhirState1794
  | MenhirState1793
  | MenhirState1789
  | MenhirState1788
  | MenhirState1784
  | MenhirState1782
  | MenhirState1781
  | MenhirState1780
  | MenhirState1775
  | MenhirState1774
  | MenhirState1773
  | MenhirState1768
  | MenhirState1767
  | MenhirState1766
  | MenhirState1765
  | MenhirState1764
  | MenhirState1762
  | MenhirState1761
  | MenhirState1760
  | MenhirState1757
  | MenhirState1751
  | MenhirState1747
  | MenhirState1743
  | MenhirState1739
  | MenhirState1738
  | MenhirState1736
  | MenhirState1733
  | MenhirState1732
  | MenhirState1731
  | MenhirState1729
  | MenhirState1728
  | MenhirState1725
  | MenhirState1723
  | MenhirState1722
  | MenhirState1720
  | MenhirState1719
  | MenhirState1717
  | MenhirState1715
  | MenhirState1714
  | MenhirState1713
  | MenhirState1712
  | MenhirState1708
  | MenhirState1707
  | MenhirState1706
  | MenhirState1705
  | MenhirState1704
  | MenhirState1703
  | MenhirState1702
  | MenhirState1701
  | MenhirState1698
  | MenhirState1697
  | MenhirState1696
  | MenhirState1694
  | MenhirState1692
  | MenhirState1689
  | MenhirState1687
  | MenhirState1686
  | MenhirState1685
  | MenhirState1684
  | MenhirState1683
  | MenhirState1682
  | MenhirState1680
  | MenhirState1679
  | MenhirState1678
  | MenhirState1677
  | MenhirState1674
  | MenhirState1672
  | MenhirState1670
  | MenhirState1668
  | MenhirState1665
  | MenhirState1663
  | MenhirState1662
  | MenhirState1661
  | MenhirState1660
  | MenhirState1658
  | MenhirState1657
  | MenhirState1656
  | MenhirState1652
  | MenhirState1651
  | MenhirState1644
  | MenhirState1642
  | MenhirState1641
  | MenhirState1637
  | MenhirState1636
  | MenhirState1635
  | MenhirState1634
  | MenhirState1633
  | MenhirState1631
  | MenhirState1629
  | MenhirState1628
  | MenhirState1627
  | MenhirState1626
  | MenhirState1625
  | MenhirState1624
  | MenhirState1621
  | MenhirState1620
  | MenhirState1617
  | MenhirState1611
  | MenhirState1609
  | MenhirState1604
  | MenhirState1603
  | MenhirState1600
  | MenhirState1597
  | MenhirState1595
  | MenhirState1593
  | MenhirState1589
  | MenhirState1588
  | MenhirState1587
  | MenhirState1586
  | MenhirState1582
  | MenhirState1581
  | MenhirState1580
  | MenhirState1575
  | MenhirState1573
  | MenhirState1570
  | MenhirState1569
  | MenhirState1567
  | MenhirState1560
  | MenhirState1557
  | MenhirState1555
  | MenhirState1553
  | MenhirState1552
  | MenhirState1550
  | MenhirState1548
  | MenhirState1547
  | MenhirState1546
  | MenhirState1545
  | MenhirState1539
  | MenhirState1538
  | MenhirState1536
  | MenhirState1533
  | MenhirState1532
  | MenhirState1530
  | MenhirState1528
  | MenhirState1526
  | MenhirState1524
  | MenhirState1522
  | MenhirState1520
  | MenhirState1519
  | MenhirState1518
  | MenhirState1516
  | MenhirState1514
  | MenhirState1513
  | MenhirState1512
  | MenhirState1510
  | MenhirState1508
  | MenhirState1507
  | MenhirState1506
  | MenhirState1503
  | MenhirState1501
  | MenhirState1500
  | MenhirState1499
  | MenhirState1497
  | MenhirState1495
  | MenhirState1494
  | MenhirState1492
  | MenhirState1490
  | MenhirState1488
  | MenhirState1487
  | MenhirState1486
  | MenhirState1485
  | MenhirState1484
  | MenhirState1483
  | MenhirState1481
  | MenhirState1479
  | MenhirState1477
  | MenhirState1476
  | MenhirState1474
  | MenhirState1472
  | MenhirState1471
  | MenhirState1470
  | MenhirState1469
  | MenhirState1468
  | MenhirState1466
  | MenhirState1465
  | MenhirState1462
  | MenhirState1461
  | MenhirState1460
  | MenhirState1458
  | MenhirState1456
  | MenhirState1454
  | MenhirState1453
  | MenhirState1451
  | MenhirState1439
  | MenhirState1438
  | MenhirState1437
  | MenhirState1436
  | MenhirState1435
  | MenhirState1432
  | MenhirState1430
  | MenhirState1429
  | MenhirState1428
  | MenhirState1427
  | MenhirState1426
  | MenhirState1424
  | MenhirState1423
  | MenhirState1422
  | MenhirState1421
  | MenhirState1419
  | MenhirState1416
  | MenhirState1415
  | MenhirState1414
  | MenhirState1413
  | MenhirState1412
  | MenhirState1408
  | MenhirState1405
  | MenhirState1403
  | MenhirState1402
  | MenhirState1401
  | MenhirState1400
  | MenhirState1399
  | MenhirState1396
  | MenhirState1395
  | MenhirState1394
  | MenhirState1392
  | MenhirState1391
  | MenhirState1387
  | MenhirState1386
  | MenhirState1384
  | MenhirState1383
  | MenhirState1381
  | MenhirState1380
  | MenhirState1378
  | MenhirState1377
  | MenhirState1374
  | MenhirState1373
  | MenhirState1371
  | MenhirState1370
  | MenhirState1369
  | MenhirState1368
  | MenhirState1366
  | MenhirState1363
  | MenhirState1361
  | MenhirState1358
  | MenhirState1355
  | MenhirState1353
  | MenhirState1352
  | MenhirState1351
  | MenhirState1346
  | MenhirState1345
  | MenhirState1341
  | MenhirState1340
  | MenhirState1339
  | MenhirState1338
  | MenhirState1332
  | MenhirState1331
  | MenhirState1326
  | MenhirState1324
  | MenhirState1322
  | MenhirState1318
  | MenhirState1314
  | MenhirState1313
  | MenhirState1310
  | MenhirState1307
  | MenhirState1306
  | MenhirState1304
  | MenhirState1303
  | MenhirState1301
  | MenhirState1300
  | MenhirState1297
  | MenhirState1296
  | MenhirState1294
  | MenhirState1293
  | MenhirState1290
  | MenhirState1289
  | MenhirState1287
  | MenhirState1286
  | MenhirState1284
  | MenhirState1283
  | MenhirState1282
  | MenhirState1281
  | MenhirState1279
  | MenhirState1278
  | MenhirState1276
  | MenhirState1274
  | MenhirState1272
  | MenhirState1271
  | MenhirState1269
  | MenhirState1266
  | MenhirState1265
  | MenhirState1264
  | MenhirState1263
  | MenhirState1262
  | MenhirState1259
  | MenhirState1255
  | MenhirState1251
  | MenhirState1249
  | MenhirState1248
  | MenhirState1247
  | MenhirState1246
  | MenhirState1245
  | MenhirState1244
  | MenhirState1243
  | MenhirState1242
  | MenhirState1241
  | MenhirState1240
  | MenhirState1239
  | MenhirState1234
  | MenhirState1233
  | MenhirState1232
  | MenhirState1231
  | MenhirState1230
  | MenhirState1229
  | MenhirState1227
  | MenhirState1222
  | MenhirState1220
  | MenhirState1219
  | MenhirState1216
  | MenhirState1214
  | MenhirState1212
  | MenhirState1210
  | MenhirState1208
  | MenhirState1206
  | MenhirState1203
  | MenhirState1202
  | MenhirState1201
  | MenhirState1200
  | MenhirState1199
  | MenhirState1194
  | MenhirState1193
  | MenhirState1192
  | MenhirState1191
  | MenhirState1190
  | MenhirState1189
  | MenhirState1188
  | MenhirState1187
  | MenhirState1184
  | MenhirState1183
  | MenhirState1180
  | MenhirState1179
  | MenhirState1178
  | MenhirState1176
  | MenhirState1171
  | MenhirState1170
  | MenhirState1169
  | MenhirState1166
  | MenhirState1162
  | MenhirState1161
  | MenhirState1156
  | MenhirState1154
  | MenhirState1153
  | MenhirState1152
  | MenhirState1151
  | MenhirState1150
  | MenhirState1149
  | MenhirState1144
  | MenhirState1142
  | MenhirState1139
  | MenhirState1137
  | MenhirState1134
  | MenhirState1133
  | MenhirState1132
  | MenhirState1130
  | MenhirState1129
  | MenhirState1126
  | MenhirState1119
  | MenhirState1118
  | MenhirState1115
  | MenhirState1112
  | MenhirState1110
  | MenhirState1108
  | MenhirState1104
  | MenhirState1103
  | MenhirState1098
  | MenhirState1097
  | MenhirState1094
  | MenhirState1093
  | MenhirState1087
  | MenhirState1084
  | MenhirState1082
  | MenhirState1081
  | MenhirState1079
  | MenhirState1078
  | MenhirState1076
  | MenhirState1073
  | MenhirState1071
  | MenhirState1069
  | MenhirState1066
  | MenhirState1063
  | MenhirState1061
  | MenhirState1060
  | MenhirState1058
  | MenhirState1056
  | MenhirState1055
  | MenhirState1054
  | MenhirState1052
  | MenhirState1051
  | MenhirState1049
  | MenhirState1048
  | MenhirState1047
  | MenhirState1046
  | MenhirState1043
  | MenhirState1042
  | MenhirState1041
  | MenhirState1040
  | MenhirState1039
  | MenhirState1038
  | MenhirState1036
  | MenhirState1033
  | MenhirState1032
  | MenhirState1031
  | MenhirState1030
  | MenhirState1029
  | MenhirState1028
  | MenhirState1027
  | MenhirState1022
  | MenhirState1020
  | MenhirState1019
  | MenhirState1018
  | MenhirState1017
  | MenhirState1016
  | MenhirState1015
  | MenhirState1014
  | MenhirState1013
  | MenhirState1011
  | MenhirState1010
  | MenhirState1009
  | MenhirState1008
  | MenhirState1007
  | MenhirState1006
  | MenhirState1004
  | MenhirState1003
  | MenhirState1002
  | MenhirState1001
  | MenhirState998
  | MenhirState997
  | MenhirState996
  | MenhirState995
  | MenhirState994
  | MenhirState993
  | MenhirState992
  | MenhirState991
  | MenhirState990
  | MenhirState989
  | MenhirState988
  | MenhirState987
  | MenhirState986
  | MenhirState985
  | MenhirState984
  | MenhirState983
  | MenhirState982
  | MenhirState981
  | MenhirState980
  | MenhirState979
  | MenhirState978
  | MenhirState977
  | MenhirState976
  | MenhirState975
  | MenhirState974
  | MenhirState973
  | MenhirState972
  | MenhirState971
  | MenhirState970
  | MenhirState969
  | MenhirState968
  | MenhirState967
  | MenhirState966
  | MenhirState965
  | MenhirState964
  | MenhirState963
  | MenhirState962
  | MenhirState961
  | MenhirState960
  | MenhirState959
  | MenhirState958
  | MenhirState956
  | MenhirState955
  | MenhirState950
  | MenhirState948
  | MenhirState946
  | MenhirState945
  | MenhirState944
  | MenhirState943
  | MenhirState942
  | MenhirState941
  | MenhirState935
  | MenhirState933
  | MenhirState931
  | MenhirState930
  | MenhirState928
  | MenhirState926
  | MenhirState925
  | MenhirState924
  | MenhirState921
  | MenhirState920
  | MenhirState919
  | MenhirState918
  | MenhirState917
  | MenhirState916
  | MenhirState915
  | MenhirState909
  | MenhirState908
  | MenhirState907
  | MenhirState906
  | MenhirState905
  | MenhirState904
  | MenhirState903
  | MenhirState902
  | MenhirState901
  | MenhirState898
  | MenhirState897
  | MenhirState894
  | MenhirState893
  | MenhirState892
  | MenhirState891
  | MenhirState890
  | MenhirState889
  | MenhirState888
  | MenhirState887
  | MenhirState886
  | MenhirState883
  | MenhirState882
  | MenhirState881
  | MenhirState880
  | MenhirState879
  | MenhirState878
  | MenhirState877
  | MenhirState876
  | MenhirState875
  | MenhirState874
  | MenhirState873
  | MenhirState872
  | MenhirState871
  | MenhirState870
  | MenhirState869
  | MenhirState868
  | MenhirState867
  | MenhirState866
  | MenhirState865
  | MenhirState864
  | MenhirState863
  | MenhirState862
  | MenhirState861
  | MenhirState860
  | MenhirState859
  | MenhirState858
  | MenhirState857
  | MenhirState856
  | MenhirState855
  | MenhirState854
  | MenhirState853
  | MenhirState852
  | MenhirState851
  | MenhirState850
  | MenhirState849
  | MenhirState848
  | MenhirState847
  | MenhirState846
  | MenhirState845
  | MenhirState843
  | MenhirState842
  | MenhirState837
  | MenhirState835
  | MenhirState833
  | MenhirState832
  | MenhirState831
  | MenhirState830
  | MenhirState829
  | MenhirState828
  | MenhirState822
  | MenhirState820
  | MenhirState818
  | MenhirState817
  | MenhirState815
  | MenhirState813
  | MenhirState812
  | MenhirState811
  | MenhirState810
  | MenhirState809
  | MenhirState808
  | MenhirState807
  | MenhirState806
  | MenhirState800
  | MenhirState796
  | MenhirState795
  | MenhirState794
  | MenhirState793
  | MenhirState792
  | MenhirState791
  | MenhirState789
  | MenhirState788
  | MenhirState786
  | MenhirState775
  | MenhirState770
  | MenhirState768
  | MenhirState767
  | MenhirState766
  | MenhirState765
  | MenhirState763
  | MenhirState762
  | MenhirState761
  | MenhirState760
  | MenhirState759
  | MenhirState758
  | MenhirState757
  | MenhirState756
  | MenhirState754
  | MenhirState753
  | MenhirState752
  | MenhirState751
  | MenhirState750
  | MenhirState749
  | MenhirState748
  | MenhirState746
  | MenhirState745
  | MenhirState744
  | MenhirState741
  | MenhirState740
  | MenhirState739
  | MenhirState738
  | MenhirState737
  | MenhirState736
  | MenhirState734
  | MenhirState733
  | MenhirState730
  | MenhirState729
  | MenhirState722
  | MenhirState719
  | MenhirState714
  | MenhirState713
  | MenhirState712
  | MenhirState711
  | MenhirState705
  | MenhirState700
  | MenhirState699
  | MenhirState692
  | MenhirState688
  | MenhirState687
  | MenhirState686
  | MenhirState681
  | MenhirState678
  | MenhirState676
  | MenhirState668
  | MenhirState666
  | MenhirState665
  | MenhirState660
  | MenhirState658
  | MenhirState657
  | MenhirState655
  | MenhirState654
  | MenhirState650
  | MenhirState649
  | MenhirState646
  | MenhirState644
  | MenhirState642
  | MenhirState635
  | MenhirState634
  | MenhirState630
  | MenhirState619
  | MenhirState615
  | MenhirState614
  | MenhirState613
  | MenhirState612
  | MenhirState611
  | MenhirState610
  | MenhirState609
  | MenhirState605
  | MenhirState603
  | MenhirState602
  | MenhirState594
  | MenhirState593
  | MenhirState592
  | MenhirState590
  | MenhirState588
  | MenhirState587
  | MenhirState579
  | MenhirState577
  | MenhirState576
  | MenhirState574
  | MenhirState573
  | MenhirState564
  | MenhirState563
  | MenhirState554
  | MenhirState553
  | MenhirState550
  | MenhirState548
  | MenhirState545
  | MenhirState540
  | MenhirState533
  | MenhirState532
  | MenhirState531
  | MenhirState526
  | MenhirState525
  | MenhirState519
  | MenhirState518
  | MenhirState516
  | MenhirState514
  | MenhirState511
  | MenhirState505
  | MenhirState504
  | MenhirState503
  | MenhirState499
  | MenhirState495
  | MenhirState487
  | MenhirState486
  | MenhirState484
  | MenhirState481
  | MenhirState480
  | MenhirState477
  | MenhirState476
  | MenhirState475
  | MenhirState474
  | MenhirState473
  | MenhirState472
  | MenhirState469
  | MenhirState468
  | MenhirState464
  | MenhirState463
  | MenhirState456
  | MenhirState455
  | MenhirState454
  | MenhirState453
  | MenhirState452
  | MenhirState451
  | MenhirState450
  | MenhirState448
  | MenhirState443
  | MenhirState442
  | MenhirState441
  | MenhirState440
  | MenhirState439
  | MenhirState438
  | MenhirState437
  | MenhirState436
  | MenhirState435
  | MenhirState433
  | MenhirState431
  | MenhirState425
  | MenhirState424
  | MenhirState422
  | MenhirState421
  | MenhirState419
  | MenhirState418
  | MenhirState417
  | MenhirState416
  | MenhirState415
  | MenhirState414
  | MenhirState412
  | MenhirState411
  | MenhirState410
  | MenhirState407
  | MenhirState404
  | MenhirState403
  | MenhirState402
  | MenhirState400
  | MenhirState399
  | MenhirState398
  | MenhirState396
  | MenhirState395
  | MenhirState394
  | MenhirState393
  | MenhirState391
  | MenhirState388
  | MenhirState336
  | MenhirState333
  | MenhirState332
  | MenhirState331
  | MenhirState328
  | MenhirState327
  | MenhirState326
  | MenhirState324
  | MenhirState323
  | MenhirState316
  | MenhirState312
  | MenhirState306
  | MenhirState305
  | MenhirState302
  | MenhirState300
  | MenhirState295
  | MenhirState294
  | MenhirState291
  | MenhirState290
  | MenhirState287
  | MenhirState286
  | MenhirState283
  | MenhirState276
  | MenhirState272
  | MenhirState271
  | MenhirState270
  | MenhirState269
  | MenhirState262
  | MenhirState254
  | MenhirState253
  | MenhirState250
  | MenhirState249
  | MenhirState248
  | MenhirState245
  | MenhirState244
  | MenhirState243
  | MenhirState242
  | MenhirState241
  | MenhirState239
  | MenhirState238
  | MenhirState233
  | MenhirState232
  | MenhirState215
  | MenhirState214
  | MenhirState211
  | MenhirState209
  | MenhirState195
  | MenhirState194
  | MenhirState192
  | MenhirState190
  | MenhirState189
  | MenhirState188
  | MenhirState186
  | MenhirState183
  | MenhirState180
  | MenhirState179
  | MenhirState177
  | MenhirState176
  | MenhirState174
  | MenhirState173
  | MenhirState172
  | MenhirState169
  | MenhirState163
  | MenhirState162
  | MenhirState161
  | MenhirState160
  | MenhirState158
  | MenhirState157
  | MenhirState155
  | MenhirState153
  | MenhirState152
  | MenhirState146
  | MenhirState142
  | MenhirState140
  | MenhirState137
  | MenhirState136
  | MenhirState134
  | MenhirState133
  | MenhirState129
  | MenhirState128
  | MenhirState127
  | MenhirState123
  | MenhirState121
  | MenhirState120
  | MenhirState119
  | MenhirState117
  | MenhirState115
  | MenhirState112
  | MenhirState111
  | MenhirState109
  | MenhirState108
  | MenhirState106
  | MenhirState105
  | MenhirState102
  | MenhirState98
  | MenhirState92
  | MenhirState90
  | MenhirState88
  | MenhirState87
  | MenhirState84
  | MenhirState83
  | MenhirState82
  | MenhirState81
  | MenhirState79
  | MenhirState78
  | MenhirState77
  | MenhirState76
  | MenhirState74
  | MenhirState73
  | MenhirState71
  | MenhirState19
  | MenhirState17
  | MenhirState15
  | MenhirState10
  | MenhirState9
  | MenhirState7
  | MenhirState6
  | MenhirState2
  | MenhirState1
  | MenhirState0

# 52 "reason_parser.mly"
  
open Location
open Asttypes
open Longident
open Parsetree
open Ast_helper
open Ast_mapper
open Syntax_util

(*
   TODO:
   - Remove all [open]s from the top of this file one by one and fix compilation
   failures that ensue by specifying the appropriate long identifiers. That
   will make the parser much easier to reason about.
   - Go back to trunk, do the same (remove [open]s, and fully specify long
   idents), to perform a clean diff.

*)

(**

   location.ml:
   ------------
   let mkloc txt loc = { txt ; loc }
   let rhs_loc n = {
     loc_start = Parsing.rhs_start_pos n;
     loc_end = Parsing.rhs_end_pos n;
     loc_ghost = false;
   }
   let symbol_rloc () = {
     loc_start = Parsing.symbol_start_pos ();
     loc_end = Parsing.symbol_end_pos ();
     loc_ghost = false;
   }

   let symbol_gloc () = {
     loc_start = Parsing.symbol_start_pos ();
     loc_end = Parsing.symbol_end_pos ();
     loc_ghost = true;
   }

   let rhs_loc n = {
     loc_start = Parsing.rhs_start_pos n;
     loc_end = Parsing.rhs_end_pos n;
     loc_ghost = false;
   }

   ast_helper.ml:
   ------------
   module Typ = struct
    val mk: ?loc:loc -> ?attrs:attrs -> core_type_desc -> core_type
    let mk ?(loc = !default_loc) ?(attrs = []) d =
       {ptyp_desc = d; ptyp_loc = loc; ptyp_attributes = attrs}
     ..
   end

   parse_tree.mli
   --------------
   and core_type = {
     ptyp_desc: core_type_desc;
     ptyp_loc: Location.t;
     ptyp_attributes: attributes; (* ... [@id1] [@id2] *)
   }

   and core_type_desc =
     | Ptyp_any
           (*  _ *)
     | Ptyp_var of string
           (* 'a *)
     | Ptyp_arrow of label * core_type * core_type
           (* T1 -> T2       (label = "")
              ~l:T1 -> T2    (label = "l")
              ?l:T1 -> T2    (label = "?l")
            *)
     | Ptyp_tuple of core_type list
           (* T1 * ... * Tn   (n >= 2) *)

   reason_parser.mly
   ---------------
   In general:

                                          syntax variant          {pblah_desc: core_blah_desc
                                                                   pblah_loc: {txt, loc}
                                                                   pblah_attributes: ... }
                                         /              \            /       \
   val mkblah: ~loc -> ~attributes ->     core_blah_desc     ->      core_blah
   let mkblah = Blah.mk

*)


let dummy_loc () = {
  loc_start = Lexing.dummy_pos;
  loc_end = Lexing.dummy_pos;
  loc_ghost = false;
}

let mklocation loc_start loc_end = {
  loc_start = loc_start;
  loc_end = loc_end;
  loc_ghost = false;
}

let with_txt a txt = {
    a with txt=txt;
}

let make_real_loc loc = {
    loc with loc_ghost = false
}

let make_ghost_loc loc = {
    loc with loc_ghost = true
}

let ghloc ?(loc=dummy_loc ()) d = { txt = d; loc = (make_ghost_loc loc) }

(**
  * turn an object into a real
  *)
let make_real_exp exp = {
    exp with pexp_loc = make_real_loc exp.pexp_loc
}
let make_real_pat pat = {
    pat with ppat_loc = make_real_loc pat.ppat_loc
}
let make_real_cf cf = {
    cf with pcf_loc = make_real_loc cf.pcf_loc
}

(**
  * turn a object into ghost
  *)
let make_ghost_cf cf = {
    cf with pcf_loc = make_ghost_loc cf.pcf_loc
}
let make_ghost_exp exp = {
    exp with pexp_loc = make_ghost_loc exp.pexp_loc
}

let make_ghost_pat pat = {
    pat with ppat_loc = make_ghost_loc pat.ppat_loc
}

(**
  * change the location state to be a ghost location or real location
  *)
let set_loc_state is_ghost loc =
    if is_ghost then make_ghost_loc loc else make_real_loc loc

let mktyp ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Typ.mk ~loc d

let mkpat ?(attrs=[]) ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Pat.mk ~loc ~attrs d

let mkexp ?(attrs=[]) ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Exp.mk ~loc ~attrs d

let mkmty ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Mty.mk ~loc d

let mksig ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Sig.mk ~loc d

let mkmod ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Mod.mk ~loc d

let mkstr ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Str.mk ~loc d

let mkclass ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cl.mk ~loc d

let mkcty ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cty.mk ~loc d

let mkctf ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Ctf.mk ~loc d

let mkcf ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cf.mk ~loc d

let mkoption d =
  let loc = {d.ptyp_loc with loc_ghost = true} in
  Typ.mk ~loc (Ptyp_constr(mkloc (Ldot (Lident "*predef*", "option")) loc,[d]))

let simple_ghost_text_attr ?(loc=dummy_loc ()) txt =
  let loc = set_loc_state true loc in
  [({txt; loc}, PStr [])]

let mkExplicitArityTuplePat ?(loc=dummy_loc ()) pat =
  (* Tell OCaml type system that what this tuple construction represents is
     not actually a tuple, and should represent several constructor
     arguments.  This allows the syntax the ability to distinguish between:

     X (10, 20)  -- One argument constructor
     X 10 20     -- Multi argument constructor
  *)
  mkpat
    ~loc
    ~attrs:(simple_ghost_text_attr ~loc "explicit_arity")
    pat

let mkExplicitArityTupleExp ?(loc=dummy_loc ()) exp =
  mkexp
    ~loc
    ~attrs:(simple_ghost_text_attr ~loc "explicit_arity")
    exp

let is_pattern_list_single_any = function
  | [{ppat_desc=Ppat_any; ppat_attributes=[]} as onlyItem] -> Some onlyItem
  | _ -> None

let set_structure_item_location x loc = {x with pstr_loc = loc};;

let mkoperator name =
  let loc = name.loc in
  Exp.mk ~loc (Pexp_ident(mkloc (Lident name.txt) loc))


(*
  Ghost expressions and patterns:
  expressions and patterns that do not appear explicitly in the
  source file they have the loc_ghost flag set to true.
  Then the profiler will not try to instrument them and the
  -annot option will not try to display their type.

  Every grammar rule that generates an element with a location must
  make at most one non-ghost element, the topmost one.

  How to tell whether your location must be ghost:
  A location corresponds to a range of characters in the source file.
  If the location contains a piece of code that is syntactically
  valid (according to the documentation), and corresponds to the
  AST node, then the location must be real; in all other cases,
  it must be ghost.

  jordwalke: Noticed that ghost expressions are often used when inserting
   additional AST nodes from a parse rule. Either an extra wrapping one, or an
   additional inner node. This is consistent with the above description, I
   believe.
*)


let ghunit ?(loc=dummy_loc ()) () =
  mkexp ~ghost:true ~loc (Pexp_construct (mknoloc (Lident "()"), None))

let mkinfix arg1 name_operator arg2 =
  mkexp (Pexp_apply(mkoperator name_operator, ["", arg1; "", arg2]))

let neg_float_string f =
  if String.length f > 0 && f.[0] = '-'
  then String.sub f 1 (String.length f - 1)
  else "-" ^ f

let mkuminus name arg =
  match name.txt, arg.pexp_desc with
  | "-", Pexp_constant(Const_int n) ->
      mkexp(Pexp_constant(Const_int(-n)))
  | "-", Pexp_constant(Const_int32 n) ->
      mkexp(Pexp_constant(Const_int32(Int32.neg n)))
  | "-", Pexp_constant(Const_int64 n) ->
      mkexp(Pexp_constant(Const_int64(Int64.neg n)))
  | "-", Pexp_constant(Const_nativeint n) ->
      mkexp(Pexp_constant(Const_nativeint(Nativeint.neg n)))
  | ("-" | "-."), Pexp_constant(Const_float f) ->
      mkexp(Pexp_constant(Const_float(neg_float_string f)))
  | _ ->
      let name = {name with txt=("~" ^ name.txt)} in
      mkexp(Pexp_apply(mkoperator name, ["", arg]))

let mkFunctorThatReturns functorArgs returns =
  List.fold_left (
    fun acc (n, t) -> mkmod (Pmod_functor(n, t, acc))
  ) returns functorArgs

let mkuplus name arg =
  let desc = arg.pexp_desc in
  match name.txt, desc with
  | "+", Pexp_constant(Const_int _)
  | "+", Pexp_constant(Const_int32 _)
  | "+", Pexp_constant(Const_int64 _)
  | "+", Pexp_constant(Const_nativeint _)
  | ("+" | "+."), Pexp_constant(Const_float _) -> mkexp desc
  | _ ->
      let name = {name with txt=("~" ^ name.txt)} in
      mkexp(Pexp_apply(mkoperator name, ["", arg]))

let mkexp_cons consloc args loc =
  mkexp ~loc (Pexp_construct(mkloc (Lident "::") consloc, Some args))

let ghexp_cons consloc args loc =
  mkexp ~ghost:true ~loc (Pexp_construct(mkloc (Lident "::") loc, Some args))

let mkpat_cons consloc args loc =
  mkpat ~loc (Ppat_construct(mkloc (Lident "::") loc, Some args))

let ghpat_cons consloc args loc =
  mkpat ~ghost:true ~loc (Ppat_construct(mkloc (Lident "::") loc, Some args))

let simple_pattern_list_to_tuple ?(loc=dummy_loc ()) lst =
  match lst with
    | [] -> assert false
    | _ -> mkpat ~loc (Ppat_tuple (List.rev lst))

let mktailexp_extension loc seq ext_opt =
  let rec handle_seq = function
    [] ->
      let base_case = match ext_opt with
        | Some ext ->
          ext
        | None ->
          let loc = make_ghost_loc loc in
          let nil = { txt = Lident "[]"; loc } in
          Exp.mk ~loc (Pexp_construct (nil, None)) in
      base_case
  | e1 :: el ->
      let exp_el = handle_seq el in
      let loc = mklocation e1.pexp_loc.loc_start exp_el.pexp_loc.loc_end in
      let arg = mkexp ~ghost:true ~loc (Pexp_tuple [e1; exp_el]) in
      ghexp_cons loc arg loc in
  handle_seq seq

let mktailpat_extension loc seq ext_opt =
  let rec handle_seq = function
    [] ->
      let base_case = match ext_opt with
        | Some ext ->
          ext
        | None ->
          let loc = make_ghost_loc loc in
          let nil = { txt = Lident "[]"; loc } in
          mkpat ~loc (Ppat_construct (nil, None)) in
      base_case
  | p1 :: pl ->
      let pat_pl = handle_seq pl in
      let loc = mklocation p1.ppat_loc.loc_start pat_pl.ppat_loc.loc_end in
      let arg = mkpat ~ghost:true ~loc (Ppat_tuple [p1; pat_pl]) in
      ghpat_cons loc arg loc in
  handle_seq seq

(* Applies attributes to the structure item, not the expression itself. Makes
 * structure item have same location as expression. *)
let mkstrexp e attrs =
  { pstr_desc = Pstr_eval (e, attrs); pstr_loc = e.pexp_loc }

let ghexp_constraint loc e (t1, t2) =
  match t1, t2 with
  | Some t, None -> mkexp ~ghost:true ~loc (Pexp_constraint(e, t))
  | _, Some t -> mkexp ~ghost:true ~loc (Pexp_coerce(e, t1, t))
  | None, None -> assert false

let array_function ?(loc=dummy_loc()) str name =
  ghloc ~loc (Ldot(Lident str, (if !Clflags.fast then "unsafe_" ^ name else name)))

let syntax_error_str loc msg =
  if !Reason_config.recoverable then
    Str.mk ~loc:loc (Pstr_extension (Syntax_util.syntax_error_extension_node loc msg, []))
  else
    raise(Syntaxerr.Error(Syntaxerr.Other loc))

let syntax_error () =
  raise Syntaxerr.Escape_error

let syntax_error_exp loc msg =
  if !Reason_config.recoverable then
    Exp.mk ~loc (Pexp_extension (Syntax_util.syntax_error_extension_node loc msg))
  else
    syntax_error ()

let unclosed opening closing =
  raise(Syntaxerr.Error(Syntaxerr.Unclosed(opening.loc, opening.txt,
                                           closing.loc, closing.txt)))

let unclosed_extension closing =
  Syntax_util.syntax_error_extension_node closing.loc ("Expecting \"" ^ closing.txt ^ "\"")

let unclosed_mod opening closing =
  if !Reason_config.recoverable then
    mkmod(Pmod_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_cl opening closing =
  if !Reason_config.recoverable then
    mkclass(Pcl_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_mty opening closing =
  if !Reason_config.recoverable then
    mkmty(Pmty_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_cty opening closing =
  if !Reason_config.recoverable then
    mkcty(Pcty_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_exp opening closing =
  if !Reason_config.recoverable then
    mkexp(Pexp_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_pat opening closing =
  if !Reason_config.recoverable then
    mkpat(Ppat_extension (unclosed_extension closing))
  else
    unclosed opening closing

let expecting nonterm =
    raise Syntaxerr.(Error(Expecting(nonterm.loc, nonterm.txt)))

let expecting_pat nonterm =
  if !Reason_config.recoverable then
    mkpat(Ppat_extension (Syntax_util.syntax_error_extension_node nonterm.loc ("Expecting " ^ nonterm.txt)))
  else
    expecting nonterm

let not_expecting pos nonterm =
    raise Syntaxerr.(Error(Not_expecting(rhs_loc pos, nonterm)))

let bigarray_function ?(loc=dummy_loc()) str name =
  ghloc ~loc (Ldot(Ldot(Lident "Bigarray", str), name))

let bigarray_untuplify = function
    { pexp_desc = Pexp_tuple explist; pexp_loc = _ } -> explist
  | exp -> [exp]

let bigarray_get ?(loc=dummy_loc()) arr arg =
  let get = if !Clflags.fast then "unsafe_get" else "get" in
  match bigarray_untuplify arg with
    [c1] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array1" get)),
                       ["", arr; "", c1]))
  | [c1;c2] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array2" get)),
                       ["", arr; "", c1; "", c2]))
  | [c1;c2;c3] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array3" get)),
                       ["", arr; "", c1; "", c2; "", c3]))
  | coords ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Genarray" "get")),
                       ["", arr; "", mkexp ~ghost:true ~loc (Pexp_array coords)]))

let bigarray_set ?(loc=dummy_loc()) arr arg newval =
  let set = if !Clflags.fast then "unsafe_set" else "set" in
  match bigarray_untuplify arg with
    [c1] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array1" set)),
                       ["", arr; "", c1; "", newval]))
  | [c1;c2] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array2" set)),
                       ["", arr; "", c1; "", c2; "", newval]))
  | [c1;c2;c3] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array3" set)),
                       ["", arr; "", c1; "", c2; "", c3; "", newval]))
  | coords ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Genarray" "set")),
                       ["", arr;
                        "", mkexp ~ghost:true ~loc (Pexp_array coords);
                        "", newval]))

let lapply p1 p2 =
  if !Clflags.applicative_functors
  then Lapply(p1, p2)
  else raise (Syntaxerr.Error(Syntaxerr.Applicative_path (symbol_rloc())))

let exp_of_label label =
  mkexp ~loc:label.loc (Pexp_ident {label with txt=Lident(Longident.last label.txt)})

let pat_of_label label =
  mkpat ~loc:label.loc (Ppat_var {label with txt=(Longident.last label.txt)})

let check_variable vl loc v =
  if List.mem v vl then
    raise Syntaxerr.(Error(Variable_in_scope(loc,v)))

let varify_constructors var_names t =
  let rec loop t =
    let desc =
      match t.ptyp_desc with
      | Ptyp_any -> Ptyp_any
      | Ptyp_var x ->
          check_variable var_names t.ptyp_loc x;
          Ptyp_var x
      | Ptyp_arrow (label,core_type,core_type') ->
          Ptyp_arrow(label, loop core_type, loop core_type')
      | Ptyp_tuple lst -> Ptyp_tuple (List.map loop lst)
      | Ptyp_constr( { txt = Lident s }, []) when List.mem s var_names ->
          Ptyp_var s
      | Ptyp_constr(longident, lst) ->
          Ptyp_constr(longident, List.map loop lst)
      | Ptyp_object (lst, o) ->
          Ptyp_object
            (List.map (fun (s, attrs, t) -> (s, attrs, loop t)) lst, o)
      | Ptyp_class (longident, lst) ->
          Ptyp_class (longident, List.map loop lst)
      | Ptyp_alias(core_type, string) ->
          check_variable var_names t.ptyp_loc string;
          Ptyp_alias(loop core_type, string)
      | Ptyp_variant(row_field_list, flag, lbl_lst_option) ->
          Ptyp_variant(List.map loop_row_field row_field_list,
                       flag, lbl_lst_option)
      | Ptyp_poly(string_lst, core_type) ->
          List.iter (check_variable var_names t.ptyp_loc) string_lst;
          Ptyp_poly(string_lst, loop core_type)
      | Ptyp_package(longident,lst) ->
          Ptyp_package(longident,List.map (fun (n,typ) -> (n,loop typ) ) lst)
      | Ptyp_extension (s, arg) ->
          Ptyp_extension (s, arg)
    in
    {t with ptyp_desc = desc}
  and loop_row_field  =
    function
      | Rtag(label,attrs,flag,lst) ->
          Rtag(label,attrs,flag,List.map loop lst)
      | Rinherit t ->
          Rinherit (loop t)
  in
  loop t

(**
  I believe that wrap_type_annotation will automatically generate the type
  arguments (type a) (type b) based on what was listed before the dot in a
  polymorphic type annotation that uses locally abstract types.
 *)
let wrap_type_annotation newtypes core_type body =
  let exp = mkexp(Pexp_constraint(body,core_type)) in
  let exp =
    List.fold_right (fun newtype exp -> mkexp (Pexp_newtype (newtype, exp)))
      newtypes exp
  in
  let typ = mktyp ~ghost:true (Ptyp_poly(newtypes,varify_constructors newtypes core_type)) in
  (exp, typ)


let struct_item_extension (ext_attrs, ext_id) structure_item =
  mkstr ~ghost:true (Pstr_extension ((ext_id, PStr [structure_item]), ext_attrs))

let extension_expression (ext_attrs, ext_id) item_expr =
  mkexp ~ghost:true ~attrs:ext_attrs (Pexp_extension (ext_id, PStr [mkstrexp item_expr []]))

(* There's no more need for these functions - this was for the following:
 *
 *     fun % ext [@foo] arg => arg;
 *
 *   Becoming
 *
 *     [%ext  (fun arg => arg) [@foo]]
 *
 *   Which we no longer support.
 *)
(* Applies the attributes to the body, then wraps entire thing in an extension
 * expression, whose payload consists of a single structure item that is body
 *)
(* let wrap_exp_attrs body (ext, attrs) = *)
(*   (* todo: keep exact location for the entire attribute *) *)
(*   let body = {body with pexp_attributes = attrs @ body.pexp_attributes} in *)
(*   match ext with *)
(*   | None -> body *)
(*   | Some id -> mkexp ~ghost:true (Pexp_extension (id, PStr [mkstrexp body []])) *)

(* Why not just mkexp with the right attributes in the first place? *)
(* let mkexp_attrs d attrs = *)
(*   wrap_exp_attrs (mkexp d) attrs *)

let mkcf_attrs ?(loc=dummy_loc()) d attrs =
  Cf.mk ~loc ~attrs d

let mkctf_attrs d attrs =
  Ctf.mk ~attrs d

let add_nonrec rf attrs pos =
  match rf with
  | Recursive -> attrs
  | Nonrecursive ->
      let name = { txt = "nonrec"; loc = rhs_loc pos } in
        (name, PStr []) :: attrs

type let_binding =
  { lb_pattern: pattern;
    lb_expression: expression;
    (* The meaning of lb_leading_attributes and lbs_extension are dependent on
     * the context of the let binding (module/expression etc) *)
    lb_attributes: attributes;
    (* lb_docs: docs Lazy.t; *)
    (* lb_text: text Lazy.t; *)
    lb_loc: Location.t; }

type let_bindings =
  { lbs_bindings: let_binding list;
    lbs_rec: rec_flag;
    lbs_extension: string Asttypes.loc option;
    (* In Reason, we use this field to represent
       extension attributes attached to the extension on a series of "let/and"
       bindings As in: let [@extAttrs ] [%id] [@attribute] x = ...; It only
       makes sense to have [lbs_attributes] when there is an [lbs_extension].
     *)
    lbs_attributes: attributes;
    lbs_loc: Location.t }

let mklb (p, e) attrs loc =
  { lb_pattern = p;
    lb_expression = e;
    (* Only some individual let bindings are allowed to have attributes
     * depending on the context *)
    lb_attributes = attrs;
    lb_loc = loc; }

let mklbs (extAttrs, extId) rf lb loc =
  { lbs_bindings = [lb];
    lbs_rec = rf;
    lbs_extension = extId ;
    lbs_attributes = extAttrs;
    lbs_loc = loc; }

let addlb lbs lb =
  { lbs with lbs_bindings = lb :: lbs.lbs_bindings }

let val_of_let_bindings lbs =
  let bindings =
    List.map
      (fun lb ->
         Vb.mk ~loc:lb.lb_loc ~attrs:lb.lb_attributes
           (* ~docs:(Lazy.force lb.lb_docs) *)
           (* ~text:(Lazy.force lb.lb_text) *)
           lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
  let str = mkstr(Pstr_value(lbs.lbs_rec, List.rev bindings)) in
  (* Note that for value bindings, when there's an extension, the
   * lbs_attributes are attributes on the extension *)
  match (lbs.lbs_extension) with
    | None -> str
    | Some ext_id -> struct_item_extension (lbs.lbs_attributes, ext_id) str

let expr_of_let_bindings lbs body =
  let bindings =
    List.map
      (fun lb ->
         (* Individual let bindings in an *expression* can't have item attributes. *)
         if lb.lb_attributes <> [] then
           raise Syntaxerr.(Error(Not_expecting(lb.lb_loc, "item attribute")));
         Vb.mk ~loc:lb.lb_loc lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
  (* The location of this expression unfortunately includes the entire rule,
   * which will include any preceeding extensions. *)
  let item_expr = mkexp (Pexp_let(lbs.lbs_rec, List.rev bindings, body)) in
  (* Note that for let expression bindings, when there's an extension, the
   * lbs_attributes are attributes on the entire [let ..in x] expression. *)
  match lbs.lbs_extension with
    | None -> item_expr
    | Some ext_id -> extension_expression (lbs.lbs_attributes, ext_id) item_expr

let class_of_let_bindings lbs body =
  let bindings =
    List.map
      (fun lb ->
         if lb.lb_attributes <> [] then
           raise Syntaxerr.(Error(Not_expecting(lb.lb_loc, "item attribute")));
         Vb.mk ~loc:lb.lb_loc lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
    if lbs.lbs_extension <> None then
      raise Syntaxerr.(Error(Not_expecting(lbs.lbs_loc, "extension")));
    if lbs.lbs_attributes <> [] then
      raise Syntaxerr.(Error(Not_expecting(lbs.lbs_loc, "attributes")));
    mkclass(Pcl_let (lbs.lbs_rec, List.rev bindings, body))

(*
 * arity_conflict_resolving_mapper is triggered when both "implicit_arity" "explicit_arity"
 * are in the attribtues. In that case we have to remove "explicit_arity"
 *
 * However, if we simply remove explicit_arity, we would end up with a
 * wrapping tuple which has only one component (inner tuple).
 * This is against the invariance where tuples must have 2+ components.
 * Therefore, in the case we have to remove explicit_arity, we also need to
 * unwrap the tuple to expose the inner tuple directly.
 *
 *)
let arity_conflict_resolving_mapper =
{ default_mapper with
  expr = begin fun mapper expr ->
    match expr with
      | {pexp_desc=Pexp_construct(lid, args);
         pexp_loc;
         pexp_attributes} when attributes_conflicted "implicit_arity" "explicit_arity" pexp_attributes ->
         let new_args =
           match args with
             | Some {pexp_desc = Pexp_tuple [sp]} -> Some sp
             | _ -> args in
         default_mapper.expr mapper
         {pexp_desc=Pexp_construct(lid, new_args); pexp_loc; pexp_attributes=
          normalized_attributes "explicit_arity" pexp_attributes}
      | x -> default_mapper.expr mapper x
  end;
  pat = begin fun mapper pattern ->
    match pattern with
      | {ppat_desc=Ppat_construct(lid, args);
         ppat_loc;
         ppat_attributes} when attributes_conflicted "implicit_arity" "explicit_arity" ppat_attributes ->
         let new_args =
           match args with
             | Some {ppat_desc = Ppat_tuple [sp]} -> Some sp
             | _ -> args in
         default_mapper.pat mapper
         {ppat_desc=Ppat_construct(lid, new_args); ppat_loc; ppat_attributes=
          normalized_attributes "explicit_arity" ppat_attributes}
      | x -> default_mapper.pat mapper x
  end;
}

let default_mapper_chain = [arity_conflict_resolving_mapper; reason_to_ml_swap_operator_mapper; unescape_stars_slashes_mapper]

let rec string_of_longident = function
    | Lident s -> s
    | Ldot(longPrefix, s) ->
        s
    | Lapply (y,s) -> string_of_longident s

let built_in_explicit_arity_constructors = ["Some"; "Assert_failure"; "Match_failure"]


# 2278 "reason_parser.ml"

let rec _menhir_error1721 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce70 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2527 "reason_parser.mly"
      ( mkexp (Pexp_try(_2, List.rev _4)) )
# 2296 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce69 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2525 "reason_parser.mly"
      ( mkexp (Pexp_match(_2, List.rev _4)) )
# 2311 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce670 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__poly_type_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__poly_type_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_poly_type = 
# 3735 "reason_parser.mly"
                                         (_1)
# 2321 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7993) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_poly_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7975 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 2335 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1326
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1326
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1326) : 'freshtv7976)
    | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7977 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1575
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1575
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1575) : 'freshtv7978)
    | MenhirState1689 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7981 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7979 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_private_flag), _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_poly_type)) = _menhir_stack in
        let _5 = () in
        let _2 = () in
        let _endpos = _endpos__6_ in
        let _v : 'tv_method_ = let _4 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2376 "reason_parser.ml"
          
        in
        
# 2040 "reason_parser.mly"
      ( if _1 = Override then syntax_error ();
        _4, _3, Cfk_virtual _6 )
# 2383 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7980)) : 'freshtv7982)
    | MenhirState1694 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7985 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7983 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_poly_type)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _2 = () in
        let _endpos = _endpos__6_ in
        let _v : 'tv_method_ = let _4 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2405 "reason_parser.ml"
          
        in
        
# 2037 "reason_parser.mly"
      ( if _1 = Override then syntax_error ();
        _4, Private, Cfk_virtual _6 )
# 2412 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7984)) : 'freshtv7986)
    | MenhirState1714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7989 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1721 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7987 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1722) : 'freshtv7988)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1721 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7990)
    | MenhirState152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7991 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2453
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2453
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2453) : 'freshtv7992)
    | _ ->
        _menhir_fail ()) : 'freshtv7994)

and _menhir_reduce209 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_semi_terminated_seq_expr), _startpos_x0_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2536 "reason_parser.ml"
      
    in
    
# 2663 "reason_parser.mly"
      ( syntax_error_exp _2.loc "SyntaxError in block" )
# 2542 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_leading_bar_match_cases_to_sequence_body : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_leading_bar_match_cases_to_sequence_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7961 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7959 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2143 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce69 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7960)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2143) : 'freshtv7962)
    | MenhirState2340 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7965 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2341 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7963 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2341 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce70 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7964)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2341) : 'freshtv7966)
    | MenhirState2427 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2428 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7967 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2428 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce69 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7968)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2428) : 'freshtv7970)
    | MenhirState2545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7973 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2546 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2546 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce70 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7972)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2546) : 'freshtv7974)
    | _ ->
        _menhir_fail ()

and _menhir_goto_simple_pattern_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7955 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState564
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | BARRBRACKET | COLON | COLONCOLON | COMMA | LBRACKETAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce166 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState564)) : 'freshtv7956)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7957 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState650
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLON | COLONCOLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | LBRACKETAT | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce166 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState650)) : 'freshtv7958)
    | _ ->
        _menhir_fail ()

and _menhir_reduce557 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__poly_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__poly_type) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__poly_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 2756 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7953) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__poly_type_) = _v in
    ((match _menhir_s with
    | MenhirState152 | MenhirState1714 | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7949) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__poly_type_) = _v in
        (_menhir_reduce670 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7950)
    | MenhirState1694 | MenhirState1689 | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7951) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__poly_type_) = _v in
        (_menhir_reduce670 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7952)
    | _ ->
        _menhir_fail ()) : 'freshtv7954)

and _menhir_goto_optional_type_parameter_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_parameter_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2070 | MenhirState1967 | MenhirState1415 | MenhirState1361 | MenhirState1249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7945 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1358
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState1358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce639 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7946)
    | MenhirState90 | MenhirState1849 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7947 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1858
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState1858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce639 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7948)
    | _ ->
        _menhir_fail ()

and _menhir_reduce337 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 2827 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters -> Lexing.position -> _menhir_state -> 'tv_class_fun_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_class_fun_binding) ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 2833 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2845 "reason_parser.ml"
      
    in
    
# 1814 "reason_parser.mly"
      (
       (_2, _4, _1, List.rev _3)
      )
# 2853 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce28 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_binding) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_binding = 
# 1864 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 2865 "reason_parser.ml"
     in
    _menhir_goto__class_fun_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2322 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce209 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7944)

and _menhir_reduce208 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_semi_terminated_seq_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2661 "reason_parser.mly"
      ( _2 )
# 2888 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2235 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7941 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce209 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7942)

and _menhir_goto_leading_bar_match_case_to_sequence_body : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_leading_bar_match_case_to_sequence_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState2546 | MenhirState2428 | MenhirState2341 | MenhirState2143 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7935 * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7933 * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_leading_bar_match_case_to_sequence_body) : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
        let _v : 'tv_leading_bar_match_cases_to_sequence_body = 
# 3074 "reason_parser.mly"
                                                                                     ( _2 :: _1 )
# 2917 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_cases_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7934)) : 'freshtv7936)
    | MenhirState2545 | MenhirState2427 | MenhirState2340 | MenhirState761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7939) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7937) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_leading_bar_match_case_to_sequence_body) : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let _v : 'tv_leading_bar_match_cases_to_sequence_body = 
# 3073 "reason_parser.mly"
                                            ( [_1] )
# 2932 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_cases_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7938)) : 'freshtv7940)
    | _ ->
        _menhir_fail ()

and _menhir_reduce182 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 2941 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 2947 "reason_parser.ml"
    )), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_module_binding_body)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)), _endpos__6_), _endpos__7_, _, (_7 : 'tv_semi_terminated_seq_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__semi_terminated_seq_expr_row = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2963 "reason_parser.ml"
      
    in
    
# 2398 "reason_parser.mly"
                                                                                                     (
      let item_attrs = _5 in
      mkexp ~attrs:item_attrs (Pexp_letmodule(_3, _4, _7))
    )
# 2972 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce184 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)), _endpos__3_, _), _endpos__4_, _, (_4 : 'tv_semi_terminated_seq_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__semi_terminated_seq_expr_row = 
# 2406 "reason_parser.mly"
                                                             (
      let item_attrs = _2 in
      mkexp ~attrs:item_attrs (Pexp_sequence(_1, _4))
    )
# 2988 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce179 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_semi_terminated_seq_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2377 "reason_parser.mly"
                                               (
      expr_of_let_bindings _1 _3
    )
# 3003 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce183 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_override_flag), _startpos__3_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__5_, _, (_5 : 'tv_post_item_attributes)), _endpos__6_), _endpos__7_, _, (_7 : 'tv_semi_terminated_seq_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__semi_terminated_seq_expr_row = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 3024 "reason_parser.ml"
      
    in
    
# 2402 "reason_parser.mly"
                                                                                                    (
      let item_attrs = _5 in
      mkexp ~attrs:item_attrs (Pexp_open(_3, _4, _7))
    )
# 3033 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce177 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_semi_terminated_seq_expr_row) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2368 "reason_parser.mly"
                                                      (
      extension_expression _1 _2
    )
# 3047 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce178 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_semi_terminated_seq_expr_row) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2371 "reason_parser.mly"
                                 (
      _1
    )
# 3060 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce147 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_row_field)), _startpos__3_), _, (_4 : 'tv_row_field_list)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3955 "reason_parser.mly"
      ( mktyp(Ptyp_variant(_2 :: List.rev _4, Closed, None)) )
# 3075 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce148 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3957 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Open, None)) )
# 3089 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce150 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3961 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, Some [])) )
# 3103 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce146 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, _startpos__2_), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3953 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, None)) )
# 3118 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run169 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _startpos ->
    let _menhir_stack = (_menhir_stack, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState169

and _menhir_error727 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3164 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce172 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3300 "reason_parser.mly"
      ( mkpat(Ppat_lazy _2) )
# 3179 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce725 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern_list = 
# 3408 "reason_parser.mly"
    ( [_1] )
# 3190 "reason_parser.ml"
     in
    _menhir_goto_simple_pattern_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce724 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_list)), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_pattern_list = 
# 3406 "reason_parser.mly"
    ( _2::_1 )
# 3201 "reason_parser.ml"
     in
    _menhir_goto_simple_pattern_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce167 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3284 "reason_parser.mly"
    (
      mkpat (Ppat_variant(_1, Some _2))
    )
# 3215 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce163 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__pattern_without_or = 
# 3259 "reason_parser.mly"
      ( _1 )
# 3227 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_non_arrowed_simple_core_type_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7909 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState254
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND | AND | AS | BAR | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | GREATER | LBRACKETAT | LBRACKETATAT | QUESTION | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce134 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState254)) : 'freshtv7910)
    | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7911 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState316
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | BAR | BARRBRACKET | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce135 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState316)) : 'freshtv7912)
    | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7913 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState328
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | BAR | BARRBRACKET | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce134 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState328)) : 'freshtv7914)
    | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1339 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1283 | MenhirState1278 | MenhirState1264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7917 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1271
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7915 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1271 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1272) : 'freshtv7916)
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1271)) : 'freshtv7918)
    | MenhirState2515 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7921 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1494
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7919 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1494 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1495) : 'freshtv7920)
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1494)) : 'freshtv7922)
    | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7923 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1587
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce33 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1587)) : 'freshtv7924)
    | MenhirState1636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7925 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1637
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce16 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1637)) : 'freshtv7926)
    | MenhirState1793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7927 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1794
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce33 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1794)) : 'freshtv7928)
    | MenhirState120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7929 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2466
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND | AND | AS | BAR | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | GREATER | LBRACKETAT | LBRACKETATAT | QUESTION | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce135 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2466)) : 'freshtv7930)
    | MenhirState2676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7931 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2677
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce16 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2677)) : 'freshtv7932)
    | _ ->
        _menhir_fail ()

and _menhir_goto_non_arrowed_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState666 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7845 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState262
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState262
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce57 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7846)
    | MenhirState323 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7855 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3739 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState324
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7847 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3751 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState324 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2400) : 'freshtv7848)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState324
          | QUESTION ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7853 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3796 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState324 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7849 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3810 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKET ->
                      _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETGREATER ->
                      _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETLESS ->
                      _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState326) : 'freshtv7850)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7851 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 3854 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7852)) : 'freshtv7854)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState324)) : 'freshtv7856)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7857 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState331
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState331
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce57 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7858)
    | MenhirState1017 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7861 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1018
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7859 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1018 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1019) : 'freshtv7860)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1018
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1018)) : 'freshtv7862)
    | MenhirState1190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7865 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1191
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7863 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1191 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1192) : 'freshtv7864)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1191
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1191)) : 'freshtv7866)
    | MenhirState1704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7869 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1705
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7867 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1705 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1706 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1706) : 'freshtv7868)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1705
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1705)) : 'freshtv7870)
    | MenhirState1764 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7873 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4158 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1765
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7871 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4170 "reason_parser.ml"
              ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1765 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1763 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1761 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1766) : 'freshtv7872)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1765
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1765)) : 'freshtv7874)
    | MenhirState2663 | MenhirState1760 | MenhirState1775 | MenhirState1768 | MenhirState1766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7877 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1767
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7875 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1767 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1763 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1761 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1768) : 'freshtv7876)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1767
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1767)) : 'freshtv7878)
    | MenhirState1773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7881 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4276 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1774
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7879 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4288 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1774 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1763 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1761 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1775 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1775) : 'freshtv7880)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1774
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1774)) : 'freshtv7882)
    | MenhirState1799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7885 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4339 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1800
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7883 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4351 "reason_parser.ml"
              ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1800 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1801 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1801) : 'freshtv7884)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1800
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1800)) : 'freshtv7886)
    | MenhirState2692 | MenhirState2057 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1803 | MenhirState1801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7889 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1802
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7887 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1802 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1803 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1803) : 'freshtv7888)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1802
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1802)) : 'freshtv7890)
    | MenhirState1808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7893 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4457 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1809
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7891 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4469 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1809 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1810 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1810) : 'freshtv7892)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1809
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1809)) : 'freshtv7894)
    | MenhirState2214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7897 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2215
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7895 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2215 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2216) : 'freshtv7896)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2215
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2215)) : 'freshtv7898)
    | MenhirState142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7907 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4613 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2455
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7899 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4625 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2455 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2459 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2459) : 'freshtv7900)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2455
          | QUESTION ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7905 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4670 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2455 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7901 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4684 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKET ->
                      _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETGREATER ->
                      _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETLESS ->
                      _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2457) : 'freshtv7902)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7903 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 4728 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7904)) : 'freshtv7906)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2455)) : 'freshtv7908)
    | _ ->
        _menhir_fail ()

and _menhir_reduce284 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_pattern_two_or_more_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3368 "reason_parser.mly"
      ( mkpat(Ppat_tuple(List.rev _2)) )
# 4749 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run587 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_pattern_two_or_more_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos ->
    let _menhir_stack = (_menhir_stack, _endpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState587

and _menhir_reduce655 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_or_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_or_pattern) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern = 
# 3254 "reason_parser.mly"
               ( _1 )
# 4817 "reason_parser.ml"
     in
    _menhir_goto_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2463 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce381 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constrain -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_constrain) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constraints)), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constraints = 
# 3483 "reason_parser.mly"
                                                ( _3 :: _1 )
# 4834 "reason_parser.ml"
     in
    _menhir_goto_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto__poly_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__poly_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState152 | MenhirState1714 | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7841) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__poly_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce557 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7842)
    | MenhirState1694 | MenhirState1689 | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7843) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__poly_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce557 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7844)
    | _ ->
        _menhir_fail ()

and _menhir_reduce137 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_core_type_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3926 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        match _2 with
        | [] -> raise (Syntaxerr.Error(Syntaxerr.Applicative_path loc))
        | one::[] -> one
        | moreThanOne -> mktyp(Ptyp_tuple(List.rev moreThanOne)) )
# 4877 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run269 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_core_type_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos ->
    let _menhir_stack = (_menhir_stack, _endpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState269

and _menhir_error1902 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1681 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1218 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce207 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_type_constraint)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2656 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        ghexp_constraint loc _2 _3
      )
# 4948 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce419 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_type_constraint)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _v : 'tv_expr_optional_constraint = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3158 "reason_parser.mly"
                         (
      let loc = mklocation _symbolstartpos _endpos in
      ghexp_constraint loc _1 _2
    )
# 4964 "reason_parser.ml"
     in
    _menhir_goto_expr_optional_constraint _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_reduce637 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list -> Lexing.position -> _menhir_state -> 'tv_optional_type_variable_with_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_optional_type_variable_with_variance) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_optional_type_parameter_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_optional_type_parameter_list = 
# 3517 "reason_parser.mly"
                                                                         ( _2 :: _1 )
# 4975 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameter_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce636 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_variable_with_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_optional_type_variable_with_variance) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_optional_type_parameter_list = 
# 3516 "reason_parser.mly"
                                                                         ( [_1] )
# 4985 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameter_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce51 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_constructor_type)) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 5006 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 5018 "reason_parser.ml"
      
    in
    
# 1973 "reason_parser.mly"
      ( unclosed_cl (with_txt _1 "(") (with_txt _5 ")") )
# 5024 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce344 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_fun_binding_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_fun_binding_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_fun_binding = 
# 1861 "reason_parser.mly"
                                                        (_1)
# 5034 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7839) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_fun_binding) = _v in
    ((match _menhir_s with
    | MenhirState1817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7831 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce28 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7832)
    | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7833 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 5054 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce337 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7834)
    | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7835 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce28 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7836)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7837 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 5072 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce337 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7838)
    | _ ->
        _menhir_fail ()) : 'freshtv7840)

and _menhir_reduce695 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__semi_terminated_seq_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__semi_terminated_seq_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_seq_expr = 
# 2366 "reason_parser.mly"
                                                                        (_1)
# 5088 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7829) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_semi_terminated_seq_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce183 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7798)
    | MenhirState1162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7799 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce179 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7800)
    | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7801 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce184 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7802)
    | MenhirState2128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7803 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 5115 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce182 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7804)
    | MenhirState1150 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7807 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7805 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_semi_terminated_seq_expr), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _v : 'tv_leading_bar_match_case_to_sequence_body = 
# 3098 "reason_parser.mly"
      ( Exp.case _1 ~guard:_3 _5 )
# 5129 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_case_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7806)) : 'freshtv7808)
    | MenhirState2141 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7811 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7809 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _startpos__2_), _endpos__3_, _, (_3 : 'tv_semi_terminated_seq_expr), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_leading_bar_match_case_to_sequence_body = 
# 3096 "reason_parser.mly"
      ( Exp.case _1 _3 )
# 5142 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_case_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7810)) : 'freshtv7812)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7815 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2235 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7813 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce208 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7814)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2235 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7816)
    | MenhirState2247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7817 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce179 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7818)
    | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7819 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce184 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7820)
    | MenhirState2314 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce183 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7822)
    | MenhirState2320 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7823 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 5180 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce182 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7824)
    | MenhirState431 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2322 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce208 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7826)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2322 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7828)
    | _ ->
        _menhir_fail ()) : 'freshtv7830)

and _menhir_reduce696 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__semi_terminated_seq_expr_row_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_seq_expr_row = 
# 2386 "reason_parser.mly"
                                                                                (_1)
# 5211 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7795) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1171 | MenhirState1162 | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7787) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce178 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7788)
    | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7789 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce177 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7790)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState2253 | MenhirState2247 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7791) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce178 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7792)
    | MenhirState2249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7793 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce177 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7794)
    | _ ->
        _menhir_fail ()) : 'freshtv7796)

and _menhir_reduce254 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_expr = 
# 1249 "reason_parser.mly"
      ( _2 )
# 5265 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_row_field_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_row_field_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7739 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7735 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce146 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7736)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7737 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7738)) : 'freshtv7740)
    | MenhirState291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7747 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7741 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState294 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState294) : 'freshtv7742)
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7743 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce150 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7744)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7745 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7746)) : 'freshtv7748)
    | MenhirState302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7753 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7749 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce148 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7750)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7751 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7752)) : 'freshtv7754)
    | MenhirState306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7759 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7755 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce146 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7756)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7757 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7758)) : 'freshtv7760)
    | MenhirState312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7765 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7761 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce147 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7762)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7763 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7764)) : 'freshtv7766)
    | MenhirState2443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7771 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7767 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce147 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7768)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7769 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7770)) : 'freshtv7772)
    | MenhirState160 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7777 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7773 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce148 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7774)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7775 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7776)) : 'freshtv7778)
    | MenhirState157 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7785 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run169 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7779 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2450 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2450) : 'freshtv7780)
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7781 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce150 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7782)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7783 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7784)) : 'freshtv7786)
    | _ ->
        _menhir_fail ()

and _menhir_goto_str_extension_constructors : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_extension_constructors -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7729 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2533
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7727 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2533 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLONCOLON ->
                  _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2534 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run1512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2534 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2534 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run1499 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2534 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run1484 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2534 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2534) : 'freshtv7728)
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2533
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2533) : 'freshtv7730)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7733 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2606
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7731 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2606 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLONCOLON ->
                  _menhir_run2602 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2607 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run2598 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2607 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2593 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2607 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2607 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2607 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2607) : 'freshtv7732)
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2606
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2606) : 'freshtv7734)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declarations_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1263 | MenhirState1338 | MenhirState1340 | MenhirState1281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7719 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1310
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1310 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce393 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1310)) : 'freshtv7720)
    | MenhirState1313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7721 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1314
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce395 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1314)) : 'freshtv7722)
    | MenhirState106 | MenhirState2514 | MenhirState2516 | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7723 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2501
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce393 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2501)) : 'freshtv7724)
    | MenhirState2504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7725 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2505
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce395 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2505)) : 'freshtv7726)
    | _ ->
        _menhir_fail ()

and _menhir_reduce503 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_let_binding) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_let_bindings = 
# 2870 "reason_parser.mly"
                                                ( _1 )
# 5629 "reason_parser.ml"
     in
    _menhir_goto_let_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1880 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1880

and _menhir_run1178 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1178

and _menhir_reduce255 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_module_expr = 
# 1251 "reason_parser.mly"
      ( mkmod(Pmod_unpack _3) )
# 5758 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce723 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__simple_pattern_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__simple_pattern_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern = 
# 3326 "reason_parser.mly"
                                                   (_1)
# 5769 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7717) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7683 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce163 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7684)
    | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7685 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce167 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7686)
    | MenhirState564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7687 * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce724 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7688)
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7689 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce725 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7690)
    | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce172 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7692)
    | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState738 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7693 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce163 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7694)
    | MenhirState646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7695 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce167 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7696)
    | MenhirState650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7697 * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce724 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7698)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7699 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce725 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7700)
    | MenhirState532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce172 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7702)
    | MenhirState486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7711 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 5824 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error727 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7703 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 5836 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState729) : 'freshtv7704)
          | OPTIONAL_NO_DEFAULT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7707 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 5894 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7705 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 5901 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 5906 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_) = _menhir_stack in
              let _4 = () in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2486 "reason_parser.mly"
      ( ("?" ^ _1, None, _3) )
# 5914 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7706)) : 'freshtv7708)
          | BACKQUOTE | CHAR _ | COLON | EQUALGREATER | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7709 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 5922 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 5927 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_) = _menhir_stack in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2483 "reason_parser.mly"
      ( (_1, None, _3) )
# 5934 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7710)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error727 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7712)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1633 | MenhirState1631 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7715 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7713 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _v : 'tv_labeled_simple_pattern = 
# 2491 "reason_parser.mly"
      ( ("", None, _1) )
# 5951 "reason_parser.ml"
         in
        _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7714)) : 'freshtv7716)
    | _ ->
        _menhir_fail ()) : 'freshtv7718)

and _menhir_run1821 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1821 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1821

and _menhir_goto_direction_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_direction_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState741 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState744) : 'freshtv7680)
    | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState907) : 'freshtv7682)
    | _ ->
        _menhir_fail ()

and _menhir_reduce291 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 6110 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 910 "reason_parser.mly"
       (string)
# 6116 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6133 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6145 "reason_parser.ml"
      
    in
    
# 3386 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _6 ")") )
# 6151 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce594 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_simple_core_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_core_type = 
# 3910 "reason_parser.mly"
    ( _1 )
# 6162 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce598 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_simple_core_type) _startpos__1_ ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_simple_core_type_list = 
# 4013 "reason_parser.mly"
                                     ( [_1] )
# 6172 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_simple_core_type_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce599 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_non_arrowed_simple_core_type) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_non_arrowed_simple_core_type_list = 
# 4014 "reason_parser.mly"
                                                                     ( _2 :: _1 )
# 6183 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_simple_core_type_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce593 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_non_simple_core_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_core_type = 
# 3908 "reason_parser.mly"
    ( _1 )
# 6194 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_pattern_two_or_more_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_pattern_two_or_more_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7671 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run587 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7667 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce284 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7668)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7669 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7670)) : 'freshtv7672)
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7677 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run587 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7673 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce284 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7674)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7675 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7676)) : 'freshtv7678)
    | _ ->
        _menhir_fail ()

and _menhir_reduce641 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__or_pattern_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__or_pattern_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_or_pattern = 
# 3077 "reason_parser.mly"
                                           (_1)
# 6256 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7665) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_or_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7661) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce655 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7662)
    | MenhirState2572 | MenhirState190 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7663) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce655 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7664)
    | _ ->
        _menhir_fail ()) : 'freshtv7666)

and _menhir_error2644 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2566 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run2515 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 6297 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2515
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2515

and _menhir_run2517 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run2474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2517 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2517

and _menhir_goto_amper_type_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_amper_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv7659 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERSAND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7657 * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2435 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2436) : 'freshtv7658)
    | LBRACKETAT ->
        _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2435
    | AMPERAMPER | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2435) : 'freshtv7660)

and _menhir_reduce646 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _v : 'tv_package_type_cstr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6419 "reason_parser.ml"
      
    in
    
# 3974 "reason_parser.mly"
                                                 ( (_2, _4) )
# 6425 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7655) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_package_type_cstr) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2405 | MenhirState238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7649 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2404 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7647 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run239 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2405) : 'freshtv7648)
          | RPAREN ->
              _menhir_reduce647 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2404 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7650)
    | MenhirState2464 | MenhirState133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7653 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2463 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7651 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run134 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2464) : 'freshtv7652)
          | COLONGREATER | RPAREN ->
              _menhir_reduce647 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2463 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7654)
    | _ ->
        _menhir_fail ()) : 'freshtv7656)

and _menhir_run2038 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 902 "reason_parser.mly"
       (string * string option)
# 6497 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | STRING _v ->
        _menhir_run2038 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | LBRACKETATAT | SEMI ->
        _menhir_reduce675 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2038

and _menhir_error2036 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1912 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1899 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1893 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1854 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1718 : _menhir_env -> ((((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce835 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_value = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6559 "reason_parser.ml"
      
    in
    
# 2015 "reason_parser.mly"
                                                                (
      if _1 = Override then not_expecting 7 "not expecting equal - cannot specify value for virtual val"
      else _4, Mutable, Cfk_virtual _6
    )
# 6568 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1675 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce835 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce837 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6591 "reason_parser.ml"
      
    in
    
# 2022 "reason_parser.mly"
      ( _3, _2, Cfk_virtual _5 )
# 6597 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1666 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce837 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error1594 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_value_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_value_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1557
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1557
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1557) : 'freshtv7646)

and _menhir_error1540 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1477 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 902 "reason_parser.mly"
       (string * string option)
# 6633 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | STRING _v ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce675 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1477

and _menhir_error1475 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce847 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_optional_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_with_constraint = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6669 "reason_parser.ml"
      
    in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3705 "reason_parser.mly"
    (
      let last = (
        match _2.txt with
            | Lident s -> s
            | _ -> not_expecting 2 "Long type identifier"
      ) in
     let loc = mklocation _symbolstartpos _endpos in
      Pwith_typesubst
          (Type.mk {_2 with txt=last}
             ~params:_3
             ~manifest:_5
             ~loc)
    )
# 6689 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce378 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constrain = 
# 2285 "reason_parser.mly"
                                           ( _1, _3, symbol_rloc() )
# 6701 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7643) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_constrain) = _v in
    ((match _menhir_s with
    | MenhirState1353 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7639 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain) = _v in
        (_menhir_reduce381 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7640)
    | MenhirState1853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7641 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain) = _v in
        (_menhir_reduce381 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7642)
    | _ ->
        _menhir_fail ()) : 'freshtv7644)

and _menhir_error1354 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1339 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 6734 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1339 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1339
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1339

and _menhir_run1341 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run1283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1341 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1341

and _menhir_reduce818 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_kind = 
# 3506 "reason_parser.mly"
      ( (Ptype_open, Public, Some _2) )
# 6799 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1337 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce175 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__poly_type = 
# 3738 "reason_parser.mly"
          ( _1 )
# 6816 "reason_parser.ml"
     in
    _menhir_goto__poly_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce176 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_typevar_list), _startpos__1_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__poly_type = 
# 3740 "reason_parser.mly"
          ( mktyp(Ptyp_poly(List.rev _1, _3)) )
# 6829 "reason_parser.ml"
     in
    _menhir_goto__poly_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce811 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_kind = 
# 3492 "reason_parser.mly"
      ( (Ptype_abstract, Private, Some _3) )
# 6842 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce449 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3603 "reason_parser.mly"
                                                ( (List.rev _1,Some _3) )
# 6854 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce450 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3605 "reason_parser.mly"
                                                ( ([],Some _2) )
# 6866 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1228 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1215 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1207 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce807 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_constraint_right_of_colon = 
# 3242 "reason_parser.mly"
                                          ( (Some _1, Some _3) )
# 6893 "reason_parser.ml"
     in
    _menhir_goto_type_constraint_right_of_colon _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce805 : _menhir_env -> ('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_constraint = 
# 3247 "reason_parser.mly"
                                                ( (None, Some _2) )
# 6905 "reason_parser.ml"
     in
    _menhir_goto_type_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce286 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3373 "reason_parser.mly"
      ( mkpat(Ppat_constraint(_2, _4)) )
# 6920 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce287 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6941 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6953 "reason_parser.ml"
      
    in
    
# 3375 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _5 ")") )
# 6959 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce162 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__pattern_optional_constraint = 
# 3414 "reason_parser.mly"
                                ( mkpat(Ppat_constraint(_1, _3)) )
# 6972 "reason_parser.ml"
     in
    _menhir_goto__pattern_optional_constraint _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_core_type_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_core_type_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2473 | MenhirState117 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7631 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7627 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce137 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7628)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7629 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7630)) : 'freshtv7632)
    | MenhirState127 | MenhirState1282 | MenhirState1265 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7637 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7633 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce137 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7634)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7635 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7636)) : 'freshtv7638)
    | _ ->
        _menhir_fail ()

and _menhir_error1910 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1226 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce269 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_not_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern = 
# 3330 "reason_parser.mly"
                             ( _1 )
# 7045 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_simple_labeled_expr_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7619 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState822
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run816 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce65 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7620)
    | MenhirState928 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7621 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState935
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run929 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce65 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7622)
    | MenhirState1641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7623 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1642
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run816 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce13 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1642)) : 'freshtv7624)
    | MenhirState2681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7625 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2682
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run929 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce13 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2682)) : 'freshtv7626)
    | _ ->
        _menhir_fail ()

and _menhir_reduce210 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2665 "reason_parser.mly"
      ( mkexp(Pexp_tuple(List.rev _2)) )
# 7282 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1142 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1142

and _menhir_reduce214 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_expr_comma_list)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7380 "reason_parser.ml"
      
    in
    
# 2679 "reason_parser.mly"
      ( mkexp(Pexp_open(Fresh, _1, mkexp(Pexp_tuple(List.rev _4)))) )
# 7386 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1071 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1071

and _menhir_goto_type_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2173 | MenhirState1073 | MenhirState1060 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7591 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        (_menhir_reduce419 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7592)
    | MenhirState1081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7593 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        (_menhir_reduce419 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7594)
    | MenhirState1144 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7599 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv7595 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7596)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7597 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce207 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7598)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7600)
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7603 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1218 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7601 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1219) : 'freshtv7602)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1218 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7604)
    | MenhirState1678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7607 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1681 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7605 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1682) : 'freshtv7606)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1681 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7608)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7611 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1902 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7609 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1903) : 'freshtv7610)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1902 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7612)
    | MenhirState2376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7617 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv7613 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7614)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7615 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce207 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7616)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7618)
    | _ ->
        _menhir_fail ()

and _menhir_reduce223 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_record_expr)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7808 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7820 "reason_parser.ml"
      
    in
    
# 2718 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{") (with_txt _3 "}"))
# 7826 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce227 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_record_expr)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7847 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7859 "reason_parser.ml"
      
    in
    
# 2730 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "{") (with_txt _5 "}") )
# 7865 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_type_variance : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_type_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7589 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7585) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1611) : 'freshtv7586)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7587 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7588)) : 'freshtv7590)

and _menhir_reduce826 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_type_variance = 
# 3542 "reason_parser.mly"
                                                ( Covariant )
# 7908 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce827 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_type_variance = 
# 3543 "reason_parser.mly"
                                                ( Contravariant )
# 7919 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce263 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_module_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7937 "reason_parser.ml"
      
    in
    
# 1549 "reason_parser.mly"
    ( mkmty(Pmty_ident _1) )
# 7943 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run239 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState239 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState239 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState239

and _menhir_run134 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState134

and _menhir_error132 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce151 : _menhir_env -> (((('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ ->
    let (((((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)), _endpos__4_, _startpos__4_), _, (_5 : 'tv_name_tag_list)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3963 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, Some (List.rev _5))) )
# 7993 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce640 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__optional_type_variable_with_variance -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__optional_type_variable_with_variance) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_optional_type_variable_with_variance = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 3519 "reason_parser.mly"
                                                                                (
      let first = fst x in
      let second = snd x in
      ({
        first with ptyp_loc = {
        first.ptyp_loc with
        loc_start = _symbolstartpos;
        loc_end = _endpos}}, second )
)
# 8013 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7583) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_optional_type_variable_with_variance) = _v in
    ((match _menhir_s with
    | MenhirState1849 | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7575) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7576)
    | MenhirState2070 | MenhirState1967 | MenhirState1415 | MenhirState1361 | MenhirState1249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7577) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7578)
    | MenhirState1358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7579 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce637 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7580)
    | MenhirState1858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7581 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce637 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7582)
    | _ ->
        _menhir_fail ()) : 'freshtv7584)

and _menhir_reduce52 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_simple_expr = 
# 1975 "reason_parser.mly"
      ( _2 )
# 8062 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1761 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1761 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1761 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1761

and _menhir_run1763 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 8088 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7571 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 8100 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1773) : 'freshtv7572)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7573 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 8143 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1764) : 'freshtv7574)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce12 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_def) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1926 "reason_parser.mly"
      ( _2 )
# 8190 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce31 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_def) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_def = 
# 1890 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 8202 "reason_parser.ml"
     in
    _menhir_goto__class_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2693 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2664 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7569 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce51 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7570)

and _menhir_goto__non_arrowed_class_constructor_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7567) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__non_arrowed_class_constructor_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7565) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__non_arrowed_class_constructor_type) : 'tv__non_arrowed_class_constructor_type) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1161 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 8239 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7563) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7561) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7559) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let _v : 'tv_non_arrowed_class_constructor_type = 
# 2207 "reason_parser.mly"
                                                                                           (_1)
# 8256 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7557) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_class_constructor_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7549 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7545 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1825) : 'freshtv7546)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7547 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7548)) : 'freshtv7550)
    | MenhirState2701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7555 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7551 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2675 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2671 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2659 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2703) : 'freshtv7552)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7553 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7554)) : 'freshtv7556)
    | _ ->
        _menhir_fail ()) : 'freshtv7558)) : 'freshtv7560)) : 'freshtv7562)) : 'freshtv7564)) : 'freshtv7566)) : 'freshtv7568)

and _menhir_error1822 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1812 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce50 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_constructor_type)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_simple_expr = 
# 1971 "reason_parser.mly"
      ( mkclass(Pcl_constraint(_2, _4)) )
# 8366 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1777 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce51 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7544)

and _menhir_reduce9 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 8382 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 8388 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_constructor_type = 
# 2202 "reason_parser.mly"
      ( mkcty(Pcty_arrow(_1, _3, _5)) )
# 8397 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce8 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 8404 "reason_parser.ml"
) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 8410 "reason_parser.ml"
    )), _startpos__1_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_constructor_type = 
# 2200 "reason_parser.mly"
      ( mkcty(Pcty_arrow("?" ^ _1, mkoption _3, _5)) )
# 8419 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce10 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_constructor_type)) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_constructor_type = 
# 2204 "reason_parser.mly"
      ( mkcty(Pcty_arrow("", _1, _3)) )
# 8432 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce364 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 8439 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 863 "reason_parser.mly"
       (string)
# 8444 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_clty_longident = 
# 4156 "reason_parser.mly"
                                                ( Lident _1 )
# 8451 "reason_parser.ml"
     in
    _menhir_goto_clty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1539 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1539

and _menhir_reduce522 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_fun_binding) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_fun_binding_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 8500 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7541) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
    ((match _menhir_s with
    | MenhirState1817 | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7537) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
        (_menhir_reduce344 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7538)
    | MenhirState2698 | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7539) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
        (_menhir_reduce344 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7540)
    | _ ->
        _menhir_fail ()) : 'freshtv7542)

and _menhir_run1283 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1283
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1283

and _menhir_run2474 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2474
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2474

and _menhir_reduce293 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_structure_item_without_item_extension_sugar) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item = 
# 1392 "reason_parser.mly"
                                                                     (
    struct_item_extension _1 _2
  )
# 8617 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce294 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_structure_item_without_item_extension_sugar) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item = 
# 1395 "reason_parser.mly"
                                                (
    _1
  )
# 8630 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2753 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2717 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1839 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7535 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | RBRACE | RBRACKET ->
        _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState1840
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1840) : 'freshtv7536)

and _menhir_reduce497 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed) ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1216 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 8763 "reason_parser.ml"
      
    in
    
# 2912 "reason_parser.mly"
      ( (_1, _2) )
# 8769 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error86 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1966 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1966

and _menhir_run1974 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7533 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1974 in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 8807 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONEQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 8819 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run1977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1976) : 'freshtv7530)
        | DOT | EQUAL ->
            _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7531 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 8842 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7532)) : 'freshtv7534)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1974

and _menhir_run1414 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1414

and _menhir_run1424 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7527 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1424 in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 8880 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONEQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 8892 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1426 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1426) : 'freshtv7524)
        | DOT | EQUAL ->
            _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7525 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 8915 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7526)) : 'freshtv7528)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1424

and _menhir_error1431 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce261 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_type = 
# 1545 "reason_parser.mly"
      ( _2 )
# 8939 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1965 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        _menhir_run1974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1965

and _menhir_reduce252 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_module_expr = 
# 1245 "reason_parser.mly"
      ( mkmod(Pmod_constraint(_2, _4)) )
# 8969 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1999 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1999

and _menhir_goto__module_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7521) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__module_declaration) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7519) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__module_declaration) : 'tv__module_declaration) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__module_declaration_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1181 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9016 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7517) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7515) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7513) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_mty__module_declaration_) : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_module_declaration = 
# 1748 "reason_parser.mly"
                                                           (_1)
# 9037 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7511) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_declaration) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7503 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 9051 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7501 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 9057 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 9062 "reason_parser.ml"
        )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_), _endpos__5_, _, _startpos__5_), _endpos__6_, _, (_6 : 'tv_module_declaration)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__module_declaration = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 9078 "reason_parser.ml"
          
        in
        
# 1753 "reason_parser.mly"
      ( mkmty(Pmty_functor(_2, Some _4, _6)) )
# 9084 "reason_parser.ml"
         in
        _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7502)) : 'freshtv7504)
    | MenhirState1451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7507 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7505 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_declaration)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__module_declaration = 
# 1755 "reason_parser.mly"
      ( mkmty(Pmty_functor(mkloc "*" (mklocation _startpos__1_ _endpos__1_), None, _3)) )
# 9100 "reason_parser.ml"
         in
        _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7506)) : 'freshtv7508)
    | MenhirState1405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 9108 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKETATAT ->
            _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1456
        | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
            _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1456) : 'freshtv7510)
    | _ ->
        _menhir_fail ()) : 'freshtv7512)) : 'freshtv7514)) : 'freshtv7516)) : 'freshtv7518)) : 'freshtv7520)) : 'freshtv7522)

and _menhir_run1413 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        _menhir_run1424 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1413

and _menhir_run1438 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1438

and _menhir_reduce131 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_simple_module_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1536 "reason_parser.mly"
                       (_1)
# 9165 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce118 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_simple_module_expr) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__module_expr = 
# 1287 "reason_parser.mly"
      ( mkmod (Pmod_apply(_1, _2)) )
# 9177 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce116 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_simple_module_expr) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__module_expr = 
# 1277 "reason_parser.mly"
      ( _1 )
# 9188 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce532 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__semi_terminated_seq_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__semi_terminated_seq_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9201 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7499) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1171 | MenhirState1162 | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7495) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7496)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState2253 | MenhirState2247 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7497) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7498)
    | _ ->
        _menhir_fail ()) : 'freshtv7500)

and _menhir_reduce533 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__semi_terminated_seq_expr_row) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9238 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7493) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7489) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce696 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7490)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7491) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce696 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7492)
    | _ ->
        _menhir_fail ()) : 'freshtv7494)

and _menhir_reduce601 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 9269 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 9275 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_binding_body)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_nonlocal_module_binding_details = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 9287 "reason_parser.ml"
      
    in
    
# 1512 "reason_parser.mly"
                                      (
      (_1, _2)
    )
# 9295 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7487) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_nonlocal_module_binding_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1865 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7475 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1866
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1866
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1866) : 'freshtv7476)
    | MenhirState2303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7477 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2304
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2304
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2304) : 'freshtv7478)
    | MenhirState441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7479 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2306
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2306
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2306) : 'freshtv7480)
    | MenhirState2633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2634
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2634
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2634) : 'freshtv7482)
    | MenhirState2630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7483 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2636
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2636
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2636) : 'freshtv7484)
    | MenhirState2726 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2727
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2727
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2727) : 'freshtv7486)
    | _ ->
        _menhir_fail ()) : 'freshtv7488)

and _menhir_reduce577 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_functor -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_module_binding_body_functor) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body = 
# 1490 "reason_parser.mly"
                                ( _1 )
# 9385 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run2161 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce254 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run2162 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2162

and _menhir_run1958 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce254 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1959 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1959 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1959 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1959

and _menhir_run1942 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1955 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7473 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1942 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1943) : 'freshtv7474)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1942

and _menhir_reduce244 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2789 "reason_parser.mly"
      ( mkexp (Pexp_pack _3) )
# 9559 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run757 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2158 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7471 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState757 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState758) : 'freshtv7472)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState757

and _menhir_error2299 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2135 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce126 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_module_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__module_type = 
# 1591 "reason_parser.mly"
                                                     (
    (* below_EQUALGREATER to prevent following shift reduce conflict:
     *  1158: shift/reduce conflict (shift 1285, reduce 75) on EQUALGREATER
     *  state 1158
     *    module_binding_body_functor : functor_args COLON non_arrowed_module_type . EQUALGREATER module_expr  (59)
     *    module_type : non_arrowed_module_type .  (75)
     *
     *    EQUALGREATER  shift 1285
     *    LBRACKETAT  reduce 75
     *    WITH  reduce 75
     *)
    _1
  )
# 9696 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run687 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState687

and _menhir_run271 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState271

and _menhir_reduce170 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_pattern_without_or), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_pattern_without_or), _startpos__7_) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__pattern_without_or = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3292 "reason_parser.mly"
      (
         let loc_coloncolon = mklocation _startpos__2_ _endpos__2_ in
         let loc = mklocation _symbolstartpos _endpos in
         mkpat_cons loc_coloncolon (mkpat ~ghost:true ~loc (Ppat_tuple[_5;_7])) loc
      )
# 9794 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run655 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState655

and _menhir_run658 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState658 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState658

and _menhir_run574 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState574

and _menhir_run577 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState577 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7469) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState577 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run581 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState579) : 'freshtv7470)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState577

and _menhir_reduce233 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_expr_comma_seq_extension) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2746 "reason_parser.mly"
      ( let seq, ext_opt = _2 in
        let loc = mklocation _startpos__2_ _endpos__2_ in
        make_real_exp (mktailexp_extension loc seq ext_opt) )
# 10011 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce234 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_expr_comma_seq_extension) _startpos__4_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 10031 "reason_parser.ml"
      
    in
    
# 2750 "reason_parser.mly"
      ( let seq, ext_opt = _4 in
        let loc = mklocation _startpos__4_ _endpos__4_ in
        let list_exp = make_real_exp (mktailexp_extension loc seq ext_opt) in
        let list_exp = { list_exp with pexp_loc = loc } in
        mkexp (Pexp_open (Fresh, _1, list_exp)) )
# 10041 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce417 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_expr_comma_seq_extension) _startpos__3_ ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _endpos__2_, _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3148 "reason_parser.mly"
    ( let seq, ext = _3 in (_1::seq, ext) )
# 10054 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_row_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_row_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7451 * _menhir_state * 'tv_row_field_list) * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7449 * _menhir_state * 'tv_row_field_list) * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_row_field_list)), _startpos__2_), _, (_3 : 'tv_row_field)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_row_field_list = 
# 3982 "reason_parser.mly"
                                                ( _3 :: _1 )
# 10072 "reason_parser.ml"
         in
        _menhir_goto_row_field_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7450)) : 'freshtv7452)
    | MenhirState157 | MenhirState160 | MenhirState2443 | MenhirState162 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7455 * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7453 * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_row_field)) = _menhir_stack in
        let _v : 'tv_row_field_list = 
# 3981 "reason_parser.mly"
                                                ( [_1] )
# 10084 "reason_parser.ml"
         in
        _menhir_goto_row_field_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7454)) : 'freshtv7456)
    | MenhirState305 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7461 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7457 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState312) : 'freshtv7458)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7459 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7460)) : 'freshtv7462)
    | MenhirState161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7467 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7463 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2443) : 'freshtv7464)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7465 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7466)) : 'freshtv7468)
    | _ ->
        _menhir_fail ()

and _menhir_reduce780 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_rebind)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_str_extension_constructors = 
# 3658 "reason_parser.mly"
                                                          ( [_1] )
# 10203 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce782 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_str_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_rebind)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_extension_constructors = 
# 3662 "reason_parser.mly"
      ( _3 :: _1 )
# 10215 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce779 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_declaration)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_str_extension_constructors = 
# 3657 "reason_parser.mly"
                                                          ( [_1] )
# 10226 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce781 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_str_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_declaration)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_extension_constructors = 
# 3660 "reason_parser.mly"
      ( _3 :: _1 )
# 10238 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_sig_extension_constructors : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_sig_extension_constructors -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((((('freshtv7447 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7445 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState1386 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1387) : 'freshtv7446)
    | LBRACKETATAT ->
        _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1386
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1386) : 'freshtv7448)

and _menhir_reduce396 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constructor_declaration_leading_bar) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations_leading_bar = 
# 3552 "reason_parser.mly"
                                                                             ( [_1] )
# 10284 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce397 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_constructor_declaration_leading_bar) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constructor_declarations_leading_bar = 
# 3553 "reason_parser.mly"
                                                                             ( _2 :: _1 )
# 10295 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_let_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2667 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1626 | MenhirState1657 | MenhirState1156 | MenhirState1171 | MenhirState1162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7441) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce503 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7442)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState475 | MenhirState2253 | MenhirState2247 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7443) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce503 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7444)
    | _ ->
        _menhir_fail ()

and _menhir_run2081 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | VIRTUAL ->
        _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2081
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2081

and _menhir_reduce300 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_exception_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_str_exception_declaration) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1421 "reason_parser.mly"
                              (
      mkstr(Pstr_exception _1)
    )
# 10345 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run2069 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7439) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2069 in
        let (_v : (
# 863 "reason_parser.mly"
       (string)
# 10363 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MINUS ->
            _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2070 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2070 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState2070 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2070 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
            _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState2070) : 'freshtv7440)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2069

and _menhir_goto_let_bindings : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_bindings -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1171 | MenhirState1162 | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7425 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1161
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7423 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1161 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7424)
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1161) : 'freshtv7426)
    | MenhirState2667 | MenhirState1657 | MenhirState1626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7429 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1656
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1656 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7427 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1656 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AS ->
                  _menhir_run1651 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1627 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
                  _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1657
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1657) : 'freshtv7428)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1656)) : 'freshtv7430)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7431 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1879
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce295 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7432)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState2253 | MenhirState2247 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7435 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2246
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7433 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2246 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7434)
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2246) : 'freshtv7436)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7437 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2736
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2736 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce295 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7438)
    | _ ->
        _menhir_fail ()

and _menhir_run2148 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce255 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1945 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce255 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce205 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2649 "reason_parser.mly"
      ( _2 )
# 10671 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce78 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__expr = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2555 "reason_parser.mly"
      (
        let loc_colon = mklocation _startpos__2_ _endpos__2_ in
        let loc = mklocation _symbolstartpos _endpos in
        mkexp_cons loc_colon (mkexp ~ghost:true ~loc (Pexp_tuple[_5;_7])) loc
      )
# 10695 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce213 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 10716 "reason_parser.ml"
      
    in
    
# 2671 "reason_parser.mly"
      ( mkexp(Pexp_open(Fresh, _1, _4)) )
# 10722 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1008 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1008

and _menhir_run943 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState943

and _menhir_run958 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState958 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState958

and _menhir_run964 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState964

and _menhir_run966 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState966

and _menhir_run960 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState960

and _menhir_run968 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState968

and _menhir_run970 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState970

and _menhir_run972 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState972

and _menhir_run974 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState974

and _menhir_run980 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState980

and _menhir_run982 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState982

and _menhir_run945 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 842 "reason_parser.mly"
       (string)
# 11653 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState945

and _menhir_run962 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (string)
# 11734 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState962

and _menhir_run976 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (string)
# 11815 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState976

and _menhir_run978 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (string)
# 11896 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState978

and _menhir_run984 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 838 "reason_parser.mly"
       (string)
# 11977 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState984

and _menhir_run986 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState986

and _menhir_run994 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState994

and _menhir_run988 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState988

and _menhir_run990 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState990

and _menhir_run992 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState992

and _menhir_run888 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState888

and _menhir_run830 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState830 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState830

and _menhir_run845 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState845 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState845

and _menhir_run851 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState851

and _menhir_run853 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState853

and _menhir_run847 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState847

and _menhir_run855 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState855

and _menhir_run857 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState857

and _menhir_run859 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState859

and _menhir_run861 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState861

and _menhir_run867 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState867

and _menhir_run869 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState869

and _menhir_run832 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 842 "reason_parser.mly"
       (string)
# 13367 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState832 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState832

and _menhir_run849 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (string)
# 13448 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState849

and _menhir_run863 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (string)
# 13529 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState863

and _menhir_run865 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (string)
# 13610 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState865

and _menhir_run871 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 838 "reason_parser.mly"
       (string)
# 13691 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState871

and _menhir_run873 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState873

and _menhir_run881 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState881

and _menhir_run875 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState875

and _menhir_run877 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState877

and _menhir_run879 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState879

and _menhir_reduce548 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_pattern) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__simple_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 14163 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7421) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7417) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce723 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7418)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1628 | MenhirState1178 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7419) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce723 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7420)
    | _ ->
        _menhir_fail ()) : 'freshtv7422)

and _menhir_reduce201 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _menhir_s (x0 : 'tv_val_longident) _startpos_x0_ ->
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 14204 "reason_parser.ml"
      
    in
    
# 2637 "reason_parser.mly"
      ( mkexp (Pexp_ident _1) )
# 14210 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce71 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__expr = 
# 2529 "reason_parser.mly"
      ( syntax_error_exp (rhs_loc 4) "Invalid try with")
# 14225 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_labeled_simple_pattern : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_labeled_simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1036 | MenhirState897 | MenhirState1022 | MenhirState901 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7381 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1017 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run902 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1022) : 'freshtv7382)
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7383 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1017 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run902 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1036) : 'freshtv7384)
    | MenhirState1222 | MenhirState1183 | MenhirState1194 | MenhirState1187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7385 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1188 _menhir_env (Obj.magic _menhir_stack) MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1194) : 'freshtv7386)
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7387 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1188 _menhir_env (Obj.magic _menhir_stack) MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1222) : 'freshtv7388)
    | MenhirState1633 | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7391 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7389 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1633 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1634) : 'freshtv7390)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1633) : 'freshtv7392)
    | MenhirState1906 | MenhirState1885 | MenhirState1697 | MenhirState1708 | MenhirState1701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7393 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1702 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1698 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1708) : 'freshtv7394)
    | MenhirState1817 | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7399 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7395) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1817 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1784 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1820) : 'freshtv7396)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7397) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1817 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1818 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1818) : 'freshtv7398)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1817) : 'freshtv7400)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7401 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1702 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1698 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1906) : 'freshtv7402)
    | MenhirState2232 | MenhirState480 | MenhirState2218 | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7403 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run2214 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run2212 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2218) : 'freshtv7404)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7405 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run2214 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run2212 _menhir_env (Obj.magic _menhir_stack) MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2232) : 'freshtv7406)
    | MenhirState2673 | MenhirState2671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7409 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7407 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2673 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2675 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2671 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2659 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2674 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2674) : 'freshtv7408)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2673) : 'freshtv7410)
    | MenhirState2698 | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7415 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7411) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2698 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1784 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2701) : 'freshtv7412)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7413) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2698 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2675 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2671 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2659 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2699 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2699) : 'freshtv7414)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2698) : 'freshtv7416)
    | _ ->
        _menhir_fail ()

and _menhir_run742 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7379) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_direction_flag = 
# 4190 "reason_parser.mly"
                                                ( Upto )
# 15060 "reason_parser.ml"
     in
    _menhir_goto_direction_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7380)

and _menhir_run743 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7377) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_direction_flag = 
# 4191 "reason_parser.mly"
                                                ( Downto )
# 15076 "reason_parser.ml"
     in
    _menhir_goto_direction_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7378)

and _menhir_run734 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState734

and _menhir_run736 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7371 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2207 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2207) : 'freshtv7372)
    | LBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7373 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2204) : 'freshtv7374)
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState736 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7375 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState737) : 'freshtv7376)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState736 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState736

and _menhir_run469 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7369) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState469 in
        let (_v : (
# 863 "reason_parser.mly"
       (string)
# 15362 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7370)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState469

and _menhir_run472 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7363 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2271) : 'freshtv7364)
    | LBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7365 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2268) : 'freshtv7366)
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7367 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState473) : 'freshtv7368)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState472

and _menhir_reduce17 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr = 
# 1960 "reason_parser.mly"
      ( mkclass(Pcl_extension _1) )
# 15635 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce37 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_instance_type = 
# 2229 "reason_parser.mly"
      ( mkcty(Pcty_extension _1) )
# 15646 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce266 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_module_type = 
# 1555 "reason_parser.mly"
      ( mkmty(Pmty_extension _1) )
# 15657 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce249 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2806 "reason_parser.mly"
      ( mkexp (Pexp_extension _1) )
# 15668 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce260 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_module_expr = 
# 1271 "reason_parser.mly"
      ( mkmod (Pmod_extension _1) )
# 15679 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce292 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3388 "reason_parser.mly"
      ( mkpat(Ppat_extension _1) )
# 15690 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce153 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3967 "reason_parser.mly"
      ( mktyp(Ptyp_extension _1) )
# 15701 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce14 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1937 "reason_parser.mly"
      ( Cl.attr _1 _2 )
# 15713 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce36 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_instance_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_instance_type = 
# 2227 "reason_parser.mly"
      ( Cty.attr _1 _2 )
# 15725 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce133 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1540 "reason_parser.mly"
      ( Mty.attr _1 _2 )
# 15737 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce110 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2631 "reason_parser.mly"
      ( Exp.attr _1 _2 )
# 15749 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce124 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__module_expr = 
# 1299 "reason_parser.mly"
      ( Mod.attr _1 _2 )
# 15761 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce174 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3309 "reason_parser.mly"
      ( Pat.attr _1 _2 )
# 15773 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce136 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = 
# 3920 "reason_parser.mly"
      ( Typ.attr _1 _2 )
# 15785 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce310 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_floating_attribute -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_floating_attribute) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1464 "reason_parser.mly"
      ( mkstr(Pstr_attribute _1) )
# 15796 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce272 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_signed_constant), _startpos__1_), _endpos__2_), _endpos__3_, _, (_3 : 'tv_signed_constant), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3340 "reason_parser.mly"
      ( mkpat(Ppat_interval (_1, _3)) )
# 15809 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run190 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState190

and _menhir_run2021 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2021

and _menhir_reduce332 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_single_attr_id), _startpos__1_), _endpos__3_, _, (_3 : 'tv_attr_id)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_attr_id = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 4295 "reason_parser.mly"
                               ( mkloc (_1 ^ "." ^ _3.txt) (mklocation _symbolstartpos _endpos) )
# 15917 "reason_parser.ml"
     in
    _menhir_goto_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2364 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2169 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2156 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2153 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2150 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce257 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _startpos__8_ ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)), _, (_7 : 'tv_package_type)) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1258 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_coerce(_3, Some(mktyp ~ghost:true ~loc (Ptyp_package _5)),
                                      mktyp ~ghost:true ~loc (Ptyp_package _7))))) )
# 15965 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1953 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce256 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1253 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_constraint(_3, mktyp ~ghost:true ~loc (Ptyp_package _5))))) )
# 15991 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1950 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce258 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1264 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_coerce(_3, None, mktyp ~ghost:true ~loc (Ptyp_package _5))))) )
# 16017 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1947 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce247 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_), _endpos__6_, _, _startpos__6_), _, (_7 : 'tv_package_type)) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16045 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2798 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_open(Fresh, _1,
        mkexp ~loc (Pexp_constraint (mkexp ~ghost:true ~loc (Pexp_pack _5),
                                     mktyp ~ghost:true ~loc (Ptyp_package _7))))) )
# 16058 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error798 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce245 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2791 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp (Pexp_constraint (mkexp ~ghost:true ~loc (Pexp_pack _3),
                                mktyp ~ghost:true ~loc (Ptyp_package _5))) )
# 16084 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error772 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error500 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 16096 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv7361 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 16103 "reason_parser.ml"
    ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce291 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7362)

and _menhir_error246 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce290 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 16118 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 16124 "reason_parser.ml"
    )), _startpos_x0_), _endpos__4_, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16141 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3381 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkpat(Ppat_constraint(mkpat ~ghost:true ~loc (Ppat_unpack _3),
                              mktyp ~ghost:true ~loc (Ptyp_package _5))) )
# 16152 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error234 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 16159 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv7359 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 16166 "reason_parser.ml"
    ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce291 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7360)

and _menhir_reduce152 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _, (_3 : 'tv_package_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3965 "reason_parser.mly"
      ( mktyp(Ptyp_package _3) )
# 16184 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error130 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce597 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__non_arrowed_simple_core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__non_arrowed_simple_core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_simple_core_type = 
# 3923 "reason_parser.mly"
                                                                               (_1)
# 16200 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7357) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_simple_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7343) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7341) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_non_arrowed_simple_core_type) : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _v : 'tv_row_field = 
# 3986 "reason_parser.mly"
                                                            ( Rinherit _1 )
# 16225 "reason_parser.ml"
         in
        _menhir_goto_row_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7342)) : 'freshtv7344)
    | MenhirState2677 | MenhirState2466 | MenhirState1794 | MenhirState1271 | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7345 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce599 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7346)
    | MenhirState2676 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState120 | MenhirState1793 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1339 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7347) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce598 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7348)
    | MenhirState2692 | MenhirState2663 | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState142 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1808 | MenhirState1803 | MenhirState1801 | MenhirState1799 | MenhirState1760 | MenhirState1775 | MenhirState1773 | MenhirState1768 | MenhirState1766 | MenhirState1764 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState323 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7349) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce594 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7350)
    | MenhirState1637 | MenhirState1587 | MenhirState1494 | MenhirState328 | MenhirState316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7351 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce599 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7352)
    | MenhirState2515 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState1636 | MenhirState1586 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState327 | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7353) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce598 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7354)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7355) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce594 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7356)
    | _ ->
        _menhir_fail ()) : 'freshtv7358)

and _menhir_reduce596 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__non_arrowed_non_simple_core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_non_simple_core_type = 
# 3913 "reason_parser.mly"
                                                                                       (_1)
# 16286 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7339) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2692 | MenhirState2663 | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState142 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1808 | MenhirState1803 | MenhirState1801 | MenhirState1799 | MenhirState1760 | MenhirState1775 | MenhirState1773 | MenhirState1768 | MenhirState1766 | MenhirState1764 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState323 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7335) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce593 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7336)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7337) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce593 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7338)
    | _ ->
        _menhir_fail ()) : 'freshtv7340)

and _menhir_reduce277 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_lbl_pattern_list)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16330 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16342 "reason_parser.ml"
      
    in
    
# 3350 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "{") (with_txt _3 "}") )
# 16348 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__pattern_optional_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_optional_constraint -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7333) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__pattern_optional_constraint) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7331) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__pattern_optional_constraint) : 'tv__pattern_optional_constraint) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_pat__pattern_optional_constraint_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 16371 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7329) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7327) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7325) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_pat__pattern_optional_constraint_) : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let _v : 'tv_pattern_optional_constraint = 
# 3411 "reason_parser.mly"
                                                                             (_1)
# 16388 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7323) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_pattern_optional_constraint) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7311 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7309 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_two_or_more_comma_list)), _endpos__2_), _, (_3 : 'tv_pattern_optional_constraint)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_pattern_two_or_more_comma_list = 
# 3399 "reason_parser.mly"
    ( _3::_1 )
# 16406 "reason_parser.ml"
         in
        _menhir_goto_pattern_two_or_more_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7310)) : 'freshtv7312)
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7317 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7313 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState676) : 'freshtv7314)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7315 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7316)) : 'freshtv7318)
    | MenhirState676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7321 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7319 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_optional_constraint)), _endpos__2_), _, (_3 : 'tv_pattern_optional_constraint)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_pattern_two_or_more_comma_list = 
# 3401 "reason_parser.mly"
    ( [_3; _1] )
# 16490 "reason_parser.ml"
         in
        _menhir_goto_pattern_two_or_more_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7320)) : 'freshtv7322)
    | _ ->
        _menhir_fail ()) : 'freshtv7324)) : 'freshtv7326)) : 'freshtv7328)) : 'freshtv7330)) : 'freshtv7332)) : 'freshtv7334)

and _menhir_reduce545 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__or_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__or_pattern) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__or_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 16505 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7307) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7303) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce641 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7304)
    | MenhirState2572 | MenhirState190 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7305) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce641 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7306)
    | _ ->
        _menhir_fail ()) : 'freshtv7308)

and _menhir_reduce398 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_core_type = 
# 3841 "reason_parser.mly"
                                         (_1)
# 16540 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7301) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7089 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7087 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_core_type_comma_list)), _endpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_core_type_comma_list = 
# 4009 "reason_parser.mly"
                                                ( _3 :: _1 )
# 16560 "reason_parser.ml"
         in
        _menhir_goto_core_type_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7088)) : 'freshtv7090)
    | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState1282 | MenhirState1265 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7093 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7091 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
        let _v : 'tv_core_type_comma_list = 
# 4008 "reason_parser.mly"
                                                ( [_1] )
# 16572 "reason_parser.ml"
         in
        _menhir_goto_core_type_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7092)) : 'freshtv7094)
    | MenhirState666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7095 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce162 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7096)
    | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7103 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7099 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7097 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
          ((let _menhir_env = _menhir_discard _menhir_env in
          _menhir_reduce287 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7098)) : 'freshtv7100)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7101 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce286 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7102)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce162 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7104)
    | MenhirState722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7111 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7107 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7105 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
          ((let _menhir_env = _menhir_discard _menhir_env in
          _menhir_reduce287 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7106)) : 'freshtv7108)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7109 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce286 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7110)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce162 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7112)
    | MenhirState1061 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7113 * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce805 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7114)
    | MenhirState1889 | MenhirState1201 | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7119 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv7115 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7116)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7117 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1066) : 'freshtv7118)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce806 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7120)
    | MenhirState1066 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7121 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce807 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7122)
    | MenhirState1082 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7123 * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce805 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7124)
    | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7127 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1086 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7125 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1087) : 'freshtv7126)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce806 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7128)
    | MenhirState1087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7129 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce807 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7130)
    | MenhirState1206 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7133 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1207 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7131 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1208) : 'freshtv7132)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1207 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7134)
    | MenhirState1214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7137 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1215 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7135 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1216 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1216) : 'freshtv7136)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1215 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7138)
    | MenhirState1227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7141 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1228 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7139 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1229) : 'freshtv7140)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1228 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7142)
    | MenhirState1269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce450 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7144)
    | MenhirState1272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7145 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce449 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7146)
    | MenhirState1281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7147 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce811 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7148)
    | MenhirState1324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7149 * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce176 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7150)
    | MenhirState152 | MenhirState1714 | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7151 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce175 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7152)
    | MenhirState1263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7159 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1337 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7157 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONCOLON ->
                  _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7153 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1338 in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce818 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7154)
              | FALSE ->
                  _menhir_run1286 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7155 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1338 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BAR ->
                      _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONCOLON ->
                      _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run1286 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run1278 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run1339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1340 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1340) : 'freshtv7156)
              | TRUE ->
                  _menhir_run1278 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run1339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState1338
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1338) : 'freshtv7158)
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce810 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1337 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7160)
    | MenhirState1353 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7163 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1354 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7161 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1355 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1355) : 'freshtv7162)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1354 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7164)
    | MenhirState1355 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7165 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce378 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7166)
    | MenhirState1419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7167 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7168)
    | MenhirState1421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1422 : 'freshtv7170)
    | MenhirState1465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7171 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1466
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1466
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1466) : 'freshtv7172)
    | MenhirState1474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7175 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1475 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7173 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1476) : 'freshtv7174)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1475 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7176)
    | MenhirState1492 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7177 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce450 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7178)
    | MenhirState1495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7179 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce449 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7180)
    | MenhirState1539 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7185 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1540 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7183 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7181 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _endpos = _endpos__3_ in
              let _v : 'tv_class_self_type = 
# 2240 "reason_parser.mly"
      ( _2 )
# 17234 "reason_parser.ml"
               in
              _menhir_goto_class_self_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7182)) : 'freshtv7184)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1540 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7186)
    | MenhirState1550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7189 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7187 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_value_type = 
# 2278 "reason_parser.mly"
      ( _3, _2, Virtual, _5 )
# 17253 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7188)) : 'freshtv7190)
    | MenhirState1555 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7193 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7191 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_virtual_flag)), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_value_type = 
# 2280 "reason_parser.mly"
      ( _3, Mutable, _2, _5 )
# 17268 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7192)) : 'freshtv7194)
    | MenhirState1560 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7197 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7195 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_label), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_value_type = 
# 2282 "reason_parser.mly"
      ( _1, Immutable, Concrete, _3 )
# 17282 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7196)) : 'freshtv7198)
    | MenhirState1573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7199 * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce176 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7200)
    | MenhirState1694 | MenhirState1689 | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7201 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce175 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7202)
    | MenhirState1738 | MenhirState1593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7205 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1594 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7203 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1595) : 'freshtv7204)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1594 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7206)
    | MenhirState1595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7215 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7213 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_constrain_field = 
# 2288 "reason_parser.mly"
                                           ( _1, _3 )
# 17354 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7211) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain_field) = _v in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState1593 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1597
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKETATAT ->
                  _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1597
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1597) : 'freshtv7208)
        | MenhirState1738 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7209 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1739
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKETATAT ->
                  _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1739
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1739) : 'freshtv7210)
        | _ ->
            _menhir_fail ()) : 'freshtv7212)) : 'freshtv7214)) : 'freshtv7216)
    | MenhirState1665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7221 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1666 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7219 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7217 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__6_ : Lexing.position) = _endpos in
              ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
              let _6 = () in
              let _4 = () in
              let _1 = () in
              let _endpos = _endpos__6_ in
              let _v : 'tv_value = 
# 2024 "reason_parser.mly"
      ( not_expecting 6 "not expecting equal - cannot specify value for virtual val" )
# 17413 "reason_parser.ml"
               in
              _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7218)) : 'freshtv7220)
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce837 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1666 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7222)
    | MenhirState1674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7227 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1675 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7225 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7223 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__7_ : Lexing.position) = _endpos in
              ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_label), _startpos__4_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_) = _menhir_stack in
              let _7 = () in
              let _5 = () in
              let _3 = () in
              let _2 = () in
              let _endpos = _endpos__7_ in
              let _v : 'tv_value = 
# 2020 "reason_parser.mly"
      ( not_expecting 7 "not expecting equal - cannot specify value for virtual val" )
# 17447 "reason_parser.ml"
               in
              _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7224)) : 'freshtv7226)
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce835 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1675 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7228)
    | MenhirState1717 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv7231 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1718 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((((('freshtv7229 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1719) : 'freshtv7230)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1718 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7232)
    | MenhirState1853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7235 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1854 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7233 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1855) : 'freshtv7234)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1854 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7236)
    | MenhirState1855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7237 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce378 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7238)
    | MenhirState1892 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7241 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1893 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7239 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1894) : 'freshtv7240)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1893 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7242)
    | MenhirState1898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7245 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1899 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7243 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1900) : 'freshtv7244)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1899 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7246)
    | MenhirState1911 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7249 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1912 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7247 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1913) : 'freshtv7248)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1912 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7250)
    | MenhirState1969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7251 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7252)
    | MenhirState1971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7253 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1972 : 'freshtv7254)
    | MenhirState2021 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7257 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7255 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_payload = 
# 4346 "reason_parser.mly"
                    ( PTyp _2 )
# 17888 "reason_parser.ml"
         in
        _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7256)) : 'freshtv7258)
    | MenhirState2035 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7261 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2036 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7259 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run2038 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2037) : 'freshtv7260)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2036 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7262)
    | MenhirState241 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce646 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7264)
    | MenhirState2433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7267 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7265 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
        let _endpos = _endpos__1_ in
        let _v : 'tv_amper_type_list = 
# 3999 "reason_parser.mly"
                                                ( [_1] )
# 17931 "reason_parser.ml"
         in
        _menhir_goto_amper_type_list _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7266)) : 'freshtv7268)
    | MenhirState2436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7271 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7269 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_amper_type_list)), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_amper_type_list = 
# 4000 "reason_parser.mly"
                                                ( _3 :: _1 )
# 17945 "reason_parser.ml"
         in
        _menhir_goto_amper_type_list _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7270)) : 'freshtv7272)
    | MenhirState140 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce646 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7274)
    | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7275 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce811 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7276)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7283 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2513 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7281 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONCOLON ->
                  _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7277 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2514 in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce818 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7278)
              | FALSE ->
                  _menhir_run2477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7279 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2514 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BAR ->
                      _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2516 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONCOLON ->
                      _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run2477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run2517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run2469 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run2515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2516) : 'freshtv7280)
              | TRUE ->
                  _menhir_run2469 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run2515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState2514
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2514) : 'freshtv7282)
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce810 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2513 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7284)
    | MenhirState2564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7295 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2566 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7293 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7291 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 18049 "reason_parser.ml"
              ) = 
# 1142 "reason_parser.mly"
                  ( apply_mapper_chain_to_type _1 default_mapper_chain )
# 18053 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7289) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 18061 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7287) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 18069 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7285) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 18077 "reason_parser.ml"
              )) : (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 18081 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv7286)) : 'freshtv7288)) : 'freshtv7290)) : 'freshtv7292)) : 'freshtv7294)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2566 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7296)
    | MenhirState2643 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7299 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2644 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7297 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run2038 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2645 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2645) : 'freshtv7298)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2644 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7300)
    | _ ->
        _menhir_fail ()) : 'freshtv7302)

and _menhir_reduce726 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__simple_pattern_not_ident_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__simple_pattern_not_ident_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern_not_ident = 
# 3333 "reason_parser.mly"
                                                                       (_1)
# 18124 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7085) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_pattern_not_ident) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState564 | MenhirState563 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7073 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce269 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7074)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState738 | MenhirState486 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState650 | MenhirState649 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7075 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce269 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7076)
    | MenhirState1628 | MenhirState1178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7079 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1226 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7077 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1227) : 'freshtv7078)
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce269 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1226 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7080)
    | MenhirState2308 | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7083 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1910 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7081 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1911) : 'freshtv7082)
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce269 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1910 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7084)
    | _ ->
        _menhir_fail ()) : 'freshtv7086)

and _menhir_reduce716 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_labeled_simple_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_labeled_expr_list = 
# 2834 "reason_parser.mly"
      ( [_1] )
# 18257 "reason_parser.ml"
     in
    _menhir_goto_simple_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce717 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_labeled_simple_expr) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_labeled_expr_list = 
# 2836 "reason_parser.mly"
      ( _2 :: _1 )
# 18268 "reason_parser.ml"
     in
    _menhir_goto_simple_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce467 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_label_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_labeled_simple_expr = 
# 2843 "reason_parser.mly"
      ( _1 )
# 18278 "reason_parser.ml"
     in
    _menhir_goto_labeled_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce72 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_simple_non_labeled_expr_list_as_tuple -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_simple_non_labeled_expr_list_as_tuple) ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 18296 "reason_parser.ml"
      
    in
    
# 2531 "reason_parser.mly"
    (
      if List.mem (string_of_longident _1.txt)
         built_in_explicit_arity_constructors then
        (* unboxing the inner tupple *)
        match _2 with
          | {pexp_desc=Pexp_tuple [inner]; pexp_loc; pexp_attributes} -> mkexp (Pexp_construct(_1, Some inner))
          | _ -> assert false
      else
        mkExplicitArityTupleExp (Pexp_construct(_1, Some _2))
    )
# 18311 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1079 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_seq -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1079 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1079

and _menhir_goto_expr_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7059 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1069
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1071 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1069
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7057 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1069 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce214 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7058)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1069)) : 'freshtv7060)
    | MenhirState15 | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7063 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1139
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7061 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1139 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce210 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7062)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1139)) : 'freshtv7064)
    | MenhirState750 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7067 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2171
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1071 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2171
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7065 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2171 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce214 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7066)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2171)) : 'freshtv7068)
    | MenhirState183 | MenhirState407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7071 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2373
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2373
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7069 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2373 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce210 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7070)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2373)) : 'freshtv7072)
    | _ ->
        _menhir_fail ()

and _menhir_reduce804 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_type_constraint_right_of_colon -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_type_constraint_right_of_colon) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_constraint = 
# 3246 "reason_parser.mly"
                                                ( _2 )
# 18497 "reason_parser.ml"
     in
    _menhir_goto_type_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2325 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7055 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce223 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7056)

and _menhir_reduce222 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_record_expr)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2716 "reason_parser.mly"
      ( let (exten, fields) = _2 in mkexp (Pexp_record(fields, exten)) )
# 18520 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2238 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7053 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce223 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7054)

and _menhir_error2198 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7051 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce227 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7052)

and _menhir_reduce226 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_record_expr)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 18559 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2725 "reason_parser.mly"
      ( let (exten, fields) = _4 in
        let loc = mklocation _symbolstartpos _endpos in
        let rec_exp = mkexp ~loc (Pexp_record (fields, exten)) in
        mkexp(Pexp_open(Fresh, _1, rec_exp)) )
# 18571 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1120 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7049 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce227 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7050)

and _menhir_run2659 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2659 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2659 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2659

and _menhir_run2667 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1651 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1627 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2667 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState2667
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2667

and _menhir_run2671 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2671

and _menhir_run2675 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2675 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2675 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2675

and _menhir_run1783 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 18713 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce364 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1784 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1784 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | SEMI | VAL ->
        _menhir_reduce354 _menhir_env (Obj.magic _menhir_stack) MenhirState1784
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1784

and _menhir_run1622 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run201 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run200 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run199 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_reduce826 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7047 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7048)

and _menhir_run1623 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run207 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run206 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run205 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_reduce827 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7045 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7046)

and _menhir_run1782 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1784 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1782 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1782 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1782 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1782 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1782

and _menhir_run1798 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 18808 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7041 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 18820 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1808) : 'freshtv7042)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7043 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 18863 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1799) : 'freshtv7044)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce825 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_type_variance = 
# 3541 "reason_parser.mly"
                                                ( Invariant )
# 18906 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce826 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run1535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce827 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run1212 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_typevar_list * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _startpos ->
    let _menhir_stack = (_menhir_stack, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1212

and _menhir_goto_mty_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mty_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2385 | MenhirState2383 | MenhirState2362 | MenhirState2356 | MenhirState2354 | MenhirState2167 | MenhirState2152 | MenhirState2155 | MenhirState2149 | MenhirState1949 | MenhirState1952 | MenhirState1946 | MenhirState796 | MenhirState770 | MenhirState245 | MenhirState128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7031 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error132 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7029 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run134 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState133) : 'freshtv7030)
          | COLONGREATER | RPAREN ->
              _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error132 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7032)
    | MenhirState499 | MenhirState232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7035 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error237 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7033 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run239 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState238) : 'freshtv7034)
          | RPAREN ->
              _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error237 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7036)
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2119 | MenhirState2116 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1438 | MenhirState1436 | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7037 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce263 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7038)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1999 | MenhirState1959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7039 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce263 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7040)
    | _ ->
        _menhir_fail ()

and _menhir_goto_name_tag_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_name_tag_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7023 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (((('freshtv7021 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState295 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce151 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7022)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState295) : 'freshtv7024)
    | MenhirState2450 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7027 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (((('freshtv7025 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2451 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce151 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7026)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2451) : 'freshtv7028)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_ampersand : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_ampersand -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7019 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2433) : 'freshtv7020)

and _menhir_goto__optional_type_variable_with_variance : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__optional_type_variable_with_variance -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2070 | MenhirState1967 | MenhirState1415 | MenhirState1361 | MenhirState1249 | MenhirState1358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7015) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__optional_type_variable_with_variance) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7016)
    | MenhirState90 | MenhirState1849 | MenhirState1858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7017) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__optional_type_variable_with_variance) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7018)
    | _ ->
        _menhir_fail ()

and _menhir_reduce348 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cty__class_instance_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cty__class_instance_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_instance_type = 
# 2214 "reason_parser.mly"
                                                             (_1)
# 19123 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7013) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_instance_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1588
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1588
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1588
          | RBRACE | SEMI ->
              _menhir_reduce40 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1588)) : 'freshtv7000)
    | MenhirState1536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7001 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19156 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1609
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1609
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce361 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1609)) : 'freshtv7002)
    | MenhirState1761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1762
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1762
          | RPAREN ->
              _menhir_reduce7 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1762)) : 'freshtv7004)
    | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7005 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1795
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1795
          | AND | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce7 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1795)) : 'freshtv7006)
    | MenhirState2701 | MenhirState1820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7009 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1828
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1828
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7007 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_instance_type), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__non_arrowed_class_constructor_type = 
# 2210 "reason_parser.mly"
      ( _1 )
# 19222 "reason_parser.ml"
               in
              _menhir_goto__non_arrowed_class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7008)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1828)) : 'freshtv7010)
    | MenhirState2050 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7011 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19234 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2051
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2051
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce361 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2051)) : 'freshtv7012)
    | _ ->
        _menhir_fail ()) : 'freshtv7014)

and _menhir_reduce341 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_expr = 
# 1921 "reason_parser.mly"
                                          (_1)
# 19259 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6997) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6953 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1644
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1644
          | AND | AS | COLON | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce30 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1644)) : 'freshtv6954)
    | MenhirState1732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6963 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1733
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6961) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1733 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6957 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 19310 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6955 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos__2_ : Lexing.position) = _endpos in
                  let ((_2 : (
# 863 "reason_parser.mly"
       (string)
# 19320 "reason_parser.ml"
                  )) : (
# 863 "reason_parser.mly"
       (string)
# 19324 "reason_parser.ml"
                  )) = _v in
                  let (_startpos__2_ : Lexing.position) = _startpos in
                  ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
                  let _1 = () in
                  let _endpos = _endpos__2_ in
                  let _v : 'tv_parent_binder = 
# 2008 "reason_parser.mly"
          ( Some _2 )
# 19333 "reason_parser.ml"
                   in
                  _menhir_goto_parent_binder _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6956)) : 'freshtv6958)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6959 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6960)) : 'freshtv6962)
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1733
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce650 _menhir_env (Obj.magic _menhir_stack) MenhirState1733
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1733)) : 'freshtv6964)
    | MenhirState2667 | MenhirState1626 | MenhirState1657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6965 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1751
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1751
          | RBRACE ->
              _menhir_reduce19 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1751)) : 'freshtv6966)
    | MenhirState1625 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1757
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6967 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1757 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1763 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1761 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1760) : 'freshtv6968)
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1757
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1757 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce52 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6970)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1757)) : 'freshtv6972)
    | MenhirState1624 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6973 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19436 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1780
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1780
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce339 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1780)) : 'freshtv6974)
    | MenhirState1813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1814
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1814
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce54 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1814)) : 'freshtv6976)
    | MenhirState1818 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6977 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1819
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1819
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce346 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1819)) : 'freshtv6978)
    | MenhirState1825 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6979 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1826
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1826
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce347 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1826)) : 'freshtv6980)
    | MenhirState2659 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2660
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2660 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1763 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1761 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2663) : 'freshtv6982)
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2660
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2660 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce52 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6984)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2660)) : 'freshtv6986)
    | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6987 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2684
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2684
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce30 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2684)) : 'freshtv6988)
    | MenhirState2658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6989 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19584 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2691
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2691
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce339 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2691)) : 'freshtv6990)
    | MenhirState2694 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6991 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2695
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2695
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce54 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2695)) : 'freshtv6992)
    | MenhirState2699 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6993 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2700
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2700
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce346 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2700)) : 'freshtv6994)
    | MenhirState2703 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6995 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2704
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2704
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce347 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2704)) : 'freshtv6996)
    | _ ->
        _menhir_fail ()) : 'freshtv6998)

and _menhir_reduce345 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_fun_def_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_fun_def_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_fun_def = 
# 1885 "reason_parser.mly"
                                               (_1)
# 19656 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6951) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_fun_def) = _v in
    ((match _menhir_s with
    | MenhirState1633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6943 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce31 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6944)
    | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6945 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce12 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6946)
    | MenhirState2673 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6947 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce31 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6948)
    | MenhirState2671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6949 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce12 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6950)
    | _ ->
        _menhir_fail ()) : 'freshtv6952)

and _menhir_reduce225 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_self_pattern_and_structure), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19711 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19723 "reason_parser.ml"
      
    in
    
# 2723 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{") (with_txt _3 "}") )
# 19729 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce49 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19749 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19761 "reason_parser.ml"
      
    in
    
# 1969 "reason_parser.mly"
    ( unclosed_cl (with_txt _1 "{") (with_txt _3 "}") )
# 19767 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce336 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cty__class_constructor_type_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cty__class_constructor_type_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_constructor_type = 
# 2195 "reason_parser.mly"
                                                                   (_1)
# 19777 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6941) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_constructor_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6897 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce10 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6898)
    | MenhirState1766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6899 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 19795 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce8 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6900)
    | MenhirState1775 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6901 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 19803 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce9 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6902)
    | MenhirState1760 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6905 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1777 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6903 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce50 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6904)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1777 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6906)
    | MenhirState1803 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6907 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce10 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6908)
    | MenhirState1801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6909 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 19834 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce8 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6910)
    | MenhirState1810 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6911 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 19842 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce9 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6912)
    | MenhirState1781 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6915 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1812 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6913 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CLASS ->
                  _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1813) : 'freshtv6914)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1812 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6916)
    | MenhirState1821 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6921 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1822 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6919 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6917 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_constructor_type)) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__3_ in
              let _v : 'tv__non_arrowed_class_constructor_type = 
# 2211 "reason_parser.mly"
                                         ( _2)
# 19909 "reason_parser.ml"
               in
              _menhir_goto__non_arrowed_class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6918)) : 'freshtv6920)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1822 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6922)
    | MenhirState2057 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6931 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19921 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6929 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 19927 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 19932 "reason_parser.ml"
        )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
        let _4 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_class_description_details = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19945 "reason_parser.ml"
          
        in
        
# 2308 "reason_parser.mly"
                                                                                   (
    (_2, _5, _1, List.rev _3)
  )
# 19953 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6927) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_description_details) = _v in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState2045 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6923 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKETATAT ->
                _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2059
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2059) : 'freshtv6924)
        | MenhirState2086 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6925 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKETATAT ->
                _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2087
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2087) : 'freshtv6926)
        | _ ->
            _menhir_fail ()) : 'freshtv6928)) : 'freshtv6930)) : 'freshtv6932)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6935 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2664 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce50 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6934)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2664 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6936)
    | MenhirState2692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2693 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6937 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CLASS ->
                  _menhir_run2675 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run2671 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run2667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2659 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2694 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2694) : 'freshtv6938)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2693 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6940)
    | _ ->
        _menhir_fail ()) : 'freshtv6942)

and _menhir_goto_private_virtual_flags : _menhir_env -> 'ttv_tail -> 'tv_private_virtual_flags -> 'ttv_return =
  fun _menhir_env _menhir_stack _v ->
    let _menhir_stack = (_menhir_stack, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv6895 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1567 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1567) : 'freshtv6896)

and _menhir_run1537 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 20062 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce364 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1538 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1538 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | SEMI | VAL ->
        _menhir_reduce354 _menhir_env (Obj.magic _menhir_stack) MenhirState1538
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1538

and _menhir_reduce338 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 20086 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters -> Lexing.position -> _menhir_state -> 'tv_constrained_class_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_constrained_class_declaration) ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 20092 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 20104 "reason_parser.ml"
      
    in
    
# 1823 "reason_parser.mly"
    (
       (_2, _4, _1, List.rev _3)
    )
# 20112 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto__class_fun_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1621 | MenhirState1817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6891) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce522 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6892)
    | MenhirState2657 | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6893) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce522 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6894)
    | _ ->
        _menhir_fail ()

and _menhir_goto_with_type_binder : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_with_type_binder -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6887 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1421) : 'freshtv6888)
    | MenhirState1968 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6889 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1971) : 'freshtv6890)
    | _ ->
        _menhir_fail ()

and _menhir_run1264 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 20220 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6885 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 20250 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1264 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1265) : 'freshtv6886)
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1264
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1264

and _menhir_run1278 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1278
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1278

and _menhir_run1282 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1282

and _menhir_run1286 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1286
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1286

and _menhir_run1289 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1289
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1289

and _menhir_run1292 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6869 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1306
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1306) : 'freshtv6870)
    | FALSE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6871 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1303
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1303) : 'freshtv6872)
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6877 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6873 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
                _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1300
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1300) : 'freshtv6874)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6875 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6876)) : 'freshtv6878)
    | TRUE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6879 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1296
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1296) : 'freshtv6880)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6881 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 20673 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1293
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1293) : 'freshtv6882)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6883 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6884)

and _menhir_run108 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 20725 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6867 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 20755 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState108 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState117) : 'freshtv6868)
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState108
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState108

and _menhir_run2469 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2469
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2469

and _menhir_run2473 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2473

and _menhir_run2477 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2477
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2477

and _menhir_reduce814 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3498 "reason_parser.mly"
      ( (Ptype_open, Public, None) )
# 20945 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run2480 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2480
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2480

and _menhir_run2483 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6851 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2497
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2497) : 'freshtv6852)
    | FALSE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6853 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2494 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2494
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2494) : 'freshtv6854)
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6859 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6855 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
                _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2491
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2491) : 'freshtv6856)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6857 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6858)) : 'freshtv6860)
    | TRUE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6861 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2487
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2487) : 'freshtv6862)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6863 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 21191 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2484
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2484) : 'freshtv6864)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6865 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6866)

and _menhir_reduce791 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_str__structure_item_without_item_extension_sugar_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_structure_item_without_item_extension_sugar = 
# 1403 "reason_parser.mly"
                                                                  (_1)
# 21247 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6849) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1933 | MenhirState1840 | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6841) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce294 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6842)
    | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6843 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce293 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6844)
    | MenhirState2747 | MenhirState2754 | MenhirState2751 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6845) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce294 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6846)
    | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6847 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce293 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6848)
    | _ ->
        _menhir_fail ()) : 'freshtv6850)

and _menhir_reduce790 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_str__structure_item_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_str__structure_item_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_structure_item = 
# 1390 "reason_parser.mly"
                                                   (_1)
# 21298 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6839) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_structure_item) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1933 | MenhirState1840 | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6823 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1839 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6819 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETATATAT ->
                  _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENTPERCENT ->
                  _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EOF | RBRACE | RBRACKET ->
                  _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState1933
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1933) : 'freshtv6820)
          | EOF | RBRACE | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6821 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _v : 'tv_structure = 
# 1379 "reason_parser.mly"
                   ( [_1] )
# 21425 "reason_parser.ml"
               in
              _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6822)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1839 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6824)
    | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6829 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2717 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6827 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6825 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__2_ : Lexing.position) = _endpos in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : 'tv__toplevel_phrase = 
# 1127 "reason_parser.mly"
                                        ( Ptop_def [_1])
# 21452 "reason_parser.ml"
               in
              _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6826)) : 'freshtv6828)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2717 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6830)
    | MenhirState2747 | MenhirState2754 | MenhirState2751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6837 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2753 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6833 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6831 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : 'tv__use_file = 
# 1137 "reason_parser.mly"
                                                   ( [Ptop_def[_1]] )
# 21477 "reason_parser.ml"
               in
              _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6832)) : 'freshtv6834)
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6835 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EOF ->
                  _menhir_run2748 _menhir_env (Obj.magic _menhir_stack) MenhirState2754
              | EXCEPTION ->
                  _menhir_run2648 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run2641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run2638 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETATATAT ->
                  _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENTPERCENT ->
                  _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run2629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run2622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run2613 _menhir_env (Obj.magic _menhir_stack) MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run2577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2754) : 'freshtv6836)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2753 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6838)
    | _ ->
        _menhir_fail ()) : 'freshtv6840)

and _menhir_reduce405 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__curried_binding_return_typed_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_curried_binding_return_typed = 
# 3004 "reason_parser.mly"
                                                                               (_1)
# 21597 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6817) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_curried_binding_return_typed) = _v in
    ((match _menhir_s with
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6813 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed) = _v in
        (_menhir_reduce497 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6814)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6815 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed) = _v in
        (_menhir_reduce497 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6816)
    | _ ->
        _menhir_fail ()) : 'freshtv6818)

and _menhir_goto_clty_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_clty_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1761 | MenhirState1536 | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6809 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1586
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce32 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1586)) : 'freshtv6810)
    | MenhirState2701 | MenhirState2050 | MenhirState1820 | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6811 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1793
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce32 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1793)) : 'freshtv6812)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_ext2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_ext2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2701 | MenhirState2692 | MenhirState2676 | MenhirState2677 | MenhirState2663 | MenhirState2643 | MenhirState2625 | MenhirState2578 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2564 | MenhirState76 | MenhirState2515 | MenhirState106 | MenhirState2472 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2473 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState117 | MenhirState120 | MenhirState2466 | MenhirState127 | MenhirState128 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState169 | MenhirState2436 | MenhirState2433 | MenhirState232 | MenhirState241 | MenhirState323 | MenhirState2400 | MenhirState326 | MenhirState400 | MenhirState2385 | MenhirState2383 | MenhirState2362 | MenhirState2356 | MenhirState2354 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2214 | MenhirState2167 | MenhirState2162 | MenhirState2152 | MenhirState2155 | MenhirState2149 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState1243 | MenhirState2107 | MenhirState2110 | MenhirState1245 | MenhirState1401 | MenhirState2057 | MenhirState2050 | MenhirState2035 | MenhirState2030 | MenhirState2021 | MenhirState1959 | MenhirState1999 | MenhirState1988 | MenhirState1984 | MenhirState1978 | MenhirState1971 | MenhirState1969 | MenhirState1949 | MenhirState1952 | MenhirState1946 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1855 | MenhirState1853 | MenhirState1820 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1793 | MenhirState1794 | MenhirState1782 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1761 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1636 | MenhirState1637 | MenhirState1536 | MenhirState1595 | MenhirState1593 | MenhirState1586 | MenhirState1587 | MenhirState1580 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1539 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState1494 | MenhirState1495 | MenhirState1492 | MenhirState1474 | MenhirState1465 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 | MenhirState1429 | MenhirState1426 | MenhirState1421 | MenhirState1419 | MenhirState1248 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1355 | MenhirState1353 | MenhirState1339 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1282 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState1271 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState796 | MenhirState770 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState666 | MenhirState499 | MenhirState327 | MenhirState328 | MenhirState243 | MenhirState316 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState271 | MenhirState269 | MenhirState244 | MenhirState253 | MenhirState254 | MenhirState245 | MenhirState211 | MenhirState192 | MenhirState78 | MenhirState87 | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6803 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error86 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6801 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState87 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState87) : 'freshtv6802)
          | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error86 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6804)
    | MenhirState1992 | MenhirState1976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6807 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1987 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6805 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1988) : 'freshtv6806)
          | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1987 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6808)
    | _ ->
        _menhir_fail ()

and _menhir_goto_with_constraints : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1413 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6795 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1431 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6793 * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | MODULE ->
                  _menhir_run1424 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run1414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1432) : 'freshtv6794)
          | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1431 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6796)
    | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6799 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1994 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6797 * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | MODULE ->
                  _menhir_run1974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run1966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1995) : 'freshtv6798)
          | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1994 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6800)
    | _ ->
        _menhir_fail ()

and _menhir_reduce584 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_type = 
# 1571 "reason_parser.mly"
                                             (_1)
# 21850 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6791) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 21865 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1412
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState1412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1412
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 21881 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1412 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLON ->
                  _menhir_run1436 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1435 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1435 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1435) : 'freshtv6710)
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1412)) : 'freshtv6712)
    | MenhirState1436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6715 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1437
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState1437 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1437
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1437 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6713 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__2_ in
              let _v : 'tv__module_declaration = 
# 1751 "reason_parser.mly"
      ( _2 )
# 21928 "reason_parser.ml"
               in
              _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6714)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1437)) : 'freshtv6716)
    | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6717 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1439
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState1439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1439
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce128 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1439)) : 'freshtv6718)
    | MenhirState1460 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6727 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 21960 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1461
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState1461 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1461
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1461 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | LBRACKETATAT | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6725 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 21978 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 21983 "reason_parser.ml"
              )), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_type), _startpos__3_) = _menhir_stack in
              let _2 = () in
              let _endpos = _endpos__3_ in
              let _v : 'tv_module_rec_declaration_details = let _1 =
                let _endpos_x_ = _endpos_x0_ in
                let _startpos_x_ = _startpos_x0_ in
                let x = x0 in
                let _endpos = _endpos_x_ in
                let _symbolstartpos = _startpos_x_ in
                
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 21996 "reason_parser.ml"
                
              in
              
# 1759 "reason_parser.mly"
                                     ( (_1, _3) )
# 22002 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6723) = _menhir_stack in
              let (_endpos : Lexing.position) = _endpos in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : 'tv_module_rec_declaration_details) = _v in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
              match _menhir_s with
              | MenhirState1458 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv6719 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKETATAT ->
                      _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1462
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1462) : 'freshtv6720)
              | MenhirState2076 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6721 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKETATAT ->
                      _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2077
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2077) : 'freshtv6722)
              | _ ->
                  _menhir_fail ()) : 'freshtv6724)) : 'freshtv6726)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1461)) : 'freshtv6728)
    | MenhirState1959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6731 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1963
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState1963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1963
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6729 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1963 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce252 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6730)
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1963)) : 'freshtv6732)
    | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6733 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2000
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2000
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce128 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2000)) : 'freshtv6734)
    | MenhirState2030 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6735 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2031
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2031
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2031
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2031) : 'freshtv6736)
    | MenhirState1401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6737 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2097
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2097 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2097
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2097
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2097 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2097) : 'freshtv6738)
    | MenhirState1245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2103
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2103
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6739 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2103 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce261 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6740)
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2103)) : 'freshtv6742)
    | MenhirState2107 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2108
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2108
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6747 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2108 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACE ->
                      _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2110) : 'freshtv6744)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6746)) : 'freshtv6748)
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2108)) : 'freshtv6750)
    | MenhirState2110 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2111
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2111 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2111
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2111 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce127 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2111)) : 'freshtv6752)
    | MenhirState1243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2112
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2112 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2112
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6753 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2112 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce261 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6754)
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2112 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2112)) : 'freshtv6756)
    | MenhirState2116 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6763 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2117
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2117
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6761 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2117 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6757 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACE ->
                      _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2119) : 'freshtv6758)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6760)) : 'freshtv6762)
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2117)) : 'freshtv6764)
    | MenhirState2119 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2120
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2120
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce127 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2120)) : 'freshtv6766)
    | MenhirState1242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6769 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2121
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6767 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2121 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2122) : 'freshtv6768)
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2121
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2121)) : 'freshtv6770)
    | MenhirState2298 | MenhirState2134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6771 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2138
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2138
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2138)) : 'freshtv6772)
    | MenhirState2162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6775 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2163
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2163
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6773 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2163 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce252 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6774)
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2163)) : 'freshtv6776)
    | MenhirState448 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6781 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2283
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2283
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2283 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6777 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__5_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__5_ : Lexing.position) = _startpos in
              ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_functor_arg_name), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
              let _5 = () in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_functor_arg = let _2 =
                let _endpos_x_ = _endpos_x0_ in
                let _startpos_x_ = _startpos_x0_ in
                let x = x0 in
                let _endpos = _endpos_x_ in
                let _symbolstartpos = _startpos_x_ in
                
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22452 "reason_parser.ml"
                
              in
              
# 1223 "reason_parser.mly"
      ( _2, Some _4 )
# 22458 "reason_parser.ml"
               in
              _menhir_goto_functor_arg _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv6778)) : 'freshtv6780)
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2283 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2283)) : 'freshtv6782)
    | MenhirState2287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6785 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2288
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6783 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2288 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2289 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2289) : 'freshtv6784)
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2288
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2288)) : 'freshtv6786)
    | MenhirState400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6787 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2390
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1999 _menhir_env (Obj.magic _menhir_stack) MenhirState2390 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2390
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2390
          | WITH ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2390 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2390) : 'freshtv6788)
    | MenhirState2625 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2626
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1438 _menhir_env (Obj.magic _menhir_stack) MenhirState2626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2626
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2626
          | WITH ->
              _menhir_run1413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2626) : 'freshtv6790)
    | _ ->
        _menhir_fail ()) : 'freshtv6792)

and _menhir_reduce719 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__simple_module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__simple_module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_module_type = 
# 1542 "reason_parser.mly"
                                                           (_1)
# 22554 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6707) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2119 | MenhirState2116 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1438 | MenhirState1436 | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6703) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce131 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6704)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1999 | MenhirState1959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6705) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce131 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6706)
    | _ ->
        _menhir_fail ()) : 'freshtv6708)

and _menhir_reduce718 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__simple_module_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__simple_module_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_module_expr = 
# 1238 "reason_parser.mly"
                                                           (_1)
# 22589 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6701) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_module_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2638 | MenhirState410 | MenhirState450 | MenhirState751 | MenhirState2136 | MenhirState2132 | MenhirState2122 | MenhirState1240 | MenhirState794 | MenhirState767 | MenhirState754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6693) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce116 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6694)
    | MenhirState2639 | MenhirState2360 | MenhirState2282 | MenhirState2166 | MenhirState756 | MenhirState2137 | MenhirState2133 | MenhirState2123 | MenhirState1241 | MenhirState795 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6695 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce118 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6696)
    | MenhirState402 | MenhirState403 | MenhirState411 | MenhirState2300 | MenhirState2296 | MenhirState2289 | MenhirState2285 | MenhirState757 | MenhirState1470 | MenhirState1942 | MenhirState1939 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6697) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce116 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6698)
    | MenhirState2389 | MenhirState2387 | MenhirState2358 | MenhirState2301 | MenhirState2297 | MenhirState2290 | MenhirState2286 | MenhirState2159 | MenhirState2019 | MenhirState1941 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6699 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce118 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6700)
    | _ ->
        _menhir_fail ()) : 'freshtv6702)

and _menhir_goto_leading_bar_match_cases : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_cases -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1033
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce68 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6690)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2229
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce68 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6692)
    | _ ->
        _menhir_fail ()

and _menhir_reduce35 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_class_sig_body)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_instance_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22680 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22692 "reason_parser.ml"
      
    in
    
# 2223 "reason_parser.mly"
      ( unclosed_cty (with_txt _1 "{") (with_txt _3 "}") )
# 22698 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__semi_terminated_seq_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1156 | MenhirState1162 | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6685) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce532 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6686)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState475 | MenhirState2247 | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6687) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce532 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6688)
    | _ ->
        _menhir_fail ()

and _menhir_goto__semi_terminated_seq_expr_row : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1156 | MenhirState1162 | MenhirState1166 | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6681) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce533 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6682)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6683) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce533 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6684)
    | _ ->
        _menhir_fail ()

and _menhir_reduce279 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list_extension), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22762 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22774 "reason_parser.ml"
      
    in
    
# 3357 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "[") (with_txt _4 "]") )
# 22780 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce282 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22800 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22812 "reason_parser.ml"
      
    in
    
# 3364 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "[|") (with_txt _4 "|]") )
# 22818 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_module_binding_body : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 22831 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2126
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2126
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2126) : 'freshtv6674)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6675 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 22847 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce601 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6676)
    | MenhirState2317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 22855 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2318
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2318
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2318) : 'freshtv6678)
    | MenhirState2631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6679 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 22871 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce601 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6680)
    | _ ->
        _menhir_fail ()

and _menhir_reduce580 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__module_binding_body_functor_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__module_binding_body_functor_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body_functor = 
# 1477 "reason_parser.mly"
                                                                             (_1)
# 22883 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6671) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_binding_body_functor) = _v in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2317 | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6667) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_functor) = _v in
        (_menhir_reduce577 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6668)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6669) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_functor) = _v in
        (_menhir_reduce577 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6670)
    | _ ->
        _menhir_fail ()) : 'freshtv6672)

and _menhir_run425 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_field_expr_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState425 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATERRBRACE ->
        _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState425

and _menhir_reduce582 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__module_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__module_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_expr = 
# 1274 "reason_parser.mly"
                                             (_1)
# 22930 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6665) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6609 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState756
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState756
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | COLON | EOF | EQUAL | EQUALGREATER | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce117 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState756)) : 'freshtv6610)
    | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6615 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState768
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6611 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState768 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState770) : 'freshtv6612)
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState768
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6613 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState768 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce244 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6614)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState768 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState768)) : 'freshtv6616)
    | MenhirState794 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6619 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState795
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6617 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState795 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState796 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState796 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState796) : 'freshtv6618)
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState795 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState795
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState795 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState795 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState795 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState795)) : 'freshtv6620)
    | MenhirState1240 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6621 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1241
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1241
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1241 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce578 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1241)) : 'freshtv6622)
    | MenhirState1939 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6623 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1941
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1941
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | COLON | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce117 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1941)) : 'freshtv6624)
    | MenhirState1942 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6625 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1956
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1959 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1956 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1956 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1956
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1956 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1956 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1956 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1956 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1956)) : 'freshtv6626)
    | MenhirState1470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2019
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2019
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2019
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2019) : 'freshtv6628)
    | MenhirState2122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6629 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2123
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2123
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce579 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2123)) : 'freshtv6630)
    | MenhirState2132 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6631 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2133
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2133
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce111 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2133)) : 'freshtv6632)
    | MenhirState2136 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6633 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2137
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2137
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce112 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2137)) : 'freshtv6634)
    | MenhirState757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6635 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2159
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run2162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2159
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2159 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2159)) : 'freshtv6636)
    | MenhirState751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6639 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2166
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6637 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2166 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2167) : 'freshtv6638)
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2166
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2166)) : 'freshtv6640)
    | MenhirState450 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6641 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2282
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2282
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2282 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2282 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce132 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2282)) : 'freshtv6642)
    | MenhirState2285 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6643 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2286
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2286
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2286 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2286 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce578 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2286)) : 'freshtv6644)
    | MenhirState2289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2290
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2290 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2290
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2290 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2290 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2290 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce579 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2290)) : 'freshtv6646)
    | MenhirState2296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6647 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2297
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2297 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2297
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2297 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2297 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2297 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce111 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2297)) : 'freshtv6648)
    | MenhirState2300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6649 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2301
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2301 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2301
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2301 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2301 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2301 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce112 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2301)) : 'freshtv6650)
    | MenhirState411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2358
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run2162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2358
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2358)) : 'freshtv6652)
    | MenhirState410 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2360
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2360 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2362 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2362 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2362) : 'freshtv6654)
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2360 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2360
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2360 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2360 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6655 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2360 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce244 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6656)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2360 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2360)) : 'freshtv6658)
    | MenhirState403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6659 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2387
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1959 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2387
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2387)) : 'freshtv6660)
    | MenhirState402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6661 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2389
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2389
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1942 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce132 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2389)) : 'freshtv6662)
    | MenhirState2638 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6663 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2639
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2639
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2639
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2639) : 'freshtv6664)
    | _ ->
        _menhir_fail ()) : 'freshtv6666)

and _menhir_reduce595 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__non_arrowed_module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__non_arrowed_module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_module_type = 
# 1534 "reason_parser.mly"
                                                                     (_1)
# 23562 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6607) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2625 | MenhirState2287 | MenhirState448 | MenhirState1242 | MenhirState2119 | MenhirState2116 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1438 | MenhirState1436 | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6595 * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce126 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6596)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1999 | MenhirState1959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6597 * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce126 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6598)
    | MenhirState2134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6601 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2135 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6599 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2136 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2136 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2136 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2136 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2136 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2136) : 'freshtv6600)
          | LBRACKETAT | WITH ->
              _menhir_reduce126 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2135 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6602)
    | MenhirState2298 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6605 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2299 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6603 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2300 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2300) : 'freshtv6604)
          | LBRACKETAT | WITH ->
              _menhir_reduce126 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2299 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6606)
    | _ ->
        _menhir_fail ()) : 'freshtv6608)

and _menhir_reduce399 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__core_type2_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__core_type2_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_core_type2 = 
# 3868 "reason_parser.mly"
                                           (_1)
# 23662 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6593) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_core_type2) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState666 | MenhirState244 | MenhirState269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6571 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState270
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6569 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState270 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | QUOTE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6565 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState276 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState276 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState276) : 'freshtv6566)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6567 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6568)) : 'freshtv6570)
          | EQUALGREATER ->
              _menhir_run271 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce55 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6572)
    | MenhirState271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6573 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState272
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run271 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce60 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6574)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6581 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState686
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6579 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState686 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | QUOTE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6575 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState692) : 'freshtv6576)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6577 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6578)) : 'freshtv6580)
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce55 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6582)
    | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6583 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState688
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce60 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6584)
    | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6585 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 23791 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2399
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState2399 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce58 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6586)
    | MenhirState2400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6587 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 23807 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2401
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState2401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce59 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6588)
    | MenhirState2457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6589 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 23823 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2458
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run271 _menhir_env (Obj.magic _menhir_stack) MenhirState2458 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce58 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6590)
    | MenhirState2459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6591 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 23839 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2460
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run271 _menhir_env (Obj.magic _menhir_stack) MenhirState2460 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce59 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6592)
    | _ ->
        _menhir_fail ()) : 'freshtv6594)

and _menhir_reduce648 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _ (_3 : 'tv_package_type_cstrs) ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_package_type_cstr)), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_package_type_cstrs = 
# 3978 "reason_parser.mly"
                                             ( _1::_3 )
# 23860 "reason_parser.ml"
     in
    _menhir_goto_package_type_cstrs _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce645 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _ (_3 : 'tv_package_type_cstrs) ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_package_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 23877 "reason_parser.ml"
      
    in
    
# 3971 "reason_parser.mly"
                                                  ( (_1, _3) )
# 23883 "reason_parser.ml"
     in
    _menhir_goto_package_type _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce665 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__pattern_without_or_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__pattern_without_or_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_without_or = 
# 3256 "reason_parser.mly"
                                                           (_1)
# 23894 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6563) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_pattern_without_or) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6535 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState548
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState548
          | AS | BAR | BARRBRACKET | COLON | COLONCOLON | COMMA | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce173 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState548)) : 'freshtv6536)
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6537 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState573
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState573
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce654 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6538)
    | MenhirState574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6539 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState576
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState576
          | AS | BAR | BARRBRACKET | COLON | COMMA | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce168 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState576)) : 'freshtv6540)
    | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6541 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState644
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState644
          | AS | BAR | COLON | COLONCOLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce173 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState644)) : 'freshtv6542)
    | MenhirState2572 | MenhirState190 | MenhirState2308 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState738 | MenhirState676 | MenhirState668 | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6543 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState654
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState654
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce654 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6544)
    | MenhirState655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6545 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState657
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState657
          | AS | BAR | COLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce168 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState657)) : 'freshtv6546)
    | MenhirState711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6549 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState712
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6547 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState712 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState713) : 'freshtv6548)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState712
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState712)) : 'freshtv6550)
    | MenhirState713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6553 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState714
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState714
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6551 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState714 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce170 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6552)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState714)) : 'freshtv6554)
    | MenhirState2409 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6557 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2410
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6555 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2410 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2411) : 'freshtv6556)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2410
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2410)) : 'freshtv6558)
    | MenhirState2411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6561 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2412
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2412
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6559 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2412 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce170 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6560)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2412)) : 'freshtv6562)
    | _ ->
        _menhir_fail ()) : 'freshtv6564)

and _menhir_reduce229 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_expr_comma_seq)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24229 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24241 "reason_parser.ml"
      
    in
    
# 2734 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "[|") (with_txt _4 "|]") )
# 24247 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce242 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_field_expr_list)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24268 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24280 "reason_parser.ml"
      
    in
    
# 2785 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "{<") (with_txt _6 ">}") )
# 24286 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_expr_comma_seq_extension : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1098 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6523 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce417 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6524)
    | MenhirState1093 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6525 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce234 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6526)
    | MenhirState2186 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6527 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce417 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6528)
    | MenhirState2181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6529 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce234 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6530)
    | MenhirState417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6531 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce233 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6532)
    | MenhirState415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6533 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce233 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6534)
    | _ ->
        _menhir_fail ()

and _menhir_reduce232 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_expr_comma_seq)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24361 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24373 "reason_parser.ml"
      
    in
    
# 2744 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "[|") (with_txt _6 "|]") )
# 24379 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce239 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_field_expr_list)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24399 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24411 "reason_parser.ml"
      
    in
    
# 2775 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{<") (with_txt _4 ">}" ) )
# 24417 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce145 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_tag_field)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3947 "reason_parser.mly"
      ( mktyp(Ptyp_variant([_2], Closed, None)) )
# 24431 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce688 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_tag_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_tag_field)) = _menhir_stack in
    let _v : 'tv_row_field = 
# 3985 "reason_parser.mly"
                                                ( _1 )
# 24441 "reason_parser.ml"
     in
    _menhir_goto_row_field _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_extension_constructor_rebind : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1524
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1524
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1524) : 'freshtv6512)
    | MenhirState2534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6513 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce782 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6514)
    | MenhirState2532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6515 * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce780 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6516)
    | MenhirState2607 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6517 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce782 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6518)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6519 * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce780 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6520)
    | MenhirState2648 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2649
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2649
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2649) : 'freshtv6522)
    | _ ->
        _menhir_fail ()

and _menhir_goto_extension_constructor_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6491 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6489 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_sig_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_declaration)) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_sig_extension_constructors = 
# 3667 "reason_parser.mly"
      ( _3 :: _1 )
# 24507 "reason_parser.ml"
         in
        _menhir_goto_sig_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6490)) : 'freshtv6492)
    | MenhirState1369 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6495 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6493 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_declaration)) = _menhir_stack in
        let _endpos = _endpos__1_ in
        let _v : 'tv_sig_extension_constructors = 
# 3665 "reason_parser.mly"
                                                          ( [_1] )
# 24520 "reason_parser.ml"
         in
        _menhir_goto_sig_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6494)) : 'freshtv6496)
    | MenhirState1483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6497 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1526
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1526
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1526) : 'freshtv6498)
    | MenhirState2042 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6499 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2043
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2043
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2043) : 'freshtv6500)
    | MenhirState2534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6501 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce781 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6502)
    | MenhirState2532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6503 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce779 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6504)
    | MenhirState2607 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6505 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce781 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6506)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6507 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce779 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6508)
    | MenhirState2648 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2651
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2651
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2651) : 'freshtv6510)
    | _ ->
        _menhir_fail ()

and _menhir_run1332 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_label_declarations -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MUTABLE ->
        _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
    | LIDENT _ ->
        _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1332
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1332

and _menhir_goto_constructor_declaration_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1314 | MenhirState1310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6481 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce397 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6482)
    | MenhirState1263 | MenhirState1338 | MenhirState1340 | MenhirState1281 | MenhirState1313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6483) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce396 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6484)
    | MenhirState2505 | MenhirState2501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6485 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce397 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6486)
    | MenhirState106 | MenhirState2514 | MenhirState2516 | MenhirState2472 | MenhirState2504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6487) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce396 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6488)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declaration_no_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_no_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1263 | MenhirState1338 | MenhirState1340 | MenhirState1281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6477 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1313
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce394 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1313)) : 'freshtv6478)
    | MenhirState106 | MenhirState2514 | MenhirState2516 | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6479 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2504
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2504 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce394 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2504)) : 'freshtv6480)
    | _ ->
        _menhir_fail ()

and _menhir_reduce299 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_type_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_str_type_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1418 "reason_parser.mly"
                       (
      mkstr(Pstr_typext _1)
    )
# 24677 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_nonlocal_module_bindings : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_nonlocal_module_bindings -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6471 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1864
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6469) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1864 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1865) : 'freshtv6470)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce302 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1864)) : 'freshtv6472)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6475 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2725
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6473) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2725 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run2631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2726) : 'freshtv6474)
          | EOF | SEMI ->
              _menhir_reduce302 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2725)) : 'freshtv6476)
    | _ ->
        _menhir_fail ()

and _menhir_reduce508 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_class_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_class_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_class_declarations = 
# 1790 "reason_parser.mly"
                                                  (
    _2::_1
  )
# 24758 "reason_parser.ml"
     in
    _menhir_goto_many_class_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce512 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_class_type_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_class_type_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_class_type_declarations = 
# 2319 "reason_parser.mly"
                                                            (
    _2::_1
  )
# 24772 "reason_parser.ml"
     in
    _menhir_goto_many_class_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce516 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_nonlocal_module_bindings -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_nonlocal_module_bindings) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_nonlocal_module_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_nonlocal_module_bindings = 
# 1498 "reason_parser.mly"
                                                               (
    _2::_1
  )
# 24786 "reason_parser.ml"
     in
    _menhir_goto_many_nonlocal_module_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce518 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_type_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_type_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_type_declarations = 
# 3463 "reason_parser.mly"
                                                ( _2 :: _1 )
# 24798 "reason_parser.ml"
     in
    _menhir_goto_many_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_class_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6463 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1874
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6461) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1874 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState1875
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1875) : 'freshtv6462)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce306 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1874)) : 'freshtv6464)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6467 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2731
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6465) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2731 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2732
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2732) : 'freshtv6466)
          | EOF | SEMI ->
              _menhir_reduce306 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2731)) : 'freshtv6468)
    | _ ->
        _menhir_fail ()

and _menhir_reduce693 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_field)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_class_fields = 
# 2002 "reason_parser.mly"
                          ( [_1] )
# 24880 "reason_parser.ml"
     in
    _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1746 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce693 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce495 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_let_binding_impl -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_let_binding_impl) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_binding = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2896 "reason_parser.mly"
                                          (
      let (rec_flag, body, item_attrs) = _2 in
      let (ext_attrs, ext_id) = _1 in
      let loc = mklocation _symbolstartpos _endpos in
      mklbs (ext_attrs, Some ext_id) rec_flag (mklb body item_attrs loc) loc
    )
# 24903 "reason_parser.ml"
     in
    _menhir_goto_let_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce494 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding_impl -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_let_binding_impl) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_let_binding = let _endpos = _endpos__1_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2886 "reason_parser.mly"
                     (
    let (rec_flag, body, item_attrs) = _1 in
    let loc = mklocation _symbolstartpos _endpos in
    mklbs ([], None) rec_flag (mklb body item_attrs loc) loc
  )
# 24920 "reason_parser.ml"
     in
    _menhir_goto_let_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_class_type_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_type_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6453 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1869
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6451) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1869 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState1870
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1870) : 'freshtv6452)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce307 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1869)) : 'freshtv6454)
    | MenhirState2557 | MenhirState2559 | MenhirState1246 | MenhirState1402 | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6457 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2080
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2081 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2080 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6455 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__signature_item = 
# 1729 "reason_parser.mly"
                                 (
      mksig(Psig_class_type (List.rev _1))
    )
# 24981 "reason_parser.ml"
               in
              _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6456)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2080)) : 'freshtv6458)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6459 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2730
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2081 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              _menhir_reduce307 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2730)) : 'freshtv6460)
    | _ ->
        _menhir_fail ()

and _menhir_goto_str_exception_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_exception_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6447) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce300 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6448)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6449) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce300 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6450)
    | _ ->
        _menhir_fail ()

and _menhir_reduce305 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_open_statement -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_open_statement) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1443 "reason_parser.mly"
                   (
      mkstr(Pstr_open _1)
    )
# 25038 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_type_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_type_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6439 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1847
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6437) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1847 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6435) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1848 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 25072 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | MINUS ->
                      _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState1849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1849 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1849) : 'freshtv6436)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1848) : 'freshtv6438)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce298 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1847)) : 'freshtv6440)
    | MenhirState2557 | MenhirState2559 | MenhirState1246 | MenhirState1402 | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6443 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2068
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2069 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6441 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__signature_item = 
# 1682 "reason_parser.mly"
                           (
      mksig(Psig_type (List.rev _1))
    )
# 25120 "reason_parser.ml"
               in
              _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6442)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2068)) : 'freshtv6444)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6445 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2724
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2069 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2724 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              _menhir_reduce298 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2724)) : 'freshtv6446)
    | _ ->
        _menhir_fail ()

and _menhir_reduce504 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_let_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_let_binding) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_bindings = 
# 2871 "reason_parser.mly"
                                                ( addlb _1 _2 )
# 25154 "reason_parser.ml"
     in
    _menhir_goto_let_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce410 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_expr = 
# 2500 "reason_parser.mly"
                               (_1)
# 25165 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6433) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState828 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6083 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState829
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState829
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState829 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce106 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6084)
    | MenhirState830 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6085 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState831
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState831 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState831
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce88 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6086)
    | MenhirState832 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6087 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 25248 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState833
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState833
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce83 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6088)
    | MenhirState842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6089 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState843
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState843
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce101 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6090)
    | MenhirState845 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6091 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState846
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState846
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce96 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6092)
    | MenhirState847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6093 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState848
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState848 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState848
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce97 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6094)
    | MenhirState849 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6095 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 25312 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState850
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState850
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce82 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6096)
    | MenhirState851 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6097 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState852
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState852
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce85 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6098)
    | MenhirState853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6099 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState854
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState854
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce84 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6100)
    | MenhirState855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6101 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState856
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState856
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce91 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6102)
    | MenhirState857 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6103 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState858
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState858
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce87 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6104)
    | MenhirState859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6105 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState860
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState860
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce86 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6106)
    | MenhirState861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6107 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState862
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState862
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce98 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6108)
    | MenhirState863 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6109 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 25496 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState864
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState864
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce81 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6110)
    | MenhirState865 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6111 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 25520 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState866
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState866
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce80 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6112)
    | MenhirState867 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6113 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState868
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState868
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce99 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6114)
    | MenhirState869 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6115 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState870
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState870
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce89 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6116)
    | MenhirState871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6117 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 25630 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState872
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState872
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce79 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6118)
    | MenhirState873 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6119 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState874
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState874
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce90 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6120)
    | MenhirState875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6121 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState876
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState876
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce92 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6122)
    | MenhirState877 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6123 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState878
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState878
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce93 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6124)
    | MenhirState879 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6125 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState880
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState880
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce94 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6126)
    | MenhirState881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6127 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState882
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState882
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce95 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6128)
    | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6133 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState883
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState883
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6131 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState883 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6129 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState886) : 'freshtv6130)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6132)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState883)) : 'freshtv6134)
    | MenhirState886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6135 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState887
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState887
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce103 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6136)
    | MenhirState920 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6141 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState921
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState921
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6139 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState921 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6137 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState924) : 'freshtv6138)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6140)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState921)) : 'freshtv6142)
    | MenhirState941 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6143 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState942
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState942
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState942 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce106 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6144)
    | MenhirState943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6145 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState944
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState944
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce88 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6146)
    | MenhirState945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6147 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 26330 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState946
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState946
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce83 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6148)
    | MenhirState955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6149 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState956
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState956
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce101 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6150)
    | MenhirState958 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6151 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState959
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState959 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState959 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState959
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState959 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce96 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6152)
    | MenhirState960 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6153 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState961
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState961 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState961
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce97 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6154)
    | MenhirState962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6155 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 26394 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState963
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState963
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce82 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6156)
    | MenhirState964 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6157 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState965
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState965
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce85 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6158)
    | MenhirState966 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6159 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState967
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState967
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce84 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6160)
    | MenhirState968 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6161 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState969
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState969
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce91 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6162)
    | MenhirState970 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6163 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState971
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState971
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce87 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6164)
    | MenhirState972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6165 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState973
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState973
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce86 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6166)
    | MenhirState974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6167 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState975
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState975
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce98 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6168)
    | MenhirState976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6169 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 26578 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState977
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState977
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce81 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6170)
    | MenhirState978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6171 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 26602 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState979
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState979
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce80 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6172)
    | MenhirState980 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6173 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState981
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState981
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce99 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6174)
    | MenhirState982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6175 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState983
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState983
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce89 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6176)
    | MenhirState984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6177 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 26712 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState985
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState985
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce79 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6178)
    | MenhirState986 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6179 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState987
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState987
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce90 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6180)
    | MenhirState988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6181 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState989
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState989
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce92 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6182)
    | MenhirState990 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6183 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState991
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState991
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce93 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6184)
    | MenhirState992 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6185 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState993
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState993
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce94 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6186)
    | MenhirState994 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6187 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState995
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState995
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce95 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6188)
    | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6189 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState996
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState996
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce103 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6190)
    | MenhirState997 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6195 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState998
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState998
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6193 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState998 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6191 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1001) : 'freshtv6192)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6194)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState998)) : 'freshtv6196)
    | MenhirState1001 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6197 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1002
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1002
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce104 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6198)
    | MenhirState1003 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6203 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1004
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1004
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6201 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1004 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6199 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1006) : 'freshtv6200)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6202)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1004 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1004)) : 'freshtv6204)
    | MenhirState1006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6205 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1007
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1007
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce105 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6206)
    | MenhirState1008 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6209 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1009
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6207 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1009 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1010 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1010) : 'freshtv6208)
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1009
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1009)) : 'freshtv6210)
    | MenhirState1010 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6211 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1011
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1011
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce109 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6212)
    | MenhirState1013 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6213 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1014
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1014
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce102 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6214)
    | MenhirState917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6215 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1015
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1015
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce100 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6216)
    | MenhirState902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6217 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1016
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1016
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce437 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1016)) : 'freshtv6218)
    | MenhirState1019 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6219 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1020
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1020
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1020 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce438 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1020)) : 'freshtv6220)
    | MenhirState1027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6223 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1028
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6221 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1028 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1029 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1029) : 'freshtv6222)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1028
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1028 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1028)) : 'freshtv6224)
    | MenhirState1029 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6225 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1030
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1030
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce486 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1030)) : 'freshtv6226)
    | MenhirState1031 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6227 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1032
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1032
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce485 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1032)) : 'freshtv6228)
    | MenhirState892 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6229 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1038
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1038
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce74 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6230)
    | MenhirState888 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6233 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1039
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6231 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1039 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1040 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1040) : 'freshtv6232)
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1039
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1039)) : 'freshtv6234)
    | MenhirState1040 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6235 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1041
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1041
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce109 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6236)
    | MenhirState1042 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6241 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1043
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1043
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6239 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1043 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6237 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1046) : 'freshtv6238)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6240)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1043)) : 'freshtv6242)
    | MenhirState1046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6243 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1047
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1047
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce104 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6244)
    | MenhirState1048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6249 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1049
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1049
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6247 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1049 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6245 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1051) : 'freshtv6246)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6248)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1049)) : 'freshtv6250)
    | MenhirState1051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6251 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1052
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1052
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce105 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6252)
    | MenhirState1054 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6253 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1055
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1055
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce102 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6254)
    | MenhirState808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6255 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1056
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1056
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce100 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6256)
    | MenhirState415 | MenhirState417 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6257 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1060
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1060
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1060
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6258)
    | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6261 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1073
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1073
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1073
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6259 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1073 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce213 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6260)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1073 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6262)
    | MenhirState177 | MenhirState414 | MenhirState2176 | MenhirState1142 | MenhirState1137 | MenhirState1076 | MenhirState1079 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6267 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1081
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6263) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1081 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1084) : 'freshtv6264)
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6265) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1081 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1082 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1082) : 'freshtv6266)
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1081
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6268)
    | MenhirState1118 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6269 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1119
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1119
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1119 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | RBRACE ->
              _menhir_reduce472 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1119)) : 'freshtv6270)
    | MenhirState1129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6271 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1130
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1130
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              _menhir_reduce600 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce472 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1130)) : 'freshtv6272)
    | MenhirState786 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1132
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1132 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1133 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1133) : 'freshtv6274)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1132
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1132)) : 'freshtv6276)
    | MenhirState1133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6279 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1134
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1134
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1134 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce78 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6278)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1134)) : 'freshtv6280)
    | MenhirState15 | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6283 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1144
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1144
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1144
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6281 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1144 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce205 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6282)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6284)
    | MenhirState765 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6287 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1149
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6285 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1149 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1150 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1150) : 'freshtv6286)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1149
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1149)) : 'freshtv6288)
    | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1156 | MenhirState1162 | MenhirState1171 | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6289 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1169
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1169
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1169
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1169) : 'freshtv6290)
    | MenhirState1188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6291 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1189
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1189
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce402 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6292)
    | MenhirState1192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1193
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1193
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | EOF | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce407 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1193)) : 'freshtv6294)
    | MenhirState1199 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6295 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1200
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1200
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1200 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce61 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6296)
    | MenhirState1216 | MenhirState1208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6297 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1210
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1210
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | EOF | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce536 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1210)) : 'freshtv6298)
    | MenhirState1219 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6299 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1220
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1220
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce496 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6300)
    | MenhirState1229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6301 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1230
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1230
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce501 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6302)
    | MenhirState1232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6303 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1233
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1233
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce500 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6304)
    | MenhirState1679 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6305 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1680
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1680
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce839 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1680)) : 'freshtv6306)
    | MenhirState1682 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6307 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1683
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1683
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce840 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1683)) : 'freshtv6308)
    | MenhirState1702 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6309 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1703
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1703
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce402 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6310)
    | MenhirState1706 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6311 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1707
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1707
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce407 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1707)) : 'freshtv6312)
    | MenhirState1712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6313 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1713
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1713
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce566 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1713)) : 'freshtv6314)
    | MenhirState1719 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((((('freshtv6315 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1720
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1720
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce567 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1720)) : 'freshtv6316)
    | MenhirState1722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6317 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1723
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1723
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce565 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1723)) : 'freshtv6318)
    | MenhirState1728 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6319 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1729
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1729
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1729
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1729) : 'freshtv6320)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6321 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1888
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1888
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce61 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6322)
    | MenhirState1900 | MenhirState1894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6323 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1896
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1896
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce536 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1896)) : 'freshtv6324)
    | MenhirState1903 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6325 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1904
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1904
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce496 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6326)
    | MenhirState1913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6327 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1914
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1914
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce501 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6328)
    | MenhirState1916 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6329 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1917
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1917
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce500 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6330)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6331 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1929
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1929
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1929
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1929) : 'freshtv6332)
    | MenhirState1943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1944
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6333 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1944 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1949) : 'freshtv6334)
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1944 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1946) : 'freshtv6336)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1944
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1944)) : 'freshtv6338)
    | MenhirState758 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6343 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2147
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6339 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2147 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2152 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2152 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2152) : 'freshtv6340)
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6341 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2147 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2149) : 'freshtv6342)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2147
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2147)) : 'freshtv6344)
    | MenhirState750 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6347 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2173
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState2173
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2173
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6345 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2173 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce213 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6346)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6348)
    | MenhirState737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6351 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2202
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2202
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6349 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2202 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6350)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2202)) : 'freshtv6352)
    | MenhirState2204 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6355 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2205
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2205
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6353 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2205 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6354)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2205)) : 'freshtv6356)
    | MenhirState2207 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6359 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2208
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2208
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6357 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2208 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6358)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2208)) : 'freshtv6360)
    | MenhirState2212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6361 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2213
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2213
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce437 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2213)) : 'freshtv6362)
    | MenhirState2216 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2217
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2217
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce438 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2217)) : 'freshtv6364)
    | MenhirState2223 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6367 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2224
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6365 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2224 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2225) : 'freshtv6366)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2224
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2224)) : 'freshtv6368)
    | MenhirState2225 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6369 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2226
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2226
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2226 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce486 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2226)) : 'freshtv6370)
    | MenhirState2227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6371 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2228
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2228
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2228 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce485 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2228)) : 'freshtv6372)
    | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState475 | MenhirState2247 | MenhirState2253 | MenhirState2249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6373 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2251
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2251
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2251
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2251) : 'freshtv6374)
    | MenhirState2264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6375 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2265
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2265
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2265 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce74 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6376)
    | MenhirState473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6379 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2266
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2266
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6377 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2266 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6378)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2266)) : 'freshtv6380)
    | MenhirState2268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6383 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2269
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2269
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6381 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2269 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6382)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2269)) : 'freshtv6384)
    | MenhirState2271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6387 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2272
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2272
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6385 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2272 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6386)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2272)) : 'freshtv6388)
    | MenhirState421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6389 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 32016 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2331
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2331
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2331 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | GREATERRBRACE ->
              _menhir_reduce432 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2331)) : 'freshtv6390)
    | MenhirState412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2352
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6391 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2352 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2356 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2356 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2356) : 'freshtv6392)
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2352 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2354 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2354 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2354) : 'freshtv6394)
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2352
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2352)) : 'freshtv6396)
    | MenhirState2368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6399 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2369
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6397 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2369 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2370) : 'freshtv6398)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2369
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2369)) : 'freshtv6400)
    | MenhirState2370 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6403 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2371
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2371
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6401 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2371 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce78 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6402)
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2371 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2371)) : 'freshtv6404)
    | MenhirState183 | MenhirState407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2376
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState2376
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2376
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2376 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce205 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6406)
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce418 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6408)
    | MenhirState404 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2381
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run879 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run877 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run875 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2381 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2385 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2385 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2385) : 'freshtv6410)
          | COLONEQUAL ->
              _menhir_run881 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2381 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2383) : 'freshtv6412)
          | GREATER ->
              _menhir_run873 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run871 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run865 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run863 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run849 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run832 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2381
          | LESS ->
              _menhir_run869 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run867 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run861 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run857 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run855 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run847 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run851 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run845 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run888 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run830 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2381 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2381)) : 'freshtv6414)
    | MenhirState2416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2417
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2417
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_payload = 
# 4348 "reason_parser.mly"
                               ( PPat (_2, Some _4) )
# 32597 "reason_parser.ml"
               in
              _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6416)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2417)) : 'freshtv6418)
    | MenhirState2568 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6429 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2570
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6427 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2570 in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6425 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_ : _menhir_state) = _menhir_s in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 32632 "reason_parser.ml"
              ) = 
# 1145 "reason_parser.mly"
             ( apply_mapper_chain_to_expr _1 default_mapper_chain)
# 32636 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6423) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 32644 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6421) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 32652 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6419) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 32660 "reason_parser.ml"
              )) : (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 32664 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv6420)) : 'freshtv6422)) : 'freshtv6424)) : 'freshtv6426)) : 'freshtv6428)
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2570
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2570)) : 'freshtv6430)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6431 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2742
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run992 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run990 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run988 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run994 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run986 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run984 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run978 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run976 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run962 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run945 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2742
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2742
          | LESS ->
              _menhir_run982 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run974 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run970 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run968 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run960 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run966 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run958 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1008 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2742) : 'freshtv6432)
    | _ ->
        _menhir_fail ()) : 'freshtv6434)

and _menhir_reduce360 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_simple_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_simple_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_simple_expr = 
# 1963 "reason_parser.mly"
                                                        (_1)
# 32777 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6081) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_simple_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1626 | MenhirState1732 | MenhirState1657 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6077 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1641
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run816 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce11 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1641)) : 'freshtv6078)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6079 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2681
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run929 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce11 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2681)) : 'freshtv6080)
    | _ ->
        _menhir_fail ()) : 'freshtv6082)

and _menhir_reduce575 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 32905 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 910 "reason_parser.mly"
       (string)
# 32910 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mod_longident = 
# 4137 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 32919 "reason_parser.ml"
     in
    _menhir_goto_mod_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce155 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3530 "reason_parser.mly"
                                                ( mktyp (Ptyp_any), Invariant )
# 32931 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce157 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3532 "reason_parser.mly"
                                                ( mktyp (Ptyp_any), Covariant )
# 32945 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce159 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3534 "reason_parser.mly"
                                                ( mktyp Ptyp_any, Contravariant )
# 32959 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1627 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | REC ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce684 _menhir_env (Obj.magic _menhir_stack) MenhirState1627
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1627

and _menhir_goto_private_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_private_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6047 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6043 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1318 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1318
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1318) : 'freshtv6044)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6045 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6046)) : 'freshtv6048)
    | MenhirState1338 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6053 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6049 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1345
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1345) : 'freshtv6050)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6051 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6052)) : 'freshtv6054)
    | MenhirState1366 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6055 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState1368
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1368) : 'freshtv6056)
    | MenhirState1686 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6057 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1687) : 'freshtv6058)
    | MenhirState1685 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6059 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1696 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1696) : 'freshtv6060)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6065 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6061 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2509 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState2509
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2509) : 'freshtv6062)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6063 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6064)) : 'freshtv6066)
    | MenhirState2514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6071 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6067 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState2521
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2521) : 'freshtv6068)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6069 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6070)) : 'freshtv6072)
    | MenhirState2530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6073 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState2531
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2531) : 'freshtv6074)
    | MenhirState2580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6075 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2581 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState2581
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2581) : 'freshtv6076)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mutable_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mutable_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2521 | MenhirState2509 | MenhirState1345 | MenhirState1332 | MenhirState1318 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6035 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv6031 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 33188 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv6027 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 33200 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LBRACKET ->
                    _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETGREATER ->
                    _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETLESS ->
                    _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESS ->
                    _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESSDOTDOTGREATER ->
                    _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESSGREATER ->
                    _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | QUOTE ->
                    _menhir_run153 _menhir_env (Obj.magic _menhir_stack) MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1322 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1322) : 'freshtv6028)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv6029 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 33245 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6030)) : 'freshtv6032)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv6033 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6034)) : 'freshtv6036)
    | MenhirState1547 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6037 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1548 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1548) : 'freshtv6038)
    | MenhirState1662 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6039 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1663) : 'freshtv6040)
    | MenhirState1670 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6041 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1677) : 'freshtv6042)
    | _ ->
        _menhir_fail ()

and _menhir_reduce47 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 33309 "reason_parser.ml"
      
    in
    
# 1966 "reason_parser.mly"
    ( mkclass(Pcl_constr(_1, [])) )
# 33315 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce237 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 33334 "reason_parser.ml"
      
    in
    
# 2771 "reason_parser.mly"
      ( mkexp (Pexp_new _2) )
# 33340 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__simple_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6023) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce548 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6024)
    | MenhirState2657 | MenhirState2698 | MenhirState2671 | MenhirState2673 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState476 | MenhirState2232 | MenhirState480 | MenhirState484 | MenhirState2218 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1881 | MenhirState1906 | MenhirState1885 | MenhirState1621 | MenhirState1817 | MenhirState1697 | MenhirState1701 | MenhirState1708 | MenhirState1651 | MenhirState1631 | MenhirState1633 | MenhirState1628 | MenhirState1178 | MenhirState1179 | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 | MenhirState893 | MenhirState1036 | MenhirState897 | MenhirState901 | MenhirState1022 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6025) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce548 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6026)
    | _ ->
        _menhir_fail ()

and _menhir_goto_val_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2262 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1471 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1642 | MenhirState1641 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState837 | MenhirState835 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState822 | MenhirState815 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState811 | MenhirState808 | MenhirState793 | MenhirState806 | MenhirState800 | MenhirState745 | MenhirState463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6015) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce201 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6016)
    | MenhirState2754 | MenhirState2751 | MenhirState2747 | MenhirState2737 | MenhirState2576 | MenhirState2682 | MenhirState2681 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState2344 | MenhirState416 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState950 | MenhirState948 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState935 | MenhirState928 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState924 | MenhirState917 | MenhirState902 | MenhirState915 | MenhirState908 | MenhirState907 | MenhirState905 | MenhirState890 | MenhirState889 | MenhirState788 | MenhirState759 | MenhirState744 | MenhirState740 | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6017) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce201 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6018)
    | MenhirState2614 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6021 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6019 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_val_longident) : 'tv_val_longident) = _v in
        let (_startpos__3_ : Lexing.position) = _startpos in
        ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_toplevel_directive = 
# 4170 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_ident _3) )
# 33403 "reason_parser.ml"
         in
        _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6020)) : 'freshtv6022)
    | _ ->
        _menhir_fail ()

and _menhir_error387 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce715 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__simple_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__simple_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_expr = 
# 2634 "reason_parser.mly"
                                             (_1)
# 33421 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6013) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState800 | MenhirState463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5905 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState468
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState468
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce236 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5906)
    | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5907 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState733
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState733
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState733 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce236 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5908)
    | MenhirState740 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5909 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState741
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState741
          | DOWNTO ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TO ->
              _menhir_run742 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState741)) : 'freshtv5910)
    | MenhirState744 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5911 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState745
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState745
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState745 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState745)) : 'freshtv5912)
    | MenhirState745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5913 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState746
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState746
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce77 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5914)
    | MenhirState759 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5917 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState760
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState760
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5915 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState760 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState761) : 'freshtv5916)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState760)) : 'freshtv5918)
    | MenhirState788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5919 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState789
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState789
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState789 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce108 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5920)
    | MenhirState806 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5921 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState807
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState807
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce107 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5922)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 | MenhirState808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5931 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState809
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5929 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState809 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5923 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState810 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1048) : 'freshtv5924)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5925 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState810 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1042) : 'freshtv5926)
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState810 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5927 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState810 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState811) : 'freshtv5928)
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState810 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState810) : 'freshtv5930)
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce493 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5932)
    | MenhirState812 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5933 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState813
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState813
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce73 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5934)
    | MenhirState1641 | MenhirState1642 | MenhirState835 | MenhirState837 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5935 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState818
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState818
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce493 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5936)
    | MenhirState889 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5937 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState890
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState890
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState890)) : 'freshtv5938)
    | MenhirState890 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5941 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState891
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState891
          | ELSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState891 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState892) : 'freshtv5940)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce75 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5942)
    | MenhirState905 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState906
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState906
          | DOWNTO ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TO ->
              _menhir_run742 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState906)) : 'freshtv5944)
    | MenhirState907 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5945 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState908
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState908
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState908)) : 'freshtv5946)
    | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5947 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState909
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState909
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce77 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5948)
    | MenhirState915 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5949 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState916
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState916
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce107 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5950)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 | MenhirState917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5959 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState918
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5957 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState918 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5951 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState919 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1003) : 'freshtv5952)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5953 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState919 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState997) : 'freshtv5954)
              | LIDENT _v ->
                  _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5955 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState919 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState920) : 'freshtv5956)
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState919) : 'freshtv5958)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState918 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce493 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5960)
    | MenhirState925 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5961 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState926
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState926
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce73 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5962)
    | MenhirState2681 | MenhirState2682 | MenhirState948 | MenhirState950 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5963 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState931
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState931
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState931 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce493 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5964)
    | MenhirState729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5967 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 34466 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2211
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2211
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE | CHAR _ | COLON | EQUALGREATER | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5965 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 34482 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 34487 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_), _endpos__4_), _endpos__5_, _, (_5 : 'tv_simple_expr), _startpos__5_) = _menhir_stack in
              let _4 = () in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2489 "reason_parser.mly"
      ( ("?" ^ _1, Some _5, _3) )
# 34495 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv5966)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2211)) : 'freshtv5968)
    | MenhirState474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2262
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2262
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2262 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2262)) : 'freshtv5970)
    | MenhirState2262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5973 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2263
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2263
          | ELSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2263 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2264 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2264) : 'freshtv5972)
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState2263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce75 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5974)
    | MenhirState456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2275
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2275
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce108 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5976)
    | MenhirState418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2337
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2337
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5977 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2337 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2340 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2340) : 'freshtv5978)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2337 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv5981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv5979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce71 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5980)) : 'freshtv5982)) : 'freshtv5984)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2337)) : 'freshtv5986)
    | MenhirState416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5987 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2344
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2344
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2344)) : 'freshtv5988)
    | MenhirState2344 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2345
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2345
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce76 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5990)
    | MenhirState179 | MenhirState186 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5991 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 34798 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2420
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2420
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState2420 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce235 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5992)
    | MenhirState176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2426
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2426
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2426 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2427) : 'freshtv5994)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2426 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2426)) : 'freshtv5996)
    | MenhirState17 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6005 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2542
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2542
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2542 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState2545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2545) : 'freshtv5998)
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2542 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2542 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv6001 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv5999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce71 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6000)) : 'freshtv6002)) : 'freshtv6004)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2542)) : 'freshtv6006)
    | MenhirState6 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6007 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 34901 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2548
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2548
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2548 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce235 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6008)
    | MenhirState2 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6009 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2549
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2549
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2549 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2549)) : 'freshtv6010)
    | MenhirState2549 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6011 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2550
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2550
          | SHARP ->
              _menhir_run469 _menhir_env (Obj.magic _menhir_stack) MenhirState2550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce76 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6012)
    | _ ->
        _menhir_fail ()) : 'freshtv6014)

and _menhir_run1319 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce588 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error1128 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error240 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error139 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_functor_args : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_functor_args -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState752 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5887 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5885 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState753 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState754) : 'freshtv5886)
        | LPAREN ->
            _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState753) : 'freshtv5888)
    | MenhirState1937 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5891 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5889 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1938 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1939) : 'freshtv5890)
        | LPAREN ->
            _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1938) : 'freshtv5892)
    | MenhirState2631 | MenhirState2317 | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5897 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5893 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2131 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2134) : 'freshtv5894)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5895 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2131 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2132 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2132 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2132) : 'freshtv5896)
        | LPAREN ->
            _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2131 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2131) : 'freshtv5898)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5903 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5899 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2295 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2298 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2298 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2298 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2298 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2298 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2298 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2298) : 'freshtv5900)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5901 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2295 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2296) : 'freshtv5902)
        | LPAREN ->
            _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2295) : 'freshtv5904)
    | _ ->
        _menhir_fail ()

and _menhir_reduce420 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension = 
# 4337 "reason_parser.mly"
                                           ( (_2, _3) )
# 35219 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5883) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2692 | MenhirState2676 | MenhirState2677 | MenhirState2663 | MenhirState2643 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState120 | MenhirState2466 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState169 | MenhirState2436 | MenhirState2433 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1793 | MenhirState1794 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1355 | MenhirState1353 | MenhirState1339 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1282 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState1271 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState271 | MenhirState269 | MenhirState244 | MenhirState253 | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5855) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce153 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5856)
    | MenhirState2515 | MenhirState106 | MenhirState2472 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1636 | MenhirState1637 | MenhirState1595 | MenhirState1586 | MenhirState1587 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState1494 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 | MenhirState328 | MenhirState243 | MenhirState316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5857) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce153 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5858)
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5859) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce292 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5860)
    | MenhirState2657 | MenhirState2698 | MenhirState2671 | MenhirState2673 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState476 | MenhirState2232 | MenhirState480 | MenhirState484 | MenhirState2218 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1881 | MenhirState1906 | MenhirState1885 | MenhirState1621 | MenhirState1817 | MenhirState1697 | MenhirState1701 | MenhirState1708 | MenhirState1651 | MenhirState1631 | MenhirState1633 | MenhirState1628 | MenhirState1178 | MenhirState1179 | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 | MenhirState893 | MenhirState1036 | MenhirState897 | MenhirState901 | MenhirState1022 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5861) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce292 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5862)
    | MenhirState2638 | MenhirState2639 | MenhirState410 | MenhirState2360 | MenhirState450 | MenhirState2282 | MenhirState751 | MenhirState2166 | MenhirState754 | MenhirState756 | MenhirState2136 | MenhirState2137 | MenhirState2132 | MenhirState2133 | MenhirState2122 | MenhirState2123 | MenhirState1240 | MenhirState1241 | MenhirState794 | MenhirState795 | MenhirState767 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5863) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce260 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5864)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState463 | MenhirState2268 | MenhirState473 | MenhirState2262 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2253 | MenhirState2249 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState745 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1641 | MenhirState1642 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState811 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState835 | MenhirState837 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState806 | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5865) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce249 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5866)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2681 | MenhirState2682 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState416 | MenhirState2344 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState730 | MenhirState2207 | MenhirState740 | MenhirState744 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState759 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1171 | MenhirState1166 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState788 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState889 | MenhirState890 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState924 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState948 | MenhirState950 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5867) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce249 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5868)
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5869) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce266 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5870)
    | MenhirState1761 | MenhirState1536 | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5871) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce37 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5872)
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1626 | MenhirState1732 | MenhirState1657 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5873) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce17 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5874)
    | MenhirState2701 | MenhirState2050 | MenhirState1820 | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5875) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce37 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5876)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState1245 | MenhirState2110 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5877) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce266 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5878)
    | MenhirState402 | MenhirState2389 | MenhirState403 | MenhirState2387 | MenhirState411 | MenhirState2358 | MenhirState2300 | MenhirState2301 | MenhirState2296 | MenhirState2297 | MenhirState2289 | MenhirState2290 | MenhirState2285 | MenhirState2286 | MenhirState757 | MenhirState2159 | MenhirState1470 | MenhirState2019 | MenhirState1939 | MenhirState1941 | MenhirState1942 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5879) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce260 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5880)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5881) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce17 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5882)
    | _ ->
        _menhir_fail ()) : 'freshtv5884)

and _menhir_reduce333 : _menhir_env -> (('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_attribute = 
# 4298 "reason_parser.mly"
                                      ( (_2, _3) )
# 35352 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5853) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_attribute) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2455 | MenhirState324 | MenhirState2215 | MenhirState1809 | MenhirState1800 | MenhirState1802 | MenhirState1774 | MenhirState1765 | MenhirState1767 | MenhirState1705 | MenhirState1191 | MenhirState1018 | MenhirState262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5821 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce136 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5822)
    | MenhirState2412 | MenhirState714 | MenhirState573 | MenhirState576 | MenhirState548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5823 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce174 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5824)
    | MenhirState2410 | MenhirState712 | MenhirState654 | MenhirState657 | MenhirState644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5825 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce174 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5826)
    | MenhirState2639 | MenhirState2360 | MenhirState2282 | MenhirState2166 | MenhirState756 | MenhirState2137 | MenhirState2133 | MenhirState2123 | MenhirState1241 | MenhirState795 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5827 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce124 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5828)
    | MenhirState2381 | MenhirState2376 | MenhirState2352 | MenhirState2331 | MenhirState2269 | MenhirState2266 | MenhirState2265 | MenhirState2251 | MenhirState2228 | MenhirState2226 | MenhirState2217 | MenhirState2213 | MenhirState2205 | MenhirState2202 | MenhirState2173 | MenhirState1929 | MenhirState1917 | MenhirState1914 | MenhirState1904 | MenhirState1896 | MenhirState1888 | MenhirState1729 | MenhirState1723 | MenhirState1720 | MenhirState1713 | MenhirState1707 | MenhirState1703 | MenhirState1683 | MenhirState1680 | MenhirState1144 | MenhirState1130 | MenhirState1119 | MenhirState1081 | MenhirState1073 | MenhirState1056 | MenhirState1055 | MenhirState1052 | MenhirState1043 | MenhirState1047 | MenhirState883 | MenhirState1041 | MenhirState998 | MenhirState921 | MenhirState887 | MenhirState829 | MenhirState882 | MenhirState856 | MenhirState876 | MenhirState878 | MenhirState880 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState862 | MenhirState866 | MenhirState864 | MenhirState860 | MenhirState858 | MenhirState854 | MenhirState852 | MenhirState846 | MenhirState850 | MenhirState848 | MenhirState831 | MenhirState843 | MenhirState833 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5829 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce110 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5830)
    | MenhirState2742 | MenhirState2570 | MenhirState2417 | MenhirState2369 | MenhirState2371 | MenhirState2272 | MenhirState2224 | MenhirState2208 | MenhirState2147 | MenhirState1149 | MenhirState1944 | MenhirState1233 | MenhirState1230 | MenhirState1220 | MenhirState1210 | MenhirState1200 | MenhirState1193 | MenhirState1189 | MenhirState1169 | MenhirState1132 | MenhirState1134 | MenhirState1060 | MenhirState1049 | MenhirState1039 | MenhirState1038 | MenhirState1032 | MenhirState1028 | MenhirState1030 | MenhirState1020 | MenhirState1016 | MenhirState1015 | MenhirState1014 | MenhirState1004 | MenhirState1009 | MenhirState1011 | MenhirState1007 | MenhirState1002 | MenhirState996 | MenhirState942 | MenhirState995 | MenhirState969 | MenhirState989 | MenhirState991 | MenhirState993 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState975 | MenhirState979 | MenhirState977 | MenhirState973 | MenhirState971 | MenhirState967 | MenhirState965 | MenhirState959 | MenhirState963 | MenhirState961 | MenhirState944 | MenhirState956 | MenhirState946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5831 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce110 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5832)
    | MenhirState2604 | MenhirState2600 | MenhirState2596 | MenhirState2591 | MenhirState2587 | MenhirState2453 | MenhirState172 | MenhirState2435 | MenhirState1384 | MenhirState1381 | MenhirState1378 | MenhirState1374 | MenhirState1371 | MenhirState1326 | MenhirState1307 | MenhirState1304 | MenhirState1301 | MenhirState1297 | MenhirState1294 | MenhirState1290 | MenhirState1287 | MenhirState1284 | MenhirState1279 | MenhirState1276 | MenhirState1274 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5833 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1276
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1276
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1276) : 'freshtv5834)
    | MenhirState2626 | MenhirState2288 | MenhirState2283 | MenhirState2138 | MenhirState2121 | MenhirState2117 | MenhirState2120 | MenhirState2108 | MenhirState2097 | MenhirState2031 | MenhirState1461 | MenhirState1412 | MenhirState1437 | MenhirState1439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5835 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce133 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5836)
    | MenhirState2498 | MenhirState2495 | MenhirState2492 | MenhirState2488 | MenhirState2485 | MenhirState2481 | MenhirState2478 | MenhirState2475 | MenhirState2470 | MenhirState2467 | MenhirState1522 | MenhirState1520 | MenhirState1516 | MenhirState1514 | MenhirState1510 | MenhirState1508 | MenhirState1503 | MenhirState1501 | MenhirState1497 | MenhirState1490 | MenhirState1488 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5837 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1490
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1490
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1490) : 'freshtv5838)
    | MenhirState1762 | MenhirState1609 | MenhirState1588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5839 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce36 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5840)
    | MenhirState2660 | MenhirState1826 | MenhirState1819 | MenhirState1814 | MenhirState1780 | MenhirState1757 | MenhirState1751 | MenhirState1733 | MenhirState1644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5841 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce14 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5842)
    | MenhirState2051 | MenhirState1828 | MenhirState1795 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5843 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce36 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5844)
    | MenhirState2390 | MenhirState2163 | MenhirState2112 | MenhirState2111 | MenhirState2103 | MenhirState1963 | MenhirState2000 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5845 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce133 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5846)
    | MenhirState2389 | MenhirState2387 | MenhirState2358 | MenhirState2301 | MenhirState2297 | MenhirState2290 | MenhirState2286 | MenhirState2159 | MenhirState2019 | MenhirState1941 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5847 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce124 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5848)
    | MenhirState331 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5849 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce136 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5850)
    | MenhirState2704 | MenhirState2700 | MenhirState2695 | MenhirState2691 | MenhirState2684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5851 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce14 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5852)
    | _ ->
        _menhir_fail ()) : 'freshtv5854)

and _menhir_reduce455 : _menhir_env -> (('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_item_attribute = 
# 4301 "reason_parser.mly"
                                        ( (_2, _3) )
# 35453 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5819) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_item_attribute) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2742 | MenhirState2739 | MenhirState2733 | MenhirState2727 | MenhirState2711 | MenhirState2651 | MenhirState2649 | MenhirState2646 | MenhirState2639 | MenhirState2636 | MenhirState2634 | MenhirState2624 | MenhirState2626 | MenhirState2606 | MenhirState2318 | MenhirState437 | MenhirState2126 | MenhirState1400 | MenhirState2097 | MenhirState2090 | MenhirState2087 | MenhirState2082 | MenhirState2077 | MenhirState2072 | MenhirState2059 | MenhirState2052 | MenhirState2043 | MenhirState2040 | MenhirState2031 | MenhirState1629 | MenhirState1466 | MenhirState1462 | MenhirState1456 | MenhirState1454 | MenhirState1396 | MenhirState1386 | MenhirState1363 | MenhirState1154 | MenhirState1234 | MenhirState1176 | MenhirState1169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5813 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1176
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1176
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1176) : 'freshtv5814)
    | MenhirState2533 | MenhirState2527 | MenhirState394 | MenhirState399 | MenhirState2390 | MenhirState2309 | MenhirState2306 | MenhirState2304 | MenhirState2251 | MenhirState2019 | MenhirState1929 | MenhirState1926 | MenhirState1918 | MenhirState1876 | MenhirState1871 | MenhirState1866 | MenhirState1861 | MenhirState1836 | MenhirState1743 | MenhirState1739 | MenhirState1736 | MenhirState1729 | MenhirState1725 | MenhirState1668 | MenhirState1617 | MenhirState1600 | MenhirState1597 | MenhirState1589 | MenhirState1575 | MenhirState1557 | MenhirState1526 | MenhirState1524 | MenhirState1481 | MenhirState1479 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5815 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1481
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1481
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1481) : 'freshtv5816)
    | MenhirState1588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5817 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1589
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1589
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1589) : 'freshtv5818)
    | _ ->
        _menhir_fail ()) : 'freshtv5820)

and _menhir_reduce436 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_floating_attribute = 
# 4305 "reason_parser.mly"
                                            ((_2, _3))
# 35511 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5811) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_floating_attribute) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1604 | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5797) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5795) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__class_sig_field = 
# 2274 "reason_parser.mly"
      ( mkctf(Pctf_attribute _1) )
# 35538 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5796)) : 'freshtv5798)
    | MenhirState1747 | MenhirState1660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5801) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5799) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__class_field = 
# 1995 "reason_parser.mly"
      ( mkcf (Pcf_attribute _1) )
# 35559 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5800)) : 'freshtv5802)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5803) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce310 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5804)
    | MenhirState2557 | MenhirState2559 | MenhirState1246 | MenhirState1402 | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5807) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5805) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1735 "reason_parser.mly"
                       (
      mksig(Psig_attribute _1)
    )
# 35590 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5806)) : 'freshtv5808)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5809) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce310 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5810)
    | _ ->
        _menhir_fail ()) : 'freshtv5812)

and _menhir_reduce456 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_item_extension = 
# 4342 "reason_parser.mly"
                                                  ( (_2, _3) )
# 35614 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5793) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_item_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1604 | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5783 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1600
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1600
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1600) : 'freshtv5784)
    | MenhirState1747 | MenhirState1660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5785 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1743
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1743
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1743) : 'freshtv5786)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5787 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1926
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1926
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1926) : 'freshtv5788)
    | MenhirState2557 | MenhirState2559 | MenhirState1246 | MenhirState1402 | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5789 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2090
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2090
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2090) : 'freshtv5790)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5791 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2739
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2739
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2739) : 'freshtv5792)
    | _ ->
        _menhir_fail ()) : 'freshtv5794)

and _menhir_reduce265 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_signature)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 35703 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 35715 "reason_parser.ml"
      
    in
    
# 1553 "reason_parser.mly"
      ( unclosed_mty (with_txt _1 "{") (with_txt _3 "}"))
# 35721 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_functor_arg_name : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_functor_arg_name -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5769 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState448 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState448 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState448 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState448 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState448 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState448 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState448) : 'freshtv5766)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5767 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5768)) : 'freshtv5770)
    | MenhirState1245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5775 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5771 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2107 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2107 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2107 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2107 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2107 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2107 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2107) : 'freshtv5772)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5773 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5774)) : 'freshtv5776)
    | MenhirState1243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5781 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5777 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2116) : 'freshtv5778)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5780)) : 'freshtv5782)
    | _ ->
        _menhir_fail ()

and _menhir_reduce142 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_meth_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3941 "reason_parser.mly"
      ( let (f, c) = _2 in mktyp(Ptyp_object (f, c)) )
# 35856 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_signed_constant : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_signed_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState563 | MenhirState564 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5753 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error544 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOTDOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5751 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState545) : 'freshtv5752)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce271 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5754)
    | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5755 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce272 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5756)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState486 | MenhirState1880 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1628 | MenhirState1178 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState738 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5761 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv5757 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5758)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOTDOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5759 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState630) : 'freshtv5760)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce271 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5762)
    | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5763 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce272 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5764)
    | _ ->
        _menhir_fail ()

and _menhir_goto_attr_id : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_attr_id -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState71 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5711 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_reduce332 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5712)
    | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5713 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState73
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState73
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState73)) : 'freshtv5714)
    | MenhirState173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5715 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState174
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState174
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState174)) : 'freshtv5716)
    | MenhirState188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5717 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState189
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState189
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState189)) : 'freshtv5718)
    | MenhirState332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5719 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState333
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState333
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState333)) : 'freshtv5720)
    | MenhirState388 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5721 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_reduce332 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5722)
    | MenhirState336 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5737 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5735 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_item_extension_sugar = 
# 4331 "reason_parser.mly"
                    (
      ([], _2)
    )
# 36407 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5733) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
        match _menhir_s with
        | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1156 | MenhirState1171 | MenhirState1162 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5723 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1166) : 'freshtv5724)
        | MenhirState2667 | MenhirState1626 | MenhirState1657 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5725 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LET ->
                _menhir_run1627 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1658) : 'freshtv5726)
        | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5727 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1923) : 'freshtv5728)
        | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState475 | MenhirState2253 | MenhirState2247 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5729 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2249) : 'freshtv5730)
        | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5731 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2648 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2638 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run2629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run2622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run2577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2737) : 'freshtv5732)
        | _ ->
            _menhir_fail ()) : 'freshtv5734)) : 'freshtv5736)) : 'freshtv5738)
    | MenhirState395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5739 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState396
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState396
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState396)) : 'freshtv5740)
    | MenhirState438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5741 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState439
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState439
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState439)) : 'freshtv5742)
    | MenhirState452 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5743 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState453
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState453
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState453)) : 'freshtv5744)
    | MenhirState454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5745 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState455
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState455
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState455)) : 'freshtv5746)
    | MenhirState1468 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5747 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1469
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState1469
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1469)) : 'freshtv5748)
    | MenhirState2026 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5749 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2027
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2021 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState2027
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2027)) : 'freshtv5750)
    | _ ->
        _menhir_fail ()

and _menhir_goto_package_type : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error130 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce152 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5652)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error130 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5654)
    | MenhirState232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 37445 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error234 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5655 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 37457 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce290 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5656)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error234 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5658)
    | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5661 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error246 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5659 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce152 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5660)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error246 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5662)
    | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5665 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 37491 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error500 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5663 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 37503 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce290 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5664)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error500 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5666)
    | MenhirState770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error772 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5667 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce245 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5668)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error772 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5670)
    | MenhirState796 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5673 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error798 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5671 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce247 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5672)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error798 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5674)
    | MenhirState2383 | MenhirState1946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1947 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce258 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5676)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1947 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5678)
    | MenhirState2385 | MenhirState1949 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1950 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1952 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1952 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1952) : 'freshtv5680)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce256 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5682)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1950 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5684)
    | MenhirState1952 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5687 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1953 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce257 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5686)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1953 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5688)
    | MenhirState2354 | MenhirState2149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2150 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce258 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5690)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2150 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5692)
    | MenhirState2356 | MenhirState2152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2153 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2155) : 'freshtv5694)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce256 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5696)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2153 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5698)
    | MenhirState2155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2156 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce257 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5700)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2156 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5702)
    | MenhirState2167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5705 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2169 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5703 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce247 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5704)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2169 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5706)
    | MenhirState2362 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2364 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce245 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5708)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2364 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5710)
    | _ ->
        _menhir_fail ()

and _menhir_reduce275 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_type_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 37749 "reason_parser.ml"
      
    in
    
# 3346 "reason_parser.mly"
      ( mkpat(Ppat_type (_2)) )
# 37755 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce556 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_simple_core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__non_arrowed_simple_core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 37768 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5649) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2692 | MenhirState2676 | MenhirState2677 | MenhirState2663 | MenhirState2643 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState120 | MenhirState2466 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState169 | MenhirState2436 | MenhirState2433 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1793 | MenhirState1794 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1355 | MenhirState1353 | MenhirState1339 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1282 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState1271 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState271 | MenhirState269 | MenhirState244 | MenhirState253 | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5645) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce597 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5646)
    | MenhirState2515 | MenhirState106 | MenhirState2472 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1636 | MenhirState1637 | MenhirState1595 | MenhirState1586 | MenhirState1587 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState1494 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 | MenhirState328 | MenhirState243 | MenhirState316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5647) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce597 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5648)
    | _ ->
        _menhir_fail ()) : 'freshtv5650)

and _menhir_reduce555 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_non_simple_core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 37805 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5643) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2692 | MenhirState2663 | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5639) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce596 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5640)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5641) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce596 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5642)
    | _ ->
        _menhir_fail ()) : 'freshtv5644)

and _menhir_error616 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5637 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce277 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5638)

and _menhir_reduce276 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_lbl_pattern_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3348 "reason_parser.mly"
      ( let (fields, closed) = _2 in mkpat(Ppat_record(fields, closed)) )
# 37852 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error521 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5635 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce277 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5636)

and _menhir_reduce283 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3366 "reason_parser.mly"
      ( _2 )
# 37875 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce161 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__pattern_optional_constraint = 
# 3413 "reason_parser.mly"
                                ( _1 )
# 37887 "reason_parser.ml"
     in
    _menhir_goto__pattern_optional_constraint _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run668 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState668

and _menhir_reduce160 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__or_pattern = 
# 3080 "reason_parser.mly"
    ( mkpat(Ppat_or(_1, _3)) )
# 37957 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5633) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__or_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5629) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce545 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5630)
    | MenhirState2572 | MenhirState190 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5631) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce545 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5632)
    | _ ->
        _menhir_fail ()) : 'freshtv5634)

and _menhir_run594 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState594

and _menhir_run592 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState592

and _menhir_run591 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce553 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38114 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5627) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState666 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5623) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce398 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5624)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5625) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce398 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5626)
    | _ ->
        _menhir_fail ()) : 'freshtv5628)

and _menhir_reduce481 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)) = _menhir_stack in
    let _v : 'tv_lbl_pattern_list = 
# 3431 "reason_parser.mly"
                ( [_1], Closed )
# 38148 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error524 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce481 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce549 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern_not_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_pattern_not_ident) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__simple_pattern_not_ident_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38165 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5621) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5617) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce726 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5618)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1628 | MenhirState1178 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5619) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce726 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5620)
    | _ ->
        _menhir_fail ()) : 'freshtv5622)

and _menhir_goto_simple_non_labeled_expr_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_non_labeled_expr_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5613 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState837
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce267 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5614)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5615 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState950
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce267 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5616)
    | _ ->
        _menhir_fail ()

and _menhir_goto_labeled_simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1642 | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5605 * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce717 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5606)
    | MenhirState1641 | MenhirState815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5607) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce716 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5608)
    | MenhirState2682 | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5609 * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce717 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5610)
    | MenhirState2681 | MenhirState928 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5611) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce716 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5612)
    | _ ->
        _menhir_fail ()

and _menhir_goto_label_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1641 | MenhirState1642 | MenhirState815 | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5601) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_expr) = _v in
        (_menhir_reduce467 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5602)
    | MenhirState2681 | MenhirState2682 | MenhirState928 | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5603) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_expr) = _v in
        (_menhir_reduce467 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5604)
    | _ ->
        _menhir_fail ()

and _menhir_reduce722 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_non_labeled_expr_list_as_tuple = 
# 2825 "reason_parser.mly"
                                                                                                 (_1)
# 38364 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5599) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
    ((match _menhir_s with
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5595 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
        (_menhir_reduce72 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5596)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5597 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
        (_menhir_reduce72 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5598)
    | _ ->
        _menhir_fail ()) : 'freshtv5600)

and _menhir_error2183 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1136 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1111 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce415 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_optional_constraint), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3144 "reason_parser.mly"
    ( ([], Some _2) )
# 38414 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1095 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_expr_comma_seq : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_comma_seq -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5587 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1078
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1079 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078
          | BARRBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1078
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1078)) : 'freshtv5588)
    | MenhirState2176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5589 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2177
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1079 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2177
          | BARRBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2177
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2177)) : 'freshtv5590)
    | MenhirState414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5591 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2348
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1079 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2348
          | BARRBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2348
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2348)) : 'freshtv5592)
    | MenhirState177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5593 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2422
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1079 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422
          | BARRBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2422
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2422)) : 'freshtv5594)
    | _ ->
        _menhir_fail ()

and _menhir_reduce411 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_comma_list)), _endpos__2_, _), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_expr_comma_list = 
# 3127 "reason_parser.mly"
      ( _3 :: _1 )
# 38501 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce412 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _endpos__2_), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_expr_comma_list = 
# 3129 "reason_parser.mly"
      ( [_3; _1] )
# 38512 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1057 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_type_constraint_right_of_colon : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_constraint_right_of_colon -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1889 | MenhirState1201 | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5583 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_constraint_right_of_colon) = _v in
        (_menhir_reduce804 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5584)
    | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5585 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_constraint_right_of_colon) = _v in
        (_menhir_reduce804 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5586)
    | _ ->
        _menhir_fail ()

and _menhir_error1125 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce474 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)) = _menhir_stack in
    let _v : 'tv_lbl_expr_list = 
# 3169 "reason_parser.mly"
              ( [_1] )
# 38552 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1114 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce474 _menhir_env (Obj.magic _menhir_stack)

and _menhir_goto_record_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5569 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1120 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv5567 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce226 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5568)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1120 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5570)
    | MenhirState2196 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5573 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2198 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv5571 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce226 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5572)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2198 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5574)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5577 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2238 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5575 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce222 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5576)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2238 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5578)
    | MenhirState431 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5581 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2325 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5579 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce222 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5580)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2325 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5582)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_type_parameters : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_class_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5545 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38648 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5543 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38658 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1533 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1536 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1536 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1536 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1536 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1536) : 'freshtv5544)
        | MINUS ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce825 _menhir_env (Obj.magic _menhir_stack) MenhirState1533
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1533) : 'freshtv5546)
    | MenhirState1620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5551 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38693 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5547) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1621 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1781) : 'freshtv5548)
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5549 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38749 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1621 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1624 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1624 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1624) : 'freshtv5550)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run1623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce825 _menhir_env (Obj.magic _menhir_stack) MenhirState1621
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1621) : 'freshtv5552)
    | MenhirState2048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5555 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38824 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5553 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38834 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2049 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1784 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2050 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2050 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2050 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2050 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2050) : 'freshtv5554)
        | MINUS ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2049 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce825 _menhir_env (Obj.magic _menhir_stack) MenhirState2049
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2049) : 'freshtv5556)
    | MenhirState2055 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5559 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38869 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5557 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38879 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2056 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2057) : 'freshtv5558)
        | MINUS ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce825 _menhir_env (Obj.magic _menhir_stack) MenhirState2056
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2056) : 'freshtv5560)
    | MenhirState2656 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5565 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38935 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5561) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2657 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1798 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2692) : 'freshtv5562)
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5563 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 38991 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2657 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2675 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2671 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2659 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2658 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2658) : 'freshtv5564)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run1623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce825 _menhir_env (Obj.magic _menhir_stack) MenhirState2657
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2657) : 'freshtv5566)
    | _ ->
        _menhir_fail ()

and _menhir_goto_typevar_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_typevar_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1201 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5523 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5519 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1214) : 'freshtv5520)
        | QUOTE ->
            _menhir_run1212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5521 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5522)) : 'freshtv5524)
    | MenhirState152 | MenhirState1714 | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5529 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5525 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1324) : 'freshtv5526)
        | QUOTE ->
            _menhir_run1212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5527 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5528)) : 'freshtv5530)
    | MenhirState1694 | MenhirState1689 | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5535 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5531 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1573) : 'freshtv5532)
        | QUOTE ->
            _menhir_run1212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5533 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5534)) : 'freshtv5536)
    | MenhirState1889 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5541 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5537 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1898) : 'freshtv5538)
        | QUOTE ->
            _menhir_run1212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5539 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5540)) : 'freshtv5542)
    | _ ->
        _menhir_fail ()

and _menhir_reduce56 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_), _endpos__2_, _, _startpos__2_), _startpos__3_), _endpos__4_, _, (_4 : 'tv_ident), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__core_type = 
# 3856 "reason_parser.mly"
      ( mktyp(Ptyp_alias(_1, _4)) )
# 39285 "reason_parser.ml"
     in
    _menhir_goto__core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce585 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mty_longident = 
# 4152 "reason_parser.mly"
                                                ( Lident _1 )
# 39297 "reason_parser.ml"
     in
    _menhir_goto_mty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce586 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mty_longident = 
# 4153 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 39310 "reason_parser.ml"
     in
    _menhir_goto_mty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce589 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_name_tag = 
# 4179 "reason_parser.mly"
                                                ( _2 )
# 39323 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5517) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_name_tag) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5491 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState172
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERSAND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5487) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState172 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5485) = Obj.magic _menhir_stack in
              let (_endpos__1_ : Lexing.position) = _endpos in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_startpos__1_ : Lexing.position) = _startpos in
              ((let _1 = () in
              let _v : 'tv_opt_ampersand = 
# 3995 "reason_parser.mly"
                                                ( true )
# 39357 "reason_parser.ml"
               in
              _menhir_goto_opt_ampersand _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5486)) : 'freshtv5488)
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState172
          | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5489) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState172 in
              ((let _v : 'tv_opt_ampersand = 
# 3996 "reason_parser.mly"
                                                ( false )
# 39369 "reason_parser.ml"
               in
              _menhir_goto_opt_ampersand _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5490)
          | AMPERAMPER | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState172) : 'freshtv5492)
    | MenhirState2451 | MenhirState295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5495 * _menhir_state * 'tv_name_tag_list) * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5493 * _menhir_state * 'tv_name_tag_list) * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_name_tag_list)), _endpos__2_, _, (_2 : 'tv_name_tag), _startpos__2_) = _menhir_stack in
        let _v : 'tv_name_tag_list = 
# 4004 "reason_parser.mly"
                                                ( _2 :: _1 )
# 39383 "reason_parser.ml"
         in
        _menhir_goto_name_tag_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5494)) : 'freshtv5496)
    | MenhirState2450 | MenhirState294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5499 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5497 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
        let _v : 'tv_name_tag_list = 
# 4003 "reason_parser.mly"
                                                ( [_1] )
# 39395 "reason_parser.ml"
         in
        _menhir_goto_name_tag_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5498)) : 'freshtv5500)
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5501 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState550
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce274 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5502)
    | MenhirState532 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5503 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5504)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState486 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1633 | MenhirState1631 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5505 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5506)
    | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5507 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState646
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce274 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5508)
    | MenhirState2549 | MenhirState179 | MenhirState186 | MenhirState456 | MenhirState463 | MenhirState2262 | MenhirState1641 | MenhirState1642 | MenhirState835 | MenhirState837 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState806 | MenhirState800 | MenhirState745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5509 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5510)
    | MenhirState2681 | MenhirState2682 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState416 | MenhirState2344 | MenhirState418 | MenhirState474 | MenhirState729 | MenhirState730 | MenhirState740 | MenhirState744 | MenhirState759 | MenhirState889 | MenhirState890 | MenhirState948 | MenhirState950 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 | MenhirState788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5511 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5512)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5513 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState812
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce204 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5514)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5515 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState925
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce204 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5516)
    | _ ->
        _menhir_fail ()) : 'freshtv5518)

and _menhir_reduce828 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _v : 'tv_typevar_list = 
# 3731 "reason_parser.mly"
                                                ( [_2] )
# 39633 "reason_parser.ml"
     in
    _menhir_goto_typevar_list _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_reduce138 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3933 "reason_parser.mly"
      ( mktyp(Ptyp_var _2) )
# 39646 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce158 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3533 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _3), Contravariant )
# 39660 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce156 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3531 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _3), Covariant)
# 39674 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce154 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3529 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _2), Invariant )
# 39687 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce676 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 39694 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 902 "reason_parser.mly"
       (string * string option)
# 39700 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, (_2 : 'tv_primitive_declaration)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_primitive_declaration = 
# 3450 "reason_parser.mly"
                                                ( fst _1 :: _2 )
# 39706 "reason_parser.ml"
     in
    _menhir_goto_primitive_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce528 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_instance_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_instance_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cty__class_instance_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1161 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 39719 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5483) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1761 | MenhirState1536 | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5479) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce348 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5480)
    | MenhirState2701 | MenhirState2050 | MenhirState1820 | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5481) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce348 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5482)
    | _ ->
        _menhir_fail ()) : 'freshtv5484)

and _menhir_reduce520 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 39756 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5477) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1732 | MenhirState1657 | MenhirState1626 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5473) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce341 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5474)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5475) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce341 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5476)
    | _ ->
        _menhir_fail ()) : 'freshtv5478)

and _menhir_reduce523 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_def -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_fun_def) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_fun_def_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 39792 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5471) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
    ((match _menhir_s with
    | MenhirState1633 | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5467) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
        (_menhir_reduce345 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5468)
    | MenhirState2673 | MenhirState2671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5469) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
        (_menhir_reduce345 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5470)
    | _ ->
        _menhir_fail ()) : 'freshtv5472)

and _menhir_reduce406 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_curried_binding) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_curried_binding_return_typed_ = 
# 3023 "reason_parser.mly"
    (_1)
# 39823 "reason_parser.ml"
     in
    _menhir_goto_curried_binding_return_typed_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce62 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed_) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__curried_binding_return_typed = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3013 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         let (l, o, p) = _1 in mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 39840 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce63 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 39847 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_curried_binding_return_typed_) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 863 "reason_parser.mly"
       (string)
# 39853 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__curried_binding_return_typed = 
# 3018 "reason_parser.mly"
      ( mkexp(Pexp_newtype(_3, _5)) )
# 39863 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce404 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 39870 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_curried_binding_return_typed_) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 863 "reason_parser.mly"
       (string)
# 39876 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_curried_binding = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3061 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_newtype(_3, _5))
      )
# 39890 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce403 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed_) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_curried_binding = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3055 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let (l, o, p) = _1 in
        mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 39907 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2328 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce225 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5466)

and _menhir_reduce224 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_self_pattern_and_structure), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2721 "reason_parser.mly"
      ( mkexp (Pexp_object _2) )
# 39930 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2259 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5463 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce225 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5464)

and _menhir_error2668 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5461 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce49 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5462)

and _menhir_reduce48 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_simple_expr = 
# 1967 "reason_parser.mly"
                                           ( _2 )
# 39962 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1754 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce49 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5460)

and _menhir_reduce527 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_constructor_type) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cty__class_constructor_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1161 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 39983 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5457) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
    ((match _menhir_s with
    | MenhirState2663 | MenhirState1760 | MenhirState1775 | MenhirState1766 | MenhirState1768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5453) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
        (_menhir_reduce336 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5454)
    | MenhirState2692 | MenhirState2057 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1801 | MenhirState1803 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5455) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
        (_menhir_reduce336 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5456)
    | _ ->
        _menhir_fail ()) : 'freshtv5458)

and _menhir_run1546 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1546 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5449) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1546 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1552 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState1552
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1552) : 'freshtv5450)
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5451) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1546 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1547 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1547
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1547) : 'freshtv5452)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1546

and _menhir_run1562 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5433) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5427 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5425 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            let (_startpos__2_ : Lexing.position) = _startpos in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _2 = () in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4210 "reason_parser.mly"
                    ( Private, Virtual )
# 40088 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5426)) : 'freshtv5428)
        | LIDENT _ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5429 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4208 "reason_parser.mly"
            ( Private, Concrete )
# 40099 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5430)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5431 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            (raise _eRR : 'freshtv5432)) : 'freshtv5434)
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5443) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PRIVATE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5437 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5435 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            let (_startpos__2_ : Lexing.position) = _startpos in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _2 = () in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4211 "reason_parser.mly"
                    ( Private, Virtual )
# 40133 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5436)) : 'freshtv5438)
        | LIDENT _ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5439 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4209 "reason_parser.mly"
            ( Public, Virtual )
# 40144 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5440)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5441 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            (raise _eRR : 'freshtv5442)) : 'freshtv5444)
    | LIDENT _ ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5445) = Obj.magic _menhir_stack in
        ((let _v : 'tv_private_virtual_flags = 
# 4207 "reason_parser.mly"
                 ( Public, Concrete )
# 40159 "reason_parser.ml"
         in
        _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5446)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5447 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5448)

and _menhir_run1580 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1580 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1580 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1580 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1580 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1580

and _menhir_run1593 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1593

and _menhir_reduce380 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__constrained_class_declaration_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__constrained_class_declaration_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_constrained_class_declaration = 
# 1804 "reason_parser.mly"
                                                                                (_1)
# 40232 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5423) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_constrained_class_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5419 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 40245 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrained_class_declaration) = _v in
        (_menhir_reduce338 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5420)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5421 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 40256 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrained_class_declaration) = _v in
        (_menhir_reduce338 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5422)
    | _ ->
        _menhir_fail ()) : 'freshtv5424)

and _menhir_reduce29 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_return -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_return) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_binding = 
# 1866 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 40273 "reason_parser.ml"
     in
    _menhir_goto__class_fun_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1417 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5413 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5411 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _v : 'tv_with_type_binder = 
# 3725 "reason_parser.mly"
                   ( Private )
# 40299 "reason_parser.ml"
         in
        _menhir_goto_with_type_binder _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5412)) : 'freshtv5414)
    | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5415 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_with_type_binder = 
# 3724 "reason_parser.mly"
                   ( Public )
# 40310 "reason_parser.ml"
         in
        _menhir_goto_with_type_binder _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5416)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5417 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5418)

and _menhir_goto_potentially_long_ident_and_optional_type_parameters : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_potentially_long_ident_and_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5397 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5393 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1366 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState1366
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1366) : 'freshtv5394)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5395 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5396)) : 'freshtv5398)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5403 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5399 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2530 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState2530
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2530) : 'freshtv5400)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5401 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5402)) : 'freshtv5404)
    | MenhirState2578 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5409 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5405 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2580 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState2580
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2580) : 'freshtv5406)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5407 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5408)) : 'freshtv5410)
    | _ ->
        _menhir_fail ()

and _menhir_run1263 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5389 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1263 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce814 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5390)
    | FALSE ->
        _menhir_run1286 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5391 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1263 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON ->
            _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1286 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1278 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1264 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1281) : 'freshtv5392)
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1278 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1264 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1263 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState1263
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1263

and _menhir_reduce673 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 40525 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 40531 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)) = _menhir_stack in
    let _v : 'tv_potentially_long_ident_and_optional_type_parameters = 
# 3626 "reason_parser.mly"
                                  (
      let loc = mklocation _startpos__1_ _endpos__1_ in
      let lident_lident_loc = mkloc (Lident _1) loc in
      (lident_lident_loc, _2)
  )
# 40540 "reason_parser.ml"
     in
    _menhir_goto_potentially_long_ident_and_optional_type_parameters _menhir_env _menhir_stack _menhir_s _v

and _menhir_run106 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5385 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState106 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce814 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5386)
    | FALSE ->
        _menhir_run2477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run2473 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5387 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState106 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON ->
            _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run2477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2473 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run2469 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run108 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2472) : 'freshtv5388)
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run2469 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run108 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState106
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState106

and _menhir_reduce552 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__structure_item_without_item_extension_sugar) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1191 "reason_parser.mly"
 ( {x with pstr_loc = {x.pstr_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40657 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5383) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5379) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce791 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5380)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5381) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce791 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5382)
    | _ ->
        _menhir_fail ()) : 'freshtv5384)

and _menhir_reduce551 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__structure_item) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_str__structure_item_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1191 "reason_parser.mly"
 ( {x with pstr_loc = {x.pstr_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40694 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5377) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_str__structure_item_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5373) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce790 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5374)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5375) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce790 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5376)
    | _ ->
        _menhir_fail ()) : 'freshtv5378)

and _menhir_reduce530 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__curried_binding_return_typed -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__curried_binding_return_typed) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__curried_binding_return_typed_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40730 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5371) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
    ((match _menhir_s with
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5367) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
        (_menhir_reduce405 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5368)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5369) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
        (_menhir_reduce405 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5370)
    | _ ->
        _menhir_fail ()) : 'freshtv5372)

and _menhir_run1853 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constraints -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1853 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1853

and _menhir_run1353 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constraints -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1353 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1353

and _menhir_run1982 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5365 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1982 in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 40843 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5363 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 40855 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1983 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1984) : 'freshtv5364)
        | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
            _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1983) : 'freshtv5366)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1982

and _menhir_reduce570 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 40884 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 910 "reason_parser.mly"
       (string)
# 40890 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_ext_longident), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_mod_ext2 = 
# 4148 "reason_parser.mly"
                                            ( lapply (Lident(_1)) _3 )
# 40899 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce365 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 40906 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 863 "reason_parser.mly"
       (string)
# 40911 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_clty_longident = 
# 4157 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 40920 "reason_parser.ml"
     in
    _menhir_goto_clty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1250 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce155 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1251 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1251

and _menhir_run1253 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5357 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1255) : 'freshtv5358)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5359 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce157 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5360)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5361 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5362)

and _menhir_run1257 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5351 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1259) : 'freshtv5352)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5353 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce159 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5354)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5355 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5356)

and _menhir_run83 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 41021 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState83 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState83

and _menhir_run82 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 41038 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState82 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState82

and _menhir_reduce821 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 41057 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 863 "reason_parser.mly"
       (string)
# 41062 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_longident = 
# 4123 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 41071 "reason_parser.ml"
     in
    _menhir_goto_type_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run195 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5349 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState195 in
        let (_v : (
# 863 "reason_parser.mly"
       (string)
# 41089 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce821 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv5350)
    | UIDENT _v ->
        _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState195 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState195

and _menhir_reduce569 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext2), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_ext_longident), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_mod_ext2 = 
# 4147 "reason_parser.mly"
                                              ( lapply _1 _3 )
# 41111 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce568 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 41118 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 910 "reason_parser.mly"
       (string)
# 41124 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_mod_ext_longident), _startpos__5_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv_mod_ext2 = 
# 4146 "reason_parser.mly"
                                                                  ( lapply ( Ldot(_1, _3)) _5 )
# 41134 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run80 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 41141 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce570 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run81 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState81 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState81

and _menhir_reduce850 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_with_constraint) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_with_constraints = 
# 3686 "reason_parser.mly"
                                                ( [_1] )
# 41166 "reason_parser.ml"
     in
    _menhir_goto_with_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce851 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_with_constraint) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_with_constraints)), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_with_constraints = 
# 3687 "reason_parser.mly"
                                                ( _3 :: _1 )
# 41178 "reason_parser.ml"
     in
    _menhir_goto_with_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce541 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1181 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41191 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5347) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5343) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce584 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5344)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5345) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce584 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5346)
    | _ ->
        _menhir_fail ()) : 'freshtv5348)

and _menhir_reduce543 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__simple_module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1181 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41228 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5341) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5337) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce719 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5338)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5339) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce719 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5340)
    | _ ->
        _menhir_fail ()) : 'freshtv5342)

and _menhir_reduce539 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_module_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__simple_module_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1156 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41265 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5335) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2638 | MenhirState2639 | MenhirState410 | MenhirState2360 | MenhirState450 | MenhirState2282 | MenhirState751 | MenhirState2166 | MenhirState754 | MenhirState756 | MenhirState2136 | MenhirState2137 | MenhirState2132 | MenhirState2133 | MenhirState2122 | MenhirState2123 | MenhirState1240 | MenhirState1241 | MenhirState794 | MenhirState795 | MenhirState767 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5331) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce718 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5332)
    | MenhirState402 | MenhirState2389 | MenhirState403 | MenhirState2387 | MenhirState411 | MenhirState2358 | MenhirState2300 | MenhirState2301 | MenhirState2296 | MenhirState2297 | MenhirState2289 | MenhirState2290 | MenhirState2285 | MenhirState2286 | MenhirState757 | MenhirState2159 | MenhirState1470 | MenhirState2019 | MenhirState1939 | MenhirState1941 | MenhirState1942 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5333) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce718 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5334)
    | _ ->
        _menhir_fail ()) : 'freshtv5336)

and _menhir_reduce66 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_fun_def) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_labeled_simple_pattern), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2513 "reason_parser.mly"
                                       (
      let (l,o,p) = _2 in
      mkexp (Pexp_fun(l, o, p, _3))
    )
# 41305 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce67 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 41312 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ (_6 : 'tv_fun_def) ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, (_4 : (
# 863 "reason_parser.mly"
       (string)
# 41318 "reason_parser.ml"
    )), _startpos__4_), _endpos__5_, _startpos__5_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__expr = 
# 2518 "reason_parser.mly"
      ( mkexp (Pexp_newtype(_4, _6)) )
# 41329 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce440 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 41336 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_fun_def) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 863 "reason_parser.mly"
       (string)
# 41342 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3115 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_newtype(_3, _5))
      )
# 41356 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce439 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_fun_def) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3109 "reason_parser.mly"
      (
       let (l,o,p) = _1 in
       let loc = mklocation _symbolstartpos _endpos in
       mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 41373 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce489 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_leading_bar_match_case) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_leading_bar_match_cases = 
# 3068 "reason_parser.mly"
                           ( [_1] )
# 41383 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_cases _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce490 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_leading_bar_match_case) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_leading_bar_match_cases)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_leading_bar_match_cases = 
# 3069 "reason_parser.mly"
                                                   ( _2 :: _1 )
# 41394 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_cases _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1785 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5329 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce35 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5330)

and _menhir_reduce34 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_class_sig_body)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_instance_type = 
# 2221 "reason_parser.mly"
      ( mkcty(Pcty_signature _2) )
# 41417 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1542 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5327 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce35 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5328)

and _menhir_reduce180 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _, (_2 : 'tv_opt_semi)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__semi_terminated_seq_expr = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2380 "reason_parser.mly"
                          (
      let loc = mklocation _symbolstartpos _endpos in
      expr_of_let_bindings _1 @@ ghunit ~loc ()
    )
# 41443 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce181 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__semi_terminated_seq_expr_row = 
# 2391 "reason_parser.mly"
                                        (
      let expr = _1 in
      let item_attrs = _2 in
      (* Final item in the sequence - just append item attributes to the
       * expression attributes *)
      {expr with pexp_attributes = item_attrs @ expr.pexp_attributes}
    )
# 41461 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error706 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5325 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce282 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5326)

and _menhir_error701 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5323 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce279 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5324)

and _menhir_reduce278 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list_extension), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3353 "reason_parser.mly"
      ( let seq, ext_opt = _2 in
        let loc_rbracket = mklocation _startpos__2_ _endpos__2_ in
        make_real_pat (mktailpat_extension loc_rbracket (List.rev seq) ext_opt) )
# 41495 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error606 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5321 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce279 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5322)

and _menhir_reduce280 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3360 "reason_parser.mly"
      ( mkpat(Ppat_array(List.rev _2)) )
# 41518 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error599 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5319 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce282 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5320)

and _menhir_reduce576 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_module_binding_body_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body = 
# 1489 "reason_parser.mly"
                             ( _1 )
# 41537 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce537 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_binding_body_functor -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_binding_body_functor) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__module_binding_body_functor_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1156 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41549 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5317) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2317 | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5313) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
        (_menhir_reduce580 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5314)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5315) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
        (_menhir_reduce580 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5316)
    | _ ->
        _menhir_fail ()) : 'freshtv5318)

and _menhir_goto_field_expr_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_field_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5305 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState424
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run425 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState424
          | GREATERRBRACE ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState424
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState424)) : 'freshtv5306)
    | MenhirState1103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5307 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1104
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run425 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1104
          | GREATERRBRACE ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1104
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1104)) : 'freshtv5308)
    | MenhirState2191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5309 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2192
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run425 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192
          | GREATERRBRACE ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2192
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2192)) : 'freshtv5310)
    | MenhirState419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5311 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2333
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run425 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2333
          | GREATERRBRACE ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2333
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2333)) : 'freshtv5312)
    | _ ->
        _menhir_fail ()

and _menhir_reduce538 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__module_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1156 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41654 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5303) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__module_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2638 | MenhirState410 | MenhirState450 | MenhirState751 | MenhirState754 | MenhirState2136 | MenhirState2132 | MenhirState2122 | MenhirState1240 | MenhirState794 | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5299) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce582 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5300)
    | MenhirState402 | MenhirState403 | MenhirState411 | MenhirState2300 | MenhirState2296 | MenhirState2289 | MenhirState2285 | MenhirState757 | MenhirState1470 | MenhirState1939 | MenhirState1942 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5301) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce582 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5302)
    | _ ->
        _menhir_fail ()) : 'freshtv5304)

and _menhir_reduce542 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__non_arrowed_module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1181 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41691 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5297) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5293) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce595 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5294)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5295) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce595 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5296)
    | _ ->
        _menhir_fail ()) : 'freshtv5298)

and _menhir_reduce554 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__core_type2) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__core_type2_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41728 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5291) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__core_type2_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState666 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5287) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type2_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce399 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5288)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5289) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type2_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce399 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5290)
    | _ ->
        _menhir_fail ()) : 'freshtv5292)

and _menhir_goto_package_type_cstrs : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5279 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce645 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5280)
    | MenhirState2405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5281 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce648 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5282)
    | MenhirState133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5283 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce645 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5284)
    | MenhirState2464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5285 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce648 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5286)
    | _ ->
        _menhir_fail ()

and _menhir_reduce547 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_without_or -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__pattern_without_or) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__pattern_without_or_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41795 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5277) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5273) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce665 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5274)
    | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5275) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce665 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5276)
    | _ ->
        _menhir_fail ()) : 'freshtv5278)

and _menhir_error2523 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2511 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2423 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5271 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce229 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5272)

and _menhir_reduce228 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_comma_seq)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2732 "reason_parser.mly"
      ( mkexp (Pexp_array(List.rev _2)) )
# 41852 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2349 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5269 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce229 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5270)

and _menhir_error2334 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5267 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce239 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5268)

and _menhir_error2193 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5265 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce242 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5266)

and _menhir_error2188 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2178 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5263 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce232 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5264)

and _menhir_reduce819 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_), _endpos__4_, _, (_4 : 'tv_private_flag), _startpos__4_), _endpos__5_, _startpos__5_), _, (_6 : 'tv_label_declarations)), _, (_7 : 'tv_opt_comma)) = _menhir_stack in
    let _8 = () in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__8_ in
    let _v : 'tv_type_kind = 
# 3508 "reason_parser.mly"
      ( (Ptype_record(List.rev _6), _4, Some _2) )
# 41908 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1347 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce815 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos__3_, _startpos__3_), _, (_4 : 'tv_label_declarations)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_type_kind = 
# 3500 "reason_parser.mly"
      ( (Ptype_record(List.rev _4), _2, None) )
# 41927 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1334 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce241 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_field_expr_list)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 41953 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2779 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let exp = Exp.mk ~loc ~attrs:[] (Pexp_override(List.rev _4)) in
        mkexp(Pexp_open(Fresh, _1, exp))
      )
# 41966 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1105 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5261 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce242 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5262)

and _menhir_reduce416 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _, (_2 : 'tv_opt_comma)) = _menhir_stack in
    let _3 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3146 "reason_parser.mly"
    ( ([_1], None) )
# 41988 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1100 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce231 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_expr_comma_seq)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42014 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2738 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let rec_exp = Exp.mk ~loc ~attrs:[] (Pexp_array(List.rev _4)) in
        mkexp(Pexp_open(Fresh, _1, rec_exp))
      )
# 42027 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1090 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5259 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce232 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5260)

and _menhir_reduce238 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_field_expr_list)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2773 "reason_parser.mly"
      ( mkexp (Pexp_override(List.rev _2)) )
# 42050 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error427 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5257 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce239 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5258)

and _menhir_goto_tag_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_tag_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState157 | MenhirState160 | MenhirState2443 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState169 | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5243 * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        (_menhir_reduce688 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5244)
    | MenhirState305 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5249 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5245 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce145 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5246)
        | BAR ->
            _menhir_reduce688 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5247 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5248)) : 'freshtv5250)
    | MenhirState161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5255 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5251 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce145 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5252)
        | BAR ->
            _menhir_reduce688 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5253 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5254)) : 'freshtv5256)
    | _ ->
        _menhir_fail ()

and _menhir_reduce428 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42131 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4099 "reason_parser.mly"
                                                ( "::" )
# 42143 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42153 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42165 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce429 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42184 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4101 "reason_parser.mly"
                                                ( "false" )
# 42196 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42206 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42218 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce427 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let ((((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__2_, _), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42238 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4098 "reason_parser.mly"
                                                ( "()" )
# 42252 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42262 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42274 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce430 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42293 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( "true" )
# 42305 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42315 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42327 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce426 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 42334 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 910 "reason_parser.mly"
       (string)
# 42340 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42353 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4096 "reason_parser.mly"
                                                ( _1 )
# 42365 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42375 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42387 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce423 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4099 "reason_parser.mly"
                                                ( "::" )
# 42405 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42415 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3671 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42426 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce424 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4101 "reason_parser.mly"
                                                ( "false" )
# 42444 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42454 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3671 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42465 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce422 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4098 "reason_parser.mly"
                                                ( "()" )
# 42486 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42496 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3671 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42507 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce425 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( "true" )
# 42525 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42535 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3671 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42546 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce421 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 42553 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 910 "reason_parser.mly"
       (string)
# 42559 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4096 "reason_parser.mly"
                                                ( _1 )
# 42571 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42581 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3671 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42592 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_label_declarations : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_label_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1318 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5235 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1332 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1331
        | RBRACE ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1331
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1331) : 'freshtv5236)
    | MenhirState1345 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5237 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1332 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346
        | RBRACE ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1346
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1346) : 'freshtv5238)
    | MenhirState2509 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5239 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1332 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2510
        | RBRACE ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2510
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2510) : 'freshtv5240)
    | MenhirState2521 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5241 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1332 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522
        | RBRACE ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2522
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2522) : 'freshtv5242)
    | _ ->
        _menhir_fail ()

and _menhir_reduce385 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4099 "reason_parser.mly"
                                                ( "::" )
# 42674 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42684 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3573 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 42696 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce386 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4101 "reason_parser.mly"
                                                ( "false" )
# 42715 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42725 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3573 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 42737 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce384 : _menhir_env -> ((('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let ((((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4098 "reason_parser.mly"
                                                ( "()" )
# 42759 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42769 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3573 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 42781 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce387 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( "true" )
# 42800 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42810 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3573 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 42822 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce383 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 42829 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, (_100 : (
# 910 "reason_parser.mly"
       (string)
# 42835 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4096 "reason_parser.mly"
                                                ( _1 )
# 42848 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42858 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3573 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 42870 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce390 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4099 "reason_parser.mly"
                                                ( "::" )
# 42888 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42898 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3564 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 42911 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce391 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4101 "reason_parser.mly"
                                                ( "false" )
# 42929 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42939 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3564 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 42952 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce389 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _, _startpos__200_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4098 "reason_parser.mly"
                                                ( "()" )
# 42973 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42983 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3564 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 42996 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce392 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( "true" )
# 43014 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43024 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3564 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43037 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce335 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_attribute -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_attributes) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_attribute)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_attributes = 
# 4310 "reason_parser.mly"
                         ( _1 :: _2 )
# 43048 "reason_parser.ml"
     in
    _menhir_goto_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce388 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 43055 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 910 "reason_parser.mly"
       (string)
# 43061 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4096 "reason_parser.mly"
                                                ( _1 )
# 43073 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43083 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3564 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43096 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce812 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_constructor_declarations) ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3494 "reason_parser.mly"
      ( (Ptype_variant(_2),  Public, None) )
# 43108 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce816 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_constructor_declarations) ->
    let (((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_kind = 
# 3502 "reason_parser.mly"
      ( (Ptype_variant(_4), Public,  Some _2) )
# 43121 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce817 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_constructor_declarations) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_type_kind = 
# 3504 "reason_parser.mly"
      ( (Ptype_variant(_5), Private, Some _2) )
# 43135 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce813 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_constructor_declarations) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_kind = 
# 3496 "reason_parser.mly"
      ( (Ptype_variant(_3), Private, None) )
# 43148 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce783 : _menhir_env -> ((((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_nonrec_flag)), _, (_3 : 'tv_potentially_long_ident_and_optional_type_parameters)), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_private_flag), _startpos__5_), _, (_6 : 'tv_opt_bar)), _endpos__7_, _, (_7 : 'tv_str_extension_constructors)), _endpos__8_, _, (_8 : 'tv_post_item_attributes)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv_str_type_extension = 
# 3638 "reason_parser.mly"
  (
    if _2 <> Recursive then not_expecting 2 "nonrec flag";
    let (potentially_long_ident, optional_type_parameters) = _3 in
    Te.mk potentially_long_ident (List.rev _7)
          ~params:optional_type_parameters ~priv:_5 ~attrs:_8
  )
# 43167 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5233) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_str_type_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5229) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce299 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5230)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5231) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce299 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5232)
    | _ ->
        _menhir_fail ()) : 'freshtv5234)

and _menhir_reduce303 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43211 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1432 "reason_parser.mly"
                                                   (
      let item_attrs = _4 in
      let ident = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_modtype (Mtd.mk ident ~attrs:item_attrs ~loc))
    )
# 43224 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce304 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_module_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43245 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1438 "reason_parser.mly"
                                                                     (
      let ident = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_modtype (Mtd.mk ident ~typ:_5 ~attrs:_6 ~loc))
    )
# 43257 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2319 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 43264 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2313 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce301 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_nonlocal_module_binding_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1424 "reason_parser.mly"
                                                                    (
      let (ident, body) = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_module (Mb.mk ident body ~attrs:_4 ~loc))
    )
# 43291 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce515 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_nonlocal_module_binding_details)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv_many_nonlocal_module_bindings = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1493 "reason_parser.mly"
                                                                        (
    let (ident, body) = _4 in
    let loc = mklocation _symbolstartpos _endpos in
    [Mb.mk ident body ~attrs:_5 ~loc]
  )
# 43312 "reason_parser.ml"
     in
    _menhir_goto_many_nonlocal_module_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2127 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 43319 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_many_class_descriptions : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_descriptions -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5227 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5221) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2086
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2086) : 'freshtv5222)
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5223 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_descriptions), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1726 "reason_parser.mly"
                            (
      mksig(Psig_class (List.rev _1))
    )
# 43361 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5224)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5225 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5226)) : 'freshtv5228)

and _menhir_reduce308 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1454 "reason_parser.mly"
                                             (
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_include (Incl.mk _2 ~attrs:_3 ~loc))
    )
# 43386 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce296 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1407 "reason_parser.mly"
                              (
      mkstrexp _1 _2
    )
# 43400 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce309 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1458 "reason_parser.mly"
                                        (
    (* No sense in having item_extension_sugar for something that's already an
     * item_extension *)
    mkstr(Pstr_extension (_1, _2))
  )
# 43416 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce324 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_class_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1796 "reason_parser.mly"
                                                     (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc
  )
# 43434 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5219) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_class_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1874 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5215 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_declaration) = _v in
        (_menhir_reduce508 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5216)
    | MenhirState2731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5217 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_declaration) = _v in
        (_menhir_reduce508 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5218)
    | _ ->
        _menhir_fail ()) : 'freshtv5220)

and _menhir_reduce326 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_type_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_class_type_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2325 "reason_parser.mly"
                                                          (
    let (ident, instance_type, virt, class_type_params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident instance_type ~virt:virt ~params:class_type_params ~attrs:_3 ~loc
  )
# 43473 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5213) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_class_type_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1869 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5209 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_type_declaration) = _v in
        (_menhir_reduce512 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5210)
    | MenhirState2730 | MenhirState2080 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5211 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_type_declaration) = _v in
        (_menhir_reduce512 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5212)
    | _ ->
        _menhir_fail ()) : 'freshtv5214)

and _menhir_reduce329 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_nonlocal_module_binding_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_nonlocal_module_bindings = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1504 "reason_parser.mly"
                                                             (
    let (ident, body) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Mb.mk ident body ~attrs:_3 ~loc
  )
# 43512 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5207) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_nonlocal_module_bindings) = _v in
    ((match _menhir_s with
    | MenhirState1864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5203 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_nonlocal_module_bindings) = _v in
        (_menhir_reduce516 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5204)
    | MenhirState2725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5205 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_nonlocal_module_bindings) = _v in
        (_menhir_reduce516 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5206)
    | _ ->
        _menhir_fail ()) : 'freshtv5208)

and _menhir_reduce330 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_type_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_type_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3467 "reason_parser.mly"
                                                      (
    let (ident, params, constraints, kind, priv, manifest) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Type.mk ident
      ~params:params ~cstrs:constraints
      ~kind ~priv ?manifest ~attrs:_3 ~loc
  )
# 43553 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5201) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_type_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5197 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_type_declaration) = _v in
        (_menhir_reduce518 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5198)
    | MenhirState2724 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5199 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_type_declaration) = _v in
        (_menhir_reduce518 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5200)
    | _ ->
        _menhir_fail ()) : 'freshtv5202)

and _menhir_reduce507 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_many_class_declarations = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1785 "reason_parser.mly"
                                                         (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc]
  )
# 43593 "reason_parser.ml"
     in
    _menhir_goto_many_class_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__class_field : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5195) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_field) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5193) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_field) : 'tv__class_field) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cf__class_field_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1201 "reason_parser.mly"
 ( {x with pcf_loc = {x.pcf_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 43617 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5191) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cf__class_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5189) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cf__class_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5187) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cf__class_field_) : 'tv_mark_position_cf__class_field_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_field = 
# 1980 "reason_parser.mly"
                                            (_1)
# 43638 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5185) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_field) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5183 * Lexing.position * _menhir_state * 'tv_class_field) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_error1746 _menhir_env (Obj.magic _menhir_stack)
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | SEMI ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv5181 * Lexing.position * _menhir_state * 'tv_class_field) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          ((let _menhir_stack = (_menhir_stack, _endpos) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INHERIT ->
              _menhir_run1731 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INITIALIZER ->
              _menhir_run1727 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | METHOD ->
              _menhir_run1684 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | VAL ->
              _menhir_run1661 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              _menhir_reduce692 _menhir_env (Obj.magic _menhir_stack) MenhirState1747
          | _ ->
              assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1747) : 'freshtv5182)
      | RBRACE ->
          _menhir_reduce693 _menhir_env (Obj.magic _menhir_stack)
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_error1746 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5184)) : 'freshtv5186)) : 'freshtv5188)) : 'freshtv5190)) : 'freshtv5192)) : 'freshtv5194)) : 'freshtv5196)

and _menhir_reduce502 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_rec_flag)), _endpos__3_, _, (_3 : 'tv_let_binding_body)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_let_binding_impl = 
# 2879 "reason_parser.mly"
                                                       (
      (_2, _3, _4)
    )
# 43699 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5179) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_let_binding_impl) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2667 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1626 | MenhirState1657 | MenhirState1156 | MenhirState1171 | MenhirState1162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5171) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce494 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5172)
    | MenhirState2737 | MenhirState1658 | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5173 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce495 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5174)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState475 | MenhirState2253 | MenhirState2247 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5175) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce494 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5176)
    | MenhirState2249 | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5177 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce495 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5178)
    | _ ->
        _menhir_fail ()) : 'freshtv5180)

and _menhir_reduce511 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_type_declaration_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_many_class_type_declarations = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2314 "reason_parser.mly"
                                                                   (
    let (ident, instance_type, virt, class_type_params) = _3 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident instance_type ~virt:virt ~params:class_type_params ~attrs:_4 ~loc]
  )
# 43759 "reason_parser.ml"
     in
    _menhir_goto_many_class_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce777 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_declaration)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_exception_declaration = 
# 3582 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 43775 "reason_parser.ml"
     in
    _menhir_goto_str_exception_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce778 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_rebind)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_exception_declaration = 
# 3587 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 43791 "reason_parser.ml"
     in
    _menhir_goto_str_exception_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce297 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_), _endpos__6_, _, (_6 : 'tv_primitive_declaration)), _endpos__7_, _, (_7 : 'tv_post_item_attributes)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43812 "reason_parser.ml"
      
    in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1410 "reason_parser.mly"
                                                                                                (
      let loc = mklocation _symbolstartpos _endpos in
      mkstr
        (Pstr_primitive (Val.mk _2 _4 ~prim:_6 ~attrs:_7 ~loc))
    )
# 43824 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_module_rec_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_module_rec_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5169 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5163) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run1459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2076) : 'freshtv5164)
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5165 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_module_rec_declarations), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1708 "reason_parser.mly"
                                 (
      mksig(Psig_recmodule (List.rev _1))
    )
# 43862 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5166)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5167 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5168)) : 'freshtv5170)

and _menhir_reduce604 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_override_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_open_statement = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43888 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1742 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        Opn.mk _3 ~override:_2 ~attrs:_4 ~loc
      )
# 43899 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5161) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_open_statement) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5153) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce305 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5154)
    | MenhirState2557 | MenhirState2559 | MenhirState1246 | MenhirState1402 | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5157) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5155) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_open_statement) : 'tv_open_statement) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1719 "reason_parser.mly"
                   (
      mksig(Psig_open _1)
    )
# 43936 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5156)) : 'freshtv5158)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5159) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce305 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5160)
    | _ ->
        _menhir_fail ()) : 'freshtv5162)

and _menhir_goto__signature_item : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__signature_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5151) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__signature_item) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5149) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__signature_item) : 'tv__signature_item) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_sig__signature_item_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1186 "reason_parser.mly"
 ( {x with psig_loc = {x.psig_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 43969 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5147) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_sig__signature_item_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5145) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_sig__signature_item_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5143) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_sig__signature_item_) : 'tv_mark_position_sig__signature_item_) = _v in
    ((let _v : 'tv_signature_item = 
# 1672 "reason_parser.mly"
                                                   ( _1 )
# 43986 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5141) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_signature_item) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1246 | MenhirState2062 | MenhirState1402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5133 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5129 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2045 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2042 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2033 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2030 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run1398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | RBRACE ->
                _menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState2062
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2062) : 'freshtv5130)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5131 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5132)) : 'freshtv5134)
    | MenhirState2559 | MenhirState2557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5139 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5135 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2045 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2042 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2033 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2030 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run1398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2559 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EOF ->
                _menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState2559
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2559) : 'freshtv5136)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5137 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5138)) : 'freshtv5140)
    | _ ->
        _menhir_fail ()) : 'freshtv5142)) : 'freshtv5144)) : 'freshtv5146)) : 'freshtv5148)) : 'freshtv5150)) : 'freshtv5152)

and _menhir_reduce517 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_nonrec_flag)), _endpos__3_, _, (_3 : 'tv_type_declaration_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_many_type_declarations = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3456 "reason_parser.mly"
                                                                   (
      let (ident, params, constraints, kind, priv, manifest) = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      [Type.mk ident
       ~params:params ~cstrs:constraints
       ~kind ~priv ?manifest ~attrs:(add_nonrec _2 _4 2) ~loc]
  )
# 44108 "reason_parser.ml"
     in
    _menhir_goto_many_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce327 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_let_binding_body)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_let_binding = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2864 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         mklb _2 _3 loc
      )
# 44125 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5127) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_let_binding) = _v in
    ((match _menhir_s with
    | MenhirState2736 | MenhirState1656 | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5123 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_let_binding) = _v in
        (_menhir_reduce504 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5124)
    | MenhirState2246 | MenhirState1879 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5125 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_let_binding) = _v in
        (_menhir_reduce504 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5126)
    | _ ->
        _menhir_fail ()) : 'freshtv5128)

and _menhir_reduce672 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_attribute)), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_post_item_attributes = 
# 4315 "reason_parser.mly"
                                        ( _1 :: _2 )
# 44157 "reason_parser.ml"
     in
    _menhir_goto_post_item_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run1151 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5119 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1151 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5115 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 910 "reason_parser.mly"
       (string)
# 44184 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1242 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1239 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                _menhir_run1240 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1239
            | LPAREN ->
                _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1239 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1239) : 'freshtv5116)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5117 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5118)) : 'freshtv5120)
    | OPEN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5121 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1151 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BANG ->
            _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _ ->
            _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState1152
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1152) : 'freshtv5122)
    | REC ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1151 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce684 _menhir_env (Obj.magic _menhir_stack) MenhirState1151
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1151

and _menhir_error1155 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce531 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 44249 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5113) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5109) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce410 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5110)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5111) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce410 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5112)
    | _ ->
        _menhir_fail ()) : 'freshtv5114)

and _menhir_reduce524 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_simple_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_simple_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 44286 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5107) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1732 | MenhirState1657 | MenhirState1626 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5103) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce360 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5104)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5105) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce360 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5106)
    | _ ->
        _menhir_fail ()) : 'freshtv5108)

and _menhir_goto_toplevel_directive : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_toplevel_directive -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5091 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5087 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5085 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)) = _menhir_stack in
            let _2 = () in
            let _v : 'tv__toplevel_phrase = 
# 1129 "reason_parser.mly"
                                        ( _1 )
# 44336 "reason_parser.ml"
             in
            _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5086)) : 'freshtv5088)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5089 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5090)) : 'freshtv5092)
    | MenhirState2754 | MenhirState2751 | MenhirState2747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5101 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EOF ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5095 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5093 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)) = _menhir_stack in
            let _2 = () in
            let _v : 'tv__use_file = 
# 1138 "reason_parser.mly"
                                                   ( [_1] )
# 44362 "reason_parser.ml"
             in
            _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5094)) : 'freshtv5096)
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5097 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EOF ->
                _menhir_run2748 _menhir_env (Obj.magic _menhir_stack) MenhirState2751
            | EXCEPTION ->
                _menhir_run2648 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2638 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run2629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run2622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PERCENT ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run2613 _menhir_env (Obj.magic _menhir_stack) MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run2577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2751) : 'freshtv5098)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5099 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5100)) : 'freshtv5102)
    | _ ->
        _menhir_fail ()

and _menhir_run1977 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 44482 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5083 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 44494 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1977 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1978) : 'freshtv5084)
    | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1977

and _menhir_reduce216 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 44533 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2683 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let pat = mkpat (Ppat_var (mkloc "this" loc)) in
        mkexp(Pexp_open (Fresh, _1,
                         mkexp(Pexp_object(Cstr.mk pat [])))) )
# 44546 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run135 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 44553 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce464 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run555 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 44563 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce575 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run750 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5081 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState750 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState751 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState751 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState751) : 'freshtv5082)
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState750 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState750

and _menhir_run2176 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2176

and _menhir_run2181 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run2182 _menhir_env (Obj.magic _menhir_stack) MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2181 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2181

and _menhir_run2191 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2191

and _menhir_run2196 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOTDOT ->
        _menhir_run1110 _menhir_env (Obj.magic _menhir_stack) MenhirState2196 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2196 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5079 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2196 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce216 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5080)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2196 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2196

and _menhir_run635 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState635 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState635

and _menhir_run554 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState554 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState554

and _menhir_reduce465 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 44926 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 863 "reason_parser.mly"
       (string)
# 44932 "reason_parser.ml"
    )), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_longident = 
# 4119 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 44940 "reason_parser.ml"
     in
    _menhir_goto_label_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run11 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 44947 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce575 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_reduce350 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 44956 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 863 "reason_parser.mly"
       (string)
# 44961 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_class_longident = 
# 4161 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 44970 "reason_parser.ml"
     in
    _menhir_goto_class_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1249 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 44977 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MINUS ->
        _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState1249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1249

and _menhir_run91 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce155 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run92 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState92 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState92 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState92

and _menhir_run96 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5073 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState98 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState98 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState98) : 'freshtv5074)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5075 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce157 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5076)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5077 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5078)

and _menhir_run100 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5067 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState102 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState102 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState102) : 'freshtv5068)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5069 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce159 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5070)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5071 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5072)

and _menhir_run1625 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1625 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1625

and _menhir_run1626 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1651 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1627 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1626 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1626
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1626

and _menhir_run1631 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1631

and _menhir_run1635 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1635 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1635 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1635

and _menhir_reduce677 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_private_flag = 
# 4194 "reason_parser.mly"
                                                ( Public )
# 45222 "reason_parser.ml"
     in
    _menhir_goto_private_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1367 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce678 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_private_flag = 
# 4195 "reason_parser.mly"
                                                ( Private )
# 45241 "reason_parser.ml"
     in
    _menhir_goto_private_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce587 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_mutable_flag = 
# 4198 "reason_parser.mly"
                                                ( Immutable )
# 45252 "reason_parser.ml"
     in
    _menhir_goto_mutable_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce588 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mutable_flag = 
# 4199 "reason_parser.mly"
                                                ( Mutable )
# 45265 "reason_parser.ml"
     in
    _menhir_goto_mutable_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_class_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5047 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce237 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5048)
    | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5049 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5050)
    | MenhirState119 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5051 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState120
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce139 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5052)
    | MenhirState121 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5053 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5054)
    | MenhirState180 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5055 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce237 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5056)
    | MenhirState242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5057 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState243
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce139 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5058)
    | MenhirState1635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5059 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1636
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1636 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce15 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1636)) : 'freshtv5060)
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1626 | MenhirState1657 | MenhirState1732 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5061 * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce47 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5062)
    | MenhirState2675 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5063 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2676
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce15 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2676)) : 'freshtv5064)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5065 * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce47 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5066)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lident_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lident_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1203 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5027 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 45461 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5025 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 45467 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 45472 "reason_parser.ml"
        )), _startpos__1_), _, (_2 : 'tv_lident_list)) = _menhir_stack in
        let _v : 'tv_lident_list = 
# 2875 "reason_parser.mly"
                                      ( _1 :: _2 )
# 45477 "reason_parser.ml"
         in
        _menhir_goto_lident_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5026)) : 'freshtv5028)
    | MenhirState1202 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5033 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5029 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1206) : 'freshtv5030)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5031 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5032)) : 'freshtv5034)
    | MenhirState1715 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5039 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((((('freshtv5035 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1717 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1717) : 'freshtv5036)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((((('freshtv5037 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5038)) : 'freshtv5040)
    | MenhirState1890 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5045 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5041 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1892) : 'freshtv5042)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5043 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5044)) : 'freshtv5046)
    | _ ->
        _menhir_fail ()

and _menhir_run245 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState245

and _menhir_error2642 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2034 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1704 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1704

and _menhir_error1473 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1464 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1184 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5023 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1184 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5019 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 45813 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv5015 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 45825 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run1188 _menhir_env (Obj.magic _menhir_stack) MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run1184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1187) : 'freshtv5016)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv5017 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 45890 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5018)) : 'freshtv5020)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5022)) : 'freshtv5024)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1184

and _menhir_run1188 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1188

and _menhir_run1190 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1190

and _menhir_reduce834 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos__3_, _, (_3 : 'tv_val_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_val_longident = 
# 4108 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 46033 "reason_parser.ml"
     in
    _menhir_goto_val_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce164 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 46052 "reason_parser.ml"
      
    in
    
# 3261 "reason_parser.mly"
      ( mkpat(Ppat_alias(_1, _3)) )
# 46058 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce268 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 46076 "reason_parser.ml"
      
    in
    
# 3329 "reason_parser.mly"
      ( mkpat(Ppat_var _1) )
# 46082 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce833 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_val_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_val_longident = 
# 4107 "reason_parser.mly"
                                                ( Lident _1 )
# 46094 "reason_parser.ml"
     in
    _menhir_goto_val_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_single_attr_id : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_single_attr_id -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2026 | MenhirState1468 | MenhirState454 | MenhirState452 | MenhirState438 | MenhirState395 | MenhirState332 | MenhirState188 | MenhirState173 | MenhirState71 | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5009 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error70 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5007 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AND ->
                  _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AS ->
                  _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ASSERT ->
                  _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BEGIN ->
                  _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT ->
                  _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DO ->
                  _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DONE ->
                  _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOWNTO ->
                  _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ELSE ->
                  _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | END ->
                  _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTION ->
                  _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTOR ->
                  _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IN ->
                  _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INHERIT ->
                  _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INITIALIZER ->
                  _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | METHOD ->
                  _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MUTABLE ->
                  _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NONREC ->
                  _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OBJECT ->
                  _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | REC ->
                  _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SIG ->
                  _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRUCT ->
                  _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | THEN ->
                  _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TO ->
                  _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VAL ->
                  _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VIRTUAL ->
                  _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHEN ->
                  _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WITH ->
                  _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState71) : 'freshtv5008)
          | ASSERT | BACKQUOTE | BANG | CHAR _ | CLASS | COLON | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | IF | INCLUDE | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LET | LIDENT _ | LPAREN | MINUS | MINUSDOT | MODULE | NATIVEINT _ | NEW | OPEN | PERCENT | PLUS | PLUSDOT | PREFIXOP _ | QUESTION | RBRACKET | STRING _ | SWITCH | TRUE | TRY | TYPE | UIDENT _ | WHILE ->
              _menhir_reduce331 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error70 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5010)
    | MenhirState388 | MenhirState336 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5013 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error387 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5011 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AND ->
                  _menhir_run386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AS ->
                  _menhir_run385 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ASSERT ->
                  _menhir_run384 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BEGIN ->
                  _menhir_run383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT ->
                  _menhir_run381 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DO ->
                  _menhir_run380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DONE ->
                  _menhir_run379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOWNTO ->
                  _menhir_run378 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ELSE ->
                  _menhir_run377 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | END ->
                  _menhir_run376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run375 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run374 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run372 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run371 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTION ->
                  _menhir_run370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTOR ->
                  _menhir_run369 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run368 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IN ->
                  _menhir_run367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run366 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INHERIT ->
                  _menhir_run365 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INITIALIZER ->
                  _menhir_run364 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run363 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run362 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run361 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | METHOD ->
                  _menhir_run360 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MUTABLE ->
                  _menhir_run358 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run357 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NONREC ->
                  _menhir_run356 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OBJECT ->
                  _menhir_run355 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run354 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  _menhir_run352 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | REC ->
                  _menhir_run351 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SIG ->
                  _menhir_run350 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRUCT ->
                  _menhir_run349 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run348 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | THEN ->
                  _menhir_run347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TO ->
                  _menhir_run346 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run344 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run343 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run342 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VAL ->
                  _menhir_run341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VIRTUAL ->
                  _menhir_run340 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHEN ->
                  _menhir_run339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WITH ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState388 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState388) : 'freshtv5012)
          | ASSERT | BACKQUOTE | BANG | CHAR _ | CLASS | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | IF | INCLUDE | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LET | LIDENT _ | LPAREN | MINUS | MINUSDOT | MODULE | NATIVEINT _ | NEW | OPEN | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | TYPE | UIDENT _ | WHILE ->
              _menhir_reduce331 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error387 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5014)
    | _ ->
        _menhir_fail ()

and _menhir_reduce534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__simple_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 46362 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5005) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState463 | MenhirState2268 | MenhirState473 | MenhirState2262 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState745 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1641 | MenhirState1642 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState835 | MenhirState837 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState811 | MenhirState812 | MenhirState806 | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5001) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce715 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5002)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2681 | MenhirState2682 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState416 | MenhirState2344 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState730 | MenhirState2207 | MenhirState740 | MenhirState744 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState759 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState788 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState889 | MenhirState890 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState948 | MenhirState950 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState924 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5003) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce715 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5004)
    | _ ->
        _menhir_fail ()) : 'freshtv5006)

and _menhir_run1661 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1661 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4999) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1661 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            _menhir_run1319 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1662 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1662
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1662) : 'freshtv5000)
    | LIDENT _ | MUTABLE ->
        _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState1661
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1661

and _menhir_run1684 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1684 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ | PRIVATE | VIRTUAL ->
        _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState1684
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1684

and _menhir_run1727 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | EQUALGREATER ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4995 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1728) : 'freshtv4996)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4997 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4998)

and _menhir_run1731 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1731 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS | FUN | LBRACE | LBRACKETPERCENT | LIDENT _ | LPAREN | UIDENT _ ->
        _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState1731
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1731

and _menhir_run1738 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1738

and _menhir_goto_label_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4959 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error139 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4957 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState140) : 'freshtv4958)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error139 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4960)
    | MenhirState239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4963 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error240 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4961 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState241 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState241) : 'freshtv4962)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error240 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4964)
    | MenhirState516 | MenhirState615 | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4967 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error530 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4965 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState531) : 'freshtv4966)
          | COMMA | RBRACE ->
              _menhir_reduce480 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error530 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4968)
    | MenhirState919 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4973 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv4969 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4970)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4971 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1013) : 'freshtv4972)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4974)
    | MenhirState810 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4977 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1053 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4975 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1054) : 'freshtv4976)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4978)
    | MenhirState1126 | MenhirState1112 | MenhirState1115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4981 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1117 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4979 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1118) : 'freshtv4980)
          | COMMA | RBRACE ->
              _menhir_reduce473 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1117 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4982)
    | MenhirState431 | MenhirState475 | MenhirState2196 | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4985 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1128 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4983 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1129) : 'freshtv4984)
          | COMMA ->
              _menhir_reduce473 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1128 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4986)
    | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4987 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1415
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState1415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1415) : 'freshtv4988)
    | MenhirState1966 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1967
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState1967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1967) : 'freshtv4990)
    | MenhirState736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4991 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce212 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4992)
    | MenhirState472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4993 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce212 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4994)
    | _ ->
        _menhir_fail ()

and _menhir_reduce259 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_module_expr = 
# 1269 "reason_parser.mly"
      ( mkmod (Pmod_structure []) )
# 47181 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce443 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 47188 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 910 "reason_parser.mly"
       (string)
# 47194 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_functor_arg_name = 
# 1227 "reason_parser.mly"
               ( _1 )
# 47201 "reason_parser.ml"
     in
    _menhir_goto_functor_arg_name _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_functor_arg : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_functor_arg -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2295 | MenhirState753 | MenhirState2131 | MenhirState1938 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4951 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_functor_arg) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4949 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_functor_arg) : 'tv_functor_arg) = _v in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _v : 'tv_functor_args = 
# 1233 "reason_parser.mly"
      ( _2 :: _1 )
# 47224 "reason_parser.ml"
         in
        _menhir_goto_functor_args _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4950)) : 'freshtv4952)
    | MenhirState2631 | MenhirState2317 | MenhirState442 | MenhirState752 | MenhirState1239 | MenhirState1937 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4955) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_functor_arg) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4953) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_functor_arg) : 'tv_functor_arg) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _v : 'tv_functor_args = 
# 1235 "reason_parser.mly"
      ( [ _1 ] )
# 47242 "reason_parser.ml"
         in
        _menhir_goto_functor_args _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4954)) : 'freshtv4956)
    | _ ->
        _menhir_fail ()

and _menhir_reduce832 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_operator)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_val_ident = 
# 4064 "reason_parser.mly"
                                                ( _2 )
# 47258 "reason_parser.ml"
     in
    _menhir_goto_val_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run173 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState173 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState173

and _menhir_run332 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState332 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState332

and _menhir_reduce362 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_class_type_parameters = 
# 1880 "reason_parser.mly"
              ( [] )
# 47489 "reason_parser.ml"
     in
    _menhir_goto_class_type_parameters _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce202 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constant) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2638 "reason_parser.mly"
             ( mkexp (Pexp_constant _1) )
# 47500 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce704 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constant) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_signed_constant = 
# 4042 "reason_parser.mly"
                                           ( _1 )
# 47511 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_payload : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_payload -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4893 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4889 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce456 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4890)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4891 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4892)) : 'freshtv4894)
    | MenhirState2027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4899 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4895 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce436 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4896)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4897 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4898)) : 'freshtv4900)
    | MenhirState455 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4905 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4901 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce436 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4902)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4903 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4904)) : 'freshtv4906)
    | MenhirState453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4911 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4907 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce456 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4908)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4909 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4910)) : 'freshtv4912)
    | MenhirState439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4917 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4913 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce455 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4914)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4915 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4916)) : 'freshtv4918)
    | MenhirState396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4923 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4919 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce455 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4920)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4921 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4922)) : 'freshtv4924)
    | MenhirState333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4929 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4925 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce333 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4926)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4927 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4928)) : 'freshtv4930)
    | MenhirState189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4935 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4931 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce420 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4932)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4933 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4934)) : 'freshtv4936)
    | MenhirState174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4941 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4937 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce333 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4938)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4939 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4940)) : 'freshtv4942)
    | MenhirState73 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4947 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4943 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce420 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4944)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4945 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4946)) : 'freshtv4948)
    | _ ->
        _menhir_fail ()

and _menhir_reduce251 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_structure)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_expr = 
# 1243 "reason_parser.mly"
      ( mkmod(Pmod_structure(_2)) )
# 47722 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1955 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce259 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run109 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState109 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState109 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState109

and _menhir_run1486 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1486

and _menhir_run1487 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1487
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1487

and _menhir_run1492 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1492

and _menhir_error2561 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2100 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4887 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce265 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4888)

and _menhir_reduce264 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_signature)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_type = 
# 1551 "reason_parser.mly"
      ( mkmty(Pmty_signature _2) )
# 47833 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2094 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4885 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce265 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4886)

and _menhir_reduce702 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_signature_item)), _endpos__2_), _, (_3 : 'tv_signature)) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_signature = 
# 1669 "reason_parser.mly"
                                  ( _1 :: _3 )
# 47853 "reason_parser.ml"
     in
    _menhir_goto_signature _menhir_env _menhir_stack _menhir_s _v

and _menhir_run444 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4883) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_functor_arg_name = 
# 1228 "reason_parser.mly"
               ( "_" )
# 47871 "reason_parser.ml"
     in
    _menhir_goto_functor_arg_name _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4884)

and _menhir_run1244 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 47878 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | EQUALGREATER | LBRACKETAT | RPAREN | WITH ->
        _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)
    | COLON ->
        _menhir_reduce443 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1244

and _menhir_run401 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4879 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState402 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState402) : 'freshtv4880)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4881 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4882)

and _menhir_run1245 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1245

and _menhir_run1246 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2045 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2042 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2033 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2030 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState1246
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1246

and _menhir_goto_meth_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_meth_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState146 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4867 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4863 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce142 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4864)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4865 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4866)) : 'freshtv4868)
    | MenhirState283 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4873 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4869 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce142 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4870)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4871 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4872)) : 'freshtv4874)
    | MenhirState287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4877 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4875 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_field)), _endpos__2_, _), _, (_3 : 'tv_meth_list)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_meth_list = 
# 4018 "reason_parser.mly"
                                              ( let (f, c) = _3 in (_1 :: f, c) )
# 48047 "reason_parser.ml"
         in
        _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4876)) : 'freshtv4878)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_bar : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4849 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState157) : 'freshtv4850)
    | MenhirState158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4851 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState160 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState160) : 'freshtv4852)
    | MenhirState290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4853 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState291) : 'freshtv4854)
    | MenhirState300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4855 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState302) : 'freshtv4856)
    | MenhirState1368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4857 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1369 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1369 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1369) : 'freshtv4858)
    | MenhirState2531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4859 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1499 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1484 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2532) : 'freshtv4860)
    | MenhirState2581 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4861 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run2602 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run2598 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2593 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2582) : 'freshtv4862)
    | _ ->
        _menhir_fail ()

and _menhir_reduce714 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 48275 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 876 "reason_parser.mly"
       (nativeint)
# 48280 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4052 "reason_parser.mly"
                                           ( Const_nativeint _2 )
# 48289 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce713 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 48296 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 847 "reason_parser.mly"
       (int64)
# 48301 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4051 "reason_parser.mly"
                                           ( Const_int64 _2 )
# 48310 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce712 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 48317 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 846 "reason_parser.mly"
       (int32)
# 48322 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4050 "reason_parser.mly"
                                           ( Const_int32 _2 )
# 48331 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce710 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 48338 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 845 "reason_parser.mly"
       (int)
# 48343 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4048 "reason_parser.mly"
                                           ( Const_int _2 )
# 48352 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce711 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 48359 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 827 "reason_parser.mly"
       (string)
# 48364 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4049 "reason_parser.mly"
                                           ( Const_float _2 )
# 48373 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce709 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 48380 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 876 "reason_parser.mly"
       (nativeint)
# 48385 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4047 "reason_parser.mly"
                                           ( Const_nativeint(Nativeint.neg _2) )
# 48394 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce708 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 48401 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 847 "reason_parser.mly"
       (int64)
# 48406 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4046 "reason_parser.mly"
                                           ( Const_int64(Int64.neg _2) )
# 48415 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce707 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 48422 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 846 "reason_parser.mly"
       (int32)
# 48427 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4045 "reason_parser.mly"
                                           ( Const_int32(Int32.neg _2) )
# 48436 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce705 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 48443 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 845 "reason_parser.mly"
       (int)
# 48448 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4043 "reason_parser.mly"
                                           ( Const_int(- _2) )
# 48457 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce706 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 48464 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 827 "reason_parser.mly"
       (string)
# 48469 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4044 "reason_parser.mly"
                                           ( Const_float("-" ^ _2) )
# 48478 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce331 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_single_attr_id), _startpos_x0_) = _menhir_stack in
    let _endpos = _endpos_x0_ in
    let _v : 'tv_attr_id = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 48495 "reason_parser.ml"
      
    in
    
# 4294 "reason_parser.mly"
                           ( _1 )
# 48501 "reason_parser.ml"
     in
    _menhir_goto_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce644 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_package_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 48517 "reason_parser.ml"
      
    in
    
# 3970 "reason_parser.mly"
                          ( (_1, []) )
# 48523 "reason_parser.ml"
     in
    _menhir_goto_package_type _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_type_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2677 | MenhirState2676 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2466 | MenhirState120 | MenhirState157 | MenhirState160 | MenhirState2443 | MenhirState161 | MenhirState1794 | MenhirState1793 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1339 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1283 | MenhirState1278 | MenhirState1271 | MenhirState1264 | MenhirState312 | MenhirState305 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState254 | MenhirState253 | MenhirState169 | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4837 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4838)
    | MenhirState192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4839 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce275 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4840)
    | MenhirState211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4841 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce275 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4842)
    | MenhirState2692 | MenhirState2663 | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState142 | MenhirState152 | MenhirState2436 | MenhirState2433 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1808 | MenhirState1803 | MenhirState1801 | MenhirState1799 | MenhirState1760 | MenhirState1775 | MenhirState1773 | MenhirState1768 | MenhirState1766 | MenhirState1764 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1355 | MenhirState1353 | MenhirState1322 | MenhirState1324 | MenhirState1263 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState323 | MenhirState271 | MenhirState269 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4843 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState253
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce141 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4844)
    | MenhirState2515 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState1637 | MenhirState1636 | MenhirState1587 | MenhirState1586 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1494 | MenhirState1484 | MenhirState328 | MenhirState327 | MenhirState316 | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4845 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4846)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4847 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState327
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce141 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4848)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_simple_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2692 | MenhirState2676 | MenhirState2677 | MenhirState2663 | MenhirState2643 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState120 | MenhirState2466 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState169 | MenhirState2433 | MenhirState2436 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1793 | MenhirState1794 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1353 | MenhirState1355 | MenhirState1263 | MenhirState1339 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1282 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState1271 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1063 | MenhirState1066 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState244 | MenhirState269 | MenhirState271 | MenhirState253 | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4833) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce556 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4834)
    | MenhirState106 | MenhirState2515 | MenhirState2472 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1636 | MenhirState1637 | MenhirState1595 | MenhirState1586 | MenhirState1587 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState1494 | MenhirState1495 | MenhirState1492 | MenhirState1084 | MenhirState1087 | MenhirState1082 | MenhirState722 | MenhirState681 | MenhirState687 | MenhirState327 | MenhirState328 | MenhirState243 | MenhirState316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4835) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce556 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4836)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_non_simple_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2692 | MenhirState2663 | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState2433 | MenhirState2436 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1353 | MenhirState1355 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1063 | MenhirState1066 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState244 | MenhirState269 | MenhirState271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4829) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce555 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4830)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1084 | MenhirState1087 | MenhirState1082 | MenhirState722 | MenhirState681 | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4831) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce555 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4832)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_pattern_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4819 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error521 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4817 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce276 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4818)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error521 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4820)
    | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4823 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4821 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_), _, (_3 : 'tv_lbl_pattern_list)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_pattern_list = 
# 3435 "reason_parser.mly"
      ( let (fields, closed) = _3 in _1 :: fields, closed )
# 48698 "reason_parser.ml"
         in
        _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4822)) : 'freshtv4824)
    | MenhirState615 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4827 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error616 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4825 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce276 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4826)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error616 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4828)
    | _ ->
        _menhir_fail ()

and _menhir_reduce480 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_lbl_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 48734 "reason_parser.ml"
      
    in
    
# 3441 "reason_parser.mly"
      (
        (_1, pat_of_label _1)
      )
# 48742 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce271 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_signed_constant), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3338 "reason_parser.mly"
      ( mkpat(Ppat_constant _1) )
# 48754 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4715 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState593
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARRBRACKET | COMMA | SEMI ->
              _menhir_reduce657 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState593)) : 'freshtv4716)
    | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4717 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce160 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4718)
    | MenhirState511 | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4719 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState602
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARRBRACKET | COMMA | SEMI ->
              _menhir_reduce656 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState602)) : 'freshtv4720)
    | MenhirState611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4721 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState612
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState612 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce659 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState612)) : 'freshtv4722)
    | MenhirState610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4723 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState613
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce660 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce657 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState613)) : 'freshtv4724)
    | MenhirState514 | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4725 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState614
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState614 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce658 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce656 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState614)) : 'freshtv4726)
    | MenhirState676 | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4729 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState665
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4727 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState665 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState666) : 'freshtv4728)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4730)
    | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4731 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce160 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4732)
    | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4737 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState678
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState678 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState678 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState681) : 'freshtv4734)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4735 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState678 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce283 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4736)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4738)
    | MenhirState531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4739 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState699
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | RBRACE ->
              _menhir_reduce479 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState699)) : 'freshtv4740)
    | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState719
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState719 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState722) : 'freshtv4742)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState719 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce283 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4744)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4746)
    | MenhirState738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState739
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState739 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4747 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState739 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState740) : 'freshtv4748)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState739)) : 'freshtv4750)
    | MenhirState762 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4779 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState763
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER | WHEN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4777 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_pattern_with_bar = 
# 3082 "reason_parser.mly"
                               (
  _2
)
# 49132 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4775) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : 'tv_pattern_with_bar) = _v in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
              match _menhir_s with
              | MenhirState2546 | MenhirState2545 | MenhirState2428 | MenhirState2427 | MenhirState2341 | MenhirState2340 | MenhirState2143 | MenhirState761 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4757 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4751 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LET ->
                          _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PERCENT ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2141) : 'freshtv4752)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4753 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState765) : 'freshtv4754)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4755 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4756)) : 'freshtv4758)
              | MenhirState1033 | MenhirState893 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4765 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4759 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1031 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1031) : 'freshtv4760)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4761 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1027) : 'freshtv4762)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4763 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4764)) : 'freshtv4766)
              | MenhirState2229 | MenhirState476 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4773 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4767 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2227) : 'freshtv4768)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4769 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2223) : 'freshtv4770)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4771 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4772)) : 'freshtv4774)
              | _ ->
                  _menhir_fail ()) : 'freshtv4776)) : 'freshtv4778)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState763)) : 'freshtv4780)
    | MenhirState903 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4783 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState904
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4781 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState904 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState905 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState905) : 'freshtv4782)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState904)) : 'freshtv4784)
    | MenhirState1628 | MenhirState1178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4787 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1231
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4785 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1231 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1232) : 'freshtv4786)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1231)) : 'freshtv4788)
    | MenhirState1651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4793 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1652
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1652 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1652 in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__3_ in
              let _v : 'tv__class_self_pattern = 
# 1913 "reason_parser.mly"
      ( _2 )
# 49851 "reason_parser.ml"
               in
              _menhir_goto__class_self_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4790)) : 'freshtv4792)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1652)) : 'freshtv4794)
    | MenhirState2308 | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4797 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1915
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4795 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1915 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1916 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1916) : 'freshtv4796)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1915)) : 'freshtv4798)
    | MenhirState190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4803 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2415
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHEN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4799 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2415 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2416) : 'freshtv4800)
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4801 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_payload = 
# 4347 "reason_parser.mly"
                     ( PPat (_2, None) )
# 50049 "reason_parser.ml"
               in
              _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4802)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2415)) : 'freshtv4804)
    | MenhirState2572 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4815 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2573
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState2573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4813 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2573 in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4811 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_ : _menhir_state) = _menhir_s in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 50078 "reason_parser.ml"
              ) = 
# 1148 "reason_parser.mly"
                ( apply_mapper_chain_to_pattern _1 default_mapper_chain)
# 50082 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4809) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 50090 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4807) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 50098 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4805) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 50106 "reason_parser.ml"
              )) : (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 50110 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv4806)) : 'freshtv4808)) : 'freshtv4810)) : 'freshtv4812)) : 'freshtv4814)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2573)) : 'freshtv4816)
    | _ ->
        _menhir_fail ()

and _menhir_goto_pattern_comma_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState590
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState590
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState590
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState590) : 'freshtv4706)
    | MenhirState514 | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4711 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState609
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4709 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState609 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4707 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_menhir_s : _menhir_state) = MenhirState610 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | EXCEPTION ->
                      _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState611) : 'freshtv4708)
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState610) : 'freshtv4710)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState609)) : 'freshtv4712)
    | MenhirState511 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4713 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState705
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState705
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState705
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState705) : 'freshtv4714)
    | _ ->
        _menhir_fail ()

and _menhir_goto_pattern_comma_list_extension : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern_comma_list_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState605
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState605
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState605) : 'freshtv4702)
    | MenhirState514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState700
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState700
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState700) : 'freshtv4704)
    | _ ->
        _menhir_fail ()

and _menhir_goto__core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState152 | MenhirState2433 | MenhirState2436 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1353 | MenhirState1355 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1063 | MenhirState1066 | MenhirState1061 | MenhirState666 | MenhirState244 | MenhirState269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4697) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce553 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4698)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1084 | MenhirState1087 | MenhirState1082 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4699) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce553 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4700)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_pattern : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4695 * _menhir_state * 'tv_lbl_pattern) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_error524 _menhir_env (Obj.magic _menhir_stack)
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | COMMA ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv4693 * _menhir_state * 'tv_lbl_pattern) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          ((let _menhir_stack = (_menhir_stack, _endpos) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LIDENT _v ->
              _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState525 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState525 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4691 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState525 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COMMA ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4689) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState526 in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4690)
              | RBRACE ->
                  _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState526
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState526) : 'freshtv4692)
          | RBRACE ->
              _menhir_reduce482 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState525) : 'freshtv4694)
      | RBRACE ->
          _menhir_reduce481 _menhir_env (Obj.magic _menhir_stack)
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_error524 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4696)

and _menhir_goto__simple_pattern_not_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern_not_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4685) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern_not_ident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce549 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4686)
    | MenhirState2657 | MenhirState2698 | MenhirState2671 | MenhirState2673 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState476 | MenhirState2232 | MenhirState480 | MenhirState484 | MenhirState2218 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1881 | MenhirState1906 | MenhirState1885 | MenhirState1621 | MenhirState1817 | MenhirState1697 | MenhirState1701 | MenhirState1708 | MenhirState1651 | MenhirState1631 | MenhirState1633 | MenhirState1628 | MenhirState1178 | MenhirState1179 | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 | MenhirState893 | MenhirState1036 | MenhirState897 | MenhirState901 | MenhirState1022 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4687) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern_not_ident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce549 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4688)
    | _ ->
        _menhir_fail ()

and _menhir_run929 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 50422 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4681 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50434 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState933) : 'freshtv4682)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4683 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50493 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState930) : 'freshtv4684)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce720 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_non_labeled_expr_list = 
# 2820 "reason_parser.mly"
      ( [_1] )
# 50555 "reason_parser.ml"
     in
    _menhir_goto_simple_non_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce721 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_non_labeled_expr_list), _startpos__1_), _endpos__2_, _, (_2 : 'tv_less_aggressive_simple_expression), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_non_labeled_expr_list = 
# 2822 "reason_parser.mly"
      ( _2 :: _1 )
# 50567 "reason_parser.ml"
     in
    _menhir_goto_simple_non_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce466 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_labeled_simple_expr = 
# 2841 "reason_parser.mly"
      ( ("", _1) )
# 50578 "reason_parser.ml"
     in
    _menhir_goto_labeled_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce462 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50585 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 50591 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_less_aggressive_simple_expression), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_expr = 
# 2849 "reason_parser.mly"
      ( (_1, _3) )
# 50598 "reason_parser.ml"
     in
    _menhir_goto_label_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce463 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50605 "reason_parser.ml"
) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 50611 "reason_parser.ml"
    )), _startpos__1_), _endpos__3_, _, (_3 : 'tv_less_aggressive_simple_expression), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_expr = 
# 2854 "reason_parser.mly"
      ( ("?" ^ _1, _3) )
# 50618 "reason_parser.ml"
     in
    _menhir_goto_label_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run816 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 50625 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4677 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50637 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState820) : 'freshtv4678)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4679 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 50696 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState817) : 'freshtv4680)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_non_labeled_expr_list_as_tuple -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_non_labeled_expr_list_as_tuple) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 50758 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4675) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
    ((match _menhir_s with
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4671) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
        (_menhir_reduce722 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4672)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4673) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
        (_menhir_reduce722 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4674)
    | _ ->
        _menhir_fail ()) : 'freshtv4676)

and _menhir_reduce212 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 50798 "reason_parser.ml"
      
    in
    
# 2669 "reason_parser.mly"
      ( mkexp(Pexp_field(_1, _3)) )
# 50804 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_expr_optional_constraint : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_optional_constraint -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState750 | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4629 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1057 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4627 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1058 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1058) : 'freshtv4628)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1057 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4630)
    | MenhirState1058 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4631 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce412 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4632)
    | MenhirState1071 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4633 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce411 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4634)
    | MenhirState177 | MenhirState414 | MenhirState2176 | MenhirState1076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4637 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4635 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_) = _menhir_stack in
        let _v : 'tv_expr_comma_seq = 
# 3138 "reason_parser.mly"
      ( [_1] )
# 50919 "reason_parser.ml"
         in
        _menhir_goto_expr_comma_seq _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4636)) : 'freshtv4638)
    | MenhirState1079 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4641 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4639 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_comma_seq)), _endpos__2_, _), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_expr_comma_seq = 
# 3136 "reason_parser.mly"
      ( _3::_1 )
# 50932 "reason_parser.ml"
         in
        _menhir_goto_expr_comma_seq _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4640)) : 'freshtv4642)
    | MenhirState1094 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4645 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1095 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4643 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce415 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4644)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1095 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4646)
    | MenhirState417 | MenhirState1098 | MenhirState1093 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4649 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1097
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4647 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1097 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  _menhir_run1094 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACKET ->
                  _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1098) : 'freshtv4648)
          | RBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1097
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1097)) : 'freshtv4650)
    | MenhirState1110 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4653 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1111 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4651 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1112 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1112 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1112) : 'freshtv4652)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1111 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4654)
    | MenhirState15 | MenhirState183 | MenhirState407 | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4657 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1136 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4655 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1137) : 'freshtv4656)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1136 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4658)
    | MenhirState1137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4659 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce412 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4660)
    | MenhirState1142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4661 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce411 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4662)
    | MenhirState2182 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4665 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2183 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4663 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce415 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4664)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2183 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4666)
    | MenhirState415 | MenhirState2186 | MenhirState2181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4669 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2185
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4667 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2185 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  _menhir_run2182 _menhir_env (Obj.magic _menhir_stack) MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACKET ->
                  _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2186) : 'freshtv4668)
          | RBRACKET ->
              _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2185
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2185)) : 'freshtv4670)
    | _ ->
        _menhir_fail ()

and _menhir_reduce806 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_type_constraint_right_of_colon = 
# 3241 "reason_parser.mly"
                                          ( (Some _1, None) )
# 51301 "reason_parser.ml"
     in
    _menhir_goto_type_constraint_right_of_colon _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_lbl_expr_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState1112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4617 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4615 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_optional_constraint), _startpos__2_), _endpos__3_) = _menhir_stack in
        let _3 = () in
        let _1 = () in
        let _v : 'tv_record_expr = 
# 3165 "reason_parser.mly"
                                                             ( (Some _2, _4) )
# 51323 "reason_parser.ml"
         in
        _menhir_goto_record_expr _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4616)) : 'freshtv4618)
    | MenhirState1115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4621 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4619 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_expr_list = 
# 3170 "reason_parser.mly"
                                  ( _1 :: _3 )
# 51340 "reason_parser.ml"
         in
        _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4620)) : 'freshtv4622)
    | MenhirState1126 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4625 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4623 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field = 
# 3206 "reason_parser.mly"
      ( _1::_3 )
# 51357 "reason_parser.ml"
         in
        _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4624)) : 'freshtv4626)
    | _ ->
        _menhir_fail ()

and _menhir_reduce473 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_lbl_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 51375 "reason_parser.ml"
      
    in
    
# 3177 "reason_parser.mly"
      (
        (_1, exp_of_label _1)
      )
# 51383 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_lbl_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1126 | MenhirState1115 | MenhirState1112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4609 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1114 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4607 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACE ->
                  _menhir_reduce476 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1115) : 'freshtv4608)
          | RBRACE ->
              _menhir_reduce474 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1114 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4610)
    | MenhirState431 | MenhirState475 | MenhirState2196 | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4613 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1125 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4611 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1126 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1126 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1126) : 'freshtv4612)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1125 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4614)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr_list_that_is_not_a_single_punned_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4605) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4603) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) = _v in
    ((let _v : 'tv_record_expr = 
# 3166 "reason_parser.mly"
                                                            ( (None, _1) )
# 51467 "reason_parser.ml"
     in
    _menhir_goto_record_expr _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4604)) : 'freshtv4606)

and _menhir_goto_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState92 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4499 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce154 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4500)
    | MenhirState98 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4501 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce156 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4502)
    | MenhirState102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4503 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce158 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4504)
    | MenhirState115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4505 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4506)
    | MenhirState123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4507 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4508)
    | MenhirState153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4511 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ('freshtv4509 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4510)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT | QUOTE ->
              _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce138 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4512)
    | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4513 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4514)
    | MenhirState249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4515 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce586 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4516)
    | MenhirState2625 | MenhirState128 | MenhirState2385 | MenhirState2383 | MenhirState2362 | MenhirState2356 | MenhirState2354 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2167 | MenhirState2152 | MenhirState2155 | MenhirState2149 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1949 | MenhirState1952 | MenhirState1946 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 | MenhirState796 | MenhirState770 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4517 * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce585 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4518)
    | MenhirState276 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4519 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce56 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4520)
    | MenhirState398 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState399
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState399 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState400 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState400 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState400) : 'freshtv4522)
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState399
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState399) : 'freshtv4524)
    | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4525 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4526)
    | MenhirState504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4527 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce586 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4528)
    | MenhirState232 | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1959 | MenhirState1999 | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4529 * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce585 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4530)
    | MenhirState692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4531 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce56 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4532)
    | MenhirState1212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4535 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4533 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_typevar_list), _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_typevar_list = 
# 3732 "reason_parser.mly"
                                                ( _3 :: _1 )
# 51590 "reason_parser.ml"
         in
        _menhir_goto_typevar_list _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4534)) : 'freshtv4536)
    | MenhirState1251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4537 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce154 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4538)
    | MenhirState1255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4539 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce156 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4540)
    | MenhirState1259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4541 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce158 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4542)
    | MenhirState1399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4545 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1400
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1400 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1401) : 'freshtv4544)
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1400
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1400) : 'freshtv4546)
    | MenhirState1570 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4549 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ('freshtv4547 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
          (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4548)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT | QUOTE ->
              _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce138 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4550)
    | MenhirState1611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4577 * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4575 * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__type_variable = 
# 3548 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _2) )
# 51668 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4573) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv__type_variable) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4571) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv__type_variable) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4569) = Obj.magic _menhir_stack in
        let (_endpos_x_ : Lexing.position) = _endpos in
        let ((x : 'tv__type_variable) : 'tv__type_variable) = _v in
        let (_startpos_x_ : Lexing.position) = _startpos in
        ((let _v : 'tv_mark_position_typ__type_variable_ = let _endpos = _endpos_x_ in
        let _symbolstartpos = _startpos_x_ in
        
# 1176 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 51690 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4567) = _menhir_stack in
        let (_v : 'tv_mark_position_typ__type_variable_) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4565) = Obj.magic _menhir_stack in
        let (_v : 'tv_mark_position_typ__type_variable_) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4563) = Obj.magic _menhir_stack in
        let ((_1 : 'tv_mark_position_typ__type_variable_) : 'tv_mark_position_typ__type_variable_) = _v in
        ((let _v : 'tv_type_variable = 
# 3546 "reason_parser.mly"
                                                 (_1)
# 51704 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4561) = _menhir_stack in
        let (_v : 'tv_type_variable) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4559 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        let (_v : 'tv_type_variable) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4557 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        let ((_2 : 'tv_type_variable) : 'tv_type_variable) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_type_variance)) = _menhir_stack in
        let _v : 'tv_type_parameter = 
# 3538 "reason_parser.mly"
                                                  ( _2, _1 )
# 51719 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4555) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_parameter) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4553 * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_parameter) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4551 * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_type_parameter) : 'tv_type_parameter) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_class_type_parameters)) = _menhir_stack in
        let _v : 'tv_class_type_parameters = 
# 1881 "reason_parser.mly"
                                         (
      _2::_1
    )
# 51739 "reason_parser.ml"
         in
        _menhir_goto_class_type_parameters _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4552)) : 'freshtv4554)) : 'freshtv4556)) : 'freshtv4558)) : 'freshtv4560)) : 'freshtv4562)) : 'freshtv4564)) : 'freshtv4566)) : 'freshtv4568)) : 'freshtv4570)) : 'freshtv4572)) : 'freshtv4574)) : 'freshtv4576)) : 'freshtv4578)
    | MenhirState2613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4597 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2614
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | FALSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4581 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2614 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4579 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4172 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_bool false) )
# 51768 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4580)) : 'freshtv4582)
          | INT _v ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4585 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2614 in
              let (_v : (
# 845 "reason_parser.mly"
       (int)
# 51779 "reason_parser.ml"
              )) = _v in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4583 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let ((_3 : (
# 845 "reason_parser.mly"
       (int)
# 51790 "reason_parser.ml"
              )) : (
# 845 "reason_parser.mly"
       (int)
# 51794 "reason_parser.ml"
              )) = _v in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4169 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_int _3) )
# 51802 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4584)) : 'freshtv4586)
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2614 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2614 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4589 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2614 in
              let (_v : (
# 902 "reason_parser.mly"
       (string * string option)
# 51817 "reason_parser.ml"
              )) = _v in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4587 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let ((_3 : (
# 902 "reason_parser.mly"
       (string * string option)
# 51828 "reason_parser.ml"
              )) : (
# 902 "reason_parser.mly"
       (string * string option)
# 51832 "reason_parser.ml"
              )) = _v in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4168 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_string (fst _3)) )
# 51840 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4588)) : 'freshtv4590)
          | TRUE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4593 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2614 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4591 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4173 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_bool true) )
# 51861 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4592)) : 'freshtv4594)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2614 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4595 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4167 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_none) )
# 51874 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4596)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2614)) : 'freshtv4598)
    | MenhirState2623 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4601 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2624
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4599 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2624 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2625 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2625 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2625 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2625) : 'freshtv4600)
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2624
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2624) : 'freshtv4602)
    | _ ->
        _menhir_fail ()

and _menhir_goto_primitive_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_primitive_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4489 * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 51930 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce676 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4490)
    | MenhirState1476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1479
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1479
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1479) : 'freshtv4492)
    | MenhirState2038 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4493 * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 51950 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce676 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4494)
    | MenhirState2037 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4495 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2040
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2040
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2040) : 'freshtv4496)
    | MenhirState2645 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4497 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2646
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2646
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2646) : 'freshtv4498)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_sig_fields : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_sig_fields -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4487 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1603
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | SEMI ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv4485 * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_menhir_s : _menhir_state) = MenhirState1603 in
          ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1593 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INHERIT ->
              _menhir_run1580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | METHOD ->
              _menhir_run1562 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | VAL ->
              _menhir_run1546 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4486)
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1603) : 'freshtv4488)

and _menhir_goto__class_instance_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_instance_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1761 | MenhirState1536 | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4481) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_instance_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce528 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4482)
    | MenhirState2701 | MenhirState2050 | MenhirState1820 | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4483) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_instance_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce528 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4484)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_sig_field : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_sig_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4479) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_sig_field) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4477) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_sig_field) : 'tv__class_sig_field) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_ctf__class_sig_field_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1166 "reason_parser.mly"
 ( {x with pctf_loc = {x.pctf_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52058 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4475) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4473) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4471) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_ctf__class_sig_field_) : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_sig_field = 
# 2253 "reason_parser.mly"
                                                      (_1)
# 52079 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4469) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_sig_field) = _v in
    ((match _menhir_s with
    | MenhirState1604 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4463 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_field) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4461 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_class_sig_field) : 'tv_class_sig_field) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_sig_fields)), _endpos__2_, _) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_class_sig_fields = 
# 2250 "reason_parser.mly"
                                        ( _3 :: _1 )
# 52104 "reason_parser.ml"
         in
        _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4462)) : 'freshtv4464)
    | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4467) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_field) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4465) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_class_sig_field) : 'tv_class_sig_field) = _v in
        ((let _endpos = _endpos__1_ in
        let _v : 'tv_class_sig_fields = 
# 2249 "reason_parser.mly"
                  ( [_1] )
# 52122 "reason_parser.ml"
         in
        _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4466)) : 'freshtv4468)
    | _ ->
        _menhir_fail ()) : 'freshtv4470)) : 'freshtv4472)) : 'freshtv4474)) : 'freshtv4476)) : 'freshtv4478)) : 'freshtv4480)

and _menhir_goto__class_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1626 | MenhirState1657 | MenhirState1732 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4457) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce520 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4458)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4459) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce520 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4460)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_fun_def : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_def -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1631 | MenhirState1633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4453) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_def) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce523 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4454)
    | MenhirState2671 | MenhirState2673 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4455) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_def) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce523 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4456)
    | _ ->
        _menhir_fail ()

and _menhir_goto_value : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_value -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1668
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1668
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1668) : 'freshtv4452)

and _menhir_goto_curried_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4443) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        (_menhir_reduce406 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4444)
    | MenhirState1906 | MenhirState1885 | MenhirState1701 | MenhirState1708 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4445) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        (_menhir_reduce406 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4446)
    | MenhirState1697 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4449 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4447 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_curried_binding) : 'tv_curried_binding) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_) = _menhir_stack in
        let _endpos = _endpos__4_ in
        let _v : 'tv_method_ = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 52226 "reason_parser.ml"
          
        in
        let _startpos__3_ = _startpos_x0_ in
        let _endpos = _endpos__4_ in
        let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
          _startpos__1_
        else
          if Pervasives.(!=) _startpos__2_ _endpos__2_ then
            _startpos__2_
          else
            _startpos__3_ in
        
# 2043 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly (_4, None))) )
# 52242 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4448)) : 'freshtv4450)
    | _ ->
        _menhir_fail ()

and _menhir_goto_curried_binding_return_typed_ : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1194 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4427 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce403 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4428)
    | MenhirState1187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4429 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 52263 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce404 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4430)
    | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4431 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 52274 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce63 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4432)
    | MenhirState1222 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4433 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce62 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4434)
    | MenhirState1708 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4435 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce403 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4436)
    | MenhirState1701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 52299 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce404 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4438)
    | MenhirState1885 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 52310 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce63 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4440)
    | MenhirState1906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4441 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce62 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4442)
    | _ ->
        _menhir_fail ()

and _menhir_goto_method_ : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_method_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1725
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1725
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1725) : 'freshtv4426)

and _menhir_goto_parent_binder : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_parent_binder -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv4423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1736
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1736
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1736) : 'freshtv4424)

and _menhir_goto_semi_terminated_class_fields : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_class_fields -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4417 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4415 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_semi_terminated_class_fields) : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_pattern), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_class_self_pattern_and_structure = 
# 1918 "reason_parser.mly"
    ( Cstr.mk _1 _2 )
# 52376 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4413) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_self_pattern_and_structure) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        match _menhir_s with
        | MenhirState2667 | MenhirState1626 | MenhirState1657 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4403 * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4401 * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_pattern_and_structure), _startpos__1_) = _menhir_stack in
            let _startpos = _startpos__1_ in
            let _endpos = _endpos__1_ in
            let _v : 'tv__class_expr_lets_and_rest = 
# 1895 "reason_parser.mly"
                                     ( mkclass(Pcl_structure _1))
# 52397 "reason_parser.ml"
             in
            _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4402)) : 'freshtv4404)
        | MenhirState475 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error2259 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce224 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4406)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error2259 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4408)
        | MenhirState431 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error2328 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce224 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4410)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error2328 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4412)
        | _ ->
            _menhir_fail ()) : 'freshtv4414)) : 'freshtv4416)) : 'freshtv4418)
    | MenhirState1747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4421 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4419 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_semi_terminated_class_fields) : 'tv_semi_terminated_class_fields) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_field)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_semi_terminated_class_fields = 
# 2003 "reason_parser.mly"
                                                    ( _1::_3 )
# 52455 "reason_parser.ml"
         in
        _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4420)) : 'freshtv4422)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_expr_lets_and_rest : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr_lets_and_rest -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4399) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_expr_lets_and_rest) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4397) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_expr_lets_and_rest) : 'tv__class_expr_lets_and_rest) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_expr_lets_and_rest_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos_x_ _endpos_x_ then
      _startpos_x_
    else
      _endpos in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52484 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4395) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4393) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4391) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cl__class_expr_lets_and_rest_) : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_expr_lets_and_rest = 
# 1893 "reason_parser.mly"
                                                                      (_1)
# 52505 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4389) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_expr_lets_and_rest) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4379 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4377 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
        let _2 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_expr_lets_and_rest = 
# 1897 "reason_parser.mly"
                                               (
    class_of_let_bindings _1 _3
  )
# 52528 "reason_parser.ml"
         in
        _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4378)) : 'freshtv4380)
    | MenhirState1626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4383 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1754 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4381 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce48 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4382)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1754 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4384)
    | MenhirState2667 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2668 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4385 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce48 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4386)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2668 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4388)
    | _ ->
        _menhir_fail ()) : 'freshtv4390)) : 'freshtv4392)) : 'freshtv4394)) : 'freshtv4396)) : 'freshtv4398)) : 'freshtv4400)

and _menhir_goto__class_constructor_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2663 | MenhirState1760 | MenhirState1775 | MenhirState1766 | MenhirState1768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4373) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_constructor_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce527 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4374)
    | MenhirState2692 | MenhirState2057 | MenhirState1821 | MenhirState1781 | MenhirState1810 | MenhirState1801 | MenhirState1803 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4375) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_constructor_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce527 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4376)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_declaration_details : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_declaration_details -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4365 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1836
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1836
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1836) : 'freshtv4366)
    | MenhirState1875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4367 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1876
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1876
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1876) : 'freshtv4368)
    | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4369 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2711
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2711
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2711) : 'freshtv4370)
    | MenhirState2732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4371 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2733
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2733
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2733) : 'freshtv4372)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_self_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_self_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4363 * Lexing.position * _menhir_state * 'tv_class_self_type) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1545
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | CONSTRAINT ->
          _menhir_run1593 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INHERIT ->
          _menhir_run1580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETATATAT ->
          _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETPERCENTPERCENT ->
          _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | METHOD ->
          _menhir_run1562 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | VAL ->
          _menhir_run1546 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | RBRACE | SEMI ->
          _menhir_reduce357 _menhir_env (Obj.magic _menhir_stack) MenhirState1545
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1545)) : 'freshtv4364)

and _menhir_reduce525 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__constrained_class_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__constrained_class_declaration) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__constrained_class_declaration_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1196 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52684 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4361) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
    ((match _menhir_s with
    | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4357) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
        (_menhir_reduce380 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4358)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4359) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
        (_menhir_reduce380 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4360)
    | _ ->
        _menhir_fail ()) : 'freshtv4362)

and _menhir_goto_class_fun_return : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_fun_return -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4353 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_return) = _v in
        (_menhir_reduce29 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4354)
    | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4355 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_return) = _v in
        (_menhir_reduce29 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4356)
    | _ ->
        _menhir_fail ()

and _menhir_goto_type_kind : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_kind -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2071 | MenhirState1262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4349 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 52738 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1351 : 'freshtv4350)
    | MenhirState105 | MenhirState1850 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4351 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 52746 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1851 : 'freshtv4352)
    | _ ->
        _menhir_fail ()

and _menhir_run395 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState395

and _menhir_run438 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState438

and _menhir_goto_optional_type_parameters : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4329 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 52983 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState105
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run106 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState105
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState105
          | PLUSEQ ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState105)) : 'freshtv4330)
    | MenhirState1249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4331 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 53005 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1262
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run1263 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1262
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState1262
          | PLUSEQ ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1262)) : 'freshtv4332)
    | MenhirState1361 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4335 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4333 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_strictly_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)) = _menhir_stack in
        let _v : 'tv_potentially_long_ident_and_optional_type_parameters = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 53037 "reason_parser.ml"
          
        in
        
# 3631 "reason_parser.mly"
                                                             ((_1, _2))
# 53043 "reason_parser.ml"
         in
        _menhir_goto_potentially_long_ident_and_optional_type_parameters _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4334)) : 'freshtv4336)
    | MenhirState1415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4339 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1416
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONEQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1416 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1419) : 'freshtv4338)
          | EQUAL ->
              _menhir_run1417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1416
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1416)) : 'freshtv4340)
    | MenhirState1849 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4341 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 53105 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1850
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run106 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1850
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState1850
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1850)) : 'freshtv4342)
    | MenhirState1967 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4345 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1968
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONEQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4343 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1968 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1969) : 'freshtv4344)
          | EQUAL ->
              _menhir_run1417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1968
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1968)) : 'freshtv4346)
    | MenhirState2070 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4347 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 53179 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2071
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run1263 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2071
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState2071
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2071)) : 'freshtv4348)
    | _ ->
        _menhir_fail ()

and _menhir_goto__structure_item_without_item_extension_sugar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4325) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce552 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4326)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4327) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce552 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4328)
    | _ ->
        _menhir_fail ()

and _menhir_goto__structure_item : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4321) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce551 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4322)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4323) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce551 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4324)
    | _ ->
        _menhir_fail ()

and _menhir_goto__curried_binding_return_typed : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__curried_binding_return_typed -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4317) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__curried_binding_return_typed) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce530 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4318)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4319) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__curried_binding_return_typed) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce530 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4320)
    | _ ->
        _menhir_fail ()

and _menhir_reduce498 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp_expr_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__7_ _ (_7 : 'tv_mark_position_exp_expr_) ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, _startpos__2_), _, (_3 : 'tv_typevar_list), _startpos__3_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_), _endpos__6_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__7_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1216 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 53280 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2915 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (mkpat ~ghost:true ~loc (Ppat_constraint(_1, mktyp ~ghost:true (Ptyp_poly(List.rev _3,_5)))), _7)
      )
# 53292 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce499 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp_expr_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ (_8 : 'tv_mark_position_exp_expr_) ->
    let ((((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_lident_list)), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_), _endpos__7_) = _menhir_stack in
    let _7 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _endpos = _endpos__8_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1216 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 53314 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2967 "reason_parser.mly"
     (
       let exp, poly = wrap_type_annotation _4 _6 _8 in
       let loc = mklocation _symbolstartpos _endpos in
       (mkpat ~ghost:true ~loc (Ppat_constraint(_1, poly)), exp)
     )
# 53327 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_let_binding_body : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4309 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1234
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1234
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1234) : 'freshtv4310)
    | MenhirState1628 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4311 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1629
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1629
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1629) : 'freshtv4312)
    | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4313 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1918
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1918
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1918) : 'freshtv4314)
    | MenhirState2308 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4315 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2309
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2309
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2309) : 'freshtv4316)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constraints : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1351 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4301 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 53395 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1352
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1352 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce808 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1352)) : 'freshtv4302)
    | MenhirState1422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4303 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1423
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce846 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1423)) : 'freshtv4304)
    | MenhirState1851 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4305 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 53431 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1852
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce808 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1852)) : 'freshtv4306)
    | MenhirState1972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4307 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1973
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1853 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce846 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1973)) : 'freshtv4308)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_ext_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_ext_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState78 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4225 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53474 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState79
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState79
          | RPAREN ->
              _menhir_run80 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState79 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState79)) : 'freshtv4226)
    | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4229 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53494 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState84
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState84
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4227 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53508 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState84 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce568 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4228)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState84)) : 'freshtv4230)
    | MenhirState87 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4233 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState88
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState88
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4231 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState88 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce569 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4232)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState88)) : 'freshtv4234)
    | MenhirState2692 | MenhirState2676 | MenhirState2677 | MenhirState2663 | MenhirState2643 | MenhirState2602 | MenhirState2598 | MenhirState2594 | MenhirState2589 | MenhirState2583 | MenhirState2564 | MenhirState2473 | MenhirState120 | MenhirState2466 | MenhirState127 | MenhirState140 | MenhirState142 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState157 | MenhirState160 | MenhirState161 | MenhirState2443 | MenhirState162 | MenhirState169 | MenhirState2436 | MenhirState2433 | MenhirState323 | MenhirState2214 | MenhirState2057 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1821 | MenhirState1781 | MenhirState1808 | MenhirState1810 | MenhirState1799 | MenhirState1801 | MenhirState1803 | MenhirState1793 | MenhirState1794 | MenhirState1760 | MenhirState1773 | MenhirState1775 | MenhirState1764 | MenhirState1766 | MenhirState1768 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1704 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1383 | MenhirState1380 | MenhirState1377 | MenhirState1373 | MenhirState1370 | MenhirState1355 | MenhirState1353 | MenhirState1339 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1306 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1289 | MenhirState1286 | MenhirState1282 | MenhirState1283 | MenhirState1278 | MenhirState1264 | MenhirState1271 | MenhirState1272 | MenhirState1269 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1190 | MenhirState1066 | MenhirState1063 | MenhirState1061 | MenhirState1017 | MenhirState666 | MenhirState305 | MenhirState312 | MenhirState306 | MenhirState302 | MenhirState291 | MenhirState271 | MenhirState269 | MenhirState244 | MenhirState253 | MenhirState254 | MenhirState192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4235 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState194
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run195 _menhir_env (Obj.magic _menhir_stack) MenhirState194
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState194)) : 'freshtv4236)
    | MenhirState2515 | MenhirState106 | MenhirState2472 | MenhirState2497 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2480 | MenhirState2477 | MenhirState2474 | MenhirState2469 | MenhirState108 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1636 | MenhirState1637 | MenhirState1595 | MenhirState1586 | MenhirState1587 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1518 | MenhirState1512 | MenhirState1506 | MenhirState1499 | MenhirState1484 | MenhirState1494 | MenhirState1495 | MenhirState1492 | MenhirState1087 | MenhirState1084 | MenhirState1082 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 | MenhirState328 | MenhirState243 | MenhirState316 | MenhirState211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4241 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState214
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4239 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState214 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4237 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState215 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 53579 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce821 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4238)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState215) : 'freshtv4240)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState214)) : 'freshtv4242)
    | MenhirState2625 | MenhirState128 | MenhirState2385 | MenhirState2383 | MenhirState2362 | MenhirState2356 | MenhirState2354 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2167 | MenhirState2152 | MenhirState2155 | MenhirState2149 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1949 | MenhirState1952 | MenhirState1946 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 | MenhirState796 | MenhirState770 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4247 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState248
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4245 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState248 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState249 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4243 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState249 in
                  let (_v : (
# 910 "reason_parser.mly"
       (string)
# 53620 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LPAREN ->
                      _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOT ->
                      _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack)
                  | AND | COLONGREATER | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
                      _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState250) : 'freshtv4244)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState249) : 'freshtv4246)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState248)) : 'freshtv4248)
    | MenhirState232 | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState2110 | MenhirState1245 | MenhirState1959 | MenhirState1999 | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4253 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState503
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4251 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState503 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState504 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4249 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState504 in
                  let (_v : (
# 910 "reason_parser.mly"
       (string)
# 53671 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LPAREN ->
                      _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOT ->
                      _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack)
                  | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
                      _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState505) : 'freshtv4250)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState504) : 'freshtv4252)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState503)) : 'freshtv4254)
    | MenhirState117 | MenhirState1265 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4255 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53701 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1266
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run195 _menhir_env (Obj.magic _menhir_stack) MenhirState1266
          | RPAREN ->
              _menhir_run80 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1266)) : 'freshtv4256)
    | MenhirState2578 | MenhirState76 | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4267 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1391
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4265 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1391 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4263 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1392 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 53740 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4261 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos__3_ : Lexing.position) = _endpos in
                  let (_ : _menhir_state) = _menhir_s in
                  let ((_3 : (
# 863 "reason_parser.mly"
       (string)
# 53751 "reason_parser.ml"
                  )) : (
# 863 "reason_parser.mly"
       (string)
# 53755 "reason_parser.ml"
                  )) = _v in
                  let (_startpos__3_ : Lexing.position) = _startpos in
                  ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
                  let _2 = () in
                  let _startpos = _startpos__1_ in
                  let _endpos = _endpos__3_ in
                  let _v : 'tv_type_strictly_longident = 
# 4133 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 53765 "reason_parser.ml"
                   in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4259) = _menhir_stack in
                  let (_endpos : Lexing.position) = _endpos in
                  let (_menhir_s : _menhir_state) = _menhir_s in
                  let (_v : 'tv_type_strictly_longident) = _v in
                  let (_startpos : Lexing.position) = _startpos in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4257 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | MINUS ->
                      _menhir_run1257 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run1253 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1361) : 'freshtv4258)) : 'freshtv4260)) : 'freshtv4262)) : 'freshtv4264)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1392 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1392) : 'freshtv4266)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1391)) : 'freshtv4268)
    | MenhirState1426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4269 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53804 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1427
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1427
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce849 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1427)) : 'freshtv4270)
    | MenhirState1429 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1430
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1430
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce848 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1430)) : 'freshtv4272)
    | MenhirState1761 | MenhirState1536 | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4277 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1581
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4275 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1581 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4273 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1582 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 53859 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce365 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4274)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1582) : 'freshtv4276)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1581)) : 'freshtv4278)
    | MenhirState2701 | MenhirState2050 | MenhirState1820 | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4283 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1788
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4281 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1788 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4279 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1789 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 53898 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce365 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4280)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1789 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1789) : 'freshtv4282)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1788)) : 'freshtv4284)
    | MenhirState1978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4287 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53918 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1979
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1979
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4285 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53932 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1979 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce570 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4286)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1979)) : 'freshtv4288)
    | MenhirState1976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53948 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1981
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run1982 _menhir_env (Obj.magic _menhir_stack) MenhirState1981
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce849 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1981)) : 'freshtv4290)
    | MenhirState1984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4293 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53968 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1985
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1985
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4291 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 53982 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1985 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce568 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4292)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1985)) : 'freshtv4294)
    | MenhirState1988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4297 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1989
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1989
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4295 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1989 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce569 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4296)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1989)) : 'freshtv4298)
    | MenhirState1992 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4299 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1993
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run1982 _menhir_env (Obj.magic _menhir_stack) MenhirState1993
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce848 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1993)) : 'freshtv4300)
    | _ ->
        _menhir_fail ()

and _menhir_reduce573 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext2), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_ext_longident = 
# 4142 "reason_parser.mly"
             ( _1 )
# 54042 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_with_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4217 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce851 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4218)
    | MenhirState1413 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4219) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce850 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4220)
    | MenhirState1995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4221 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce851 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4222)
    | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4223) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce850 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4224)
    | _ ->
        _menhir_fail ()

and _menhir_reduce125 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_with_constraints)) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_type = 
# 1573 "reason_parser.mly"
                                      (
    (* See note above about why WITH constraints aren't considered
     * non-arrowed.
     * We might just consider unifying the syntax for record extension with
     * module extension/WITH constraints.
     *
     *    mod MyModule = {
     *       ModuleToInclude...
     *    };
     *
     *    let module CreateFactory
     *               (Spec: ContainerSpec)
     *               :{DescriptorFactoryIntf.S with
     *                  type props = Spec.props and type dependencies = Spec.props} =>
     *
     *)
    mkmty(Pmty_with(_1, List.rev _3))
  )
# 54106 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4213) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce541 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4214)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState1245 | MenhirState2110 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4215) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce541 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4216)
    | _ ->
        _menhir_fail ()

and _menhir_goto__simple_module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4209) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce543 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4210)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState1245 | MenhirState2110 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4211) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce543 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4212)
    | _ ->
        _menhir_fail ()

and _menhir_goto__simple_module_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2638 | MenhirState2639 | MenhirState410 | MenhirState2360 | MenhirState450 | MenhirState2282 | MenhirState751 | MenhirState2166 | MenhirState754 | MenhirState756 | MenhirState2136 | MenhirState2137 | MenhirState2132 | MenhirState2133 | MenhirState2122 | MenhirState2123 | MenhirState1240 | MenhirState1241 | MenhirState794 | MenhirState795 | MenhirState767 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4205) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce539 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4206)
    | MenhirState402 | MenhirState2389 | MenhirState403 | MenhirState2387 | MenhirState411 | MenhirState2358 | MenhirState2300 | MenhirState2301 | MenhirState2296 | MenhirState2297 | MenhirState2289 | MenhirState2290 | MenhirState2285 | MenhirState2286 | MenhirState757 | MenhirState2159 | MenhirState1470 | MenhirState2019 | MenhirState1939 | MenhirState1941 | MenhirState1942 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4207) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce539 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4208)
    | _ ->
        _menhir_fail ()

and _menhir_goto_fun_def : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1022 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4189 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce439 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4190)
    | MenhirState901 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4191 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 54191 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce440 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4192)
    | MenhirState897 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4193 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 54202 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce67 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4194)
    | MenhirState1036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4195 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce66 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4196)
    | MenhirState2218 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4197 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce439 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4198)
    | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 54227 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce440 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4200)
    | MenhirState480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 54238 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce67 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4202)
    | MenhirState2232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4203 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce66 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4204)
    | _ ->
        _menhir_fail ()

and _menhir_goto_leading_bar_match_case : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4181 * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce490 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4182)
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4183) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce489 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4184)
    | MenhirState2229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4185 * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce490 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4186)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4187) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce489 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4188)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_semi : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4145 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error599 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce280 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4144)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error599 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4146)
    | MenhirState605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4149 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error606 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4147 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce278 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4148)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error606 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4150)
    | MenhirState700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4153 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error701 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4151 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce278 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4152)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error701 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4154)
    | MenhirState705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error706 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4155 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce280 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4156)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error706 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4158)
    | MenhirState1170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4159 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce181 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4160)
    | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4161 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce180 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4162)
    | MenhirState1603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4175 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4173 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_type)), _endpos__2_, _, (_2 : 'tv_class_sig_fields)), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
        let _v : 'tv_class_sig_body = 
# 2235 "reason_parser.mly"
    ( Csig.mk _1 (List.rev _2 ))
# 54381 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4171) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_body) = _v in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState1538 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4165 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error1542 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4163 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce34 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4164)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error1542 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4166)
        | MenhirState1784 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4169 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error1785 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4167 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce34 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4168)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error1785 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4170)
        | _ ->
            _menhir_fail ()) : 'freshtv4172)) : 'freshtv4174)) : 'freshtv4176)
    | MenhirState2252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4177 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce181 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4178)
    | MenhirState2246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4179 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce180 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4180)
    | _ ->
        _menhir_fail ()

and _menhir_reduce218 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54455 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54467 "reason_parser.ml"
      
    in
    
# 2694 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _5 ")") )
# 54473 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce220 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54494 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54506 "reason_parser.ml"
      
    in
    
# 2701 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "[") (with_txt _5 "]") )
# 54512 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_module_binding_body_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState2631 | MenhirState2317 | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4139) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_expr) = _v in
        (_menhir_reduce576 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4140)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4141) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_expr) = _v in
        (_menhir_reduce576 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4142)
    | _ ->
        _menhir_fail ()

and _menhir_goto__module_binding_body_functor : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_binding_body_functor -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2631 | MenhirState2317 | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4135) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_binding_body_functor) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce537 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4136)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4137) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_binding_body_functor) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce537 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4138)
    | _ ->
        _menhir_fail ()

and _menhir_goto_field_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_field_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4129 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field_expr) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4127 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_field_expr) : 'tv_field_expr) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_field_expr_list)), _endpos__2_, _) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_field_expr_list = 
# 3236 "reason_parser.mly"
      ( _3::_1 )
# 54575 "reason_parser.ml"
         in
        _menhir_goto_field_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4128)) : 'freshtv4130)
    | MenhirState419 | MenhirState2191 | MenhirState1103 | MenhirState422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4133) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field_expr) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4131) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_field_expr) : 'tv_field_expr) = _v in
        ((let _v : 'tv_field_expr_list = 
# 3234 "reason_parser.mly"
     ( [_1] )
# 54590 "reason_parser.ml"
         in
        _menhir_goto_field_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4132)) : 'freshtv4134)
    | _ ->
        _menhir_fail ()

and _menhir_goto__module_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2638 | MenhirState410 | MenhirState450 | MenhirState751 | MenhirState754 | MenhirState2136 | MenhirState2132 | MenhirState2122 | MenhirState1240 | MenhirState794 | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4123) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce538 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4124)
    | MenhirState402 | MenhirState403 | MenhirState411 | MenhirState2300 | MenhirState2296 | MenhirState2289 | MenhirState2285 | MenhirState757 | MenhirState1470 | MenhirState1939 | MenhirState1942 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4125) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce538 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4126)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2625 | MenhirState2298 | MenhirState2287 | MenhirState448 | MenhirState2134 | MenhirState1242 | MenhirState2116 | MenhirState2119 | MenhirState2107 | MenhirState1401 | MenhirState2030 | MenhirState1460 | MenhirState1408 | MenhirState1436 | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4119) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce542 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4120)
    | MenhirState400 | MenhirState2162 | MenhirState1243 | MenhirState1245 | MenhirState2110 | MenhirState1959 | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4121) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce542 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4122)
    | _ ->
        _menhir_fail ()

and _menhir_goto__core_type2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2643 | MenhirState2564 | MenhirState2473 | MenhirState117 | MenhirState127 | MenhirState140 | MenhirState2459 | MenhirState2457 | MenhirState152 | MenhirState2433 | MenhirState2436 | MenhirState2035 | MenhirState2021 | MenhirState1911 | MenhirState1889 | MenhirState1898 | MenhirState1892 | MenhirState1853 | MenhirState1738 | MenhirState1714 | MenhirState1717 | MenhirState1593 | MenhirState1539 | MenhirState1474 | MenhirState1465 | MenhirState1421 | MenhirState1419 | MenhirState1353 | MenhirState1355 | MenhirState1263 | MenhirState1322 | MenhirState1324 | MenhirState1281 | MenhirState1282 | MenhirState1272 | MenhirState1269 | MenhirState1265 | MenhirState1227 | MenhirState1201 | MenhirState1214 | MenhirState1206 | MenhirState1063 | MenhirState1066 | MenhirState1061 | MenhirState666 | MenhirState244 | MenhirState269 | MenhirState271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4115) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type2) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce554 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4116)
    | MenhirState106 | MenhirState2472 | MenhirState241 | MenhirState2400 | MenhirState326 | MenhirState1971 | MenhirState1969 | MenhirState1855 | MenhirState1694 | MenhirState1689 | MenhirState1674 | MenhirState1665 | MenhirState1595 | MenhirState1569 | MenhirState1573 | MenhirState1560 | MenhirState1555 | MenhirState1550 | MenhirState1495 | MenhirState1492 | MenhirState1084 | MenhirState1087 | MenhirState1082 | MenhirState722 | MenhirState681 | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4117) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type2) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce554 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4118)
    | _ ->
        _menhir_fail ()

and _menhir_reduce647 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_package_type_cstr)) = _menhir_stack in
    let _v : 'tv_package_type_cstrs = 
# 3977 "reason_parser.mly"
                      ( [_1] )
# 54668 "reason_parser.ml"
     in
    _menhir_goto_package_type_cstrs _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto__pattern_without_or : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_without_or -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4111) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__pattern_without_or) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce547 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4112)
    | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4113) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__pattern_without_or) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce547 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4114)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_comma : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState286 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4049 * _menhir_state * 'tv_field) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4047 * _menhir_state * 'tv_field) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_field)), _, (_2 : 'tv_opt_comma)) = _menhir_stack in
        let _v : 'tv_meth_list = 
# 4019 "reason_parser.mly"
                                                 ( [_1], Closed )
# 54707 "reason_parser.ml"
         in
        _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4048)) : 'freshtv4050)
    | MenhirState424 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4053 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error427 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4051 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce238 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4052)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error427 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4054)
    | MenhirState526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4057 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4055 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_opt_comma)) = _menhir_stack in
        let _3 = () in
        let _2 = () in
        let _v : 'tv_lbl_pattern_list = 
# 3433 "reason_parser.mly"
                                           ( [_1], Open )
# 54739 "reason_parser.ml"
         in
        _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4056)) : 'freshtv4058)
    | MenhirState1078 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4061 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1090 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4059 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce231 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4060)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1090 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4062)
    | MenhirState1097 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4065 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1100 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4063 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce416 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4064)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1100 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4066)
    | MenhirState1104 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4069 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1105 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4067 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce241 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4068)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1105 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4070)
    | MenhirState1331 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4073 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1334 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4071 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce815 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4072)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1334 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4074)
    | MenhirState1346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv4077 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1347 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv4075 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce819 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4076)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1347 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4078)
    | MenhirState2177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4081 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2178 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4079 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce231 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4080)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2178 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4082)
    | MenhirState2185 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4085 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2188 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4083 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce416 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4084)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2188 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4086)
    | MenhirState2192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4089 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2193 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4087 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce241 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4088)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2193 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4090)
    | MenhirState2333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4093 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2334 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4091 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce238 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4092)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2334 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4094)
    | MenhirState2348 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4097 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2349 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4095 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce228 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4096)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2349 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4098)
    | MenhirState2422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4101 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2423 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4099 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce228 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4100)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2423 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4102)
    | MenhirState2510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4105 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2511 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4103 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce815 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4104)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2511 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4106)
    | MenhirState2522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv4109 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2523 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv4107 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce819 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4108)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2523 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4110)
    | _ ->
        _menhir_fail ()

and _menhir_goto_attributes : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1274 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3931 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 54987 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce388 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3932)
    | MenhirState1276 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3933 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3934)
    | MenhirState1279 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3935 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce392 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3936)
    | MenhirState1284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3937 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce389 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3938)
    | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce391 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3940)
    | MenhirState1290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3941 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce390 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3942)
    | MenhirState1294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3943 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 55033 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3944)
    | MenhirState1297 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3945 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce387 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3946)
    | MenhirState1301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3947 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce384 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3948)
    | MenhirState1304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3949 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce386 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3950)
    | MenhirState1307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3951 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce385 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3952)
    | MenhirState1326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3965 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 55072 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3963 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 55081 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos__5_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_5 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mutable_flag), _startpos__1_), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 55089 "reason_parser.ml"
        )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_poly_type)) = _menhir_stack in
        let _3 = () in
        let _v : 'tv_label_declaration = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 55101 "reason_parser.ml"
          
        in
        let _startpos__2_ = _startpos_x0_ in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
          _startpos__1_
        else
          _startpos__2_ in
        
# 3617 "reason_parser.mly"
      (
       let loc = mklocation _symbolstartpos _endpos in
       Type.field _2 _4 ~mut:_1 ~attrs:_5 ~loc
      )
# 55116 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3961) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_declaration) = _v in
        ((match _menhir_s with
        | MenhirState1332 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3955 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : 'tv_label_declaration) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3953 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
            let (_ : _menhir_state) = _menhir_s in
            let ((_3 : 'tv_label_declaration) : 'tv_label_declaration) = _v in
            ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_label_declarations)), _endpos__2_, _) = _menhir_stack in
            let _2 = () in
            let _v : 'tv_label_declarations = 
# 3612 "reason_parser.mly"
                                                 ( _3 :: _1 )
# 55137 "reason_parser.ml"
             in
            _menhir_goto_label_declarations _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3954)) : 'freshtv3956)
        | MenhirState2521 | MenhirState2509 | MenhirState1345 | MenhirState1318 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3959) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : 'tv_label_declaration) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3957) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let ((_1 : 'tv_label_declaration) : 'tv_label_declaration) = _v in
            ((let _v : 'tv_label_declarations = 
# 3611 "reason_parser.mly"
                                                ( [_1] )
# 55152 "reason_parser.ml"
             in
            _menhir_goto_label_declarations _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3958)) : 'freshtv3960)
        | _ ->
            _menhir_fail ()) : 'freshtv3962)) : 'freshtv3964)) : 'freshtv3966)
    | MenhirState1371 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3967 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55162 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce421 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3968)
    | MenhirState1374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce425 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3970)
    | MenhirState1378 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce422 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3972)
    | MenhirState1381 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3973 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce424 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3974)
    | MenhirState1384 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce423 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3976)
    | MenhirState1488 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3977 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55201 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce426 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3978)
    | MenhirState1490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3979 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3980)
    | MenhirState1497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3981 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55219 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce421 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3982)
    | MenhirState1501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce430 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3984)
    | MenhirState1503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce425 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3986)
    | MenhirState1508 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3987 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce427 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3988)
    | MenhirState1510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce422 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3990)
    | MenhirState1514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3991 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce429 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3992)
    | MenhirState1516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce424 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3994)
    | MenhirState1520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce428 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3996)
    | MenhirState1522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce423 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3998)
    | MenhirState2435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4001 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3999 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _, (_2 : 'tv_opt_ampersand)), _endpos__3_, _, (_3 : 'tv_amper_type_list)) = _menhir_stack in
        let _v : 'tv_tag_field = 
# 3990 "reason_parser.mly"
      ( Rtag (_1, _4, _2, List.rev _3) )
# 55296 "reason_parser.ml"
         in
        _menhir_goto_tag_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4000)) : 'freshtv4002)
    | MenhirState172 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4005 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4003 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
        let _v : 'tv_tag_field = 
# 3992 "reason_parser.mly"
      ( Rtag (_1, _2, true, []) )
# 55314 "reason_parser.ml"
         in
        _menhir_goto_tag_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4004)) : 'freshtv4006)
    | MenhirState2453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4015 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4013 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_label), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_poly_type)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_field = 
# 4024 "reason_parser.mly"
                                               ( (_1, _4, _3) )
# 55333 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4011) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field) = _v in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4009 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4007 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState286 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | DOTDOT ->
                _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState287
            | LIDENT _v ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState287 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | GREATER ->
                _menhir_reduce633 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState287) : 'freshtv4008)
        | GREATER ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState286
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState286) : 'freshtv4010)) : 'freshtv4012)) : 'freshtv4014)) : 'freshtv4016)
    | MenhirState2467 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4017 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55375 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce388 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4018)
    | MenhirState2470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4019 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce392 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4020)
    | MenhirState2475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce389 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4022)
    | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4023 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce391 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4024)
    | MenhirState2481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4025 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce390 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4026)
    | MenhirState2485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4027 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 55414 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4028)
    | MenhirState2488 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4029 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce387 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4030)
    | MenhirState2492 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4031 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce384 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4032)
    | MenhirState2495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4033 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce386 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4034)
    | MenhirState2498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4035 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce385 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4036)
    | MenhirState2587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4037 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55453 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce426 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4038)
    | MenhirState2591 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4039 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce430 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4040)
    | MenhirState2596 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4041 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce427 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4042)
    | MenhirState2600 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4043 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce429 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4044)
    | MenhirState2604 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4045 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce428 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4046)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3915 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce813 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3916)
    | MenhirState1340 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3917 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce817 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3918)
    | MenhirState1338 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3919 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce816 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3920)
    | MenhirState1263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3921 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3922)
    | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3923 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce813 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3924)
    | MenhirState2516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3925 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce817 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3926)
    | MenhirState2514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3927 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce816 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3928)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3929 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3930)
    | _ ->
        _menhir_fail ()

and _menhir_reduce810 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3490 "reason_parser.mly"
      ( (Ptype_abstract, Public, Some _2) )
# 55560 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_generalized_constructor_arguments : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_generalized_constructor_arguments -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1339 | MenhirState1264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3855 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55573 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1274
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1274
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1274) : 'freshtv3856)
    | MenhirState1278 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3857 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1279
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1279
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1279) : 'freshtv3858)
    | MenhirState1283 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3859 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1284
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1284
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1284) : 'freshtv3860)
    | MenhirState1286 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3861 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1287
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1287
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1287) : 'freshtv3862)
    | MenhirState1289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3863 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1290
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1290
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1290) : 'freshtv3864)
    | MenhirState1293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3865 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 55637 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1294
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1294
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1294) : 'freshtv3866)
    | MenhirState1296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3867 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1297
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1297
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1297) : 'freshtv3868)
    | MenhirState1300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3869 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1301
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1301
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1301) : 'freshtv3870)
    | MenhirState1303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3871 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1304
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1304
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1304) : 'freshtv3872)
    | MenhirState1306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3873 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1307
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1307
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1307) : 'freshtv3874)
    | MenhirState2583 | MenhirState1370 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3875 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55701 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1371
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1371
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1371) : 'freshtv3876)
    | MenhirState2589 | MenhirState1373 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3877 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1374
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1374
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1374) : 'freshtv3878)
    | MenhirState2594 | MenhirState1377 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3879 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1378
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1378
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1378) : 'freshtv3880)
    | MenhirState2598 | MenhirState1380 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3881 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1381
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1381
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1381) : 'freshtv3882)
    | MenhirState2602 | MenhirState1383 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3883 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1384
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState1384
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1384) : 'freshtv3884)
    | MenhirState1484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3885 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55765 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1497
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1497
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1497) : 'freshtv3886)
    | MenhirState1499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3887 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1503
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1503
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1503) : 'freshtv3888)
    | MenhirState1506 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3889 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1510
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1510
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1510) : 'freshtv3890)
    | MenhirState1512 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3891 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1516
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1516
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1516) : 'freshtv3892)
    | MenhirState1518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3893 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1522
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1522
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1522) : 'freshtv3894)
    | MenhirState2515 | MenhirState108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3895 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 55829 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2467
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2467
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2467) : 'freshtv3896)
    | MenhirState2469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3897 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2470
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2470
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2470) : 'freshtv3898)
    | MenhirState2474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3899 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2475
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2475
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2475) : 'freshtv3900)
    | MenhirState2477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3901 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2478
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2478
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2478) : 'freshtv3902)
    | MenhirState2480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3903 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2481
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2481
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2481) : 'freshtv3904)
    | MenhirState2484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3905 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 55893 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2485
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2485
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2485) : 'freshtv3906)
    | MenhirState2487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3907 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2488
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2488
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2488) : 'freshtv3908)
    | MenhirState2491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3909 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2492
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2492
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2492) : 'freshtv3910)
    | MenhirState2494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3911 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2495
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2495
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2495) : 'freshtv3912)
    | MenhirState2497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3913 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2498
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState2498
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2498) : 'freshtv3914)
    | _ ->
        _menhir_fail ()

and _menhir_goto_post_item_attributes : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1154 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3621 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1155 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3619 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1156) : 'freshtv3620)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1155 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3622)
    | MenhirState1169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3625 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1170
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3623 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1170 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3624)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1170) : 'freshtv3626)
    | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3627 * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce672 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3628)
    | MenhirState1234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3629 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce327 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3630)
    | MenhirState1363 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3631 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce517 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3632)
    | MenhirState1386 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3641 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3639 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_nonrec_flag)), _, (_3 : 'tv_potentially_long_ident_and_optional_type_parameters)), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_private_flag), _startpos__5_), _, (_6 : 'tv_opt_bar)), _endpos__7_, _, (_7 : 'tv_sig_extension_constructors)), _endpos__8_, _, (_8 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__8_ in
        let _v : 'tv_sig_type_extension = 
# 3649 "reason_parser.mly"
  (
    if _2 <> Recursive then not_expecting 2 "nonrec flag";
    let (potentially_long_ident, optional_type_parameters) = _3 in
    Te.mk potentially_long_ident (List.rev _7)
          ~params:optional_type_parameters ~priv:_5 ~attrs:_8
  )
# 56174 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3637) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3635) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3633) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_sig_type_extension) : 'tv_sig_type_extension) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1685 "reason_parser.mly"
                       (
      mksig(Psig_typext _1)
    )
# 56201 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3634)) : 'freshtv3636)) : 'freshtv3638)) : 'freshtv3640)) : 'freshtv3642)
    | MenhirState1396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3643 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce604 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3644)
    | MenhirState1454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3647 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56213 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56219 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 56224 "reason_parser.ml"
        )), _startpos_x0_), _endpos__4_, _), _endpos_x1_, _, (x1 : 'tv_mod_longident), _startpos_x1_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__signature_item = let _5 =
          let _endpos_x_ = _endpos_x1_ in
          let _startpos_x_ = _startpos_x1_ in
          let x = x1 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56240 "reason_parser.ml"
          
        in
        let _endpos__5_ = _endpos_x1_ in
        let _startpos__5_ = _startpos_x1_ in
        let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56254 "reason_parser.ml"
          
        in
        let _endpos = _endpos__6_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1695 "reason_parser.mly"
                                                                               (
      let loc = mklocation _symbolstartpos _endpos in
      let loc_mod = mklocation _startpos__5_ _endpos__5_ in
      mksig(
        Psig_module (
          Md.mk
            _3
            (Mty.alias ~loc:loc_mod _5)
            ~attrs:_6
            ~loc
        )
      )
    )
# 56274 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3646)) : 'freshtv3648)
    | MenhirState1456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56282 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3649 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56288 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 56293 "reason_parser.ml"
        )), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_module_declaration)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56308 "reason_parser.ml"
          
        in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1691 "reason_parser.mly"
                                                                      (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_module (Md.mk _3 _4 ~attrs:_5 ~loc))
    )
# 56319 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3650)) : 'freshtv3652)
    | MenhirState1462 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3655 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_rec_declaration_details)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _3 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv_many_module_rec_declarations = let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1763 "reason_parser.mly"
                                                                       (
      let (ident, body) = _4 in
      let loc = mklocation _symbolstartpos _endpos in
      [Md.mk ident body ~attrs:_5 ~loc]
    )
# 56342 "reason_parser.ml"
         in
        _menhir_goto_many_module_rec_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3654)) : 'freshtv3656)
    | MenhirState1466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3659 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__signature_item = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56364 "reason_parser.ml"
          
        in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1674 "reason_parser.mly"
                                                               (
        let loc = mklocation _symbolstartpos _endpos in
        mksig(Psig_value (Val.mk _2 _4 ~attrs:_5 ~loc))
    )
# 56375 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3658)) : 'freshtv3660)
    | MenhirState1479 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3661 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce297 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3662)
    | MenhirState1481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3663 * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce672 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3664)
    | MenhirState1524 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3665 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce778 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3666)
    | MenhirState1526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3667 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce777 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3668)
    | MenhirState1557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3671 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_value_type)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_sig_field = 
# 2261 "reason_parser.mly"
                                        (
      mkctf_attrs (Pctf_val _2) _3
    )
# 56408 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3670)) : 'freshtv3672)
    | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3675 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3673 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), (_2 : 'tv_private_virtual_flags)), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_poly_type)), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__class_sig_field = 
# 2265 "reason_parser.mly"
       (
        let (p, v) = _2 in
        mkctf_attrs (Pctf_method (_3, p, v, _5)) _6
       )
# 56427 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3674)) : 'freshtv3676)
    | MenhirState1589 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_), _endpos__3_, _, (_3 : 'tv_item_attribute)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__class_sig_field = 
# 2260 "reason_parser.mly"
      ( mkctf_attrs (Pctf_inherit _2) (_3::_4) )
# 56442 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3678)) : 'freshtv3680)
    | MenhirState1597 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_constrain_field)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_sig_field = 
# 2270 "reason_parser.mly"
       ( mkctf_attrs (Pctf_constraint _2) _3 )
# 56457 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3682)) : 'freshtv3684)
    | MenhirState1600 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3687 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3685 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__class_sig_field = 
# 2272 "reason_parser.mly"
       ( mkctf_attrs (Pctf_extension _1) _2 )
# 56471 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3686)) : 'freshtv3688)
    | MenhirState1617 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce511 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3690)
    | MenhirState1629 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3691 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce502 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3692)
    | MenhirState1668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_value)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1985 "reason_parser.mly"
      ( mkcf_attrs (Pcf_val _2) _3 )
# 56494 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3694)) : 'freshtv3696)
    | MenhirState1725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_method_)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1987 "reason_parser.mly"
      ( mkcf_attrs (Pcf_method _2) _3 )
# 56509 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3698)) : 'freshtv3700)
    | MenhirState1729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__class_field = 
# 1991 "reason_parser.mly"
      ( mkcf_attrs (Pcf_initializer _3) _4 )
# 56525 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3702)) : 'freshtv3704)
    | MenhirState1736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_override_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_expr), _startpos__3_), _endpos__4_, _, (_4 : 'tv_parent_binder)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__class_field = 
# 1983 "reason_parser.mly"
      ( mkcf_attrs (Pcf_inherit (_2, _3, _4)) _5 )
# 56540 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3706)) : 'freshtv3708)
    | MenhirState1739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_constrain_field)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1989 "reason_parser.mly"
      ( mkcf_attrs (Pcf_constraint _2) _3 )
# 56555 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3710)) : 'freshtv3712)
    | MenhirState1743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3715 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3713 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__class_field = 
# 1993 "reason_parser.mly"
      ( mkcf_attrs (Pcf_extension _1) _2 )
# 56569 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3714)) : 'freshtv3716)
    | MenhirState1836 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3717 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce507 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3718)
    | MenhirState1861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3719 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce330 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3720)
    | MenhirState1866 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3721 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce329 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3722)
    | MenhirState1871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3723 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce326 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3724)
    | MenhirState1876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3725 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce324 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3726)
    | MenhirState1918 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3727 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce327 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3728)
    | MenhirState1926 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3729 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce309 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3730)
    | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3731 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce296 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3732)
    | MenhirState2019 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce308 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3734)
    | MenhirState2031 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3737 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3735 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__signature_item = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1722 "reason_parser.mly"
                                             (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_include (Incl.mk _2 ~attrs:_3 ~loc))
    )
# 56625 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3736)) : 'freshtv3738)
    | MenhirState2040 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3739 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_), _endpos__6_, _, (_6 : 'tv_primitive_declaration)), _endpos__7_, _, (_7 : 'tv_post_item_attributes)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__7_ in
        let _v : 'tv__signature_item = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56648 "reason_parser.ml"
          
        in
        let _endpos = _endpos__7_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1678 "reason_parser.mly"
                                                                                                (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_value (Val.mk _2 _4 ~prim:_6 ~attrs:_7 ~loc))
    )
# 56659 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3740)) : 'freshtv3742)
    | MenhirState2043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_declaration)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv_sig_exception_declaration = 
# 3594 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 56677 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3747) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3745) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3743) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_sig_exception_declaration) : 'tv_sig_exception_declaration) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1688 "reason_parser.mly"
                              (
      mksig(Psig_exception _1)
    )
# 56704 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3744)) : 'freshtv3746)) : 'freshtv3748)) : 'freshtv3750)) : 'freshtv3752)
    | MenhirState2052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3753 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce511 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3754)
    | MenhirState2059 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3757 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_description_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv_many_class_descriptions = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 2291 "reason_parser.mly"
                                                         (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc]
  )
# 56729 "reason_parser.ml"
         in
        _menhir_goto_many_class_descriptions _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3756)) : 'freshtv3758)
    | MenhirState2072 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce330 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3760)
    | MenhirState2077 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3769 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3767 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_rec_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_and_module_rec_declaration = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1774 "reason_parser.mly"
                                                            (
      let (ident, body) = _2 in
      let loc = mklocation _symbolstartpos _endpos in
      Md.mk ident body ~attrs:_3 ~loc
    )
# 56753 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3765) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3763 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3761 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let ((_2 : 'tv_and_module_rec_declaration) : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_module_rec_declarations), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_many_module_rec_declarations = 
# 1768 "reason_parser.mly"
                                                             (
      _2::_1
    )
# 56775 "reason_parser.ml"
         in
        _menhir_goto_many_module_rec_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3762)) : 'freshtv3764)) : 'freshtv3766)) : 'freshtv3768)) : 'freshtv3770)
    | MenhirState2082 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3771 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce326 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3772)
    | MenhirState2087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3781 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3779 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_description_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_and_class_description = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 2301 "reason_parser.mly"
                                                       (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc
  )
# 56799 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3777) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_class_description) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3775 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_class_description) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3773 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let ((_2 : 'tv_and_class_description) : 'tv_and_class_description) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_descriptions), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_many_class_descriptions = 
# 2296 "reason_parser.mly"
                                                  (
    _2 :: _1
  )
# 56821 "reason_parser.ml"
         in
        _menhir_goto_many_class_descriptions _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3774)) : 'freshtv3776)) : 'freshtv3778)) : 'freshtv3780)) : 'freshtv3782)
    | MenhirState2090 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3785 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3783 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__signature_item = 
# 1732 "reason_parser.mly"
                                        (
      mksig(Psig_extension (_1, _2))
    )
# 56837 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3784)) : 'freshtv3786)
    | MenhirState2097 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3787 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_module_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56860 "reason_parser.ml"
          
        in
        let _endpos = _endpos__6_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1715 "reason_parser.mly"
                                                                     (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_modtype (Mtd.mk _3 ~typ:_5 ~loc ~attrs:_6))
    )
# 56871 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3788)) : 'freshtv3790)
    | MenhirState1400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3793 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56893 "reason_parser.ml"
          
        in
        let _endpos = _endpos__4_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1711 "reason_parser.mly"
                                                   (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_modtype (Mtd.mk _3 ~attrs:_4 ~loc))
    )
# 56904 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3792)) : 'freshtv3794)
    | MenhirState2126 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56912 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2127 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3795 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 56924 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1151 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2128) : 'freshtv3796)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2127 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3798)
    | MenhirState2251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3801 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2252
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3799 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2252 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3800)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2252) : 'freshtv3802)
    | MenhirState2304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3803 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce515 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3804)
    | MenhirState2306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3805 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce301 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3806)
    | MenhirState2309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3807 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce502 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3808)
    | MenhirState437 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3811 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2313 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2314) : 'freshtv3810)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2313 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3812)
    | MenhirState2318 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3815 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 57211 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2319 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3813 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 57223 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2320) : 'freshtv3814)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2319 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3816)
    | MenhirState2390 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3817 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce304 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3818)
    | MenhirState399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3819 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce303 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3820)
    | MenhirState394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce604 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3822)
    | MenhirState2527 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3823 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce517 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3824)
    | MenhirState2533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3825 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce783 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3826)
    | MenhirState2606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3827 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce783 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3828)
    | MenhirState2626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce304 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3830)
    | MenhirState2624 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3831 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce303 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3832)
    | MenhirState2634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3833 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce515 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3834)
    | MenhirState2636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce301 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3836)
    | MenhirState2639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce308 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3838)
    | MenhirState2646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce297 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3840)
    | MenhirState2649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3841 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce778 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3842)
    | MenhirState2651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3843 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce777 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3844)
    | MenhirState2711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3845 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce507 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3846)
    | MenhirState2727 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3847 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce329 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3848)
    | MenhirState2733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce324 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3850)
    | MenhirState2739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3851 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce309 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3852)
    | MenhirState2742 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3853 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce296 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3854)
    | _ ->
        _menhir_fail ()

and _menhir_goto__expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState766 | MenhirState1142 | MenhirState1137 | MenhirState1129 | MenhirState1118 | MenhirState1076 | MenhirState1079 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1042 | MenhirState1046 | MenhirState811 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState828 | MenhirState881 | MenhirState855 | MenhirState875 | MenhirState877 | MenhirState879 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState861 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState853 | MenhirState851 | MenhirState845 | MenhirState849 | MenhirState847 | MenhirState830 | MenhirState832 | MenhirState842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3615) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce531 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3616)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2568 | MenhirState2416 | MenhirState2368 | MenhirState2370 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState765 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1166 | MenhirState1171 | MenhirState786 | MenhirState1133 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1027 | MenhirState1029 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1003 | MenhirState1008 | MenhirState1010 | MenhirState1006 | MenhirState1001 | MenhirState924 | MenhirState941 | MenhirState994 | MenhirState968 | MenhirState988 | MenhirState990 | MenhirState992 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState974 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState966 | MenhirState964 | MenhirState958 | MenhirState962 | MenhirState960 | MenhirState943 | MenhirState945 | MenhirState955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3617) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce531 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3618)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1626 | MenhirState1657 | MenhirState1732 | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3611) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce524 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3612)
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3613) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce524 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3614)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2703 | MenhirState2699 | MenhirState2694 | MenhirState2658 | MenhirState2674 | MenhirState2675 | MenhirState121 | MenhirState119 | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3531 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState9
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3529 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState9 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3527 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState10 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 57459 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce350 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv3528)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState10 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState10) : 'freshtv3530)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState9)) : 'freshtv3532)
    | MenhirState2667 | MenhirState2659 | MenhirState1825 | MenhirState1818 | MenhirState1813 | MenhirState1624 | MenhirState1625 | MenhirState1732 | MenhirState1657 | MenhirState1626 | MenhirState1634 | MenhirState1635 | MenhirState242 | MenhirState180 | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3537 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState111
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3535 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState111 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3533 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState112 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 57498 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce350 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv3534)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState112 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState112) : 'freshtv3536)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState111)) : 'freshtv3538)
    | MenhirState736 | MenhirState2196 | MenhirState1966 | MenhirState1414 | MenhirState1108 | MenhirState919 | MenhirState239 | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3543 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState136
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3541 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState136 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3539 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState137 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 57537 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce465 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3540)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState137) : 'freshtv3542)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState136)) : 'freshtv3544)
    | MenhirState393 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3545 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState394
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState394
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState394
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState394) : 'freshtv3546)
    | MenhirState436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3547 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState437
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState437
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState437
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState437) : 'freshtv3548)
    | MenhirState472 | MenhirState1126 | MenhirState1115 | MenhirState1112 | MenhirState810 | MenhirState615 | MenhirState525 | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3553 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState518
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3551 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState518 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3549 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState519 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 57605 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce465 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3550)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState519 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState519) : 'freshtv3552)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState518)) : 'freshtv3554)
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1519 | MenhirState1513 | MenhirState1507 | MenhirState1500 | MenhirState1485 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3555 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState553
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState553
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3556)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2603 | MenhirState2599 | MenhirState2595 | MenhirState2590 | MenhirState2584 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1628 | MenhirState1178 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3557 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState634
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState634
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3558)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState463 | MenhirState2268 | MenhirState473 | MenhirState2262 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1641 | MenhirState1642 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState835 | MenhirState837 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState811 | MenhirState812 | MenhirState806 | MenhirState800 | MenhirState745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3561 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState748
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3559 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState748 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run2196 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run2191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run2181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run2176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run750 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState749 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState749) : 'freshtv3560)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3562)
    | MenhirState2638 | MenhirState2639 | MenhirState410 | MenhirState2360 | MenhirState450 | MenhirState2282 | MenhirState751 | MenhirState2166 | MenhirState754 | MenhirState756 | MenhirState2136 | MenhirState2137 | MenhirState2132 | MenhirState2133 | MenhirState2122 | MenhirState2123 | MenhirState1240 | MenhirState1241 | MenhirState794 | MenhirState795 | MenhirState767 | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3563 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState775
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState775
          | AND | COLON | EOF | EQUAL | EQUALGREATER | LBRACE | LBRACKETAT | LBRACKETATAT | LBRACKETPERCENT | LPAREN | RPAREN | SEMI | UIDENT _ | WITH ->
              _menhir_reduce250 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState775)) : 'freshtv3564)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2681 | MenhirState2682 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState416 | MenhirState2344 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState730 | MenhirState2207 | MenhirState740 | MenhirState744 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState759 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState889 | MenhirState890 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState948 | MenhirState950 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState924 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 | MenhirState788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3581 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState791
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3579 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState791 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3567 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState792 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | DOTDOTDOT ->
                      _menhir_run1110 _menhir_env (Obj.magic _menhir_stack) MenhirState1108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | RBRACE ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : (('freshtv3565 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_menhir_s : _menhir_state) = MenhirState1108 in
                      ((let _menhir_env = _menhir_discard _menhir_env in
                      _menhir_reduce216 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv3566)
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1108) : 'freshtv3568)
              | LBRACELESS ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3569 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState792 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1103) : 'freshtv3570)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3571 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState792 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOTDOTDOT ->
                      _menhir_run1094 _menhir_env (Obj.magic _menhir_stack) MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1093 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1093) : 'freshtv3572)
              | LBRACKETBAR ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3573 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState792 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1076) : 'freshtv3574)
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState792 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3577 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState792 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | AMPERAMPER ->
                      _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERSAND ->
                      _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | ASSERT ->
                      _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BARBAR ->
                      _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONEQUAL ->
                      _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | GREATER ->
                      _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP0 _v ->
                      _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP1 _v ->
                      _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP2 _v ->
                      _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP3 _v ->
                      _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP4 _v ->
                      _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : (('freshtv3575 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_menhir_s : _menhir_state) = MenhirState793 in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | FUN ->
                          _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState794 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState794) : 'freshtv3576)
                  | NATIVEINT _v ->
                      _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | OR ->
                      _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PERCENT ->
                      _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSEQ ->
                      _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STAR ->
                      _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState793) : 'freshtv3578)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState792 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState792) : 'freshtv3580)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3582)
    | MenhirState1153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3583 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1154
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1154
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1154
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1154) : 'freshtv3584)
    | MenhirState1395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3585 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1396
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1396
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1396
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1396) : 'freshtv3586)
    | MenhirState1424 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3589 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1428
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1428
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3587 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1428 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1429 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1429) : 'freshtv3588)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1428)) : 'freshtv3590)
    | MenhirState1453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3591 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 58128 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1454
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1454
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1454
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1454) : 'freshtv3592)
    | MenhirState1974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3595 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1991
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1991
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3593 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1991 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run1977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1992) : 'freshtv3594)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1991)) : 'freshtv3596)
    | MenhirState402 | MenhirState2389 | MenhirState403 | MenhirState2387 | MenhirState411 | MenhirState2358 | MenhirState2300 | MenhirState2301 | MenhirState2296 | MenhirState2297 | MenhirState2289 | MenhirState2290 | MenhirState2285 | MenhirState2286 | MenhirState757 | MenhirState2159 | MenhirState1470 | MenhirState2019 | MenhirState1939 | MenhirState1941 | MenhirState1942 | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3597 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2010
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState2010
          | AND | COLON | EOF | EQUALGREATER | LBRACE | LBRACKETAT | LBRACKETATAT | LBRACKETPERCENT | LPAREN | RBRACE | RBRACKET | RPAREN | SEMI | UIDENT _ | WITH ->
              _menhir_reduce250 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2010)) : 'freshtv3598)
    | MenhirState431 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3603 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2241
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3601 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2241 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run2196 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run2191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run2181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run2176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3599 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2242 in
                  let (_v : (
# 863 "reason_parser.mly"
       (string)
# 58218 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
                  | COLON | COMMA ->
                      _menhir_reduce465 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3600)
              | LPAREN ->
                  _menhir_run750 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2242) : 'freshtv3602)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3604)
    | MenhirState2614 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3609 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2620
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3605 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2620 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2621 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2621 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2621) : 'freshtv3606)
          | EOF | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3607 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_longident), _startpos__3_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4171 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_ident _3) )
# 58273 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3608)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2620)) : 'freshtv3610)
    | _ ->
        _menhir_fail ()

and _menhir_goto_nonrec_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_nonrec_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState74 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3521 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3519) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState76 in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 58301 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MINUS ->
                _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState90) : 'freshtv3520)
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState76 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState76) : 'freshtv3522)
    | MenhirState1247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3523 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run1249 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1248) : 'freshtv3524)
    | MenhirState2577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3525 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run1249 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2578) : 'freshtv3526)
    | _ ->
        _menhir_fail ()

and _menhir_goto_additive : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_additive -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState766 | MenhirState1142 | MenhirState1137 | MenhirState1129 | MenhirState1118 | MenhirState1076 | MenhirState1079 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1042 | MenhirState1046 | MenhirState811 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState828 | MenhirState881 | MenhirState855 | MenhirState875 | MenhirState877 | MenhirState879 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState861 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState853 | MenhirState851 | MenhirState845 | MenhirState849 | MenhirState847 | MenhirState830 | MenhirState842 | MenhirState832 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3515 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState842) : 'freshtv3516)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2568 | MenhirState2416 | MenhirState2368 | MenhirState2370 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState765 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1166 | MenhirState1171 | MenhirState786 | MenhirState1133 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1027 | MenhirState1029 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1003 | MenhirState1008 | MenhirState1010 | MenhirState1006 | MenhirState1001 | MenhirState924 | MenhirState941 | MenhirState994 | MenhirState968 | MenhirState988 | MenhirState990 | MenhirState992 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState974 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState966 | MenhirState964 | MenhirState958 | MenhirState962 | MenhirState960 | MenhirState943 | MenhirState955 | MenhirState945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3517 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState955) : 'freshtv3518)
    | _ ->
        _menhir_fail ()

and _menhir_goto_override_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_override_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState391 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3487 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState393 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState393) : 'freshtv3488)
    | MenhirState435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3489 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState436 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState436) : 'freshtv3490)
    | MenhirState1152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1153) : 'freshtv3492)
    | MenhirState1394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1395 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1395) : 'freshtv3494)
    | MenhirState1661 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3501 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3499 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1670 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | VIRTUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3495 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LIDENT _v ->
                    _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1672 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1672) : 'freshtv3496)
            | LIDENT _ ->
                _menhir_reduce588 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3497 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3498)) : 'freshtv3500)
        | LIDENT _ ->
            _menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) MenhirState1670
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1670) : 'freshtv3502)
    | MenhirState1684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3511 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PRIVATE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3507 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1685 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | VIRTUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3503 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LIDENT _v ->
                    _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1692) : 'freshtv3504)
            | LIDENT _ ->
                _menhir_reduce678 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3505 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3506)) : 'freshtv3508)
        | VIRTUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3509 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1685 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState1686
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1686) : 'freshtv3510)
        | LIDENT _ ->
            _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) MenhirState1685
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1685) : 'freshtv3512)
    | MenhirState1731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3513 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | CLASS ->
            _menhir_run1635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1626 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1732 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1732) : 'freshtv3514)
    | _ ->
        _menhir_fail ()

and _menhir_reduce349 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 58706 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 863 "reason_parser.mly"
       (string)
# 58711 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_longident = 
# 4160 "reason_parser.mly"
                                                ( Lident _1 )
# 58718 "reason_parser.ml"
     in
    _menhir_goto_class_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_subtractive : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_subtractive -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1471 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 | MenhirState808 | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3483 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState808 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState808) : 'freshtv3484)
    | MenhirState2747 | MenhirState2754 | MenhirState2751 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 | MenhirState917 | MenhirState902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3485 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState917) : 'freshtv3486)
    | _ ->
        _menhir_fail ()

and _menhir_run1698 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3481 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1698 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3477 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 58992 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 59004 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run1702 _menhir_env (Obj.magic _menhir_stack) MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run1698 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1701 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1701) : 'freshtv3474)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3475 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 59069 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3476)) : 'freshtv3478)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3479 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3480)) : 'freshtv3482)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1698

and _menhir_run1702 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1702

and _menhir_run1203 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 59169 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run1203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1203 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3471 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 59183 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 59188 "reason_parser.ml"
        )), _startpos__1_) = _menhir_stack in
        let _v : 'tv_lident_list = 
# 2874 "reason_parser.mly"
                                      ( [_1] )
# 59193 "reason_parser.ml"
         in
        _menhir_goto_lident_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3472)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1203

and _menhir_error1693 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1688 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1061 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1061 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1061

and _menhir_run1063 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1063

and _menhir_error1673 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1664 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1570 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1570

and _menhir_error1568 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1559 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1554 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run107 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce140 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run242 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState242

and _menhir_run115 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState115

and _menhir_run244 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState244

and _menhir_run322 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 59402 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3469 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 59414 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState323) : 'freshtv3470)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run281 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce144 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run282 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce143 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run283 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOT ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState283
    | LIDENT _v ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState283 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState283

and _menhir_run290 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState290 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState290
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState290

and _menhir_run300 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState300 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3467 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState300 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce149 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv3468)
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState300
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState300

and _menhir_run305 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3465 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState305 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState306) : 'freshtv3466)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState305 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState305

and _menhir_error1549 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error940 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error827 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce243 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2787 "reason_parser.mly"
      ( mkexp(Pexp_send(_1, _3)) )
# 59621 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run153 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState153

and _menhir_error151 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_val_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState2262 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState475 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1933 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1840 | MenhirState1471 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1642 | MenhirState1641 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState837 | MenhirState835 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState822 | MenhirState815 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState811 | MenhirState808 | MenhirState793 | MenhirState806 | MenhirState800 | MenhirState745 | MenhirState463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3393 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce833 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3394)
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState564 | MenhirState563 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3395 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce268 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3396)
    | MenhirState577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3397 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce164 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3398)
    | MenhirState2698 | MenhirState2657 | MenhirState2673 | MenhirState2671 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2232 | MenhirState476 | MenhirState480 | MenhirState2218 | MenhirState484 | MenhirState1906 | MenhirState1881 | MenhirState1885 | MenhirState1817 | MenhirState1621 | MenhirState1697 | MenhirState1708 | MenhirState1701 | MenhirState1651 | MenhirState1633 | MenhirState1631 | MenhirState1222 | MenhirState1179 | MenhirState1183 | MenhirState1194 | MenhirState1187 | MenhirState1036 | MenhirState893 | MenhirState897 | MenhirState1022 | MenhirState901 | MenhirState903 | MenhirState762 | MenhirState738 | MenhirState486 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState650 | MenhirState649 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3399 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce268 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3400)
    | MenhirState658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3401 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce164 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3402)
    | MenhirState2754 | MenhirState2751 | MenhirState2747 | MenhirState2737 | MenhirState2576 | MenhirState2682 | MenhirState2681 | MenhirState2614 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState2344 | MenhirState416 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState1156 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState950 | MenhirState948 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState935 | MenhirState928 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState924 | MenhirState917 | MenhirState902 | MenhirState915 | MenhirState908 | MenhirState907 | MenhirState905 | MenhirState890 | MenhirState889 | MenhirState788 | MenhirState759 | MenhirState744 | MenhirState740 | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3403 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce833 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3404)
    | MenhirState2621 | MenhirState792 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3405 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce834 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3406)
    | MenhirState1628 | MenhirState1178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3425 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1179
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3409 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1179 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run153 _menhir_env (Obj.magic _menhir_stack) MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3407 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1201 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1202) : 'freshtv3408)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1201 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1201) : 'freshtv3410)
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1179
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3411) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1179 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1199) : 'freshtv3412)
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3423) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1179 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AMPERAMPER ->
                  _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERSAND ->
                  _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BARBAR ->
                  _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONEQUAL ->
                  _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | GREATER ->
                  _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP0 _v ->
                  _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP1 _v ->
                  _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP2 _v ->
                  _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP3 _v ->
                  _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP4 _v ->
                  _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSEQ ->
                  _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RPAREN ->
                  _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STAR ->
                  _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv3421 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1180 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_v : (
# 863 "reason_parser.mly"
       (string)
# 59961 "reason_parser.ml"
                      )) = _v in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | RPAREN ->
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 59973 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                          ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                          let _menhir_env = _menhir_discard _menhir_env in
                          let _tok = _menhir_env._menhir_token in
                          match _tok with
                          | BACKQUOTE ->
                              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | CHAR _v ->
                              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | COLON ->
                              _menhir_run1190 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | EQUALGREATER ->
                              _menhir_run1188 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FALSE ->
                              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FLOAT _v ->
                              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT _v ->
                              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT32 _v ->
                              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT64 _v ->
                              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACE ->
                              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKET ->
                              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETBAR ->
                              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETPERCENT ->
                              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LIDENT _v ->
                              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LPAREN ->
                              _menhir_run1184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | MINUS ->
                              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | NATIVEINT _v ->
                              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | PLUS ->
                              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | SHARP ->
                              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | STRING _v ->
                              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | TRUE ->
                              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UIDENT _v ->
                              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UNDERSCORE ->
                              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | _ ->
                              assert (not _menhir_env._menhir_error);
                              _menhir_env._menhir_error <- true;
                              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1183) : 'freshtv3414)
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 60038 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3416)) : 'freshtv3418)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3420)) : 'freshtv3422)
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1180) : 'freshtv3424)
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce268 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1179)) : 'freshtv3426)
    | MenhirState1403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3429 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1464 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1465) : 'freshtv3428)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1464 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3430)
    | MenhirState1472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1473 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3431 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1474) : 'freshtv3432)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1473 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3434)
    | MenhirState2308 | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3453 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1881
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3437 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1881 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run153 _menhir_env (Obj.magic _menhir_stack) MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3435 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1889 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1890 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1890) : 'freshtv3436)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1889) : 'freshtv3438)
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1881
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3439) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1881 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1887) : 'freshtv3440)
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3451) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1881 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AMPERAMPER ->
                  _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERSAND ->
                  _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BARBAR ->
                  _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONEQUAL ->
                  _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | GREATER ->
                  _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP0 _v ->
                  _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP1 _v ->
                  _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP2 _v ->
                  _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP3 _v ->
                  _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP4 _v ->
                  _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSEQ ->
                  _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RPAREN ->
                  _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STAR ->
                  _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv3449 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1882 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3445 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_v : (
# 863 "reason_parser.mly"
       (string)
# 60465 "reason_parser.ml"
                      )) = _v in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | RPAREN ->
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3441 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 60477 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                          ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                          let _menhir_env = _menhir_discard _menhir_env in
                          let _tok = _menhir_env._menhir_token in
                          match _tok with
                          | BACKQUOTE ->
                              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | CHAR _v ->
                              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | COLON ->
                              _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | EQUALGREATER ->
                              _menhir_run1702 _menhir_env (Obj.magic _menhir_stack) MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FALSE ->
                              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FLOAT _v ->
                              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT _v ->
                              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT32 _v ->
                              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT64 _v ->
                              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACE ->
                              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKET ->
                              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETBAR ->
                              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETPERCENT ->
                              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LIDENT _v ->
                              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LPAREN ->
                              _menhir_run1698 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | MINUS ->
                              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | NATIVEINT _v ->
                              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | PLUS ->
                              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | SHARP ->
                              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | STRING _v ->
                              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | TRUE ->
                              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UIDENT _v ->
                              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UNDERSCORE ->
                              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | _ ->
                              assert (not _menhir_env._menhir_error);
                              _menhir_env._menhir_error <- true;
                              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1885) : 'freshtv3442)
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3443 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 60542 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3444)) : 'freshtv3446)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3448)) : 'freshtv3450)
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1882) : 'freshtv3452)
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce268 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1881)) : 'freshtv3454)
    | MenhirState2033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3457 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2034 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3455 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2035) : 'freshtv3456)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2034 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3458)
    | MenhirState2242 | MenhirState749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3459 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce834 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3460)
    | MenhirState2641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3463 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2642 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3461 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2643) : 'freshtv3462)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2642 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3464)
    | _ ->
        _menhir_fail ()

and _menhir_goto_rec_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_rec_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2629 | MenhirState1151 | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3389 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1628 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1628) : 'freshtv3390)
    | MenhirState433 | MenhirState440 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3391 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2308 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2308) : 'freshtv3392)
    | _ ->
        _menhir_fail ()

and _menhir_reduce776 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4289 "reason_parser.mly"
         ( "with" )
# 60819 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce775 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4288 "reason_parser.mly"
          ( "while" )
# 60831 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce774 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4287 "reason_parser.mly"
         ( "when" )
# 60843 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce773 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4286 "reason_parser.mly"
            ( "virtual" )
# 60855 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce772 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4285 "reason_parser.mly"
        ( "val" )
# 60867 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce728 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 60874 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 910 "reason_parser.mly"
       (string)
# 60879 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4241 "reason_parser.mly"
           ( _1 )
# 60886 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce771 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4284 "reason_parser.mly"
         ( "type" )
# 60898 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce770 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4283 "reason_parser.mly"
        ( "try" )
# 60910 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce769 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4282 "reason_parser.mly"
         ( "true" )
# 60922 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce768 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4281 "reason_parser.mly"
       ( "to" )
# 60934 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce767 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4280 "reason_parser.mly"
         ( "then" )
# 60946 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce754 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4267 "reason_parser.mly"
           ( "match" )
# 60958 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce766 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4279 "reason_parser.mly"
           ( "struct" )
# 60970 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce765 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4278 "reason_parser.mly"
        ( "sig" )
# 60982 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce764 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4277 "reason_parser.mly"
        ( "rec" )
# 60994 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce763 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4276 "reason_parser.mly"
            ( "private" )
# 61006 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce762 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4275 "reason_parser.mly"
       ( "or" )
# 61018 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce761 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4274 "reason_parser.mly"
         ( "open" )
# 61030 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce760 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4273 "reason_parser.mly"
           ( "object" )
# 61042 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce759 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4272 "reason_parser.mly"
           ( "nonrec" )
# 61054 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce758 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4271 "reason_parser.mly"
        ( "new" )
# 61066 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce757 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4270 "reason_parser.mly"
            ( "mutable" )
# 61078 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce756 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4269 "reason_parser.mly"
           ( "module" )
# 61090 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce755 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4268 "reason_parser.mly"
           ( "method" )
# 61102 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce727 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 61109 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 863 "reason_parser.mly"
       (string)
# 61114 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4240 "reason_parser.mly"
           ( _1 )
# 61121 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce753 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4266 "reason_parser.mly"
        ( "let" )
# 61133 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce752 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4265 "reason_parser.mly"
         ( "lazy" )
# 61145 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce751 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4264 "reason_parser.mly"
                ( "initializer" )
# 61157 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce750 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4263 "reason_parser.mly"
            ( "inherit" )
# 61169 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce749 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4262 "reason_parser.mly"
            ( "include" )
# 61181 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce748 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4261 "reason_parser.mly"
       ( "in" )
# 61193 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce747 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4260 "reason_parser.mly"
       ( "if" )
# 61205 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce746 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4259 "reason_parser.mly"
            ( "functor" )
# 61217 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce745 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4258 "reason_parser.mly"
             ( "function" )
# 61229 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce744 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4257 "reason_parser.mly"
        ( "fun" )
# 61241 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce743 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4256 "reason_parser.mly"
        ( "for" )
# 61253 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce742 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4255 "reason_parser.mly"
          ( "false" )
# 61265 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce741 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4254 "reason_parser.mly"
             ( "external" )
# 61277 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce740 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4253 "reason_parser.mly"
              ( "exception" )
# 61289 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce739 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4252 "reason_parser.mly"
        ( "end" )
# 61301 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce738 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4251 "reason_parser.mly"
         ( "else" )
# 61313 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce737 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4250 "reason_parser.mly"
           ( "downto" )
# 61325 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce736 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4249 "reason_parser.mly"
         ( "done" )
# 61337 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce735 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4248 "reason_parser.mly"
       ( "do" )
# 61349 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce734 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4247 "reason_parser.mly"
               ( "constraint" )
# 61361 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce733 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4246 "reason_parser.mly"
          ( "class" )
# 61373 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce732 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4245 "reason_parser.mly"
          ( "begin" )
# 61385 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce731 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4244 "reason_parser.mly"
           ( "assert" )
# 61397 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce730 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4243 "reason_parser.mly"
       ( "as" )
# 61409 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce729 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4242 "reason_parser.mly"
        ( "and" )
# 61421 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce374 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constr_longident = 
# 4112 "reason_parser.mly"
                                                ( Lident "[]" )
# 61435 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce433 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 61442 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 61448 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _v : 'tv_field_expr = let _endpos = _endpos__1_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3224 "reason_parser.mly"
    (
      let loc = mklocation _symbolstartpos _endpos in
      let lident_loc = mkloc _1 loc in
      let lident_lident_loc = mkloc (Lident _1) loc in
      (lident_loc, mkexp (Pexp_ident lident_lident_loc))
    )
# 61460 "reason_parser.ml"
     in
    _menhir_goto_field_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto__simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState463 | MenhirState2268 | MenhirState473 | MenhirState2262 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState745 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1641 | MenhirState1642 | MenhirState766 | MenhirState1142 | MenhirState1137 | MenhirState1129 | MenhirState1118 | MenhirState1076 | MenhirState1079 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1042 | MenhirState1046 | MenhirState811 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState828 | MenhirState881 | MenhirState855 | MenhirState875 | MenhirState877 | MenhirState879 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState861 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState853 | MenhirState851 | MenhirState845 | MenhirState849 | MenhirState847 | MenhirState830 | MenhirState832 | MenhirState842 | MenhirState835 | MenhirState837 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState806 | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3385) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce534 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3386)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2681 | MenhirState2682 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2368 | MenhirState2370 | MenhirState415 | MenhirState416 | MenhirState2344 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState730 | MenhirState2207 | MenhirState740 | MenhirState744 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState758 | MenhirState759 | MenhirState2141 | MenhirState765 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1166 | MenhirState1171 | MenhirState786 | MenhirState1133 | MenhirState788 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState889 | MenhirState890 | MenhirState892 | MenhirState1031 | MenhirState1027 | MenhirState1029 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1003 | MenhirState1008 | MenhirState1010 | MenhirState1006 | MenhirState1001 | MenhirState924 | MenhirState941 | MenhirState994 | MenhirState968 | MenhirState988 | MenhirState990 | MenhirState992 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState974 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState966 | MenhirState964 | MenhirState958 | MenhirState962 | MenhirState960 | MenhirState943 | MenhirState945 | MenhirState955 | MenhirState948 | MenhirState950 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3387) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce534 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3388)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_self_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_self_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3383) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_self_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3381) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_self_pattern) : 'tv__class_self_pattern) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__class_self_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos_x_ _endpos_x_ then
      _startpos_x_
    else
      _endpos in
    
# 1206 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 61510 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3379) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3377) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3375) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_pat__class_self_pattern_) : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_self_pattern = 
# 1902 "reason_parser.mly"
                                                           (_1)
# 61535 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3373) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_self_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3371 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1660
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | CONSTRAINT ->
          _menhir_run1738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INHERIT ->
          _menhir_run1731 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INITIALIZER ->
          _menhir_run1727 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETATATAT ->
          _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETPERCENTPERCENT ->
          _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | METHOD ->
          _menhir_run1684 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | VAL ->
          _menhir_run1661 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | RBRACE ->
          _menhir_reduce692 _menhir_env (Obj.magic _menhir_stack) MenhirState1660
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1660)) : 'freshtv3372)) : 'freshtv3374)) : 'freshtv3376)) : 'freshtv3378)) : 'freshtv3380)) : 'freshtv3382)) : 'freshtv3384)

and _menhir_reduce464 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 61575 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 61581 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_label_longident = 
# 4118 "reason_parser.mly"
                                                ( Lident _1 )
# 61588 "reason_parser.ml"
     in
    _menhir_goto_label_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1240 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1240 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1240 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1240 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1240 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1240 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1240

and _menhir_run1242 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1242

and _menhir_run2158 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce259 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run443 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3367 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState443 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3365 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_functor_arg = 
# 1221 "reason_parser.mly"
 ( mkloc "*" (mklocation _startpos__2_ _endpos__2_), None )
# 61666 "reason_parser.ml"
         in
        _menhir_goto_functor_arg _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv3366)) : 'freshtv3368)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3369) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState443 in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 61677 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce443 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3370)
    | UNDERSCORE ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState443

and _menhir_run212 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 61693 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce375 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constr_longident = 
# 4113 "reason_parser.mly"
                                                ( Lident "()" )
# 61710 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce605 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 61717 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 887 "reason_parser.mly"
       (string)
# 61723 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _v : 'tv_operator = 
# 4091 "reason_parser.mly"
                                                ( _1 )
# 61728 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce613 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4074 "reason_parser.mly"
                                                ( "+." )
# 61741 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 61747 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce612 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4073 "reason_parser.mly"
                                                ( "+" )
# 61760 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 61766 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_run222 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 61773 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce714 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run223 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 61782 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce713 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run224 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 61791 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce712 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run225 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 61800 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce710 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run226 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 61809 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce711 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_reduce289 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 61818 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 61824 "reason_parser.ml"
    )), _startpos_x0_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 61840 "reason_parser.ml"
      
    in
    
# 3379 "reason_parser.mly"
      ( mkpat(Ppat_unpack _3) )
# 61846 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run233 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 61853 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
        _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState233

and _menhir_reduce615 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4076 "reason_parser.mly"
                                                ( "-." )
# 61880 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 61886 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce614 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4075 "reason_parser.mly"
                                                ( "-" )
# 61899 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 61905 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_run490 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 61912 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce709 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run491 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 61921 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce708 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run492 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 61930 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce707 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run493 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 61939 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce705 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run494 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 61948 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce706 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run496 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3361 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 61967 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3355 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 61979 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState499) : 'freshtv3356)
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3357 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 62000 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce289 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3358)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3359 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 62013 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3360)) : 'freshtv3362)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3364)

and _menhir_run533 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run496 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState533

and _menhir_reduce606 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_operator = 
# 4092 "reason_parser.mly"
                                                ( "!" )
# 62135 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_operator : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_operator -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState183 | MenhirState407 | MenhirState750 | MenhirState495 | MenhirState533 | MenhirState579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3347 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3343 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce832 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3344)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3345 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3346)) : 'freshtv3348)
    | MenhirState15 | MenhirState209 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState766 | MenhirState793 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3353 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3349 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce832 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3350)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3351 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3352)) : 'freshtv3354)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constr_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constr_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState532 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3307 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce273 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3308)
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3309 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState563
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce273 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3310)
    | MenhirState2657 | MenhirState2698 | MenhirState2671 | MenhirState2673 | MenhirState476 | MenhirState2232 | MenhirState480 | MenhirState484 | MenhirState2218 | MenhirState486 | MenhirState1881 | MenhirState1906 | MenhirState1885 | MenhirState1621 | MenhirState1817 | MenhirState1697 | MenhirState1701 | MenhirState1708 | MenhirState1631 | MenhirState1633 | MenhirState1179 | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 | MenhirState893 | MenhirState1036 | MenhirState897 | MenhirState901 | MenhirState1022 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3311 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce273 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3312)
    | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState738 | MenhirState1880 | MenhirState1651 | MenhirState1628 | MenhirState1178 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3313 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState649
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce273 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3314)
    | MenhirState2549 | MenhirState179 | MenhirState186 | MenhirState456 | MenhirState463 | MenhirState2262 | MenhirState745 | MenhirState1641 | MenhirState1642 | MenhirState835 | MenhirState837 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState806 | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3315 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3316)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState766 | MenhirState1142 | MenhirState1137 | MenhirState1129 | MenhirState1118 | MenhirState1076 | MenhirState1079 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1042 | MenhirState1046 | MenhirState811 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState828 | MenhirState881 | MenhirState855 | MenhirState875 | MenhirState877 | MenhirState879 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState861 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState853 | MenhirState851 | MenhirState845 | MenhirState849 | MenhirState847 | MenhirState830 | MenhirState842 | MenhirState832 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3317 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState835
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce203 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3318)
    | MenhirState2681 | MenhirState2682 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState416 | MenhirState2344 | MenhirState418 | MenhirState474 | MenhirState729 | MenhirState730 | MenhirState740 | MenhirState744 | MenhirState759 | MenhirState788 | MenhirState889 | MenhirState890 | MenhirState948 | MenhirState950 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3319 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3320)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2568 | MenhirState2416 | MenhirState2368 | MenhirState2370 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState765 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1166 | MenhirState1171 | MenhirState786 | MenhirState1133 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1027 | MenhirState1029 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1003 | MenhirState1008 | MenhirState1010 | MenhirState1006 | MenhirState1001 | MenhirState924 | MenhirState941 | MenhirState994 | MenhirState968 | MenhirState988 | MenhirState990 | MenhirState992 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState974 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState966 | MenhirState964 | MenhirState958 | MenhirState962 | MenhirState960 | MenhirState943 | MenhirState955 | MenhirState945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3321 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState948
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce203 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3322)
    | MenhirState1485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3323 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 62419 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1488
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1488
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1488) : 'freshtv3324)
    | MenhirState1500 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3325 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1501
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1501
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1501) : 'freshtv3326)
    | MenhirState1507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3327 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1508
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1508
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1508) : 'freshtv3328)
    | MenhirState1513 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3329 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1514
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1514
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1514) : 'freshtv3330)
    | MenhirState1519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1520
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run332 _menhir_env (Obj.magic _menhir_stack) MenhirState1520
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1520) : 'freshtv3332)
    | MenhirState2584 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3333 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 62483 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2587
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2587
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2587) : 'freshtv3334)
    | MenhirState2590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2591
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2591
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2591) : 'freshtv3336)
    | MenhirState2595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2596
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2596
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2596) : 'freshtv3338)
    | MenhirState2599 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3339 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2600
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2600
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2600) : 'freshtv3340)
    | MenhirState2603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3341 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2604
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run173 _menhir_env (Obj.magic _menhir_stack) MenhirState2604
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2604) : 'freshtv3342)
    | _ ->
        _menhir_fail ()

and _menhir_run580 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce612 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run581 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run2585 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2585 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2585

and _menhir_run2586 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2586
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2586

and _menhir_run1269 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1269

and _menhir_fail : unit -> 'a =
  fun () ->
    Printf.fprintf Pervasives.stderr "Internal failure -- please contact the parser generator's developers.\n%!";
    assert false

and _menhir_goto_virtual_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_virtual_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1870 | MenhirState1530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3279 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3275 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 62642 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) MenhirState1532) : 'freshtv3276)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3277 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3278)) : 'freshtv3280)
    | MenhirState1552 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3281 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1553 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1553) : 'freshtv3282)
    | MenhirState1875 | MenhirState1528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3287 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3283 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 62680 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) MenhirState1620) : 'freshtv3284)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3285 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3286)) : 'freshtv3288)
    | MenhirState2081 | MenhirState2046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3293 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3289 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 62706 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) MenhirState2048) : 'freshtv3290)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3291 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3292)) : 'freshtv3294)
    | MenhirState2086 | MenhirState2045 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3299 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3295 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 62732 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) MenhirState2055) : 'freshtv3296)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3297 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3298)) : 'freshtv3300)
    | MenhirState2732 | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3305 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3301 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 62758 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) MenhirState2656) : 'freshtv3302)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3303 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3304)) : 'freshtv3306)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constant : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState209 | MenhirState2411 | MenhirState477 | MenhirState481 | MenhirState1882 | MenhirState1698 | MenhirState1180 | MenhirState1184 | MenhirState894 | MenhirState898 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3267) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3268)
    | MenhirState2657 | MenhirState2698 | MenhirState2671 | MenhirState2673 | MenhirState2572 | MenhirState190 | MenhirState2409 | MenhirState2308 | MenhirState476 | MenhirState2232 | MenhirState480 | MenhirState484 | MenhirState2218 | MenhirState486 | MenhirState738 | MenhirState1880 | MenhirState1881 | MenhirState1906 | MenhirState1885 | MenhirState1621 | MenhirState1817 | MenhirState1697 | MenhirState1701 | MenhirState1708 | MenhirState1651 | MenhirState1631 | MenhirState1633 | MenhirState1628 | MenhirState1178 | MenhirState1179 | MenhirState1222 | MenhirState1183 | MenhirState1187 | MenhirState1194 | MenhirState893 | MenhirState1036 | MenhirState897 | MenhirState901 | MenhirState1022 | MenhirState903 | MenhirState762 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3269) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3270)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState2549 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState179 | MenhirState183 | MenhirState186 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState456 | MenhirState463 | MenhirState2268 | MenhirState473 | MenhirState2262 | MenhirState2264 | MenhirState475 | MenhirState2247 | MenhirState2249 | MenhirState2253 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState745 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1840 | MenhirState1923 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1641 | MenhirState1642 | MenhirState766 | MenhirState1142 | MenhirState1137 | MenhirState1129 | MenhirState1118 | MenhirState1076 | MenhirState1079 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1042 | MenhirState1046 | MenhirState811 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState828 | MenhirState881 | MenhirState855 | MenhirState875 | MenhirState877 | MenhirState879 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState861 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState853 | MenhirState851 | MenhirState845 | MenhirState849 | MenhirState847 | MenhirState830 | MenhirState842 | MenhirState832 | MenhirState835 | MenhirState837 | MenhirState815 | MenhirState822 | MenhirState820 | MenhirState817 | MenhirState812 | MenhirState806 | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3271) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce202 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3272)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2576 | MenhirState2737 | MenhirState2681 | MenhirState2682 | MenhirState2568 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState176 | MenhirState2416 | MenhirState2368 | MenhirState2370 | MenhirState415 | MenhirState416 | MenhirState2344 | MenhirState417 | MenhirState418 | MenhirState2271 | MenhirState474 | MenhirState2223 | MenhirState729 | MenhirState730 | MenhirState2207 | MenhirState740 | MenhirState744 | MenhirState2181 | MenhirState2186 | MenhirState2182 | MenhirState758 | MenhirState759 | MenhirState2141 | MenhirState765 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1162 | MenhirState1166 | MenhirState1171 | MenhirState786 | MenhirState1133 | MenhirState788 | MenhirState1110 | MenhirState1093 | MenhirState1098 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState889 | MenhirState890 | MenhirState892 | MenhirState1031 | MenhirState1027 | MenhirState1029 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1003 | MenhirState1008 | MenhirState1010 | MenhirState1006 | MenhirState1001 | MenhirState924 | MenhirState941 | MenhirState994 | MenhirState968 | MenhirState988 | MenhirState990 | MenhirState992 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState974 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState966 | MenhirState964 | MenhirState958 | MenhirState962 | MenhirState960 | MenhirState943 | MenhirState955 | MenhirState945 | MenhirState948 | MenhirState950 | MenhirState928 | MenhirState935 | MenhirState933 | MenhirState930 | MenhirState925 | MenhirState915 | MenhirState905 | MenhirState907 | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3273) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce202 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3274)
    | _ ->
        _menhir_fail ()

and _menhir_reduce452 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 62815 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 863 "reason_parser.mly"
       (string)
# 62820 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_ident = 
# 4059 "reason_parser.mly"
                                                ( _1 )
# 62827 "reason_parser.ml"
     in
    _menhir_goto_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_structure : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_structure -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3223 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3221 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_structure_item), _startpos_x0_), _endpos__2_, _startpos__2_), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 62851 "reason_parser.ml"
          
        in
        
# 1380 "reason_parser.mly"
                                           (
    syntax_error_str _1.loc "Statement has to end with a semicolon" :: _3
  )
# 62859 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3222)) : 'freshtv3224)
    | MenhirState1933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3227 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3225 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_), _endpos__2_), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_structure = 
# 1383 "reason_parser.mly"
                                  (
      let effective_loc = mklocation _startpos__1_ _endpos__2_ in
      set_structure_item_location _1 effective_loc :: _3
  )
# 62875 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3226)) : 'freshtv3228)
    | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3233 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3229 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce251 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3230)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3231 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3232)) : 'freshtv3234)
    | MenhirState73 | MenhirState174 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState439 | MenhirState453 | MenhirState455 | MenhirState2027 | MenhirState1469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3237 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3235 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_structure)) = _menhir_stack in
        let _v : 'tv_payload = 
# 4345 "reason_parser.mly"
              ( PStr _1 )
# 62906 "reason_parser.ml"
         in
        _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3236)) : 'freshtv3238)
    | MenhirState451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3243 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3239 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce251 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3240)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3241 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3242)) : 'freshtv3244)
    | MenhirState2551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3247 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3245 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let x0 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 62945 "reason_parser.ml"
          
        in
        
# 1378 "reason_parser.mly"
                                 ( syntax_error_str _1.loc "Invalid statement" :: _3 )
# 62951 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3246)) : 'freshtv3248)
    | MenhirState1 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3251 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3249 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_structure)) = _menhir_stack in
        let x0 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 62970 "reason_parser.ml"
          
        in
        
# 1377 "reason_parser.mly"
                            ( syntax_error_str _1.loc "Invalid statement" :: _2 )
# 62976 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3250)) : 'freshtv3252)
    | MenhirState0 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3265 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EOF ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3261 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3259 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_structure)) = _menhir_stack in
            let _2 = () in
            let _v : (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 62995 "reason_parser.ml"
            ) = 
# 1119 "reason_parser.mly"
                                         ( apply_mapper_chain_to_structure _1 default_mapper_chain )
# 62999 "reason_parser.ml"
             in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3257) = _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 63007 "reason_parser.ml"
            )) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3255) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 63015 "reason_parser.ml"
            )) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3253) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let ((_1 : (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 63023 "reason_parser.ml"
            )) : (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 63027 "reason_parser.ml"
            )) = _v in
            (Obj.magic _1 : 'freshtv3254)) : 'freshtv3256)) : 'freshtv3258)) : 'freshtv3260)) : 'freshtv3262)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3263 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3264)) : 'freshtv3266)
    | _ ->
        _menhir_fail ()

and _menhir_run110 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 63043 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce349 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run185 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce375 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run410 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState410 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState410 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState410

and _menhir_run442 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 63078 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3217) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState442 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2287 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2287 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2287) : 'freshtv3218)
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3219) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState442 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2285 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2285 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2285 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2285 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2285 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2285) : 'freshtv3220)
    | LPAREN ->
        _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState442

and _menhir_run515 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s

and _menhir_run2182 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2182

and _menhir_run403 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1955 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3215 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState403 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState404) : 'freshtv3216)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState403

and _menhir_run1471 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState1471
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1471

and _menhir_run1937 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1937 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1937

and _menhir_run481 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3213 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState481 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3209 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 63551 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3205 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 63563 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run2214 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run2212 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState484) : 'freshtv3206)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 63628 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3208)) : 'freshtv3210)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3211 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3212)) : 'freshtv3214)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState481

and _menhir_run2212 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2212 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2212

and _menhir_run2214 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2214

and _menhir_run1484 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 63765 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3203 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 63779 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1484 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1485 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1485 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1485 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1485 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1485 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1485) : 'freshtv3204)
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1484
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1484

and _menhir_run1499 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3201 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1499 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1500) : 'freshtv3202)
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1499 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1499
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1499

and _menhir_run1505 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3197 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1506 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FALSE ->
                _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1507 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run1487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1507 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1507 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1507 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1507 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1507) : 'freshtv3196)
        | LBRACKET ->
            _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1506
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1506) : 'freshtv3198)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3199 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3200)

and _menhir_run1512 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3193 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1512 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1513) : 'freshtv3194)
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1512
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1512

and _menhir_run1518 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3191 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1518 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1519) : 'freshtv3192)
    | LBRACKET ->
        _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1518
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1518

and _menhir_run93 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 64109 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run94 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 64119 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce452 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run183 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState183

and _menhir_goto_signature : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3167 * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        (_menhir_reduce702 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3168)
    | MenhirState1402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3171 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2094 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3169 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce264 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3170)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2094 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3172)
    | MenhirState1246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3175 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2100 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3173 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce264 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3174)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2100 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3176)
    | MenhirState2559 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3177 * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        (_menhir_reduce702 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3178)
    | MenhirState2557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3189 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2561 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3187 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3185 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _menhir_s, (_1 : 'tv_signature)) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 64306 "reason_parser.ml"
              ) = 
# 1122 "reason_parser.mly"
                                         ( apply_mapper_chain_to_signature _1 default_mapper_chain )
# 64310 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3183) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 64318 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3181) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 64326 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3179) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 64334 "reason_parser.ml"
              )) : (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 64338 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv3180)) : 'freshtv3182)) : 'freshtv3184)) : 'freshtv3186)) : 'freshtv3188)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2561 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3190)
    | _ ->
        _menhir_fail ()

and _menhir_run1406 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3157 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1436 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1451) : 'freshtv3158)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3163 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 64378 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3159 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 64390 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1408) : 'freshtv3160)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3161 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 64421 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3162)) : 'freshtv3164)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3165 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3166)

and _menhir_run1436 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1436 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1436 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1436

and _menhir_run1459 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 64459 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3153 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 64471 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1460 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1460 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1460 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1460 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1460 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1460 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1460) : 'freshtv3154)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3155 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 64502 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3156)

and _menhir_run129 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 64510 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | AND | COLONGREATER | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState129

and _menhir_run449 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3149 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState450 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState450 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState450 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState450 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState450 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState450) : 'freshtv3150)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3151 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3152)

and _menhir_run1243 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run401 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1243 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1243

and _menhir_run1402 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2045 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2042 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2033 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2030 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1402 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState1402
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1402

and _menhir_run1370 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 64626 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1370 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1370
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1370

and _menhir_run1373 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1373 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1373
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1373

and _menhir_run1376 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3145 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1377
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1377) : 'freshtv3146)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3148)

and _menhir_run1380 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1380 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1380
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1380

and _menhir_run1383 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1383
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1383

and _menhir_reduce140 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3937 "reason_parser.mly"
      ( mktyp(Ptyp_any) )
# 64856 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run78 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 64863 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState78 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState78

and _menhir_run128 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState128

and _menhir_reduce144 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3945 "reason_parser.mly"
      ( mktyp(Ptyp_object ([], Closed)) )
# 64900 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce143 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3943 "reason_parser.mly"
      ( mktyp(Ptyp_object ([], Open)) )
# 64912 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run147 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 64919 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run148 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3143) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    ((let _1 = () in
    let _v : 'tv_meth_list = 
# 4020 "reason_parser.mly"
                                                ( [], Open )
# 64937 "reason_parser.ml"
     in
    _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3144)

and _menhir_reduce630 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_opt_bar = 
# 4218 "reason_parser.mly"
                                                ( () )
# 64946 "reason_parser.ml"
     in
    _menhir_goto_opt_bar _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce149 : _menhir_env -> 'ttv_tail * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3959 "reason_parser.mly"
      ( mktyp(Ptyp_variant([], Open, None)) )
# 64960 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run156 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3141) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_opt_bar = 
# 4219 "reason_parser.mly"
                                                ( () )
# 64975 "reason_parser.ml"
     in
    _menhir_goto_opt_bar _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3142)

and _menhir_run121 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState121

and _menhir_reduce270 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3336 "reason_parser.mly"
      ( mkpat(Ppat_any) )
# 65002 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run143 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 65009 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run198 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 65019 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce714 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run199 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 65028 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce713 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run200 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 65037 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce712 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run201 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 65046 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce710 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run202 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 65055 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce711 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run204 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 876 "reason_parser.mly"
       (nativeint)
# 65064 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce709 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run205 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 847 "reason_parser.mly"
       (int64)
# 65073 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce708 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run206 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 846 "reason_parser.mly"
       (int32)
# 65082 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce707 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run207 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 845 "reason_parser.mly"
       (int)
# 65091 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce705 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run208 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 827 "reason_parser.mly"
       (string)
# 65100 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce706 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_reduce281 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3362 "reason_parser.mly"
      ( mkpat(Ppat_array []) )
# 65116 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run512 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run226 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run225 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run224 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run223 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3139 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3140)

and _menhir_run513 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run494 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run491 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3137 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3138)

and _menhir_run517 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 65171 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce464 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run487 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState487

and _menhir_error70 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce331 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error237 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce139 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65304 "reason_parser.ml"
      
    in
    
# 3935 "reason_parser.mly"
      ( mktyp(Ptyp_class(_2, [])) )
# 65310 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce135 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65329 "reason_parser.ml"
      
    in
    
# 3918 "reason_parser.mly"
      ( mktyp(Ptyp_class(_2, List.rev _3)) )
# 65335 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce820 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 65342 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 65348 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_type_longident = 
# 4122 "reason_parser.mly"
                                                ( Lident _1 )
# 65355 "reason_parser.ml"
     in
    _menhir_goto_type_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce141 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65373 "reason_parser.ml"
      
    in
    
# 3939 "reason_parser.mly"
      ( mktyp(Ptyp_constr(_1, [])) )
# 65379 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce134 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65397 "reason_parser.ml"
      
    in
    
# 3916 "reason_parser.mly"
      ( mktyp(Ptyp_constr(_1, List.rev _2)) )
# 65403 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce57 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__core_type2 = 
# 3873 "reason_parser.mly"
      ( _1 )
# 65415 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error420 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 65422 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce433 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce236 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65442 "reason_parser.ml"
      
    in
    
# 2766 "reason_parser.mly"
      (
        let bang = {_1 with txt="!"} in
        mkexp(Pexp_apply(mkoperator bang, ["",_2]))
      )
# 65451 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce482 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_lbl_pattern) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_pattern_list = 
# 3432 "reason_parser.mly"
                      ( [_1], Closed )
# 65462 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error530 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce480 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error544 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce271 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce173 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_without_or), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3302 "reason_parser.mly"
      ( mkpat(Ppat_exception _2) )
# 65483 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce274 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3344 "reason_parser.mly"
      ( mkpat(Ppat_variant(_1, None)) )
# 65495 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce273 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65513 "reason_parser.ml"
      
    in
    
# 3342 "reason_parser.mly"
      ( mkpat(Ppat_construct(_1, None)) )
# 65519 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce166 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_pattern_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65537 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3273 "reason_parser.mly"
    (
      match is_pattern_list_single_any _2 with
        | Some singleAnyPat ->
            mkpat
              (Ppat_construct(_1, Some singleAnyPat))
        | None ->
          let loc = mklocation _symbolstartpos _endpos in
          let argPattern = simple_pattern_list_to_tuple ~loc _2 in
          mkExplicitArityTuplePat (Ppat_construct(_1, Some argPattern))
    )
# 65555 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce654 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern = 
# 3253 "reason_parser.mly"
                       ( _1 )
# 65567 "reason_parser.ml"
     in
    _menhir_goto_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce168 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos_x0_, _, _startpos_x0_), _endpos__3_, _, (_3 : 'tv_pattern_without_or), _startpos__3_) = _menhir_stack in
    let x0 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__pattern_without_or = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65586 "reason_parser.ml"
      
    in
    
# 3288 "reason_parser.mly"
      ( Location.raise_errorf ~loc:_2.loc ":: is not supported in Reason, please use [hd, ...tl] instead" )
# 65592 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce657 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_pattern_comma_list = 
# 3420 "reason_parser.mly"
                                                  ( _3 :: _1 )
# 65605 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce656 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_comma_list = 
# 3419 "reason_parser.mly"
                                                ( [_1] )
# 65617 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce659 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_pattern), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3426 "reason_parser.mly"
                                                ( (_1, Some _4) )
# 65631 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce660 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3427 "reason_parser.mly"
                                                ( (_3 :: _1, None) )
# 65644 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce658 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3425 "reason_parser.mly"
                                                ( [_1], None )
# 65656 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce169 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65676 "reason_parser.ml"
      
    in
    
# 3290 "reason_parser.mly"
      ( expecting_pat (with_txt _3 "pattern") )
# 65682 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce165 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65702 "reason_parser.ml"
      
    in
    
# 3263 "reason_parser.mly"
      ( expecting_pat (with_txt _3 "identifier") )
# 65708 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce55 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__core_type = 
# 3854 "reason_parser.mly"
      ( _1 )
# 65720 "reason_parser.ml"
     in
    _menhir_goto__core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce60 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type2), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__core_type2 = 
# 3879 "reason_parser.mly"
      ( mktyp(Ptyp_arrow("", _1, _3)) )
# 65733 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce479 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65750 "reason_parser.ml"
      
    in
    
# 3439 "reason_parser.mly"
      ( (_1,_3) )
# 65756 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce285 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65776 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65788 "reason_parser.ml"
      
    in
    
# 3370 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _3 ")") )
# 65794 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce288 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let x0 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65815 "reason_parser.ml"
      
    in
    
# 3377 "reason_parser.mly"
      ( expecting_pat (with_txt _4 "type") )
# 65821 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce77 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_simple_expr), _startpos__4_), _, (_5 : 'tv_direction_flag)), _endpos__6_, _, (_6 : 'tv_simple_expr), _startpos__6_), _endpos__7_, _, (_7 : 'tv_simple_expr), _startpos__7_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = 
# 2552 "reason_parser.mly"
      ( mkexp(Pexp_for(_2, _4, _6, _5, _7)) )
# 65835 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce107 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2590 "reason_parser.mly"
      ( mkexp (Pexp_assert _2) )
# 65848 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce204 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2647 "reason_parser.mly"
      ( mkexp (Pexp_variant (_1, None)) )
# 65860 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce73 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2542 "reason_parser.mly"
    (
      mkexp(Pexp_variant(_1, Some _2))
    )
# 65874 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce64 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__expr = 
# 2510 "reason_parser.mly"
      ( _1 )
# 65886 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce493 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_less_aggressive_simple_expression = 
# 2815 "reason_parser.mly"
              (_1)
# 65898 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3135) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_less_aggressive_simple_expression) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState475 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3111 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState815
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run816 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce64 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3112)
    | MenhirState817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3113 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 65965 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce463 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3114)
    | MenhirState820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3115 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 65973 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce462 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3116)
    | MenhirState1641 | MenhirState1642 | MenhirState815 | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3117 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce466 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3118)
    | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3119 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce721 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3120)
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3121 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce720 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3122)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3123 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState928
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run929 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState928 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce64 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3124)
    | MenhirState930 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3125 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 66045 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce463 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3126)
    | MenhirState933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3127 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 66053 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce462 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3128)
    | MenhirState2681 | MenhirState2682 | MenhirState928 | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3129 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce466 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3130)
    | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3131 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce721 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3132)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3133 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce720 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3134)
    | _ ->
        _menhir_fail ()) : 'freshtv3136)

and _menhir_error816 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 66074 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce65 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2512 "reason_parser.mly"
      ( mkexp(Pexp_apply(_1, List.rev _2)) )
# 66087 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce106 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label), _startpos_x0_), _endpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66106 "reason_parser.ml"
      
    in
    
# 2588 "reason_parser.mly"
      ( mkexp(Pexp_setinstvar(_1, _3)) )
# 66112 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce88 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4077 "reason_parser.mly"
                                                ( "*" )
# 66131 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66141 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66147 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce83 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 66154 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 842 "reason_parser.mly"
       (string)
# 66160 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4072 "reason_parser.mly"
                                                ( _1 )
# 66173 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66183 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66189 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce203 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66207 "reason_parser.ml"
      
    in
    
# 2642 "reason_parser.mly"
    (
      mkexp (Pexp_construct (_1, None))
    )
# 66215 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce267 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_non_labeled_expr_list), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_non_labeled_expr_list_as_tuple = 
# 2827 "reason_parser.mly"
                               (
    mkexp (Pexp_tuple(List.rev _1))
  )
# 66229 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3109) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3105) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce535 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3106)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3107) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce535 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3108)
    | _ ->
        _menhir_fail ()) : 'freshtv3110)

and _menhir_reduce101 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_additive), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66271 "reason_parser.ml"
      
    in
    
# 2567 "reason_parser.mly"
      (
        mkuplus _1 _2
      )
# 66279 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce96 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4085 "reason_parser.mly"
                                                ( "+=" )
# 66298 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66308 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66314 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce97 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4086 "reason_parser.mly"
                                                ( "%" )
# 66333 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66343 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66349 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce82 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 66356 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 841 "reason_parser.mly"
       (string)
# 66362 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4071 "reason_parser.mly"
                                                ( _1 )
# 66375 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66385 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66391 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce85 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4074 "reason_parser.mly"
                                                ( "+." )
# 66410 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66420 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66426 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce84 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4073 "reason_parser.mly"
                                                ( "+" )
# 66445 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66455 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66461 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce91 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4080 "reason_parser.mly"
                                                ( "or" )
# 66480 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66490 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66496 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce87 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4076 "reason_parser.mly"
                                                ( "-." )
# 66515 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66525 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66531 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce86 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4075 "reason_parser.mly"
                                                ( "-" )
# 66550 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66560 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66566 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce98 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4087 "reason_parser.mly"
                                                ( "<>" )
# 66585 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66595 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66601 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce81 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 66608 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 840 "reason_parser.mly"
       (string)
# 66614 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4070 "reason_parser.mly"
                                                ( _1 )
# 66627 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66637 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66643 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce80 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 66650 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 839 "reason_parser.mly"
       (string)
# 66656 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4069 "reason_parser.mly"
                                                ( _1 )
# 66669 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66679 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66685 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce99 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4088 "reason_parser.mly"
                                                ( "<..>" )
# 66704 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66714 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66720 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce89 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4078 "reason_parser.mly"
                                                ( "<" )
# 66739 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66749 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66755 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce79 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 66762 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 838 "reason_parser.mly"
       (string)
# 66768 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4068 "reason_parser.mly"
                                                ( _1 )
# 66781 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66791 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66797 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce90 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4079 "reason_parser.mly"
                                                ( ">" )
# 66816 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66826 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66832 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce92 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4081 "reason_parser.mly"
                                                ( "||" )
# 66851 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66861 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66867 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce93 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4082 "reason_parser.mly"
                                                ( "&" )
# 66886 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66896 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66902 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce94 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4083 "reason_parser.mly"
                                                ( "&&" )
# 66921 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66931 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66937 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce95 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4084 "reason_parser.mly"
                                                ( ":=" )
# 66956 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66966 "reason_parser.ml"
      
    in
    
# 2561 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66972 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce217 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _, _startpos__5_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2689 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "Array" "get")),
                         ["",_1; "",_4])) )
# 66992 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce103 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _, _startpos__5_), _endpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2573 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "Array" "set")),
                         ["",_1; "",_4; "",_7])) )
# 67013 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce109 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2617 "reason_parser.mly"
                                                       (
      (* Should use ghost expressions, but not sure how that would work with source maps *)
      (* So ? will become true and : becomes false for now*)
      let loc_question = mklocation _startpos__2_ _endpos__2_ in
      let loc_colon = mklocation _startpos__4_ _endpos__4_ in
      let fauxTruePat =
        Pat.mk ~loc:loc_question (Ppat_construct({txt = Lident "true"; loc = loc_question}, None)) in
      let fauxFalsePat =
        Pat.mk ~loc:loc_colon (Ppat_construct({txt = Lident "false"; loc = loc_colon}, None)) in
      let fauxMatchCaseTrue = Exp.case fauxTruePat _3 in
      let fauxMatchCaseFalse = Exp.case fauxFalsePat _5 in
      mkexp (Pexp_match (_1, [fauxMatchCaseTrue; fauxMatchCaseFalse]))
    )
# 67039 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce219 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2696 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "String" "get")),
                         ["",_1; "",_4])) )
# 67059 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce104 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _), _endpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2578 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "String" "set")),
                         ["",_1; "",_4; "",_7])) )
# 67080 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce221 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2703 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        bigarray_get ~loc _1 _4
      )
# 67100 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce105 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _), _endpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2583 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        bigarray_set ~loc _1 _4 _7
      )
# 67121 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1053 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce102 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67145 "reason_parser.ml"
      
    in
    
# 2571 "reason_parser.mly"
      ( mkexp(Pexp_setfield(_1, _3, _5)) )
# 67151 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce100 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_subtractive), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67169 "reason_parser.ml"
      
    in
    
# 2563 "reason_parser.mly"
      (
        mkuminus _1 _2
      )
# 67177 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce633 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_opt_comma = 
# 4223 "reason_parser.mly"
                                                ( () )
# 67188 "reason_parser.ml"
     in
    _menhir_goto_opt_comma _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce418 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _v : 'tv_expr_optional_constraint = 
# 3157 "reason_parser.mly"
                         ( _1 )
# 67199 "reason_parser.ml"
     in
    _menhir_goto_expr_optional_constraint _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_error1086 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce806 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce476 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_lbl_expr) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_expr_list = 
# 3171 "reason_parser.mly"
                    ( [_1] )
# 67214 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1117 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce473 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce472 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67235 "reason_parser.ml"
      
    in
    
# 3175 "reason_parser.mly"
      ( (_1, _3) )
# 67241 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce600 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_non_punned_lbl_expression = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67258 "reason_parser.ml"
      
    in
    
# 3184 "reason_parser.mly"
      ( (_1, _3) )
# 67264 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3103) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_punned_lbl_expression) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3101) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_punned_lbl_expression) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3099) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_non_punned_lbl_expression) : 'tv_non_punned_lbl_expression) = _v in
    ((let _v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field = 
# 3204 "reason_parser.mly"
     ( [_1] )
# 67281 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3100)) : 'freshtv3102)) : 'freshtv3104)

and _menhir_reduce451 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 67288 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 910 "reason_parser.mly"
       (string)
# 67294 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_ident = 
# 4058 "reason_parser.mly"
                                                ( _1 )
# 67301 "reason_parser.ml"
     in
    _menhir_goto_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce675 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 67308 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 902 "reason_parser.mly"
       (string * string option)
# 67314 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_primitive_declaration = 
# 3449 "reason_parser.mly"
                                                ( [fst _1] )
# 67320 "reason_parser.ml"
     in
    _menhir_goto_primitive_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce448 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3601 "reason_parser.mly"
                                                         ( (List.rev _1, None) )
# 67331 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce357 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_class_sig_fields = 
# 2248 "reason_parser.mly"
                                        ( [] )
# 67341 "reason_parser.ml"
     in
    _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce32 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_clty_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_instance_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67359 "reason_parser.ml"
      
    in
    
# 2217 "reason_parser.mly"
      ( mkcty(Pcty_constr (_1, [])) )
# 67365 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce33 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_clty_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_instance_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67383 "reason_parser.ml"
      
    in
    
# 2219 "reason_parser.mly"
      ( mkcty(Pcty_constr (_1, List.rev _2)) )
# 67389 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce40 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_sig_field = 
# 2258 "reason_parser.mly"
      ( mkctf_attrs (Pctf_inherit _2) [] )
# 67402 "reason_parser.ml"
     in
    _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce361 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 67409 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 67415 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_class_instance_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_class_type_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67428 "reason_parser.ml"
      
    in
    
# 2333 "reason_parser.mly"
                                                                                (
      (_2, _5, _1, List.rev _3)
    )
# 67436 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3097) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_type_declaration_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3089 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1617
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1617
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1617) : 'freshtv3090)
    | MenhirState1870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3091 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1871
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1871
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1871) : 'freshtv3092)
    | MenhirState2046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3093 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2052
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2052
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2052) : 'freshtv3094)
    | MenhirState2081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3095 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2082
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2082
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2082) : 'freshtv3096)
    | _ ->
        _menhir_fail ()) : 'freshtv3098)

and _menhir_reduce15 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67511 "reason_parser.ml"
      
    in
    
# 1951 "reason_parser.mly"
                                  (
      mkclass(Pcl_constr(_2, []))
    )
# 67519 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce16 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67538 "reason_parser.ml"
      
    in
    
# 1955 "reason_parser.mly"
                                                                    (
      mkclass(Pcl_constr(_2, List.rev _3))
    )
# 67546 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce11 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_simple_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr = 
# 1924 "reason_parser.mly"
      ( _1 )
# 67558 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce13 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_simple_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1935 "reason_parser.mly"
      ( mkclass(Pcl_apply(_1, List.rev _2)) )
# 67570 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce30 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_fun_def = 
# 1888 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _3)) )
# 67583 "reason_parser.ml"
     in
    _menhir_goto__class_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce839 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67601 "reason_parser.ml"
      
    in
    
# 2026 "reason_parser.mly"
      ( _3, _2, Cfk_concrete (_1, _5) )
# 67607 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce840 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_type_constraint)), _endpos__5_), _endpos__6_, _, (_6 : 'tv_expr), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67625 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2028 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let e = ghexp_constraint loc _6 _4 in
        _3, _2, Cfk_concrete (_1, e)
      )
# 67644 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce402 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_curried_binding = 
# 3053 "reason_parser.mly"
      ( _2 )
# 67656 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce407 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_non_arrowed_core_type), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_curried_binding_return_typed_ = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3042 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        ghexp_constraint loc _4 (Some _2, None)
      )
# 67674 "reason_parser.ml"
     in
    _menhir_goto_curried_binding_return_typed_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce566 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67692 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2050 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly(_5, None))) )
# 67708 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce567 : _menhir_env -> ((((((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, _startpos__5_), _, (_6 : 'tv_lident_list)), _endpos__8_, _, (_8 : 'tv_core_type), _startpos__8_), _endpos__9_), _endpos__10_, _, (_10 : 'tv_expr), _startpos__10_) = _menhir_stack in
    let _9 = () in
    let _7 = () in
    let _5 = () in
    let _4 = () in
    let _endpos = _endpos__10_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67729 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__10_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2056 "reason_parser.mly"
      (

        (* For non, methods we'd create a pattern binding:
           ((Ppat_constraint(mkpatvar ..., Ptyp_poly (typeVars, poly_type_varified))),
            exp_with_newtypes_constrained_by_non_varified)

           For methods, we create:
           Pexp_poly (Pexp_constraint (methodFunWithNewtypes, non_varified), Some (Ptyp_poly newTypes varified))
         *)
        let (exp_with_newtypes_constrained_by_non_varified, poly_type_varified) =
          wrap_type_annotation _6 _8 _10 in
        (
          _3, _2,
          Cfk_concrete (
            _1,
            let loc = mklocation _symbolstartpos _endpos in
            mkexp ~ghost:true ~loc (Pexp_poly(exp_with_newtypes_constrained_by_non_varified, Some poly_type_varified))
          )
        )
      )
# 67763 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce565 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_poly_type)), _endpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _endpos = _endpos__7_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67782 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2047 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly(_7, Some _5))) )
# 67798 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce650 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_parent_binder = 
# 2010 "reason_parser.mly"
          ( None )
# 67808 "reason_parser.ml"
     in
    _menhir_goto_parent_binder _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce692 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_semi_terminated_class_fields = 
# 2001 "reason_parser.mly"
                          ([])
# 67818 "reason_parser.ml"
     in
    _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce19 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr_lets_and_rest = 
# 1896 "reason_parser.mly"
               (_1)
# 67830 "reason_parser.ml"
     in
    _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce7 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_constructor_type = 
# 2198 "reason_parser.mly"
      ( _2 )
# 67843 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce339 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 67850 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 67856 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _), _endpos__5_, _, (_5 : 'tv_class_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67869 "reason_parser.ml"
      
    in
    
# 1831 "reason_parser.mly"
    (
       (_2, _5, _1, List.rev _3)
    )
# 67877 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce354 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos__0_) = Obj.magic _menhir_stack in
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_class_self_type = let _endpos = _endpos__0_ in
    let _symbolstartpos = _endpos in
    
# 2242 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        Typ.mk ~loc Ptyp_any
       )
# 67893 "reason_parser.ml"
     in
    _menhir_goto_class_self_type _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce54 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_constructor_type)), _endpos__3_), _endpos__4_, _, (_4 : 'tv_class_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__constrained_class_declaration = 
# 1807 "reason_parser.mly"
      ( mkclass(Pcl_constraint(_4, _2)) )
# 67907 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3087) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__constrained_class_declaration) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3083) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__constrained_class_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce525 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3084)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3085) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__constrained_class_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce525 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3086)
    | _ ->
        _menhir_fail ()) : 'freshtv3088)

and _menhir_reduce346 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_class_fun_return = 
# 1871 "reason_parser.mly"
      ( _2 )
# 67943 "reason_parser.ml"
     in
    _menhir_goto_class_fun_return _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce347 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_non_arrowed_class_constructor_type)), _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_fun_return = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1873 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkclass ~loc (Pcl_constraint(_4, _2))
      )
# 67961 "reason_parser.ml"
     in
    _menhir_goto_class_fun_return _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce298 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1415 "reason_parser.mly"
                           (
      mkstr(Pstr_type (List.rev _1))
    )
# 67975 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce638 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_optional_type_parameters = 
# 3511 "reason_parser.mly"
                                   ( [] )
# 67985 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameters _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce809 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_type_kind = 
# 3488 "reason_parser.mly"
      ( (Ptype_abstract, Public, None) )
# 67995 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce808 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 68002 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 68008 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)), _endpos__3_, _, (_3 : 'tv_type_kind)), _endpos__4_, _, (_4 : 'tv_constraints)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_declaration_details = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68020 "reason_parser.ml"
      
    in
    
# 3478 "reason_parser.mly"
      ( let (kind, priv, manifest) = _3 in
        (_1, _2, List.rev _4, kind, priv, manifest)
       )
# 68028 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3081) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_type_declaration_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2578 | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3073 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1363
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState1363
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1363) : 'freshtv3074)
    | MenhirState1848 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3075 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1861
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState1861
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1861) : 'freshtv3076)
    | MenhirState2069 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3077 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2072
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run438 _menhir_env (Obj.magic _menhir_stack) MenhirState2072
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2072) : 'freshtv3078)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3079 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2527
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run395 _menhir_env (Obj.magic _menhir_stack) MenhirState2527
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2527) : 'freshtv3080)
    | _ ->
        _menhir_fail ()) : 'freshtv3082)

and _menhir_reduce639 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_optional_type_parameter_list)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_optional_type_parameters = 
# 3512 "reason_parser.mly"
                                   ( List.rev _1 )
# 68095 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameters _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce302 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_nonlocal_module_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1429 "reason_parser.mly"
                                  (
      mkstr(Pstr_recmodule(List.rev _1))
    )
# 68109 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce307 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1450 "reason_parser.mly"
                                 (
      (* Each declaration has their own preceeding post_item_attributes *)
      mkstr(Pstr_class_type (List.rev _1))
    )
# 68124 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce306 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1446 "reason_parser.mly"
                            (
      (* Each declaration has their own preceeding post_item_attributes *)
      mkstr(Pstr_class (List.rev _1))
    )
# 68139 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce295 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item = 
# 1399 "reason_parser.mly"
                 ( val_of_let_bindings _1 )
# 68151 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce61 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__curried_binding_return_typed = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3007 "reason_parser.mly"
      (
          let loc = mklocation _symbolstartpos _endpos in
          let nil = { txt = Lident "()"; loc = make_ghost_loc loc } in
          mkexp ~ghost:true ~loc (Pexp_fun("", None, mkpat ~ghost:true ~loc (Ppat_construct (nil, None)), _2))
      )
# 68170 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce536 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x_, _menhir_s, (x : 'tv_expr), _startpos_x_) = _menhir_stack in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1171 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 68183 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3071) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp_expr_) = _v in
    ((match _menhir_s with
    | MenhirState1208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3063 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce499 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3064)
    | MenhirState1216 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3065 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce498 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3066)
    | MenhirState1894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3067 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce499 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3068)
    | MenhirState1900 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3069 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce498 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3070)
    | _ ->
        _menhir_fail ()) : 'freshtv3072)

and _menhir_reduce496 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_type_constraint)), _endpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1216 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 68237 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2906 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (_1, ghexp_constraint loc _4 _2)
      )
# 68249 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce501 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_not_ident), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_), _endpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_let_binding_body = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2986 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (mkpat ~loc (Ppat_constraint(_1, _3)), _5)
      )
# 68267 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce500 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_let_binding_body = 
# 2984 "reason_parser.mly"
      ( (_1, _3) )
# 68279 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce117 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_functor_args), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__module_expr = 
# 1285 "reason_parser.mly"
      ( mkFunctorThatReturns _2 _4 )
# 68293 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce382 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_constraints = 
# 3484 "reason_parser.mly"
                                                ( [] )
# 68303 "reason_parser.ml"
     in
    _menhir_goto_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce846 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_optional_type_parameters)), _, (_4 : 'tv_with_type_binder)), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_constraints)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_with_constraint = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68321 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3691 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in

         Pwith_type
          (_2,
           (Type.mk {_2 with txt=Longident.last _2.txt}
              ~params:_3
              ~cstrs:(List.rev _6)
              ~manifest:_5
              ~priv:_4
              ~loc)) )
# 68339 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce571 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 68346 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 910 "reason_parser.mly"
       (string)
# 68352 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_ext_longident = 
# 4140 "reason_parser.mly"
                                                ( Lident _1 )
# 68359 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce849 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 68366 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : (
# 910 "reason_parser.mly"
       (string)
# 68372 "reason_parser.ml"
    )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos_x1_, _, (x1 : 'tv_mod_ext_longident), _startpos_x1_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos_x1_ in
    let _v : 'tv_with_constraint = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68386 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68398 "reason_parser.ml"
      
    in
    
# 3721 "reason_parser.mly"
      ( Pwith_modsubst (_2, _4) )
# 68404 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce572 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 68411 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 910 "reason_parser.mly"
       (string)
# 68417 "reason_parser.ml"
    )), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mod_ext_longident = 
# 4141 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 68425 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1987 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce848 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__3_, _), _endpos_x1_, _, (x1 : 'tv_mod_ext_longident), _startpos_x1_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos_x1_ in
    let _v : 'tv_with_constraint = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68448 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68460 "reason_parser.ml"
      
    in
    
# 3719 "reason_parser.mly"
      ( Pwith_module (_2, _4) )
# 68466 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1994 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce128 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_type = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1618 "reason_parser.mly"
                                                          (
      (**
       * In OCaml, this is invalid:
       * module MyFunctor: functor MT -> (sig end) = functor MT -> (struct end);;
       *
       * Not only must curried functor args have annotations, but functor
       * annotations must include *names* for each argument in a functor type.
       *
       * module MyFunctor: functor (MT:MT) -> (sig end) = functor (MT:MT) -> (struct end)
       *
       * In Reason, we will parse the functor type:
       *
       *    (AB:MT) -> ReturnSig
       *
       * as in:
       *                   /----------------\
       * module MyFunctor: (A:B) => ReturnSig = functor (C:D) => {}
       *
       * But only for the sake of compatibility with existing OCaml code (the
       * ability to "view" OCaml code in Reason's syntax without loss of
       * information.) Do not write identifiers in functor argument type
       * positions - you wouldn't do it with functions, and they are
       * meaningless in functors.
       *
       *  But for sake of consistency (and for sake of a syntax that truly
       *  unifies functor syntax with function syntax, the following "sugars"
       *  will be parsed (and printed):
       *
       *   A => B => C
       *
       * Is parsed into:
       *
       * functor (_:A) -> functor (_:B) -> C
       *
       * And a dummy "_" is inserted into the parse tree where no name has been
       * provided.
       *
       *   {SomeSig} => {} => {}
       *
       * Is parsed into:
       *
       * (_:SomeSig) => (_:{}) => {}
       *
       *
       *)
      let loc = mklocation _symbolstartpos _endpos in
      mkmty(Pmty_functor(ghloc ~loc "_", Some _1, _3))
    )
# 68532 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce250 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_module_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68550 "reason_parser.ml"
      
    in
    
# 1241 "reason_parser.mly"
      ( mkmod(Pmod_ident _1) )
# 68556 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce127 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_functor_arg_name), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_), _endpos__5_, _, _startpos__5_), _startpos__6_), _endpos__7_, _, (_7 : 'tv_module_type), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__module_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68578 "reason_parser.ml"
      
    in
    
# 1604 "reason_parser.mly"
                                                                                                       (
    (* Why does this rule cause a conflict with core_type2? It has nothing to do
     * with it.
     *
     * Update: I'm guessing it has something to do with the fact that this isn't
     * parsed correctly because the => is considered part of "type dependenences" :
     *
     *    let module CreateFactory
     *               (Spec: ContainerSpec)
     *               :DescriptorFactoryIntf.S with
     *                  type props = Spec.props and type dependencies = Spec.props =>
     *)
      mkmty(Pmty_functor(_2, Some _4, _7))
     )
# 68597 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce262 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68617 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68629 "reason_parser.ml"
      
    in
    
# 1547 "reason_parser.mly"
      ( unclosed_mty (with_txt _1 "(") (with_txt _3 ")"))
# 68635 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce119 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_), _endpos_x0_, _, _startpos_x0_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68655 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68667 "reason_parser.ml"
      
    in
    
# 1289 "reason_parser.mly"
      ( unclosed_mod (with_txt _2 "(") (with_txt _4 ")") )
# 68673 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce253 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68694 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68706 "reason_parser.ml"
      
    in
    
# 1247 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")"))
# 68712 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce248 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_), _endpos__6_, _, _startpos__6_) = _menhir_stack in
    let x1 = () in
    let _6 = () in
    let _4 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _7 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68735 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68747 "reason_parser.ml"
      
    in
    
# 2804 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _7 ")"))
# 68753 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce215 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68774 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68786 "reason_parser.ml"
      
    in
    
# 2681 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _5 ")") )
# 68792 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run922 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce218 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run999 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce220 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce437 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_fun_def = 
# 3102 "reason_parser.mly"
                      ( _2 )
# 68814 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce438 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_non_arrowed_core_type), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3104 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_constraint(_4, _2))
      )
# 68832 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce486 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_leading_bar_match_case = 
# 3091 "reason_parser.mly"
      ( Exp.case _1 ~guard:_3 _5 )
# 68845 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_case _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce485 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_leading_bar_match_case = 
# 3087 "reason_parser.mly"
                                       (
      Exp.case _1 _3
    )
# 68859 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_case _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce68 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_leading_bar_match_cases)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2523 "reason_parser.mly"
      ( mkexp (Pexp_function(List.rev _2)) )
# 68872 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce373 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4111 "reason_parser.mly"
                                                ( _1 )
# 68884 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce634 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_opt_semi = 
# 4226 "reason_parser.mly"
                                                ( () )
# 68894 "reason_parser.ml"
     in
    _menhir_goto_opt_semi _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce635 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__1_ in
    let _v : 'tv_opt_semi = 
# 4227 "reason_parser.mly"
                                                ( () )
# 68906 "reason_parser.ml"
     in
    _menhir_goto_opt_semi _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce75 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2548 "reason_parser.mly"
      ( mkexp (Pexp_ifthenelse(_2, _3, None)) )
# 68919 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce74 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2546 "reason_parser.mly"
      ( mkexp(Pexp_ifthenelse(_2, _3, Some _5)) )
# 68933 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run884 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce218 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1044 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce220 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce108 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2592 "reason_parser.mly"
      ( mkexp (Pexp_lazy _2) )
# 68956 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce578 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_module_binding_body_expr = 
# 1469 "reason_parser.mly"
      ( _2 )
# 68968 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce579 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_), _endpos__3_, _), _endpos__4_, _, (_4 : 'tv_module_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_module_binding_body_expr = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1471 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod ~loc (Pmod_constraint(_4, _2))
      )
# 68986 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce111 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_binding_body_functor = 
# 1479 "reason_parser.mly"
                                          (
      mkFunctorThatReturns _1 _3
    )
# 69001 "reason_parser.ml"
     in
    _menhir_goto__module_binding_body_functor _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce112 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_module_type), _startpos__3_), _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__module_binding_body_functor = 
# 1482 "reason_parser.mly"
                                                                        (
      let loc = mklocation _startpos__5_ _endpos__5_ in
      mkFunctorThatReturns _1 (mkmod ~loc (Pmod_constraint(_5, _3)))
    )
# 69018 "reason_parser.ml"
     in
    _menhir_goto__module_binding_body_functor _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce432 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 69025 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 863 "reason_parser.mly"
       (string)
# 69031 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_field_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69043 "reason_parser.ml"
      
    in
    
# 3222 "reason_parser.mly"
    ( (_1, _3) )
# 69049 "reason_parser.ml"
     in
    _menhir_goto_field_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce246 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69071 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69083 "reason_parser.ml"
      
    in
    
# 2796 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _5 ")") )
# 69089 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce211 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_expr_comma_list)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69109 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69121 "reason_parser.ml"
      
    in
    
# 2667 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _3 ")") )
# 69127 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce206 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69147 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69159 "reason_parser.ml"
      
    in
    
# 2651 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _3 ")") )
# 69165 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce123 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let x1 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69186 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69198 "reason_parser.ml"
      
    in
    
# 1297 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _4 ")") )
# 69204 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce122 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69226 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69238 "reason_parser.ml"
      
    in
    
# 1295 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")") )
# 69244 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce121 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69266 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69278 "reason_parser.ml"
      
    in
    
# 1293 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")") )
# 69284 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce120 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69304 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69316 "reason_parser.ml"
      
    in
    
# 1291 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _3 ")") )
# 69322 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce132 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1538 "reason_parser.mly"
      ( mkmty(Pmty_typeof _3) )
# 69336 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce58 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 69343 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 69349 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _endpos__4_, _, _startpos__4_), _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type2), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__core_type2 = 
# 3875 "reason_parser.mly"
      ( mktyp(Ptyp_arrow("?" ^ _1 , mkoption _3, _6)) )
# 69359 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce59 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 69366 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 69372 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type2), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__core_type2 = 
# 3877 "reason_parser.mly"
      ( mktyp(Ptyp_arrow(_1, _3, _5)) )
# 69381 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2404 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce647 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce171 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_pattern_without_or), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_pattern_without_or), _startpos__7_) = _menhir_stack in
    let x1 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__pattern_without_or = let _8 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69409 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69421 "reason_parser.ml"
      
    in
    
# 3298 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _8 ")") )
# 69427 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce235 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 69434 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 887 "reason_parser.mly"
       (string)
# 69440 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69453 "reason_parser.ml"
      
    in
    
# 2756 "reason_parser.mly"
      (
        mkexp(Pexp_apply(mkoperator _1, ["",_2]))
      )
# 69461 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce632 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_opt_comma = 
# 4222 "reason_parser.mly"
                                                ( () )
# 69470 "reason_parser.ml"
     in
    _menhir_goto_opt_comma _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce334 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_attributes = 
# 4309 "reason_parser.mly"
               ( [] )
# 69480 "reason_parser.ml"
     in
    _menhir_goto_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce393 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations = 
# 3557 "reason_parser.mly"
                                                                                ( List.rev _1 )
# 69491 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce394 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declaration_no_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations = 
# 3558 "reason_parser.mly"
                                                                                ( [_1] )
# 69502 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce395 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declaration_no_leading_bar)), _endpos__2_, _, (_2 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constructor_declarations = 
# 3559 "reason_parser.mly"
                                                                                 ( _1 :: List.rev _2 )
# 69513 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2513 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce810 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce447 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_generalized_constructor_arguments = 
# 3600 "reason_parser.mly"
                                                ( ([],None) )
# 69527 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce671 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_post_item_attributes = 
# 4314 "reason_parser.mly"
               ( [] )
# 69537 "reason_parser.ml"
     in
    _menhir_goto_post_item_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce76 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2550 "reason_parser.mly"
      ( mkexp (Pexp_while(_2, _3)) )
# 69550 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3061 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CLASS ->
            _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXTERNAL ->
            _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INCLUDE ->
            _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETATATAT ->
            _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENTPERCENT ->
            _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LET ->
            _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OPEN ->
            _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EOF | RBRACE | RBRACKET ->
            _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState2551
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2551) : 'freshtv3062)
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | RBRACE | RBRACKET ->
        _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState1
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1

and _menhir_reduce53 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69774 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1211 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69786 "reason_parser.ml"
      
    in
    
# 1977 "reason_parser.mly"
      ( unclosed_cl (with_txt _1 "(") (with_txt _3 ")") )
# 69792 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce574 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 69799 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 910 "reason_parser.mly"
       (string)
# 69805 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_longident = 
# 4136 "reason_parser.mly"
                                                ( Lident _1 )
# 69812 "reason_parser.ml"
     in
    _menhir_goto_mod_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce602 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_nonrec_flag = 
# 4186 "reason_parser.mly"
                                                ( Recursive )
# 69821 "reason_parser.ml"
     in
    _menhir_goto_nonrec_flag _menhir_env _menhir_stack _menhir_s _v

and _menhir_run75 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3059) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_nonrec_flag = 
# 4187 "reason_parser.mly"
                                                ( Nonrecursive )
# 69837 "reason_parser.ml"
     in
    _menhir_goto_nonrec_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3060)

and _menhir_reduce377 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4115 "reason_parser.mly"
                                                ( Lident "true" )
# 69849 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce368 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 902 "reason_parser.mly"
       (string * string option)
# 69856 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 902 "reason_parser.mly"
       (string * string option)
# 69861 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4035 "reason_parser.mly"
                                      ( let (s, d) = _1 in Const_string (s, d) )
# 69868 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce321 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_additive = 
# 4235 "reason_parser.mly"
                                                ( "+." )
# 69881 "reason_parser.ml"
     in
    _menhir_goto_additive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce320 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_additive = 
# 4234 "reason_parser.mly"
                                                ( "+" )
# 69894 "reason_parser.ml"
     in
    _menhir_goto_additive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run337 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce776 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run338 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce775 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run339 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce774 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run340 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce773 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run341 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce772 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run342 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 69926 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce728 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run343 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce771 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run344 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce770 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run345 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce769 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run346 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce768 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run347 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce767 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run348 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce754 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run349 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce766 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run350 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce765 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run351 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce764 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run352 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce763 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run353 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce762 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run354 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce761 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run355 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce760 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run356 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce759 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run357 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce758 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run358 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce757 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run359 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce756 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run360 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce755 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run361 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 70025 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce727 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run362 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce753 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run363 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce752 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run364 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce751 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run365 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce750 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run366 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce749 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run367 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce748 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run368 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce747 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run369 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce746 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run370 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce745 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run371 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce744 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run372 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce743 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run373 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce742 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run374 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce741 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run375 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce740 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run376 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce739 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run377 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce738 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run378 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce737 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run379 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce736 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run380 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce735 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run381 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce734 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run382 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce733 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run383 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce732 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run384 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce731 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run385 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce730 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run386 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce729 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce642 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_override_flag = 
# 4214 "reason_parser.mly"
                                                ( Fresh )
# 70163 "reason_parser.ml"
     in
    _menhir_goto_override_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run392 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3057) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_override_flag = 
# 4215 "reason_parser.mly"
                                                ( Override )
# 70181 "reason_parser.ml"
     in
    _menhir_goto_override_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3058)

and _menhir_run8 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 70188 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce349 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_reduce372 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 876 "reason_parser.mly"
       (nativeint)
# 70197 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 876 "reason_parser.mly"
       (nativeint)
# 70202 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4039 "reason_parser.mly"
                                      ( Const_nativeint _1 )
# 70209 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce793 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_subtractive = 
# 4231 "reason_parser.mly"
                                                ( "-." )
# 70222 "reason_parser.ml"
     in
    _menhir_goto_subtractive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce792 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_subtractive = 
# 4230 "reason_parser.mly"
                                                ( "-" )
# 70235 "reason_parser.ml"
     in
    _menhir_goto_subtractive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run186 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 887 "reason_parser.mly"
       (string)
# 70242 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState186 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce605 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState186

and _menhir_run408 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_reduce613 _menhir_env (Obj.magic _menhir_stack)
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce321 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3055 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3056)

and _menhir_run409 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_reduce612 _menhir_env (Obj.magic _menhir_stack)
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce320 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3053 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3054)

and _menhir_run767 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState767

and _menhir_run782 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce793 _menhir_env (Obj.magic _menhir_stack)
    | RPAREN ->
        _menhir_reduce615 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3051 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3052)

and _menhir_run783 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce792 _menhir_env (Obj.magic _menhir_stack)
    | RPAREN ->
        _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3049 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3050)

and _menhir_run800 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce606 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState800

and _menhir_reduce458 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 70449 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 70455 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_label = 
# 4027 "reason_parser.mly"
                                                ( _1 )
# 70462 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3047) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_label) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState287 | MenhirState283 | MenhirState146 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2989 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error151 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv2987 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run153 _menhir_env (Obj.magic _menhir_stack) MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState152 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState152) : 'freshtv2988)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error151 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv2990)
    | MenhirState469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2991 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce243 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2992)
    | MenhirState734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2993 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce243 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2994)
    | MenhirState0 | MenhirState1 | MenhirState2551 | MenhirState15 | MenhirState73 | MenhirState174 | MenhirState177 | MenhirState183 | MenhirState189 | MenhirState333 | MenhirState396 | MenhirState404 | MenhirState407 | MenhirState412 | MenhirState414 | MenhirState421 | MenhirState431 | MenhirState2320 | MenhirState2314 | MenhirState439 | MenhirState451 | MenhirState453 | MenhirState455 | MenhirState2268 | MenhirState473 | MenhirState2264 | MenhirState475 | MenhirState2253 | MenhirState2249 | MenhirState2247 | MenhirState2227 | MenhirState2225 | MenhirState2216 | MenhirState2212 | MenhirState2204 | MenhirState737 | MenhirState2176 | MenhirState750 | MenhirState2027 | MenhirState1469 | MenhirState1471 | MenhirState1933 | MenhirState1923 | MenhirState1840 | MenhirState1916 | MenhirState1913 | MenhirState1903 | MenhirState1900 | MenhirState1894 | MenhirState1887 | MenhirState1728 | MenhirState1722 | MenhirState1719 | MenhirState1712 | MenhirState1706 | MenhirState1702 | MenhirState1682 | MenhirState1679 | MenhirState1142 | MenhirState1137 | MenhirState766 | MenhirState1129 | MenhirState1118 | MenhirState1079 | MenhirState1076 | MenhirState793 | MenhirState808 | MenhirState1054 | MenhirState1051 | MenhirState1046 | MenhirState1042 | MenhirState1040 | MenhirState997 | MenhirState920 | MenhirState886 | MenhirState881 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState871 | MenhirState869 | MenhirState867 | MenhirState865 | MenhirState863 | MenhirState861 | MenhirState859 | MenhirState857 | MenhirState855 | MenhirState853 | MenhirState851 | MenhirState849 | MenhirState847 | MenhirState845 | MenhirState842 | MenhirState832 | MenhirState830 | MenhirState828 | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2997 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error827 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv2995 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState828 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState828) : 'freshtv2996)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error827 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv2998)
    | MenhirState2747 | MenhirState2751 | MenhirState2754 | MenhirState2737 | MenhirState2576 | MenhirState2568 | MenhirState2416 | MenhirState2370 | MenhirState2368 | MenhirState415 | MenhirState417 | MenhirState2271 | MenhirState2223 | MenhirState2207 | MenhirState2186 | MenhirState2181 | MenhirState2182 | MenhirState758 | MenhirState2141 | MenhirState1150 | MenhirState2128 | MenhirState1943 | MenhirState1156 | MenhirState1232 | MenhirState1229 | MenhirState1219 | MenhirState1216 | MenhirState1208 | MenhirState1199 | MenhirState1192 | MenhirState1188 | MenhirState1171 | MenhirState1166 | MenhirState1162 | MenhirState765 | MenhirState1133 | MenhirState786 | MenhirState1110 | MenhirState1098 | MenhirState1093 | MenhirState1094 | MenhirState1071 | MenhirState1058 | MenhirState1048 | MenhirState888 | MenhirState892 | MenhirState1031 | MenhirState1029 | MenhirState1027 | MenhirState1019 | MenhirState902 | MenhirState917 | MenhirState1013 | MenhirState1010 | MenhirState1008 | MenhirState1006 | MenhirState1003 | MenhirState1001 | MenhirState994 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState984 | MenhirState982 | MenhirState980 | MenhirState978 | MenhirState976 | MenhirState974 | MenhirState972 | MenhirState970 | MenhirState968 | MenhirState966 | MenhirState964 | MenhirState962 | MenhirState960 | MenhirState958 | MenhirState955 | MenhirState945 | MenhirState943 | MenhirState941 | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3001 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error940 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv2999 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState941 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState941) : 'freshtv3000)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error940 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3002)
    | MenhirState1548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3005 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1549 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1550) : 'freshtv3004)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1549 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3006)
    | MenhirState1553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3009 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1554 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3007 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1555) : 'freshtv3008)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1554 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3010)
    | MenhirState1546 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3013 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1559 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3011 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1560 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1560) : 'freshtv3012)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1559 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3014)
    | MenhirState1567 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3017 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1568 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3015 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1570 _menhir_env (Obj.magic _menhir_stack) MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1569 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1569) : 'freshtv3016)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1568 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3018)
    | MenhirState1663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1664 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3019 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1665) : 'freshtv3020)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1664 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3022)
    | MenhirState1672 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3025 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1673 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3023 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1674 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1674) : 'freshtv3024)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1673 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3026)
    | MenhirState1677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3029 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1678
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1063 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1678 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1061 _menhir_env (Obj.magic _menhir_stack) MenhirState1678
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3027 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1678 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1679) : 'freshtv3028)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1678)) : 'freshtv3030)
    | MenhirState1687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3033 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1688 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3031 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1570 _menhir_env (Obj.magic _menhir_stack) MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1689) : 'freshtv3032)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1688 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3034)
    | MenhirState1692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3037 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1693 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3035 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run305 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run300 _menhir_env (Obj.magic _menhir_stack) MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run290 _menhir_env (Obj.magic _menhir_stack) MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run281 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run322 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run244 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1570 _menhir_env (Obj.magic _menhir_stack) MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run242 _menhir_env (Obj.magic _menhir_stack) MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1694 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1694) : 'freshtv3036)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1693 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3038)
    | MenhirState1696 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3045 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1697
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3041 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1697 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run153 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv3039 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1714 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1715 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1715) : 'freshtv3040)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1714) : 'freshtv3042)
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3043 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1697 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1712) : 'freshtv3044)
          | EQUALGREATER ->
              _menhir_run1702 _menhir_env (Obj.magic _menhir_stack) MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1698 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1697 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1697)) : 'freshtv3046)
    | _ ->
        _menhir_fail ()) : 'freshtv3048)

and _menhir_reduce831 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 71411 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 863 "reason_parser.mly"
       (string)
# 71417 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_val_ident = 
# 4063 "reason_parser.mly"
                                                ( _1 )
# 71424 "reason_parser.ml"
     in
    _menhir_goto_val_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce684 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_rec_flag = 
# 4182 "reason_parser.mly"
                                                ( Nonrecursive )
# 71433 "reason_parser.ml"
     in
    _menhir_goto_rec_flag _menhir_env _menhir_stack _menhir_s _v

and _menhir_run434 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2985) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_rec_flag = 
# 4183 "reason_parser.mly"
                                                ( Recursive )
# 71449 "reason_parser.ml"
     in
    _menhir_goto_rec_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2986)

and _menhir_run2631 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 71456 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1242 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        _menhir_run1240 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631
    | LPAREN ->
        _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2631

and _menhir_reduce230 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2736 "reason_parser.mly"
      ( mkexp (Pexp_array []) )
# 71484 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run20 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce776 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run21 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce775 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run22 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce774 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run23 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce773 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run24 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce772 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run25 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 71516 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce728 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run26 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce771 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run27 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce770 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run28 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce769 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run29 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce768 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run30 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce767 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run31 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce754 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run32 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce766 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run33 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce765 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run34 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce764 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run35 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce763 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run36 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce762 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run37 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce761 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run38 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce760 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run39 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce759 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run40 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce758 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run41 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce757 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run42 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce756 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run43 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce755 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run44 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 71615 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce727 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run45 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce753 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run46 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce752 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run47 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce751 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run48 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce750 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run49 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce749 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run50 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce748 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run51 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce747 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run52 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce746 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run53 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce745 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run54 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce744 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run55 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce743 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run56 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce742 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run57 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce741 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run58 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce740 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run59 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce739 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run60 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce738 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run61 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce737 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run62 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce736 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run63 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce735 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run64 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce734 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run65 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce733 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run66 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce732 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run67 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce731 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run68 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce730 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run69 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce729 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run604 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s

and _menhir_run1094 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1094 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1094

and _menhir_run420 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 71831 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2983 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 71843 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState421) : 'freshtv2984)
    | COMMA | GREATERRBRACE ->
        _menhir_reduce433 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_error420 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce240 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2777 "reason_parser.mly"
      ( mkexp (Pexp_override []))
# 71938 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce38 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos__0_) = Obj.magic _menhir_stack in
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv__class_self_pattern = let _endpos = _endpos__0_ in
    let _symbolstartpos = _endpos in
    
# 1905 "reason_parser.mly"
    (
    let loc = mklocation _symbolstartpos _endpos in
    mkpat (Ppat_var (mkloc "this" loc))
  )
# 71955 "reason_parser.ml"
     in
    _menhir_goto__class_self_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run432 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 71962 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
    | COLON | COMMA ->
        _menhir_reduce464 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run433 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2979 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState433 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 910 "reason_parser.mly"
       (string)
# 71999 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1242 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                _menhir_run1240 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2317
            | LPAREN ->
                _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2317) : 'freshtv2976)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2977 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2978)) : 'freshtv2980)
    | OPEN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2981 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState433 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BANG ->
            _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState435 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _ ->
            _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState435
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState435) : 'freshtv2982)
    | REC ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce684 _menhir_env (Obj.magic _menhir_stack) MenhirState433
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState433

and _menhir_run1110 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1110 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1110

and _menhir_run1651 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1651 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1651

and _menhir_reduce371 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 847 "reason_parser.mly"
       (int64)
# 72187 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 847 "reason_parser.mly"
       (int64)
# 72192 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4038 "reason_parser.mly"
                                      ( Const_int64 _1 )
# 72199 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce370 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 846 "reason_parser.mly"
       (int32)
# 72206 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 846 "reason_parser.mly"
       (int32)
# 72211 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4037 "reason_parser.mly"
                                      ( Const_int32 _1 )
# 72218 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce366 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 845 "reason_parser.mly"
       (int)
# 72225 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 845 "reason_parser.mly"
       (int)
# 72230 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4033 "reason_parser.mly"
                                      ( Const_int _1 )
# 72237 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run411 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2158 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2973 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState411 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState412) : 'freshtv2974)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState411

and _menhir_run451 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState451
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState451

and _menhir_run752 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState752 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState752

and _menhir_run210 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce270 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run898 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2971 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState898 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2967 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 863 "reason_parser.mly"
       (string)
# 72574 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2963 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 72586 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1017 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run902 _menhir_env (Obj.magic _menhir_stack) MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState901 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState901) : 'freshtv2964)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2965 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 72651 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2966)) : 'freshtv2968)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2970)) : 'freshtv2972)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState898

and _menhir_run902 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState902

and _menhir_run1017 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1017 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1017

and _menhir_run184 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2961) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4077 "reason_parser.mly"
                                                ( "*" )
# 72799 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 72805 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2962)

and _menhir_run211 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState211

and _menhir_run217 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce375 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run218 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 887 "reason_parser.mly"
       (string)
# 72832 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce605 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run219 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2959) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4085 "reason_parser.mly"
                                                ( "+=" )
# 72853 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 72859 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2960)

and _menhir_run220 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce613 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run221 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run226 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run225 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run224 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run223 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce612 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2957 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2958)

and _menhir_run227 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2955) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4086 "reason_parser.mly"
                                                ( "%" )
# 72909 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 72915 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2956)

and _menhir_run228 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2953) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4080 "reason_parser.mly"
                                                ( "or" )
# 72933 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 72939 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2954)

and _menhir_run229 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2949 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 910 "reason_parser.mly"
       (string)
# 72956 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 72968 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run233 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState232) : 'freshtv2944)
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2945 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 72989 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce289 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv2946)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2947 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 73002 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2948)) : 'freshtv2950)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2951 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2952)

and _menhir_run488 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce615 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run489 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run494 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run491 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2941 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2942)

and _menhir_run495 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2939 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState495 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2935 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2931 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EXCEPTION ->
                    _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState711) : 'freshtv2932)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2934)) : 'freshtv2936)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2937 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2938)) : 'freshtv2940)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run496 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState495

and _menhir_run187 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 73244 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run508 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2929) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4087 "reason_parser.mly"
                                                ( "<>" )
# 73265 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73271 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2930)

and _menhir_run509 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2927) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4088 "reason_parser.mly"
                                                ( "<..>" )
# 73289 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73295 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2928)

and _menhir_run510 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2925) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4078 "reason_parser.mly"
                                                ( "<" )
# 73313 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73319 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2926)

and _menhir_run511 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2923 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState511 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce281 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2924)
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState511

and _menhir_run514 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState514

and _menhir_run516 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState516 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState516 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState516

and _menhir_run532 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState532

and _menhir_run534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 842 "reason_parser.mly"
       (string)
# 73517 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2921) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 842 "reason_parser.mly"
       (string)
# 73528 "reason_parser.ml"
    )) : (
# 842 "reason_parser.mly"
       (string)
# 73532 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4072 "reason_parser.mly"
                                                ( _1 )
# 73540 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73546 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2922)

and _menhir_run535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (string)
# 73553 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2919) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 841 "reason_parser.mly"
       (string)
# 73564 "reason_parser.ml"
    )) : (
# 841 "reason_parser.mly"
       (string)
# 73568 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4071 "reason_parser.mly"
                                                ( _1 )
# 73576 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73582 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2920)

and _menhir_run536 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (string)
# 73589 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2917) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 840 "reason_parser.mly"
       (string)
# 73600 "reason_parser.ml"
    )) : (
# 840 "reason_parser.mly"
       (string)
# 73604 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4070 "reason_parser.mly"
                                                ( _1 )
# 73612 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73618 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2918)

and _menhir_run537 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (string)
# 73625 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2915) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 839 "reason_parser.mly"
       (string)
# 73636 "reason_parser.ml"
    )) : (
# 839 "reason_parser.mly"
       (string)
# 73640 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4069 "reason_parser.mly"
                                                ( _1 )
# 73648 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73654 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2916)

and _menhir_run538 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 838 "reason_parser.mly"
       (string)
# 73661 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2913) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 838 "reason_parser.mly"
       (string)
# 73672 "reason_parser.ml"
    )) : (
# 838 "reason_parser.mly"
       (string)
# 73676 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4068 "reason_parser.mly"
                                                ( _1 )
# 73684 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73690 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2914)

and _menhir_run539 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2911) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4079 "reason_parser.mly"
                                                ( ">" )
# 73708 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73714 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2912)

and _menhir_run540 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState540

and _menhir_run568 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2909) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4084 "reason_parser.mly"
                                                ( ":=" )
# 73789 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73795 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2910)

and _menhir_run569 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2907) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4081 "reason_parser.mly"
                                                ( "||" )
# 73813 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73819 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2908)

and _menhir_run570 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce606 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run571 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2905) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4082 "reason_parser.mly"
                                                ( "&" )
# 73843 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73849 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2906)

and _menhir_run572 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2903) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4083 "reason_parser.mly"
                                                ( "&&" )
# 73867 "reason_parser.ml"
      
    in
    
# 4093 "reason_parser.mly"
                                                ( _1 )
# 73873 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2904)

and _menhir_run485 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 73880 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2901 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 73892 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState486) : 'freshtv2902)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run762 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState762

and _menhir_reduce369 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 827 "reason_parser.mly"
       (string)
# 74009 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 827 "reason_parser.mly"
       (string)
# 74014 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4036 "reason_parser.mly"
                                      ( Const_float _1 )
# 74021 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce376 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4114 "reason_parser.mly"
                                                ( Lident "false" )
# 74033 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run660 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run581 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState660

and _menhir_run2583 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 74097 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2899 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 74111 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2583 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2585 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2584) : 'freshtv2900)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2583
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2583

and _menhir_run2589 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2897 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2589 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2585 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2590) : 'freshtv2898)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2589
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2589

and _menhir_run2593 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2893 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2891 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2594 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run2586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2585 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2595) : 'freshtv2892)
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2594
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2594) : 'freshtv2894)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2895 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2896)

and _menhir_run2598 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2889 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2598 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2585 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2599 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2599) : 'freshtv2890)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2598 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2598
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2598

and _menhir_run2602 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2887 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2602 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2585 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2603) : 'freshtv2888)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2602
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2602

and _menhir_goto__use_file : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv__use_file -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2885) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__use_file) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2883) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv__use_file) : 'tv__use_file) = _v in
    ((let _v : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74451 "reason_parser.ml"
    ) = 
# 1132 "reason_parser.mly"
                    (apply_mapper_chain_to_use_file _1 default_mapper_chain)
# 74455 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2881) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74463 "reason_parser.ml"
    )) = _v in
    ((match _menhir_s with
    | MenhirState2747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2871) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74473 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2869) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74481 "reason_parser.ml"
        )) : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74485 "reason_parser.ml"
        )) = _v in
        (Obj.magic _1 : 'freshtv2870)) : 'freshtv2872)
    | MenhirState2751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2875 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74495 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2873 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74503 "reason_parser.ml"
        )) : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74507 "reason_parser.ml"
        )) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv__use_file = 
# 1136 "reason_parser.mly"
                                                   ( _1 :: _3 )
# 74514 "reason_parser.ml"
         in
        _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2874)) : 'freshtv2876)
    | MenhirState2754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2879 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74524 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2877 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74532 "reason_parser.ml"
        )) : (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74536 "reason_parser.ml"
        )) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv__use_file = 
# 1135 "reason_parser.mly"
                                                   ( Ptop_def[_1] :: _3 )
# 74543 "reason_parser.ml"
         in
        _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2878)) : 'freshtv2880)
    | _ ->
        _menhir_fail ()) : 'freshtv2882)) : 'freshtv2884)) : 'freshtv2886)

and _menhir_reduce844 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_virtual_flag = 
# 4202 "reason_parser.mly"
                                                ( Concrete )
# 74554 "reason_parser.ml"
     in
    _menhir_goto_virtual_flag _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1529 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2867) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_virtual_flag = 
# 4203 "reason_parser.mly"
                                                ( Virtual )
# 74570 "reason_parser.ml"
     in
    _menhir_goto_virtual_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2868)

and _menhir_run2046 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | VIRTUAL ->
        _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2046
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2046

and _menhir_reduce367 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 806 "reason_parser.mly"
       (char)
# 74592 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 806 "reason_parser.mly"
       (char)
# 74597 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4034 "reason_parser.mly"
                                      ( Const_char _1 )
# 74604 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run124 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 74611 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce451 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run125 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 74621 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce452 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run15 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState15

and _menhir_reduce784 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_structure = 
# 1376 "reason_parser.mly"
                ( [] )
# 74747 "reason_parser.ml"
     in
    _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v

and _menhir_run2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2

and _menhir_run16 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 74807 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run74 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState74 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce602 _menhir_env (Obj.magic _menhir_stack) MenhirState74

and _menhir_run17 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState17

and _menhir_run175 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce377 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run176 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState176

and _menhir_run178 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 902 "reason_parser.mly"
       (string * string option)
# 74939 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce368 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run179 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 887 "reason_parser.mly"
       (string)
# 74948 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState179

and _menhir_run391 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState391 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _ ->
        _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState391
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState391

and _menhir_run180 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState180

and _menhir_run182 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 876 "reason_parser.mly"
       (nativeint)
# 75035 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce372 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run397 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2863 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState398 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState398 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState398) : 'freshtv2864)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2865 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2866)

and _menhir_run407 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2861 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState407 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2857 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2853 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | ASSERT ->
                    _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BANG ->
                    _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FOR ->
                    _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FUN ->
                    _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | IF ->
                    _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACELESS ->
                    _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUSDOT ->
                    _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NEW ->
                    _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUSDOT ->
                    _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PREFIXOP _v ->
                    _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SWITCH ->
                    _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRY ->
                    _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | WHILE ->
                    _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2368) : 'freshtv2854)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2855 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2856)) : 'freshtv2858)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2859 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2860)) : 'freshtv2862)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState407

and _menhir_run413 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 75302 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run440 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2851 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState440 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState441 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2303) : 'freshtv2850)
        | UIDENT _v ->
            _menhir_run442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState441 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState441) : 'freshtv2852)
    | REC ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce684 _menhir_env (Obj.magic _menhir_stack) MenhirState440
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState440

and _menhir_run452 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState452

and _menhir_run188 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState188

and _menhir_run414 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2847 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState414 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce230 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2848)
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState414

and _menhir_run454 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState454

and _menhir_run415 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run2182 _menhir_env (Obj.magic _menhir_stack) MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState415

and _menhir_run422 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | GREATERRBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2845 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState422 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce240 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2846)
    | LIDENT _v ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState422

and _menhir_run431 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1651 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1110 _menhir_env (Obj.magic _menhir_stack) MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState431 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState431
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState431

and _menhir_run456 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState456

and _menhir_run457 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 847 "reason_parser.mly"
       (int64)
# 76022 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce371 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run458 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 846 "reason_parser.mly"
       (int32)
# 76031 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce370 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run459 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 845 "reason_parser.mly"
       (int)
# 76040 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce366 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1470 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1937 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1471 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState1470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1470

and _menhir_run474 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState474

and _menhir_run476 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2843 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState476 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RPAREN ->
            _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2841 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState477 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_v : (
# 863 "reason_parser.mly"
       (string)
# 76267 "reason_parser.ml"
                )) = _v in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | RPAREN ->
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv2833 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 76279 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                    let _menhir_env = _menhir_discard _menhir_env in
                    let _tok = _menhir_env._menhir_token in
                    match _tok with
                    | BACKQUOTE ->
                        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | CHAR _v ->
                        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | COLON ->
                        _menhir_run2214 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | EQUALGREATER ->
                        _menhir_run2212 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FALSE ->
                        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FLOAT _v ->
                        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT _v ->
                        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT32 _v ->
                        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT64 _v ->
                        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACE ->
                        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKET ->
                        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETBAR ->
                        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETPERCENT ->
                        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LIDENT _v ->
                        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LPAREN ->
                        _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | MINUS ->
                        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | NATIVEINT _v ->
                        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | PLUS ->
                        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | SHARP ->
                        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | STRING _v ->
                        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | TRUE ->
                        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UIDENT _v ->
                        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UNDERSCORE ->
                        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | _ ->
                        assert (not _menhir_env._menhir_error);
                        _menhir_env._menhir_error <- true;
                        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState480) : 'freshtv2834)
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv2835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 76344 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2836)) : 'freshtv2838)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2840)) : 'freshtv2842)
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState477) : 'freshtv2844)
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState476

and _menhir_run738 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState738

and _menhir_run460 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 827 "reason_parser.mly"
       (string)
# 76444 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce369 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run461 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce376 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1472 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1472

and _menhir_run1483 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run1512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1499 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1484 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1483

and _menhir_run1528 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2831 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1528 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1530 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState1530
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1530) : 'freshtv2832)
    | VIRTUAL ->
        _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1528 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState1528
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1528

and _menhir_run462 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 806 "reason_parser.mly"
       (char)
# 76527 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce367 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run463 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState463

and _menhir_run464 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState464

and _menhir_run806 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState806 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState806

and _menhir_reduce701 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_signature = 
# 1668 "reason_parser.mly"
                         ( [] )
# 76659 "reason_parser.ml"
     in
    _menhir_goto_signature _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1247 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce602 _menhir_env (Obj.magic _menhir_stack) MenhirState1247

and _menhir_run1398 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2827 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1399 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1399 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1399) : 'freshtv2828)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2829 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2830)

and _menhir_run1403 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1403 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2825 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1403 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2817 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run1459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1458 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1458) : 'freshtv2818)
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 910 "reason_parser.mly"
       (string)
# 76747 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1436 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2819 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 76761 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_menhir_s : _menhir_state) = MenhirState1405 in
                ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1453) : 'freshtv2820)
            | LPAREN ->
                _menhir_run1406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1405) : 'freshtv2822)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2823 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2824)) : 'freshtv2826)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1403

and _menhir_run2030 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1243 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run449 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2030 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2030 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2030

and _menhir_run2033 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2033

and _menhir_run2042 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run1383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run1380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2042 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2042 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2042

and _menhir_run2045 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        _menhir_run2046 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2045
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2045

and _menhir_run118 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce140 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run77 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 76877 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState77 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState77

and _menhir_run119 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState119 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState119

and _menhir_run123 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState123

and _menhir_run127 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState127

and _menhir_run141 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 76965 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2815 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 76977 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState142) : 'freshtv2816)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce820 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run144 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce144 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run145 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce143 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run146 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOT ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState146
    | LIDENT _v ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState146 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState146

and _menhir_run155 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState155
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState155

and _menhir_run158 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2813 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState158 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce149 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2814)
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce630 _menhir_env (Obj.magic _menhir_stack) MenhirState158
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState158

and _menhir_run161 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2811 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState161 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState162) : 'freshtv2812)
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run121 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState161

and _menhir_run191 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce270 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run192 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run143 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState192

and _menhir_run197 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run201 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run200 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run199 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2809 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2810)

and _menhir_run203 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run207 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run206 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run205 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2807 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2808)

and _menhir_run209 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2805 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState209 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2801 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EXCEPTION ->
                    _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2409) : 'freshtv2798)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2799 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2800)) : 'freshtv2802)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2803 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2804)) : 'freshtv2806)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState209 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState209

and _menhir_run18 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 77426 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run588 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2795 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState588 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce281 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2796)
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState588

and _menhir_run603 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603
    | SHARP ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState603

and _menhir_run615 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState615 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState615 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState615

and _menhir_run619 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState619

and _menhir_run642 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState642

and _menhir_goto__toplevel_phrase : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv__toplevel_phrase -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2793) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__toplevel_phrase) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2791) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv__toplevel_phrase) : 'tv__toplevel_phrase) = _v in
    ((let _v : (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 77694 "reason_parser.ml"
    ) = 
# 1125 "reason_parser.mly"
                                  (apply_mapper_chain_to_toplevel_phrase _1 default_mapper_chain)
# 77698 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2789) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 77706 "reason_parser.ml"
    )) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2787) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 77714 "reason_parser.ml"
    )) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2785) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 77722 "reason_parser.ml"
    )) : (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 77726 "reason_parser.ml"
    )) = _v in
    (Obj.magic _1 : 'freshtv2786)) : 'freshtv2788)) : 'freshtv2790)) : 'freshtv2792)) : 'freshtv2794)

and _menhir_errorcase : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    match _menhir_s with
    | MenhirState2754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv55 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv56)
    | MenhirState2751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv57 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv58)
    | MenhirState2747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv59) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv60)
    | MenhirState2742 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv61 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv62)
    | MenhirState2739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv63 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv64)
    | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv65 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv66)
    | MenhirState2736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv67 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv68)
    | MenhirState2733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv69 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv70)
    | MenhirState2732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv71 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv72)
    | MenhirState2731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv73 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv74)
    | MenhirState2730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv75 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv76)
    | MenhirState2727 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv77 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv78)
    | MenhirState2726 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv79 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv80)
    | MenhirState2725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv81 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv82)
    | MenhirState2724 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv83 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv84)
    | MenhirState2711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv85 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv86)
    | MenhirState2704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv87 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv88)
    | MenhirState2703 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv89 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv90)
    | MenhirState2701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv91 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv92)
    | MenhirState2700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv93 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv94)
    | MenhirState2699 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv95 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv96)
    | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv97 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv98)
    | MenhirState2695 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv99 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv100)
    | MenhirState2694 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv101 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv102)
    | MenhirState2692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv103 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv104)
    | MenhirState2691 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv105 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 77862 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv106)
    | MenhirState2684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv107 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv108)
    | MenhirState2682 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv109 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv110)
    | MenhirState2681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv111 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv112)
    | MenhirState2677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv113 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv114)
    | MenhirState2676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv115 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv116)
    | MenhirState2675 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv117 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv118)
    | MenhirState2674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv119 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv120)
    | MenhirState2673 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv121 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv122)
    | MenhirState2671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv123 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv124)
    | MenhirState2667 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv125 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState2667 : 'freshtv126)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv127 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv128)
    | MenhirState2660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv131 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv129 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2660 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce53 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv130)) : 'freshtv132)
    | MenhirState2659 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv133 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv134)
    | MenhirState2658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv135 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 77940 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv136)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv137 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 77949 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv138)
    | MenhirState2656 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv139 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 77958 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv140)
    | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv141 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv142)
    | MenhirState2651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv144)
    | MenhirState2649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv145 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv146)
    | MenhirState2648 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv148)
    | MenhirState2646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv149 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv150)
    | MenhirState2645 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv151 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv152)
    | MenhirState2643 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv153 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv154)
    | MenhirState2641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv155 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv156)
    | MenhirState2639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv158)
    | MenhirState2638 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv160)
    | MenhirState2636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv161 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv162)
    | MenhirState2634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv163 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv164)
    | MenhirState2633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv165 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv166)
    | MenhirState2631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv167 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78032 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv168)
    | MenhirState2630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv170)
    | MenhirState2629 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv171 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv172)
    | MenhirState2626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv173 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv174)
    | MenhirState2625 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv175 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv176)
    | MenhirState2624 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv177 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv178)
    | MenhirState2623 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv179 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv180)
    | MenhirState2621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv181 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv182)
    | MenhirState2620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv183 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv184)
    | MenhirState2614 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv185 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv186)
    | MenhirState2613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv187 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv188)
    | MenhirState2607 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv189 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv190)
    | MenhirState2606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv191 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv192)
    | MenhirState2604 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv193 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv194)
    | MenhirState2603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv196)
    | MenhirState2602 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv197 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv198)
    | MenhirState2600 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv200)
    | MenhirState2599 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv202)
    | MenhirState2598 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv203 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv204)
    | MenhirState2596 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv205 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv206)
    | MenhirState2595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv208)
    | MenhirState2594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv209 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv210)
    | MenhirState2591 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv211 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv212)
    | MenhirState2590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv213 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv214)
    | MenhirState2589 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv215 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv216)
    | MenhirState2587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv217 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78161 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv218)
    | MenhirState2586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv219 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv220)
    | MenhirState2585 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv221 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv222)
    | MenhirState2584 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv223 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78180 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv224)
    | MenhirState2583 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv225 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78189 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv226)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv227 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv228)
    | MenhirState2581 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv229 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv230)
    | MenhirState2580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv231 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv232)
    | MenhirState2578 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv233 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv234)
    | MenhirState2577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv235 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv236)
    | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv237) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv238)
    | MenhirState2573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv239 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv240)
    | MenhirState2572 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv241) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv242)
    | MenhirState2570 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv243 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv244)
    | MenhirState2568 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv245) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv246)
    | MenhirState2564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv247) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv248)
    | MenhirState2559 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv249 * _menhir_state * 'tv_signature_item) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv250)
    | MenhirState2557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv251) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv252)
    | MenhirState2551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv253 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv254)
    | MenhirState2550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv255 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce76 _menhir_env (Obj.magic _menhir_stack) : 'freshtv256)
    | MenhirState2549 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv257 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv258)
    | MenhirState2548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv259 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 78271 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv260)
    | MenhirState2546 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv261 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv262)
    | MenhirState2545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv264)
    | MenhirState2542 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv265 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv266)
    | MenhirState2534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv267 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv268)
    | MenhirState2533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv269 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2533 : 'freshtv270)
    | MenhirState2532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv271 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv272)
    | MenhirState2531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv273 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv274)
    | MenhirState2530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv275 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv276)
    | MenhirState2527 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv277 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2527 : 'freshtv278)
    | MenhirState2522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv279 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv280)
    | MenhirState2521 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv281 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv282)
    | MenhirState2517 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv283 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv284)
    | MenhirState2516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv285 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv286)
    | MenhirState2515 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv287 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78343 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2515 : 'freshtv288)
    | MenhirState2514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv289 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        _menhir_error2513 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv290)
    | MenhirState2510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv291 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv292)
    | MenhirState2509 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv293 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv294)
    | MenhirState2505 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv295 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce395 _menhir_env (Obj.magic _menhir_stack) : 'freshtv296)
    | MenhirState2504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv297 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce394 _menhir_env (Obj.magic _menhir_stack) : 'freshtv298)
    | MenhirState2501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv299 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce393 _menhir_env (Obj.magic _menhir_stack) : 'freshtv300)
    | MenhirState2498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv301 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2498 : 'freshtv302)
    | MenhirState2497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv303 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 : 'freshtv304)
    | MenhirState2495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv305 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2495 : 'freshtv306)
    | MenhirState2494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv307 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 : 'freshtv308)
    | MenhirState2492 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv309 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2492 : 'freshtv310)
    | MenhirState2491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv311 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 : 'freshtv312)
    | MenhirState2488 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv313 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2488 : 'freshtv314)
    | MenhirState2487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv315 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 : 'freshtv316)
    | MenhirState2485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv317 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 78410 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2485 : 'freshtv318)
    | MenhirState2484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv319 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 78418 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 : 'freshtv320)
    | MenhirState2481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv321 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2481 : 'freshtv322)
    | MenhirState2480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv323 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 : 'freshtv324)
    | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv325 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 : 'freshtv326)
    | MenhirState2477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv327 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2477 : 'freshtv328)
    | MenhirState2475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv329 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 : 'freshtv330)
    | MenhirState2474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2474 : 'freshtv332)
    | MenhirState2473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv333 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv334)
    | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv335 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv336)
    | MenhirState2470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2470 : 'freshtv338)
    | MenhirState2469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv339 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState2469 : 'freshtv340)
    | MenhirState2467 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv341 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 78468 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState2467 : 'freshtv342)
    | MenhirState2466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv343 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv344)
    | MenhirState2464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv345 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv346)
    | MenhirState2460 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv347 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78486 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv348)
    | MenhirState2459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv349 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78495 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv350)
    | MenhirState2458 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv351 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78504 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv352)
    | MenhirState2457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv353 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78513 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv354)
    | MenhirState2455 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv355 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78522 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv356)
    | MenhirState2453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv357 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv358)
    | MenhirState2451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv359 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv360)
    | MenhirState2450 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv361 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv362)
    | MenhirState2443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv363 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv364)
    | MenhirState2436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv365 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv366)
    | MenhirState2435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv367 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv368)
    | MenhirState2433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv369 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv370)
    | MenhirState2428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv371 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv372)
    | MenhirState2427 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv374)
    | MenhirState2426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv375 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv376)
    | MenhirState2422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv377 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2422 : 'freshtv378)
    | MenhirState2420 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv379 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 78585 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce235 _menhir_env (Obj.magic _menhir_stack) : 'freshtv380)
    | MenhirState2417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv381 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv382)
    | MenhirState2416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv383 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv384)
    | MenhirState2415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv385 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv386)
    | MenhirState2412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv389 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2412 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce171 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv388)) : 'freshtv390)
    | MenhirState2411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv391 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv392)
    | MenhirState2410 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv394)
    | MenhirState2409 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv396)
    | MenhirState2405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv397 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error2404 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv398)
    | MenhirState2401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv399 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78638 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce59 _menhir_env (Obj.magic _menhir_stack) : 'freshtv400)
    | MenhirState2400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv401 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78646 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv402)
    | MenhirState2399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv403 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78655 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce58 _menhir_env (Obj.magic _menhir_stack) : 'freshtv404)
    | MenhirState2390 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2390 : 'freshtv406)
    | MenhirState2389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce132 _menhir_env (Obj.magic _menhir_stack) : 'freshtv408)
    | MenhirState2387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2387 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce120 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv410)) : 'freshtv412)
    | MenhirState2385 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2385 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce121 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv414)) : 'freshtv416)
    | MenhirState2383 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2383 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce122 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv418)) : 'freshtv420)
    | MenhirState2381 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2381 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce123 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv422)) : 'freshtv424)
    | MenhirState2376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2376 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce206 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv426)) : 'freshtv428)
    | MenhirState2373 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv431 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv429 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2373 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce211 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv430)) : 'freshtv432)
    | MenhirState2371 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv434)
    | MenhirState2370 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv435 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv436)
    | MenhirState2369 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv438)
    | MenhirState2368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv440)
    | MenhirState2362 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv443 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv441 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2362 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce246 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv442)) : 'freshtv444)
    | MenhirState2360 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv445 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv446)
    | MenhirState2358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv449 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2358 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce120 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv448)) : 'freshtv450)
    | MenhirState2356 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv453 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2356 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce121 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv452)) : 'freshtv454)
    | MenhirState2354 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv457 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv455 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2354 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce122 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv456)) : 'freshtv458)
    | MenhirState2352 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv461 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2352 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce123 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv460)) : 'freshtv462)
    | MenhirState2348 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv463 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2348 : 'freshtv464)
    | MenhirState2345 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv466)
    | MenhirState2344 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv467 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv468)
    | MenhirState2341 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv469 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv470)
    | MenhirState2340 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv471 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv472)
    | MenhirState2337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv474)
    | MenhirState2333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv475 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2333 : 'freshtv476)
    | MenhirState2331 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv477 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 78839 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce432 _menhir_env (Obj.magic _menhir_stack) : 'freshtv478)
    | MenhirState2320 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv479 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 78847 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv480)
    | MenhirState2318 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 78856 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv482)
    | MenhirState2317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv483 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 78865 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv484)
    | MenhirState2314 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv486)
    | MenhirState2309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv487 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2309 : 'freshtv488)
    | MenhirState2308 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv489 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv490)
    | MenhirState2306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2306 : 'freshtv492)
    | MenhirState2304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2304 : 'freshtv494)
    | MenhirState2303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv495 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv496)
    | MenhirState2301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv497 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce112 _menhir_env (Obj.magic _menhir_stack) : 'freshtv498)
    | MenhirState2300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv499 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv500)
    | MenhirState2298 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv501 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv502)
    | MenhirState2297 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv503 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce111 _menhir_env (Obj.magic _menhir_stack) : 'freshtv504)
    | MenhirState2296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv505 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv506)
    | MenhirState2295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv507 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv508)
    | MenhirState2290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce579 _menhir_env (Obj.magic _menhir_stack) : 'freshtv510)
    | MenhirState2289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv512)
    | MenhirState2288 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv513 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv514)
    | MenhirState2287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv515 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv516)
    | MenhirState2286 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv517 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce578 _menhir_env (Obj.magic _menhir_stack) : 'freshtv518)
    | MenhirState2285 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv519 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv520)
    | MenhirState2283 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv522)
    | MenhirState2282 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv524)
    | MenhirState2275 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv525 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce108 _menhir_env (Obj.magic _menhir_stack) : 'freshtv526)
    | MenhirState2272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv527 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv528)
    | MenhirState2271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv529 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv530)
    | MenhirState2269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv531 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1044 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv532)
    | MenhirState2268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv533 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv534)
    | MenhirState2266 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv535 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2266 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv536)
    | MenhirState2265 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce74 _menhir_env (Obj.magic _menhir_stack) : 'freshtv538)
    | MenhirState2264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv539 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv540)
    | MenhirState2263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv541 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce75 _menhir_env (Obj.magic _menhir_stack) : 'freshtv542)
    | MenhirState2262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv544)
    | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv545 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce635 _menhir_env (Obj.magic _menhir_stack) : 'freshtv546)
    | MenhirState2252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv547 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2252 : 'freshtv548)
    | MenhirState2251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv549 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2251 : 'freshtv550)
    | MenhirState2249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv551 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv552)
    | MenhirState2247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv553 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce635 _menhir_env (Obj.magic _menhir_stack) : 'freshtv554)
    | MenhirState2246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv555 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2246 : 'freshtv556)
    | MenhirState2242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv557 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv558)
    | MenhirState2241 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv559 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce373 _menhir_env (Obj.magic _menhir_stack) : 'freshtv560)
    | MenhirState2232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv561 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv562)
    | MenhirState2229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv563 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        (_menhir_reduce68 _menhir_env (Obj.magic _menhir_stack) : 'freshtv564)
    | MenhirState2228 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv565 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce485 _menhir_env (Obj.magic _menhir_stack) : 'freshtv566)
    | MenhirState2227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv567 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv568)
    | MenhirState2226 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv569 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce486 _menhir_env (Obj.magic _menhir_stack) : 'freshtv570)
    | MenhirState2225 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv571 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv572)
    | MenhirState2224 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv573 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv574)
    | MenhirState2223 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv575 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv576)
    | MenhirState2218 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv577 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv578)
    | MenhirState2217 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv579 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce438 _menhir_env (Obj.magic _menhir_stack) : 'freshtv580)
    | MenhirState2216 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv581 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv582)
    | MenhirState2215 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv583 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv584)
    | MenhirState2214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv585 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv586)
    | MenhirState2213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv587 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce437 _menhir_env (Obj.magic _menhir_stack) : 'freshtv588)
    | MenhirState2212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv589 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv590)
    | MenhirState2211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv591 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 79116 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv592)
    | MenhirState2208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv593 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv594)
    | MenhirState2207 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv595 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv596)
    | MenhirState2205 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv597 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv598)
    | MenhirState2204 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv599 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv600)
    | MenhirState2202 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv601 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run922 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv602)
    | MenhirState2196 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv603 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv604)
    | MenhirState2192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv605 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2192 : 'freshtv606)
    | MenhirState2191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv607 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv608)
    | MenhirState2186 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv609 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv610)
    | MenhirState2185 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv611 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv612)
    | MenhirState2182 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv613 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv614)
    | MenhirState2181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv615 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv616)
    | MenhirState2177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv617 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2177 : 'freshtv618)
    | MenhirState2176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv619 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv620)
    | MenhirState2173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv623 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv621 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2173 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce215 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv622)) : 'freshtv624)
    | MenhirState2171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv625 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv626)
    | MenhirState2167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv629 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv627 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2167 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce248 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv628)) : 'freshtv630)
    | MenhirState2166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv631 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv632)
    | MenhirState2163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv635 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv633 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2163 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce253 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv634)) : 'freshtv636)
    | MenhirState2162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv637 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv638)
    | MenhirState2159 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv641 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv639 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2159 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce119 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv640)) : 'freshtv642)
    | MenhirState2155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv643 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv644)
    | MenhirState2152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv646)
    | MenhirState2149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv647 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv648)
    | MenhirState2147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv649 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv650)
    | MenhirState2143 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv652)
    | MenhirState2141 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv653 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv654)
    | MenhirState2138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv655 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv656)
    | MenhirState2137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv657 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv658)
    | MenhirState2136 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv659 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv660)
    | MenhirState2134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv661 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv662)
    | MenhirState2133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv663 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv664)
    | MenhirState2132 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv665 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv666)
    | MenhirState2131 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv667 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv668)
    | MenhirState2128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 79311 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv670)
    | MenhirState2126 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv671 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 79320 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv672)
    | MenhirState2123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv674)
    | MenhirState2122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv676)
    | MenhirState2121 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv678)
    | MenhirState2120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv680)
    | MenhirState2119 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv682)
    | MenhirState2117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv684)
    | MenhirState2116 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv686)
    | MenhirState2112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv687 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2112 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce262 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv688)) : 'freshtv690)
    | MenhirState2111 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce127 _menhir_env (Obj.magic _menhir_stack) : 'freshtv692)
    | MenhirState2110 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv694)
    | MenhirState2108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv696)
    | MenhirState2107 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv698)
    | MenhirState2103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2103 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce262 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv700)) : 'freshtv702)
    | MenhirState2097 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv704)
    | MenhirState2090 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv705 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv706)
    | MenhirState2087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv707 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv708)
    | MenhirState2086 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv709 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv710)
    | MenhirState2082 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv712)
    | MenhirState2081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv713 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv714)
    | MenhirState2080 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv715 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv716)
    | MenhirState2077 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv717 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv718)
    | MenhirState2076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv719 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv720)
    | MenhirState2072 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv721 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv722)
    | MenhirState2071 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv723 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 79451 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv724)
    | MenhirState2070 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv725 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 79460 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv726)
    | MenhirState2069 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv727 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv728)
    | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv729 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv730)
    | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv731 * _menhir_state * 'tv_signature_item) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState2062 : 'freshtv732)
    | MenhirState2059 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv734)
    | MenhirState2057 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv735 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79488 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv736)
    | MenhirState2056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv737 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79497 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv738)
    | MenhirState2055 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv739 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79506 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv740)
    | MenhirState2052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv742)
    | MenhirState2051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv743 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79520 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv744)
    | MenhirState2050 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv745 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79529 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv746)
    | MenhirState2049 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv747 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79538 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv748)
    | MenhirState2048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv749 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79547 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv750)
    | MenhirState2046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv752)
    | MenhirState2045 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv753 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv754)
    | MenhirState2043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv756)
    | MenhirState2042 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv757 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv758)
    | MenhirState2040 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv760)
    | MenhirState2038 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv761 * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 79581 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv762)
    | MenhirState2037 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv763 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv764)
    | MenhirState2035 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv766)
    | MenhirState2033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv767 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv768)
    | MenhirState2031 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv769 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv770)
    | MenhirState2030 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv771 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv772)
    | MenhirState2027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv773 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv774)
    | MenhirState2026 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv775 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv776)
    | MenhirState2021 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv777 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv778)
    | MenhirState2019 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState2019 : 'freshtv780)
    | MenhirState2010 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv781 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce250 _menhir_env (Obj.magic _menhir_stack) : 'freshtv782)
    | MenhirState2000 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv783 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce128 _menhir_env (Obj.magic _menhir_stack) : 'freshtv784)
    | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv785 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv786)
    | MenhirState1995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv787 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1994 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv788)
    | MenhirState1993 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce848 _menhir_env (Obj.magic _menhir_stack) : 'freshtv790)
    | MenhirState1992 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv792)
    | MenhirState1991 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv793 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv794)
    | MenhirState1989 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv795 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv796)
    | MenhirState1988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv797 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1987 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv798)
    | MenhirState1985 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv799 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79675 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv800)
    | MenhirState1984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv801 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79684 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv802)
    | MenhirState1983 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv803 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79693 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce572 _menhir_env (Obj.magic _menhir_stack) : 'freshtv804)
    | MenhirState1982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv805 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv806)
    | MenhirState1981 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv807 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79706 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce849 _menhir_env (Obj.magic _menhir_stack) : 'freshtv808)
    | MenhirState1979 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv809 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79714 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv810)
    | MenhirState1978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv811 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79723 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv812)
    | MenhirState1977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv813 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79732 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce571 _menhir_env (Obj.magic _menhir_stack) : 'freshtv814)
    | MenhirState1976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv815 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 79740 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv816)
    | MenhirState1974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv817 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv818)
    | MenhirState1973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv819 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        (_menhir_reduce846 _menhir_env (Obj.magic _menhir_stack) : 'freshtv820)
    | MenhirState1972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1972 : 'freshtv822)
    | MenhirState1971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv823 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv824)
    | MenhirState1969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv826)
    | MenhirState1968 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv828)
    | MenhirState1967 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv830)
    | MenhirState1966 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv831 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv832)
    | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv833 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv834)
    | MenhirState1963 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1963 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce253 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv836)) : 'freshtv838)
    | MenhirState1959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv840)
    | MenhirState1956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv843 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv841 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1956 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce119 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv842)) : 'freshtv844)
    | MenhirState1952 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv845 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv846)
    | MenhirState1949 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv847 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv848)
    | MenhirState1946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv850)
    | MenhirState1944 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv851 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv852)
    | MenhirState1943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv853 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv854)
    | MenhirState1942 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv855 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv856)
    | MenhirState1941 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv857 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce117 _menhir_env (Obj.magic _menhir_stack) : 'freshtv858)
    | MenhirState1939 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv859 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv860)
    | MenhirState1938 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv861 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv862)
    | MenhirState1937 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv863 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv864)
    | MenhirState1933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv865 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv866)
    | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv867 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1929 : 'freshtv868)
    | MenhirState1926 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv869 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1926 : 'freshtv870)
    | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv871 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv872)
    | MenhirState1918 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv873 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1918 : 'freshtv874)
    | MenhirState1917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv875 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce500 _menhir_env (Obj.magic _menhir_stack) : 'freshtv876)
    | MenhirState1916 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv877 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv878)
    | MenhirState1915 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv879 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv880)
    | MenhirState1914 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv881 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce501 _menhir_env (Obj.magic _menhir_stack) : 'freshtv882)
    | MenhirState1913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv883 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv884)
    | MenhirState1911 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv885 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv886)
    | MenhirState1906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv887 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv888)
    | MenhirState1904 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv889 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce496 _menhir_env (Obj.magic _menhir_stack) : 'freshtv890)
    | MenhirState1903 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv891 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv892)
    | MenhirState1900 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv893 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv894)
    | MenhirState1898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv895 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv896)
    | MenhirState1896 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv897 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce536 _menhir_env (Obj.magic _menhir_stack) : 'freshtv898)
    | MenhirState1894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv899 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv900)
    | MenhirState1892 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv901 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv902)
    | MenhirState1890 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv903 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv904)
    | MenhirState1889 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv905 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv906)
    | MenhirState1888 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv907 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce61 _menhir_env (Obj.magic _menhir_stack) : 'freshtv908)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv909 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv910)
    | MenhirState1885 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv911 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 79972 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv912)
    | MenhirState1882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv913 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv914)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv915 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv916)
    | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv917 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv918)
    | MenhirState1879 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv919 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce295 _menhir_env (Obj.magic _menhir_stack) : 'freshtv920)
    | MenhirState1876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv921 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1876 : 'freshtv922)
    | MenhirState1875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv923 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv924)
    | MenhirState1874 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv925 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce306 _menhir_env (Obj.magic _menhir_stack) : 'freshtv926)
    | MenhirState1871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv927 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1871 : 'freshtv928)
    | MenhirState1870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv929 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv930)
    | MenhirState1869 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv931 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce307 _menhir_env (Obj.magic _menhir_stack) : 'freshtv932)
    | MenhirState1866 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1866 : 'freshtv934)
    | MenhirState1865 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv935 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv936)
    | MenhirState1864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv937 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce302 _menhir_env (Obj.magic _menhir_stack) : 'freshtv938)
    | MenhirState1861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 : 'freshtv940)
    | MenhirState1858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv941 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        (_menhir_reduce639 _menhir_env (Obj.magic _menhir_stack) : 'freshtv942)
    | MenhirState1855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv943 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv944)
    | MenhirState1853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv945 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv946)
    | MenhirState1852 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv947 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80057 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        (_menhir_reduce808 _menhir_env (Obj.magic _menhir_stack) : 'freshtv948)
    | MenhirState1851 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv949 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80065 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) MenhirState1851 : 'freshtv950)
    | MenhirState1850 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv951 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80073 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        (_menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState1850 : 'freshtv952)
    | MenhirState1849 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv953 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80081 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState1849 : 'freshtv954)
    | MenhirState1848 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv955 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv956)
    | MenhirState1847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv957 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce298 _menhir_env (Obj.magic _menhir_stack) : 'freshtv958)
    | MenhirState1840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv959 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv960)
    | MenhirState1836 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv961 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1836 : 'freshtv962)
    | MenhirState1828 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv963 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv964)
    | MenhirState1826 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv965 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce347 _menhir_env (Obj.magic _menhir_stack) : 'freshtv966)
    | MenhirState1825 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv967 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv968)
    | MenhirState1821 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv969 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv970)
    | MenhirState1820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv971 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv972)
    | MenhirState1819 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv973 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce346 _menhir_env (Obj.magic _menhir_stack) : 'freshtv974)
    | MenhirState1818 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv975 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv976)
    | MenhirState1817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv977 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv978)
    | MenhirState1814 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce54 _menhir_env (Obj.magic _menhir_stack) : 'freshtv980)
    | MenhirState1813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv982)
    | MenhirState1810 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv983 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80153 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv984)
    | MenhirState1809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv985 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80162 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv986)
    | MenhirState1808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv987 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80171 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv988)
    | MenhirState1803 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv989 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv990)
    | MenhirState1802 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv991 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv992)
    | MenhirState1801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv993 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80190 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv994)
    | MenhirState1800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv995 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80199 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv996)
    | MenhirState1799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv997 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80208 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv998)
    | MenhirState1795 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1000)
    | MenhirState1794 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1001 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1002)
    | MenhirState1793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1003 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1004)
    | MenhirState1789 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1005 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1006)
    | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1007 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1008)
    | MenhirState1784 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1009 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce354 _menhir_env (Obj.magic _menhir_stack) MenhirState1784 : 'freshtv1010)
    | MenhirState1782 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1011 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1012)
    | MenhirState1781 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1013 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1014)
    | MenhirState1780 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1015 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80256 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce339 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1016)
    | MenhirState1775 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1017 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80264 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1018)
    | MenhirState1774 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1019 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80273 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1020)
    | MenhirState1773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1021 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80282 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1022)
    | MenhirState1768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1023 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1024)
    | MenhirState1767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1025 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1026)
    | MenhirState1766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1027 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80301 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1028)
    | MenhirState1765 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1029 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80310 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1030)
    | MenhirState1764 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1031 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 80319 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1032)
    | MenhirState1762 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1033 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce7 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1034)
    | MenhirState1761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1035 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1036)
    | MenhirState1760 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1037 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1038)
    | MenhirState1757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1041 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1039 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1757 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce53 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1040)) : 'freshtv1042)
    | MenhirState1751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1043 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce19 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1044)
    | MenhirState1747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1045 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce692 _menhir_env (Obj.magic _menhir_stack) MenhirState1747 : 'freshtv1046)
    | MenhirState1743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1047 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1743 : 'freshtv1048)
    | MenhirState1739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1049 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1739 : 'freshtv1050)
    | MenhirState1738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1051 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1052)
    | MenhirState1736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1053 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1736 : 'freshtv1054)
    | MenhirState1733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1055 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce650 _menhir_env (Obj.magic _menhir_stack) MenhirState1733 : 'freshtv1056)
    | MenhirState1732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1057 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1058)
    | MenhirState1731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1059 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1060)
    | MenhirState1729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1061 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1729 : 'freshtv1062)
    | MenhirState1728 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1063 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1064)
    | MenhirState1725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1065 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1725 : 'freshtv1066)
    | MenhirState1723 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1067 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce565 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1068)
    | MenhirState1722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1069 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1070)
    | MenhirState1720 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((((('freshtv1071 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce567 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1072)
    | MenhirState1719 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((((('freshtv1073 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1074)
    | MenhirState1717 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1075 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1076)
    | MenhirState1715 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1077 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1078)
    | MenhirState1714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1079 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1080)
    | MenhirState1713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1081 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce566 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1082)
    | MenhirState1712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1083 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1084)
    | MenhirState1708 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1085 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1086)
    | MenhirState1707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1087 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce407 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1088)
    | MenhirState1706 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1089 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1090)
    | MenhirState1705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1091 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1092)
    | MenhirState1704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1093 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1094)
    | MenhirState1703 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1095 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce402 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1096)
    | MenhirState1702 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1097 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1098)
    | MenhirState1701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1099 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80479 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1100)
    | MenhirState1698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1101 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1102)
    | MenhirState1697 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1103 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1104)
    | MenhirState1696 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1105 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1106)
    | MenhirState1694 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1107 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1108)
    | MenhirState1692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1109 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1110)
    | MenhirState1689 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1111 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1112)
    | MenhirState1687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1113 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1114)
    | MenhirState1686 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1115 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1116)
    | MenhirState1685 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1117 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1118)
    | MenhirState1684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1119 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1120)
    | MenhirState1683 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1121 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce840 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1122)
    | MenhirState1682 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1123 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1124)
    | MenhirState1680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1125 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce839 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1126)
    | MenhirState1679 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1127 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1128)
    | MenhirState1678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1129 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1130)
    | MenhirState1677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1131 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1132)
    | MenhirState1674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1133 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1134)
    | MenhirState1672 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1135 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1136)
    | MenhirState1670 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1137 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1138)
    | MenhirState1668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1139 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1668 : 'freshtv1140)
    | MenhirState1665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1141 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1142)
    | MenhirState1663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1144)
    | MenhirState1662 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1145 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1146)
    | MenhirState1661 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1148)
    | MenhirState1660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1149 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce692 _menhir_env (Obj.magic _menhir_stack) MenhirState1660 : 'freshtv1150)
    | MenhirState1658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1151 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1152)
    | MenhirState1657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1153 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1657 : 'freshtv1154)
    | MenhirState1656 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1155 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1156)
    | MenhirState1652 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1158)
    | MenhirState1651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1160)
    | MenhirState1644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1161 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce30 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1162)
    | MenhirState1642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1163 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce13 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1164)
    | MenhirState1641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1165 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce11 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1166)
    | MenhirState1637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1167 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce16 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1168)
    | MenhirState1636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce15 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1170)
    | MenhirState1635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1171 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1172)
    | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1173 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1174)
    | MenhirState1633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1175 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1176)
    | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1177 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1178)
    | MenhirState1629 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1179 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1180)
    | MenhirState1628 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1181 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1182)
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1183 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1184)
    | MenhirState1626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1185 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1626 : 'freshtv1186)
    | MenhirState1625 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1187 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1188)
    | MenhirState1624 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1189 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80697 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1190)
    | MenhirState1621 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1191 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80706 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1192)
    | MenhirState1620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1193 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80715 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1194)
    | MenhirState1617 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1617 : 'freshtv1196)
    | MenhirState1611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1197 * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv1198)
    | MenhirState1609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1199 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80732 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce361 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1200)
    | MenhirState1604 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1201 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce635 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1202)
    | MenhirState1603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1203 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1603 : 'freshtv1204)
    | MenhirState1600 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1205 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1600 : 'freshtv1206)
    | MenhirState1597 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1597 : 'freshtv1208)
    | MenhirState1595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1209 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1210)
    | MenhirState1593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1211 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1212)
    | MenhirState1589 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1213 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1589 : 'freshtv1214)
    | MenhirState1588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1215 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce40 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1216)
    | MenhirState1587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1217 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce33 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1218)
    | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1219 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce32 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1220)
    | MenhirState1582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1221 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1222)
    | MenhirState1581 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1223 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1224)
    | MenhirState1580 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1225 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1226)
    | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1227 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 : 'freshtv1228)
    | MenhirState1573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1229 * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1230)
    | MenhirState1570 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1231 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1232)
    | MenhirState1569 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1233 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1234)
    | MenhirState1567 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1235 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1236)
    | MenhirState1560 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1237 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1238)
    | MenhirState1557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1239 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1557 : 'freshtv1240)
    | MenhirState1555 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1241 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1242)
    | MenhirState1553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1243 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1244)
    | MenhirState1552 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1245 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1246)
    | MenhirState1550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1247 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1248)
    | MenhirState1548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1249 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1250)
    | MenhirState1547 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1251 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1252)
    | MenhirState1546 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1253 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1254)
    | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1255 * Lexing.position * _menhir_state * 'tv_class_self_type) = Obj.magic _menhir_stack in
        (_menhir_reduce357 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 : 'freshtv1256)
    | MenhirState1539 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1257 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1258)
    | MenhirState1538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1259 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce354 _menhir_env (Obj.magic _menhir_stack) MenhirState1538 : 'freshtv1260)
    | MenhirState1536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1261 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80878 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1262)
    | MenhirState1533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1263 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80887 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1264)
    | MenhirState1532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1265 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 80896 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1266)
    | MenhirState1530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1267 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1268)
    | MenhirState1528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1269 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1270)
    | MenhirState1526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1526 : 'freshtv1272)
    | MenhirState1524 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 : 'freshtv1274)
    | MenhirState1522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1522 : 'freshtv1276)
    | MenhirState1520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1520 : 'freshtv1278)
    | MenhirState1519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1279 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1280)
    | MenhirState1518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1281 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 : 'freshtv1282)
    | MenhirState1516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1283 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1516 : 'freshtv1284)
    | MenhirState1514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1514 : 'freshtv1286)
    | MenhirState1513 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1287 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1288)
    | MenhirState1512 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1289 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 : 'freshtv1290)
    | MenhirState1510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1510 : 'freshtv1292)
    | MenhirState1508 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1508 : 'freshtv1294)
    | MenhirState1507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1295 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1296)
    | MenhirState1506 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1297 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1506 : 'freshtv1298)
    | MenhirState1503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1299 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1503 : 'freshtv1300)
    | MenhirState1501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1301 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 : 'freshtv1302)
    | MenhirState1500 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1303 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1304)
    | MenhirState1499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1305 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1499 : 'freshtv1306)
    | MenhirState1497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1307 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 80991 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1497 : 'freshtv1308)
    | MenhirState1495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1309 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1310)
    | MenhirState1494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1311 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1312)
    | MenhirState1492 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1313 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1314)
    | MenhirState1490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1315 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 : 'freshtv1316)
    | MenhirState1488 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1317 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81017 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) MenhirState1488 : 'freshtv1318)
    | MenhirState1487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1319 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1320)
    | MenhirState1486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1321 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1322)
    | MenhirState1485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1323 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81035 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1324)
    | MenhirState1484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1325 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81044 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState1484 : 'freshtv1326)
    | MenhirState1483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1327 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1328)
    | MenhirState1481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1329 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1481 : 'freshtv1330)
    | MenhirState1479 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState1479 : 'freshtv1332)
    | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1333 * Lexing.position * _menhir_state * (
# 902 "reason_parser.mly"
       (string * string option)
# 81065 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce675 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1334)
    | MenhirState1476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1336)
    | MenhirState1474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1338)
    | MenhirState1472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1339 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1340)
    | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1341 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1342)
    | MenhirState1470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1343 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1344)
    | MenhirState1469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1345 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1346)
    | MenhirState1468 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1347 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1348)
    | MenhirState1466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1349 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1350)
    | MenhirState1465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1351 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1352)
    | MenhirState1462 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1353 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1354)
    | MenhirState1461 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1355 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81121 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1356)
    | MenhirState1460 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1357 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81130 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1358)
    | MenhirState1458 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1359 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1360)
    | MenhirState1456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1361 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81144 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1362)
    | MenhirState1454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81153 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1364)
    | MenhirState1453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1365 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81162 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1366)
    | MenhirState1451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1367 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1368)
    | MenhirState1439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1369 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1370)
    | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1371 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1372)
    | MenhirState1437 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1374)
    | MenhirState1436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1375 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1376)
    | MenhirState1435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1377 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81196 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1378)
    | MenhirState1432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1379 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1380)
    | MenhirState1430 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1381 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1382)
    | MenhirState1429 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1383 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1384)
    | MenhirState1428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1385 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1386)
    | MenhirState1427 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81225 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1388)
    | MenhirState1426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1389 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81234 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1390)
    | MenhirState1424 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1391 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1392)
    | MenhirState1423 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1394)
    | MenhirState1422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1396)
    | MenhirState1421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1397 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1398)
    | MenhirState1419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1399 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1400)
    | MenhirState1416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1401 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1402)
    | MenhirState1415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1403 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1404)
    | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1405 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1406)
    | MenhirState1413 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1407 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1408)
    | MenhirState1412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81288 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1410)
    | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81297 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1412)
    | MenhirState1405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81306 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1414)
    | MenhirState1403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1415 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1416)
    | MenhirState1402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1417 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState1402 : 'freshtv1418)
    | MenhirState1401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1420)
    | MenhirState1400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1422)
    | MenhirState1399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1424)
    | MenhirState1396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1426)
    | MenhirState1395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1428)
    | MenhirState1394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1429 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1430)
    | MenhirState1392 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1431 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1432)
    | MenhirState1391 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1433 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1434)
    | MenhirState1387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1435 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1436)
    | MenhirState1386 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1437 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1438)
    | MenhirState1384 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1440)
    | MenhirState1383 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1441 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1442)
    | MenhirState1381 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1443 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1444)
    | MenhirState1380 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1445 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1446)
    | MenhirState1378 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1448)
    | MenhirState1377 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1449 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1450)
    | MenhirState1374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1452)
    | MenhirState1373 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1453 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1454)
    | MenhirState1371 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1455 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81414 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1456)
    | MenhirState1370 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1457 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81423 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1458)
    | MenhirState1369 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1459 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1460)
    | MenhirState1368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1461 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1462)
    | MenhirState1366 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1463 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1464)
    | MenhirState1363 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1465 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1466)
    | MenhirState1361 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1467 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1468)
    | MenhirState1358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1469 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1470)
    | MenhirState1355 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1471 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1472)
    | MenhirState1353 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1473 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1474)
    | MenhirState1352 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1475 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 81472 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1476)
    | MenhirState1351 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1477 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 81481 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1478)
    | MenhirState1346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1479 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1480)
    | MenhirState1345 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1481 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1482)
    | MenhirState1341 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1483 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1484)
    | MenhirState1340 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1485 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1486)
    | MenhirState1339 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1487 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81510 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1488)
    | MenhirState1338 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1489 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1490)
    | MenhirState1332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1491 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1492)
    | MenhirState1331 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1493 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1494)
    | MenhirState1326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1495 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 81534 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1496)
    | MenhirState1324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1497 * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1498)
    | MenhirState1322 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1499 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 81548 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1500)
    | MenhirState1318 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1501 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1502)
    | MenhirState1314 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1503 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1504)
    | MenhirState1313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1505 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1506)
    | MenhirState1310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1507 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1508)
    | MenhirState1307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1509 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1510)
    | MenhirState1306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1511 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1512)
    | MenhirState1304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1513 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1514)
    | MenhirState1303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1515 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1516)
    | MenhirState1301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1517 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1518)
    | MenhirState1300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1519 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1520)
    | MenhirState1297 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1521 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1522)
    | MenhirState1296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1523 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1524)
    | MenhirState1294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1525 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81617 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1526)
    | MenhirState1293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1527 * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81626 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1528)
    | MenhirState1290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1530)
    | MenhirState1289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1531 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1532)
    | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1533 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1534)
    | MenhirState1286 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1535 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1536)
    | MenhirState1284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1538)
    | MenhirState1283 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1539 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1540)
    | MenhirState1282 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1541 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1542)
    | MenhirState1281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1543 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1544)
    | MenhirState1279 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1545 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1546)
    | MenhirState1278 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1547 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1548)
    | MenhirState1276 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1549 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1550)
    | MenhirState1274 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1551 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81690 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1552)
    | MenhirState1272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1553 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1554)
    | MenhirState1271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1555 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1556)
    | MenhirState1269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1557 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1558)
    | MenhirState1266 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1559 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81714 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1560)
    | MenhirState1265 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1561 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81723 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1562)
    | MenhirState1264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1563 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81732 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1564)
    | MenhirState1263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1565 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1566)
    | MenhirState1262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1567 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 81746 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1568)
    | MenhirState1259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1569 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1570)
    | MenhirState1255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1571 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1572)
    | MenhirState1251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1573 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1574)
    | MenhirState1249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1575 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 81770 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1576)
    | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1577 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1578)
    | MenhirState1247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1579 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1580)
    | MenhirState1246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1581 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState1246 : 'freshtv1582)
    | MenhirState1245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1583 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1584)
    | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1585 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 81798 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce451 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1586)
    | MenhirState1243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1587 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1588)
    | MenhirState1242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1589 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1590)
    | MenhirState1241 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1591 * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1592)
    | MenhirState1240 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1593 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1594)
    | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1595 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 81826 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1596)
    | MenhirState1234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1597 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1598)
    | MenhirState1233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1599 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1600)
    | MenhirState1232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1601 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1602)
    | MenhirState1231 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1603 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1604)
    | MenhirState1230 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1605 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1606)
    | MenhirState1229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1607 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1608)
    | MenhirState1227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1609 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1610)
    | MenhirState1222 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1611 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1612)
    | MenhirState1220 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1613 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1614)
    | MenhirState1219 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1615 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1616)
    | MenhirState1216 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1617 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1618)
    | MenhirState1214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1619 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1620)
    | MenhirState1212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1621 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1622)
    | MenhirState1210 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1623 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1624)
    | MenhirState1208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1625 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1626)
    | MenhirState1206 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1627 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1628)
    | MenhirState1203 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1629 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 81915 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1630)
    | MenhirState1202 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1631 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1632)
    | MenhirState1201 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1633 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1634)
    | MenhirState1200 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1635 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1636)
    | MenhirState1199 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1637 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1638)
    | MenhirState1194 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1639 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1640)
    | MenhirState1193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1641 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1642)
    | MenhirState1192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1643 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1644)
    | MenhirState1191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1646)
    | MenhirState1190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1647 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1648)
    | MenhirState1189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1649 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1650)
    | MenhirState1188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1651 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1652)
    | MenhirState1187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 81979 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1654)
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1655 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1656)
    | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 81993 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1658)
    | MenhirState1180 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1659 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1660)
    | MenhirState1179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1661 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1662)
    | MenhirState1178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1663 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1664)
    | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1665 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1666)
    | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1667 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1668)
    | MenhirState1170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1669 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1670)
    | MenhirState1169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1671 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1672)
    | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1673 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1674)
    | MenhirState1162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1675 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1676)
    | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1677 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1678)
    | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1680)
    | MenhirState1154 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1682)
    | MenhirState1153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1684)
    | MenhirState1152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1686)
    | MenhirState1151 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1687 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1688)
    | MenhirState1150 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1689 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1690)
    | MenhirState1149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1691 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1692)
    | MenhirState1144 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1144 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce206 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1694)) : 'freshtv1696)
    | MenhirState1142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1697 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1698)
    | MenhirState1139 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1701 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1699 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1139 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce211 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1700)) : 'freshtv1702)
    | MenhirState1137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1703 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1704)
    | MenhirState1134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1706)
    | MenhirState1133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1708)
    | MenhirState1132 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1710)
    | MenhirState1130 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1711 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce600 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1712)
    | MenhirState1129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1713 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1714)
    | MenhirState1126 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1715 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1716)
    | MenhirState1119 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1717 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce472 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1718)
    | MenhirState1118 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1719 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1117 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv1720)
    | MenhirState1115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1721 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce476 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1722)
    | MenhirState1112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1723 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1724)
    | MenhirState1110 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1725 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1726)
    | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1727 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1728)
    | MenhirState1104 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1729 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1104 : 'freshtv1730)
    | MenhirState1103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1731 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1732)
    | MenhirState1098 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1733 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1734)
    | MenhirState1097 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1735 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1736)
    | MenhirState1094 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1737 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1738)
    | MenhirState1093 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1739 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1740)
    | MenhirState1087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1741 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error1086 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1742)
    | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1743 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1744)
    | MenhirState1082 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1745 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1746)
    | MenhirState1081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1747 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce418 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1748)
    | MenhirState1079 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1749 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce633 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1750)
    | MenhirState1078 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1751 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1078 : 'freshtv1752)
    | MenhirState1076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1753 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1754)
    | MenhirState1073 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1757 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1755 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1073 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce215 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1756)) : 'freshtv1758)
    | MenhirState1071 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1759 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1760)
    | MenhirState1069 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1761 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1762)
    | MenhirState1066 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1763 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1764)
    | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1765 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1766)
    | MenhirState1061 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1767 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1768)
    | MenhirState1060 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1769 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1770)
    | MenhirState1058 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1771 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1772)
    | MenhirState1056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1773 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce100 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1774)
    | MenhirState1055 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1775 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce102 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1776)
    | MenhirState1054 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1777 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        _menhir_error1053 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv1778)
    | MenhirState1052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1779 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce105 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1780)
    | MenhirState1051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1783 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1781 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce221 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1782)) : 'freshtv1784)
    | MenhirState1049 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1785 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1786)
    | MenhirState1048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1787 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1788)
    | MenhirState1047 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1789 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce104 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1790)
    | MenhirState1046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1793 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1791 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce219 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1792)) : 'freshtv1794)
    | MenhirState1043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1795 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1044 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1796)
    | MenhirState1042 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1797 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1798)
    | MenhirState1041 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1799 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce109 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1800)
    | MenhirState1040 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1801 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1802)
    | MenhirState1039 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1803 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1804)
    | MenhirState1038 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1805 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1806)
    | MenhirState1036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1807 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1808)
    | MenhirState1033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1810)
    | MenhirState1032 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1811 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1812)
    | MenhirState1031 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1813 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1814)
    | MenhirState1030 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1815 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1816)
    | MenhirState1029 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1817 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1818)
    | MenhirState1028 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1819 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1820)
    | MenhirState1027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1821 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1822)
    | MenhirState1022 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1823 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1824)
    | MenhirState1020 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1826)
    | MenhirState1019 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1828)
    | MenhirState1018 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1830)
    | MenhirState1017 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1831 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1832)
    | MenhirState1016 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1833 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1834)
    | MenhirState1015 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1835 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1836)
    | MenhirState1014 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1837 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1838)
    | MenhirState1013 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1839 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1840)
    | MenhirState1011 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1841 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1842)
    | MenhirState1010 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1843 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1844)
    | MenhirState1009 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1845 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1846)
    | MenhirState1008 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1847 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1848)
    | MenhirState1007 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1849 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1850)
    | MenhirState1006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1851 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1852)
    | MenhirState1004 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1853 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1854)
    | MenhirState1003 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1855 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1856)
    | MenhirState1002 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1857 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1858)
    | MenhirState1001 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1859 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1860)
    | MenhirState998 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1861 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1862)
    | MenhirState997 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1863 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1864)
    | MenhirState996 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1865 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1866)
    | MenhirState995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1867 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1868)
    | MenhirState994 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1869 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1870)
    | MenhirState993 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1871 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1872)
    | MenhirState992 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1873 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1874)
    | MenhirState991 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1875 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1876)
    | MenhirState990 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1877 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1878)
    | MenhirState989 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1879 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1880)
    | MenhirState988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1881 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1882)
    | MenhirState987 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1883 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1884)
    | MenhirState986 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1885 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1886)
    | MenhirState985 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1887 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 82551 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1888)
    | MenhirState984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1889 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 82560 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1890)
    | MenhirState983 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1891 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1892)
    | MenhirState982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1893 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1894)
    | MenhirState981 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1895 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1896)
    | MenhirState980 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1897 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1898)
    | MenhirState979 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1899 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 82589 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1900)
    | MenhirState978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1901 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 82598 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1902)
    | MenhirState977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1903 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 82607 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1904)
    | MenhirState976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1905 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 82616 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1906)
    | MenhirState975 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1907 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1908)
    | MenhirState974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1909 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1910)
    | MenhirState973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1911 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1912)
    | MenhirState972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1913 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1914)
    | MenhirState971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1915 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1916)
    | MenhirState970 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1917 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1918)
    | MenhirState969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1919 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1920)
    | MenhirState968 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1921 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1922)
    | MenhirState967 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1923 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1924)
    | MenhirState966 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1925 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1926)
    | MenhirState965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1927 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1928)
    | MenhirState964 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1929 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1930)
    | MenhirState963 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1931 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 82685 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1932)
    | MenhirState962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1933 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 82694 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1934)
    | MenhirState961 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1935 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1936)
    | MenhirState960 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1937 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1938)
    | MenhirState959 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1939 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1940)
    | MenhirState958 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1941 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1942)
    | MenhirState956 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1943 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1944)
    | MenhirState955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1945 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1946)
    | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1947 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1948)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1949 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1950)
    | MenhirState946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1951 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 82743 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1952)
    | MenhirState945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1953 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 82752 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1954)
    | MenhirState944 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1955 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1956)
    | MenhirState943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1957 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1958)
    | MenhirState942 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1959 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1960)
    | MenhirState941 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1961 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1962)
    | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1963 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1964)
    | MenhirState933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1965 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 82786 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1966)
    | MenhirState931 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1967 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1968)
    | MenhirState930 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1969 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 82800 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1970)
    | MenhirState928 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1971 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1972)
    | MenhirState926 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1973 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1974)
    | MenhirState925 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1975 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1976)
    | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1977 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1978)
    | MenhirState921 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1979 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run922 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1980)
    | MenhirState920 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1981 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1982)
    | MenhirState919 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1983 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1984)
    | MenhirState918 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1985 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1986)
    | MenhirState917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1987 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1988)
    | MenhirState916 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1990)
    | MenhirState915 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1991 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1992)
    | MenhirState909 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1994)
    | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1996)
    | MenhirState907 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1998)
    | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2000)
    | MenhirState905 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2001 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2002)
    | MenhirState904 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2004)
    | MenhirState903 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2005 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2006)
    | MenhirState902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2007 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2008)
    | MenhirState901 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2009 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 82903 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2010)
    | MenhirState898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2011 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2012)
    | MenhirState897 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2013 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 82917 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2014)
    | MenhirState894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2015 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2016)
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2017 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2018)
    | MenhirState892 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2019 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2020)
    | MenhirState891 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2022)
    | MenhirState890 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2023 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2024)
    | MenhirState889 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2025 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2026)
    | MenhirState888 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2027 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2028)
    | MenhirState887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2029 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce103 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2030)
    | MenhirState886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2033 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2031 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce217 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2032)) : 'freshtv2034)
    | MenhirState883 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2035 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2036)
    | MenhirState882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2037 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce95 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2038)
    | MenhirState881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2039 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2040)
    | MenhirState880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2041 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce94 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2042)
    | MenhirState879 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2043 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2044)
    | MenhirState878 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2045 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce93 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2046)
    | MenhirState877 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2047 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2048)
    | MenhirState876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2049 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce92 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2050)
    | MenhirState875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2051 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2052)
    | MenhirState874 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2053 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce90 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2054)
    | MenhirState873 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2055 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2056)
    | MenhirState872 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2057 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 83021 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce79 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2058)
    | MenhirState871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2059 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 838 "reason_parser.mly"
       (string)
# 83029 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2060)
    | MenhirState870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2061 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce89 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2062)
    | MenhirState869 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2063 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2064)
    | MenhirState868 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2065 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce99 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2066)
    | MenhirState867 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2067 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2068)
    | MenhirState866 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2069 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 83056 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce80 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2070)
    | MenhirState865 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2071 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 839 "reason_parser.mly"
       (string)
# 83064 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2072)
    | MenhirState864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2073 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 83073 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce81 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2074)
    | MenhirState863 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2075 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 840 "reason_parser.mly"
       (string)
# 83081 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2076)
    | MenhirState862 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2077 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce98 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2078)
    | MenhirState861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2079 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2080)
    | MenhirState860 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2081 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce86 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2082)
    | MenhirState859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2083 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2084)
    | MenhirState858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2085 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce87 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2086)
    | MenhirState857 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2087 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2088)
    | MenhirState856 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2089 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce91 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2090)
    | MenhirState855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2091 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2092)
    | MenhirState854 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2093 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce84 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2094)
    | MenhirState853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2095 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2096)
    | MenhirState852 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2097 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce85 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2098)
    | MenhirState851 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2099 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2100)
    | MenhirState850 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2101 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 83144 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce82 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2102)
    | MenhirState849 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2103 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 841 "reason_parser.mly"
       (string)
# 83152 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2104)
    | MenhirState848 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2105 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce97 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2106)
    | MenhirState847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2107 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2108)
    | MenhirState846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2109 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce96 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2110)
    | MenhirState845 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2111 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2112)
    | MenhirState843 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2113 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce101 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2114)
    | MenhirState842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2115 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2116)
    | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2117 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce267 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2118)
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2119 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2120)
    | MenhirState833 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2121 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 83196 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce83 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2122)
    | MenhirState832 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2123 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 842 "reason_parser.mly"
       (string)
# 83204 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2124)
    | MenhirState831 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2125 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce88 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2126)
    | MenhirState830 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2127 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2128)
    | MenhirState829 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2129 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce106 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2130)
    | MenhirState828 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2131 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2132)
    | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2133 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce65 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2134)
    | MenhirState820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2135 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 83235 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error816 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2136)
    | MenhirState818 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2137 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce493 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2138)
    | MenhirState817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2139 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 83248 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error816 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2140)
    | MenhirState815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2141 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce64 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2142)
    | MenhirState813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2143 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce73 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2144)
    | MenhirState812 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2145 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2146)
    | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2147 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2148)
    | MenhirState810 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2149 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2150)
    | MenhirState809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2151 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce493 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2152)
    | MenhirState808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2153 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2154)
    | MenhirState807 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2155 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce107 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2156)
    | MenhirState806 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2157 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2158)
    | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2160)
    | MenhirState796 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2163 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2161 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState796 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce248 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2162)) : 'freshtv2164)
    | MenhirState795 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2165 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2166)
    | MenhirState794 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2167 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2168)
    | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2169 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2170)
    | MenhirState792 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2171 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2172)
    | MenhirState791 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2173 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2174)
    | MenhirState789 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2175 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2176)
    | MenhirState788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2177 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2178)
    | MenhirState786 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2179 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2180)
    | MenhirState775 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2181 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2182)
    | MenhirState770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2185 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2183 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState770 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce246 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2184)) : 'freshtv2186)
    | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2187 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2188)
    | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2189 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2190)
    | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2191 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2192)
    | MenhirState765 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2193 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2194)
    | MenhirState763 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2195 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2196)
    | MenhirState762 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2197 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2198)
    | MenhirState761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2200)
    | MenhirState760 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2202)
    | MenhirState759 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2203 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2204)
    | MenhirState758 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2205 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2206)
    | MenhirState757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2207 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2208)
    | MenhirState756 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2209 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2210)
    | MenhirState754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2211 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2212)
    | MenhirState753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2213 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2214)
    | MenhirState752 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2215 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2216)
    | MenhirState751 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2217 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2218)
    | MenhirState750 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2219 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2220)
    | MenhirState749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2221 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2222)
    | MenhirState748 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2223 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce373 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2224)
    | MenhirState746 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2225 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce77 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2226)
    | MenhirState745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2227 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2228)
    | MenhirState744 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2229 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2230)
    | MenhirState741 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2231 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2232)
    | MenhirState740 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2233 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2234)
    | MenhirState739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2235 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2236)
    | MenhirState738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2237 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2238)
    | MenhirState737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2239 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2240)
    | MenhirState736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2241 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2242)
    | MenhirState734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2243 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2244)
    | MenhirState733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2245 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2246)
    | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2247 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2248)
    | MenhirState729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2249 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 83519 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2250)
    | MenhirState722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2253 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2251 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState722 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce288 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2252)) : 'freshtv2254)
    | MenhirState719 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2257 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2255 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState719 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce285 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2256)) : 'freshtv2258)
    | MenhirState714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2261 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2259 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState714 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce171 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2260)) : 'freshtv2262)
    | MenhirState713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2264)
    | MenhirState712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2265 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2266)
    | MenhirState711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2267 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2268)
    | MenhirState705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2269 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState705 : 'freshtv2270)
    | MenhirState700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState700 : 'freshtv2272)
    | MenhirState699 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2273 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce479 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2274)
    | MenhirState692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2275 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2276)
    | MenhirState688 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2277 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce60 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2278)
    | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2279 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2280)
    | MenhirState686 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2281 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce55 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2282)
    | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2283 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState681 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce288 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2284)) : 'freshtv2286)
    | MenhirState678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2287 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState678 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce285 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2288)) : 'freshtv2290)
    | MenhirState676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2291 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2292)
    | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2293 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2294)
    | MenhirState666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2295 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2296)
    | MenhirState665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2297 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2298)
    | MenhirState660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2299 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2300)
    | MenhirState658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2303 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2301 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState658 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce165 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2302)) : 'freshtv2304)
    | MenhirState657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2305 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2306)
    | MenhirState655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2309 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2307 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState655 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce169 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2308)) : 'freshtv2310)
    | MenhirState654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2311 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2312)
    | MenhirState650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2313 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2314)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2315 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2316)
    | MenhirState646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2317 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2318)
    | MenhirState644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2319 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2320)
    | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2321 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2322)
    | MenhirState635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2323 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2324)
    | MenhirState634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2325 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2326)
    | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2327 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2328)
    | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2329 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2330)
    | MenhirState615 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2331 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2332)
    | MenhirState614 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2333 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce658 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2334)
    | MenhirState613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2335 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce660 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2336)
    | MenhirState612 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2337 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce659 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2338)
    | MenhirState611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2339 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2340)
    | MenhirState610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2341 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2342)
    | MenhirState609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2343 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2344)
    | MenhirState605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2345 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState605 : 'freshtv2346)
    | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2347 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2348)
    | MenhirState602 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2349 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce656 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2350)
    | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2351 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2352)
    | MenhirState593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2353 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce657 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2354)
    | MenhirState592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2355 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2356)
    | MenhirState590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2357 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState590 : 'freshtv2358)
    | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2359 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2360)
    | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2361 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2362)
    | MenhirState579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2363 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2364)
    | MenhirState577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2367 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2365 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState577 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce165 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2366)) : 'freshtv2368)
    | MenhirState576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2369 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce168 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2370)
    | MenhirState574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2373 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2371 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState574 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce169 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2372)) : 'freshtv2374)
    | MenhirState573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2375 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce654 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2376)
    | MenhirState564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2377 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        (_menhir_reduce166 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2378)
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2379 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce273 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2380)
    | MenhirState554 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2381 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2382)
    | MenhirState553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2383 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce373 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2384)
    | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2385 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2386)
    | MenhirState548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce173 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2388)
    | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2389 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        _menhir_error544 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2390)
    | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2391 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2392)
    | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2393 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2394)
    | MenhirState532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2395 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2396)
    | MenhirState531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2397 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error530 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2398)
    | MenhirState526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2399 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState526 : 'freshtv2400)
    | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2401 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce482 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2402)
    | MenhirState519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2403 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2404)
    | MenhirState518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2405 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2406)
    | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2407 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2408)
    | MenhirState514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2409 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2410)
    | MenhirState511 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2411 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2412)
    | MenhirState505 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2413 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 83912 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce451 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2414)
    | MenhirState504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2415 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2416)
    | MenhirState503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2417 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2418)
    | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 83930 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2420)
    | MenhirState495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2421 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2422)
    | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2423 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2424)
    | MenhirState486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2425 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 83949 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2426)
    | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 83958 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2428)
    | MenhirState481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2429 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2430)
    | MenhirState480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2431 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 83972 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2432)
    | MenhirState477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2434)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2435 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2436)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2437 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState475 : 'freshtv2438)
    | MenhirState474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2439 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2440)
    | MenhirState473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2441 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2442)
    | MenhirState472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2443 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2444)
    | MenhirState469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2445 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2446)
    | MenhirState468 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce236 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2448)
    | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2449 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2450)
    | MenhirState463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2451 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2452)
    | MenhirState456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2453 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2454)
    | MenhirState455 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2455 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2456)
    | MenhirState454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2457 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2458)
    | MenhirState453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2459 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2460)
    | MenhirState452 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2461 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2462)
    | MenhirState451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2463 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2464)
    | MenhirState450 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2466)
    | MenhirState448 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2467 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2468)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2469 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2470)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2471 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84071 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2472)
    | MenhirState441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2474)
    | MenhirState440 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2475 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2476)
    | MenhirState439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2477 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2478)
    | MenhirState438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2479 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2480)
    | MenhirState437 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2482)
    | MenhirState436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2483 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2484)
    | MenhirState435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2486)
    | MenhirState433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2487 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2488)
    | MenhirState431 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2489 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState431 : 'freshtv2490)
    | MenhirState425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2491 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce633 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2492)
    | MenhirState424 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2493 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState424 : 'freshtv2494)
    | MenhirState422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2495 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2496)
    | MenhirState421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2497 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84136 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error420 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2498)
    | MenhirState419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2499 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2500)
    | MenhirState418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2501 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2502)
    | MenhirState417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2503 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2504)
    | MenhirState416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2505 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2506)
    | MenhirState415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2507 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2508)
    | MenhirState414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2509 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2510)
    | MenhirState412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2512)
    | MenhirState411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2513 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2514)
    | MenhirState410 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2515 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2516)
    | MenhirState407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2517 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2518)
    | MenhirState404 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2519 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2520)
    | MenhirState403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2521 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2522)
    | MenhirState402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2524)
    | MenhirState400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2525 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2526)
    | MenhirState399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2527 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState399 : 'freshtv2528)
    | MenhirState398 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2530)
    | MenhirState396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2531 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2532)
    | MenhirState395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2533 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2534)
    | MenhirState394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2535 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce671 _menhir_env (Obj.magic _menhir_stack) MenhirState394 : 'freshtv2536)
    | MenhirState393 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2538)
    | MenhirState391 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2539 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2540)
    | MenhirState388 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2541 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2542)
    | MenhirState336 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2543 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2544)
    | MenhirState333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2545 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2546)
    | MenhirState332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2547 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2548)
    | MenhirState331 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2549 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce57 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2550)
    | MenhirState328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2551 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce134 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2552)
    | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2553 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2554)
    | MenhirState326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2555 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84278 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2556)
    | MenhirState324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2557 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84287 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2558)
    | MenhirState323 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2561 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84296 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2559 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84303 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce820 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2560)) : 'freshtv2562)
    | MenhirState316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2563 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce135 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2564)
    | MenhirState312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2565 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2566)
    | MenhirState306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2567 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2568)
    | MenhirState305 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2569 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2570)
    | MenhirState302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2571 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2572)
    | MenhirState300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2573 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2574)
    | MenhirState295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2575 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2576)
    | MenhirState294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2577 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2578)
    | MenhirState291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2579 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2580)
    | MenhirState290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2581 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2582)
    | MenhirState287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2583 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2584)
    | MenhirState286 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2585 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2586)
    | MenhirState283 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2587 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2588)
    | MenhirState276 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2589 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2590)
    | MenhirState272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2591 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2592)
    | MenhirState271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2593 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2594)
    | MenhirState270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2595 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2596)
    | MenhirState269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2597 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2598)
    | MenhirState262 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2599 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2600)
    | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2601 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2602)
    | MenhirState253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2603 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2604)
    | MenhirState250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2605 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84415 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2606)
    | MenhirState249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2607 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2608)
    | MenhirState248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2609 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2610)
    | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2611 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2612)
    | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2613 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2614)
    | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2615 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2616)
    | MenhirState242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2617 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2618)
    | MenhirState241 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2619 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2620)
    | MenhirState239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2621 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2622)
    | MenhirState238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2623 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error237 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2624)
    | MenhirState233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2625 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84468 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce451 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2626)
    | MenhirState232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 910 "reason_parser.mly"
       (string)
# 84476 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2628)
    | MenhirState215 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2629 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2630)
    | MenhirState214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2631 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2632)
    | MenhirState211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2633 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2634)
    | MenhirState209 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2635 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2636)
    | MenhirState195 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2637 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2638)
    | MenhirState194 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2639 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2640)
    | MenhirState192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2641 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2642)
    | MenhirState190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2643 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2644)
    | MenhirState189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2645 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2646)
    | MenhirState188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2647 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2648)
    | MenhirState186 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2649 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 84534 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2650)
    | MenhirState183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2651 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2652)
    | MenhirState180 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2653 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2654)
    | MenhirState179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2655 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 84553 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2656)
    | MenhirState177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2657 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2658)
    | MenhirState176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2659 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2660)
    | MenhirState174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2661 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2662)
    | MenhirState173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2663 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2664)
    | MenhirState172 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2665 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2666)
    | MenhirState169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2667 * _menhir_state * 'tv_row_field_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2668)
    | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2669 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2670)
    | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2671 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2672)
    | MenhirState161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2673 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2674)
    | MenhirState160 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2675 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2676)
    | MenhirState158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2677 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2678)
    | MenhirState157 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2679 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2680)
    | MenhirState155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2681 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2682)
    | MenhirState153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2683 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2684)
    | MenhirState152 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2685 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2686)
    | MenhirState146 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2687 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2688)
    | MenhirState142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2689 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84641 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2690)
    | MenhirState140 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2692)
    | MenhirState137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2693 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2694)
    | MenhirState136 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2695 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2696)
    | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2697 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2698)
    | MenhirState133 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2699 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2700)
    | MenhirState129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2701 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84675 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2702)
    | MenhirState128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2704)
    | MenhirState127 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2705 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2706)
    | MenhirState123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2707 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2708)
    | MenhirState121 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2709 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2710)
    | MenhirState120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2711 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2712)
    | MenhirState119 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2713 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2714)
    | MenhirState117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2715 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84714 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2716)
    | MenhirState115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2717 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2718)
    | MenhirState112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2719 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2720)
    | MenhirState111 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2721 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2722)
    | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2723 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2724)
    | MenhirState108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2725 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84743 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce447 _menhir_env (Obj.magic _menhir_stack) MenhirState108 : 'freshtv2726)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2727 * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2728)
    | MenhirState105 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2729 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84756 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        (_menhir_reduce809 _menhir_env (Obj.magic _menhir_stack) MenhirState105 : 'freshtv2730)
    | MenhirState102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2731 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2732)
    | MenhirState98 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2734)
    | MenhirState92 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2735 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2736)
    | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2737 * Lexing.position * _menhir_state * (
# 863 "reason_parser.mly"
       (string)
# 84779 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) MenhirState90 : 'freshtv2738)
    | MenhirState88 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2739 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2740)
    | MenhirState87 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2741 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2742)
    | MenhirState84 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2743 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84797 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2744)
    | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2745 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84806 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2746)
    | MenhirState82 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2747 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84815 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2748)
    | MenhirState81 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2749 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2750)
    | MenhirState79 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2751 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84829 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2752)
    | MenhirState78 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2753 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84838 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2754)
    | MenhirState77 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2755 * Lexing.position * _menhir_state * (
# 910 "reason_parser.mly"
       (string)
# 84847 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2756)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2757 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_nonrec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2758)
    | MenhirState74 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2759 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2760)
    | MenhirState73 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2761 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2762)
    | MenhirState71 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2763 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error70 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2764)
    | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2765 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2766)
    | MenhirState17 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2767 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2768)
    | MenhirState15 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2769 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2770)
    | MenhirState10 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2771 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2772)
    | MenhirState9 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2773 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2774)
    | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2775 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2776)
    | MenhirState6 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2777 * Lexing.position * _menhir_state * (
# 887 "reason_parser.mly"
       (string)
# 84904 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2778)
    | MenhirState2 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2779 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2780)
    | MenhirState1 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2781 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2782)
    | MenhirState0 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2783) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2784)

and _menhir_run416 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState416

and _menhir_run3 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 910 "reason_parser.mly"
       (string)
# 84978 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run2577 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2577 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce602 _menhir_env (Obj.magic _menhir_stack) MenhirState2577

and _menhir_run418 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState418

and _menhir_run4 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce377 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run759 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState759

and _menhir_run5 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 902 "reason_parser.mly"
       (string * string option)
# 85110 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce368 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2613 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2613

and _menhir_run6 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 887 "reason_parser.mly"
       (string)
# 85134 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState6

and _menhir_run334 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce321 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run335 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce320 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run336 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run385 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run384 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run381 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run378 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run377 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run375 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run374 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run372 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run371 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run369 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run368 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run366 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run365 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run364 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run363 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run362 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run361 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run360 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run358 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run357 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run356 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run355 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run354 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run352 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run351 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run350 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run349 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run348 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run346 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run344 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run343 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run342 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run340 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState336 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState336

and _menhir_run1394 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1394 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _ ->
        _menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) MenhirState1394
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1394

and _menhir_run7 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState7 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState7 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState7

and _menhir_run14 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 876 "reason_parser.mly"
       (nativeint)
# 85344 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce372 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2622 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv51 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2623 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2623 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2623) : 'freshtv52)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv53 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv54)

and _menhir_run405 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce793 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run406 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce792 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run766 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run800 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv49 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState766 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv45 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv41 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | ASSERT ->
                    _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BACKQUOTE ->
                    _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BANG ->
                    _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FOR ->
                    _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FUN ->
                    _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | IF ->
                    _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACELESS ->
                    _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUSDOT ->
                    _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NEW ->
                    _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUSDOT ->
                    _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PREFIXOP _v ->
                    _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SWITCH ->
                    _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRY ->
                    _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | WHILE ->
                    _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState786 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState786) : 'freshtv42)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv43 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv44)) : 'freshtv46)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv47 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv48)) : 'freshtv50)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run783 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run782 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState766

and _menhir_run787 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 863 "reason_parser.mly"
       (string)
# 85623 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run2629 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv39 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2629 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv37 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2630 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run2631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2633 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2633) : 'freshtv38)
        | UIDENT _v ->
            _menhir_run2631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2630) : 'freshtv40)
    | REC ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2629 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce684 _menhir_env (Obj.magic _menhir_stack) MenhirState2629
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2629

and _menhir_run1468 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1468 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1468

and _menhir_run19 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState19

and _menhir_run177 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv35 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState177 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce230 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv36)
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState177

and _menhir_run2026 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2026 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2026

and _menhir_run417 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1094 _menhir_env (Obj.magic _menhir_stack) MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState417

and _menhir_run419 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | GREATERRBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv33 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState419 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce240 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv34)
    | LIDENT _v ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState419

and _menhir_run475 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1651 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1110 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState475
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState475

and _menhir_run788 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState788

and _menhir_run620 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 847 "reason_parser.mly"
       (int64)
# 86343 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce371 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run621 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 846 "reason_parser.mly"
       (int32)
# 86352 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce370 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run622 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 845 "reason_parser.mly"
       (int)
# 86361 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce366 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2638 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run752 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2638 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run451 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2638 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2638 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2638 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2638 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2638

and _menhir_run889 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState889 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState889

and _menhir_run893 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        _menhir_run762 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv31 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState893 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RPAREN ->
            _menhir_run217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run211 _menhir_env (Obj.magic _menhir_stack) MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv29 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState894 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv25 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_v : (
# 863 "reason_parser.mly"
       (string)
# 86588 "reason_parser.ml"
                )) = _v in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | RPAREN ->
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv21 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 86600 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                    let _menhir_env = _menhir_discard _menhir_env in
                    let _tok = _menhir_env._menhir_token in
                    match _tok with
                    | BACKQUOTE ->
                        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | CHAR _v ->
                        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | COLON ->
                        _menhir_run1017 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | EQUALGREATER ->
                        _menhir_run902 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FALSE ->
                        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FLOAT _v ->
                        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT _v ->
                        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT32 _v ->
                        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT64 _v ->
                        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACE ->
                        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKET ->
                        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETBAR ->
                        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETPERCENT ->
                        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LIDENT _v ->
                        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LPAREN ->
                        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | MINUS ->
                        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | NATIVEINT _v ->
                        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | PLUS ->
                        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | SHARP ->
                        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | STRING _v ->
                        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | TRUE ->
                        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UIDENT _v ->
                        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UNDERSCORE ->
                        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | _ ->
                        assert (not _menhir_env._menhir_error);
                        _menhir_env._menhir_error <- true;
                        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState897) : 'freshtv22)
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv23 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 863 "reason_parser.mly"
       (string)
# 86665 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv24)) : 'freshtv26)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv27 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv28)) : 'freshtv30)
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState894) : 'freshtv32)
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState893

and _menhir_run903 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState903

and _menhir_run623 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 827 "reason_parser.mly"
       (string)
# 86765 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce369 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run624 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce376 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run2641 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2641 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2641

and _menhir_run2648 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run2602 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run2598 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run2593 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2648

and _menhir_run2748 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv19) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    ((let _1 = () in
    let _v : 'tv__use_file = 
# 1134 "reason_parser.mly"
                                                   ( [] )
# 86821 "reason_parser.ml"
     in
    _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv20)

and _menhir_run2654 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        _menhir_run2046 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run1529 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce844 _menhir_env (Obj.magic _menhir_stack) MenhirState2654
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2654

and _menhir_run625 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 806 "reason_parser.mly"
       (char)
# 86845 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce367 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run730 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState730

and _menhir_run163 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState163

and _menhir_run915 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState915 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState915

and _menhir_discard : _menhir_env -> _menhir_env =
  fun _menhir_env ->
    let lexer = _menhir_env._menhir_lexer in
    let lexbuf = _menhir_env._menhir_lexbuf in
    let _tok = lexer lexbuf in
    {
      _menhir_lexer = lexer;
      _menhir_lexbuf = lexbuf;
      _menhir_token = _tok;
      _menhir_error = false;
    }

and _menhir_init : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> _menhir_env =
  fun lexer lexbuf ->
    let _tok = Obj.magic () in
    {
      _menhir_lexer = lexer;
      _menhir_lexbuf = lexbuf;
      _menhir_token = _tok;
      _menhir_error = false;
    }

and implementation : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1062 "reason_parser.mly"
      (Parsetree.structure)
# 86997 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv17) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run806 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1528 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1472 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run456 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run431 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run422 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run454 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run440 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run413 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run397 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run391 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run175 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) MenhirState0
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState0) : 'freshtv18))

and interface : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1064 "reason_parser.mly"
      (Parsetree.signature)
# 87104 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv15) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2045 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2042 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2033 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2030 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1403 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_reduce701 _menhir_env (Obj.magic _menhir_stack) MenhirState2557
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2557) : 'freshtv16))

and parse_core_type : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1070 "reason_parser.mly"
      (Parsetree.core_type)
# 87143 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv13) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run161 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run158 _menhir_env (Obj.magic _menhir_stack) MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run155 _menhir_env (Obj.magic _menhir_stack) MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run141 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run127 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run123 _menhir_env (Obj.magic _menhir_stack) MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run118 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2564) : 'freshtv14))

and parse_expression : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1072 "reason_parser.mly"
      (Parsetree.expression)
# 87186 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv11) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2568 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2568) : 'freshtv12))

and parse_pattern : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1074 "reason_parser.mly"
      (Parsetree.pattern)
# 87269 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv9) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2572 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2572) : 'freshtv10))

and toplevel_phrase : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1066 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 87332 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState2576 in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        ((let _1 = () in
        let _v : 'tv__toplevel_phrase = 
# 1128 "reason_parser.mly"
                                        ( raise End_of_file)
# 87362 "reason_parser.ml"
         in
        _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4)) : 'freshtv6)
    | EXCEPTION ->
        _menhir_run2648 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2638 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run2629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run2622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run2613 _menhir_env (Obj.magic _menhir_stack) MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run2577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2576) : 'freshtv8))

and use_file : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1068 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 87453 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv1) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run915 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run163 _menhir_env (Obj.magic _menhir_stack) MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_run2748 _menhir_env (Obj.magic _menhir_stack) MenhirState2747
    | EXCEPTION ->
        _menhir_run2648 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2638 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2026 _menhir_env (Obj.magic _menhir_stack) MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1468 _menhir_env (Obj.magic _menhir_stack) MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run2629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run766 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run405 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run2622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1394 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run334 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run2613 _menhir_env (Obj.magic _menhir_stack) MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run759 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run2577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2747) : 'freshtv2))

# 4354 "reason_parser.mly"
  

# 87562 "reason_parser.ml"

# 220 "/Users/hzhang295/.opam/4.02.3+local-git-master/lib/menhir/standard.mly"
  


# 87568 "reason_parser.ml"

end
module 
Reason_lexer
= struct
#1 "reason_lexer.ml"
# 52 "reason_lexer.mll"
 
open Lexing
open Misc
open Reason_parser

type error =
  | Illegal_character of char
  | Illegal_escape of string
  | Unterminated_comment of Location.t
  | Unterminated_string
  | Unterminated_string_in_comment of Location.t * Location.t
  | Keyword_as_label of string
  | Literal_overflow of string
;;

exception Error of error * Location.t;;

(* The table of keywords *)

let keyword_table =
  create_hashtable 149 [
    "and", AND;
    "as", AS;
    "assert", ASSERT;
    "begin", BEGIN;
    "class", CLASS;
    "constraint", CONSTRAINT;
    "do", DO;
    "done", DONE;
    "downto", DOWNTO;
    "else", ELSE;
    "end", END;
    "exception", EXCEPTION;
    "external", EXTERNAL;
    "false", FALSE;
    "for", FOR;
    "fun", FUN;
    "function", FUNCTION;
    "functor", FUNCTOR;
    "if", IF;
    "in", IN;
    "include", INCLUDE;
    "inherit", INHERIT;
    "initializer", INITIALIZER;
    "lazy", LAZY;
    "let", LET;
    "switch", SWITCH;
    "match", MATCH; (* Including MATCH for source transforming compat *)
    "method", METHOD;
    "module", MODULE;
    "mutable", MUTABLE;
    "new", NEW;
    "nonrec", NONREC;
    "object", OBJECT;
    "open", OPEN;
    "or", OR;
(*  "parser", PARSER; *)
    "private", PRIVATE;
    "rec", REC;
    "sig", SIG;
    "struct", STRUCT;
    "then", THEN;
    "to", TO;
    "true", TRUE;
    "try", TRY;
    "type", TYPE;
    "val", VAL;
    "virtual", VIRTUAL;
    "when", WHEN;
    "while", WHILE;
    "with", WITH;

    "mod", INFIXOP3("mod");
    "land", INFIXOP3("land");
    "lor", INFIXOP3("lor");
    "lxor", INFIXOP3("lxor");
    "lsl", INFIXOP4("lsl");
    "lsr", INFIXOP4("lsr");
    "asr", INFIXOP4("asr")
]

(* To buffer string literals *)

let initial_string_buffer = Bytes.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  if !string_index >= Bytes.length !string_buff then begin
    let new_buff = Bytes.create (Bytes.length (!string_buff) * 2) in
    Bytes.blit !string_buff 0 new_buff 0 (Bytes.length !string_buff);
    string_buff := new_buff
  end;
  Bytes.unsafe_set !string_buff !string_index c;
  incr string_index

let store_string s =
  for i = 0 to String.length s - 1 do
    store_string_char s.[i];
  done

let store_lexeme lexbuf =
  store_string (Lexing.lexeme lexbuf)

let get_stored_string () =
  let s = Bytes.sub_string !string_buff 0 !string_index in
  string_buff := initial_string_buffer;
  s

(* To store the position of the beginning of a string and comment *)
let string_start_loc = ref Location.none;;
let comment_start_loc = ref [];;
let in_comment () = !comment_start_loc <> [];;
let is_in_string = ref false
let in_string () = !is_in_string
let print_warnings = ref true

(* To translate escape sequences *)

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c   -> c

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in
  if (c < 0 || c > 255) then
    if in_comment ()
    then 'x'
    else raise (Error(Illegal_escape (Lexing.lexeme lexbuf),
                      Location.curr lexbuf))
  else Char.chr c

let char_for_hexadecimal_code lexbuf i =
  let d1 = Char.code (Lexing.lexeme_char lexbuf i) in
  let val1 = if d1 >= 97 then d1 - 87
             else if d1 >= 65 then d1 - 55
             else d1 - 48
  in
  let d2 = Char.code (Lexing.lexeme_char lexbuf (i+1)) in
  let val2 = if d2 >= 97 then d2 - 87
             else if d2 >= 65 then d2 - 55
             else d2 - 48
  in
  Char.chr (val1 * 16 + val2)

(* To convert integer literals, allowing max_int + 1 (PR#4210) *)

let cvt_int_literal s =
  - int_of_string ("-" ^ s)
let cvt_int32_literal s =
  Int32.neg (Int32.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_int64_literal s =
  Int64.neg (Int64.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_nativeint_literal s =
  Nativeint.neg (Nativeint.of_string ("-" ^ String.sub s 0
                                                       (String.length s - 1)))

(* Remove underscores from float literals *)

let remove_underscores s =
  let l = String.length s in
  let b = Bytes.create l in
  let rec remove src dst =
    if src >= l then
      if dst >= l then s else Bytes.sub_string b 0 dst
    else
      match s.[src] with
        '_' -> remove (src + 1) dst
      |  c  -> Bytes.set b dst c; remove (src + 1) (dst + 1)
  in remove 0 0

(* Update the current location with file name and line number. *)

let update_loc lexbuf file line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_file = match file with
                 | None -> pos.pos_fname
                 | Some s -> s
  in
  lexbuf.lex_curr_p <- { pos with
    pos_fname = new_file;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

let preprocessor = ref None

(* Warn about Latin-1 characters used in idents *)

let warn_latin1 lexbuf =
  Location.prerr_warning (Location.curr lexbuf)
    (Warnings.Deprecated "ISO-Latin1 characters in identifiers")
;;

(* Error report *)

open Format

let report_error ppf = function
  | Illegal_character c ->
      fprintf ppf "Illegal character (%s)" (Char.escaped c)
  | Illegal_escape s ->
      fprintf ppf "Illegal backslash escape in string or character (%s)" s
  | Unterminated_comment _ ->
      fprintf ppf "Comment not terminated"
  | Unterminated_string ->
      fprintf ppf "String literal not terminated"
  | Unterminated_string_in_comment (_, loc) ->
      fprintf ppf "This comment contains an unterminated string literal@.\
                   %aString literal begins here"
              Location.print_error loc
  | Keyword_as_label kwd ->
      fprintf ppf "`%s' is a keyword, it cannot be used as label name" kwd
  | Literal_overflow ty ->
      fprintf ppf "Integer literal exceeds the range of representable \
                   integers of type %s" ty

let () =
  Location.register_error_of_exn
    (function
      | Error (err, loc) ->
          Some (Location.error_of_printer loc report_error err)
      | _ ->
          None
    )

# 239 "reason_lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\162\255\163\255\224\000\235\000\018\001\053\001\088\001\
    \192\255\123\001\158\001\199\255\091\000\195\001\068\000\083\000\
    \230\001\219\255\221\255\222\255\224\255\009\002\122\000\044\002\
    \079\002\008\001\109\002\238\255\186\002\219\002\036\003\244\003\
    \196\004\148\005\072\000\115\006\126\006\165\006\227\006\126\000\
    \254\255\001\000\217\007\230\007\011\008\046\008\081\008\116\008\
    \151\008\255\255\005\000\090\000\186\008\221\008\000\009\094\000\
    \035\009\097\000\070\009\098\000\105\009\100\000\101\000\110\000\
    \114\000\006\000\117\000\140\009\175\009\209\255\210\255\083\000\
    \248\255\239\255\240\255\241\255\101\002\111\003\101\000\132\000\
    \121\003\063\004\015\005\116\001\196\002\126\002\197\255\237\255\
    \176\005\222\009\121\000\123\000\007\000\236\255\235\255\231\255\
    \029\005\136\002\124\000\234\255\144\003\126\000\233\255\046\007\
    \127\000\232\255\120\000\229\255\228\255\196\009\101\007\227\255\
    \011\000\178\000\023\001\040\001\025\001\227\255\012\000\247\009\
    \029\010\064\010\122\000\213\255\207\255\099\010\137\010\172\010\
    \207\010\242\010\132\000\125\000\200\255\201\255\202\255\126\000\
    \186\255\188\255\194\255\021\011\191\255\056\011\091\011\126\011\
    \161\011\216\000\163\001\243\255\244\255\013\000\245\255\096\001\
    \204\005\253\255\200\000\217\000\255\255\254\255\252\255\159\011\
    \250\011\225\000\226\000\017\000\251\255\250\255\249\255\028\012\
    \092\004\227\000\248\255\052\005\236\000\247\255\067\012\237\000\
    \246\255\018\000\005\002\247\255\248\255\249\255\035\001\138\012\
    \255\255\250\255\172\012\063\005\253\255\014\001\055\001\056\001\
    \025\006\252\255\211\012\251\255\057\001\074\001\252\255\219\012\
    \254\255\255\255\075\001\091\001\253\255\248\012\248\000\001\001\
    \048\001\049\001\011\001\083\001\017\001\019\000\255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\088\000\085\000\084\000\075\000\070\000\
    \255\255\062\000\059\000\255\255\052\000\049\000\047\000\040\000\
    \078\000\255\255\255\255\255\255\255\255\029\000\039\000\035\000\
    \090\000\032\000\057\000\255\255\012\000\012\000\011\000\010\000\
    \009\000\008\000\043\000\051\000\005\000\004\000\003\000\002\000\
    \255\255\093\000\093\000\091\000\090\000\089\000\086\000\255\255\
    \255\255\255\255\255\255\255\255\082\000\073\000\074\000\255\255\
    \083\000\255\255\089\000\255\255\087\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\038\000\006\000\255\255\255\255\044\000\
    \255\255\255\255\255\255\255\255\255\255\013\000\255\255\255\255\
    \255\255\012\000\012\000\012\000\013\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \024\000\024\000\024\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\025\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\030\000\
    \079\000\037\000\041\000\255\255\255\255\084\000\080\000\050\000\
    \084\000\081\000\066\000\065\000\255\255\255\255\255\255\068\000\
    \255\255\255\255\255\255\060\000\255\255\071\000\072\000\077\000\
    \076\000\255\255\255\255\255\255\255\255\012\000\255\255\012\000\
    \012\000\255\255\012\000\012\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\008\000\
    \255\255\255\255\005\000\005\000\255\255\001\000\005\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\003\000\
    \255\255\255\255\003\000\255\255\255\255\255\255\002\000\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\090\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \255\255\095\000\255\255\255\255\255\255\000\000\000\000\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\000\000\113\000\255\255\000\000\
    \255\255\113\000\114\000\113\000\116\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\147\000\000\000\000\000\255\255\000\000\161\000\
    \255\255\000\000\255\255\255\255\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\
    \000\000\255\255\179\000\000\000\000\000\000\000\255\255\185\000\
    \000\000\000\000\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\255\255\198\000\000\000\255\255\
    \000\000\000\000\255\255\255\255\000\000\255\255\255\255\255\255\
    \208\000\211\000\255\255\211\000\255\255\255\255\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\039\000\040\000\040\000\039\000\041\000\065\000\049\000\
    \040\000\091\000\050\000\065\000\092\000\111\000\117\000\148\000\
    \112\000\118\000\177\000\162\000\148\000\214\000\163\000\177\000\
    \039\000\007\000\027\000\022\000\005\000\003\000\021\000\025\000\
    \019\000\018\000\023\000\006\000\017\000\016\000\015\000\024\000\
    \029\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\034\000\014\000\013\000\035\000\009\000\036\000\
    \004\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\012\000\042\000\011\000\004\000\038\000\
    \020\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\026\000\010\000\008\000\037\000\124\000\
    \130\000\122\000\071\000\110\000\052\000\070\000\069\000\039\000\
    \056\000\052\000\039\000\046\000\060\000\056\000\060\000\058\000\
    \046\000\058\000\072\000\060\000\058\000\083\000\083\000\133\000\
    \044\000\132\000\110\000\131\000\043\000\044\000\039\000\005\000\
    \094\000\043\000\093\000\099\000\005\000\102\000\105\000\107\000\
    \123\000\137\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\111\000\135\000\136\000\112\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\134\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\157\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \002\000\043\000\004\000\156\000\043\000\043\000\043\000\004\000\
    \166\000\165\000\170\000\043\000\004\000\043\000\043\000\004\000\
    \004\000\004\000\091\000\173\000\176\000\092\000\004\000\189\000\
    \004\000\004\000\043\000\207\000\043\000\043\000\043\000\043\000\
    \043\000\111\000\208\000\117\000\112\000\004\000\118\000\004\000\
    \004\000\004\000\004\000\004\000\212\000\181\000\189\000\255\255\
    \196\000\115\000\111\000\005\000\213\000\112\000\005\000\005\000\
    \005\000\116\000\209\000\210\000\064\000\005\000\043\000\005\000\
    \005\000\189\000\189\000\181\000\191\000\191\000\196\000\145\000\
    \115\000\004\000\114\000\000\000\005\000\000\000\005\000\005\000\
    \005\000\005\000\005\000\000\000\201\000\201\000\046\000\202\000\
    \203\000\046\000\046\000\046\000\043\000\210\000\043\000\000\000\
    \046\000\000\000\046\000\144\000\089\000\201\000\000\000\004\000\
    \203\000\004\000\162\000\000\000\000\000\163\000\066\000\046\000\
    \005\000\046\000\143\000\046\000\046\000\046\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\052\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\052\000\164\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\057\000\052\000\046\000\052\000\141\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\083\000\083\000\005\000\000\000\
    \005\000\005\000\000\000\000\000\000\000\148\000\000\000\000\000\
    \149\000\046\000\255\255\046\000\051\000\005\000\052\000\005\000\
    \005\000\005\000\005\000\005\000\160\000\000\000\000\000\005\000\
    \074\000\000\000\005\000\005\000\005\000\153\000\199\000\000\000\
    \000\000\005\000\151\000\005\000\005\000\154\000\000\000\000\000\
    \000\000\000\000\155\000\083\000\052\000\000\000\052\000\066\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\000\000\
    \075\000\000\000\073\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \127\000\125\000\000\000\000\000\000\000\000\000\000\000\005\000\
    \140\000\005\000\066\000\138\000\005\000\005\000\000\000\005\000\
    \005\000\126\000\005\000\005\000\000\000\000\000\000\000\046\000\
    \255\255\000\000\046\000\046\000\046\000\000\000\000\000\181\000\
    \000\000\046\000\182\000\046\000\120\000\000\000\000\000\255\255\
    \000\000\255\255\139\000\000\000\005\000\000\000\152\000\066\000\
    \046\000\005\000\046\000\046\000\121\000\046\000\046\000\184\000\
    \255\255\000\000\005\000\000\000\000\000\005\000\005\000\119\000\
    \255\255\255\255\000\000\000\000\005\000\000\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\057\000\005\000\046\000\005\000\005\000\005\000\
    \005\000\005\000\200\000\000\000\000\000\058\000\000\000\000\000\
    \058\000\058\000\058\000\255\255\000\000\000\000\000\000\058\000\
    \000\000\058\000\058\000\108\000\000\000\000\000\000\000\000\000\
    \255\255\183\000\046\000\000\000\046\000\066\000\058\000\005\000\
    \058\000\058\000\058\000\058\000\058\000\000\000\000\000\000\000\
    \044\000\000\000\000\000\044\000\044\000\044\000\000\000\000\000\
    \000\000\106\000\044\000\000\000\044\000\044\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \059\000\044\000\058\000\044\000\044\000\044\000\044\000\044\000\
    \085\000\000\000\085\000\000\000\000\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\000\000\
    \000\000\000\000\000\000\150\000\000\000\000\000\000\000\000\000\
    \058\000\086\000\058\000\063\000\000\000\044\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \100\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\044\000\088\000\044\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \077\000\087\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\000\000\076\000\
    \000\000\000\000\000\000\000\000\000\000\180\000\074\000\000\000\
    \000\000\077\000\000\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\000\000\000\000\000\000\
    \000\000\028\000\000\000\000\000\000\000\078\000\000\000\076\000\
    \076\000\000\000\000\000\084\000\000\000\000\000\075\000\074\000\
    \073\000\000\000\079\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\080\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\028\000\000\000\000\000\078\000\000\000\000\000\
    \076\000\000\000\000\000\000\000\000\000\000\000\000\000\075\000\
    \000\000\073\000\079\000\030\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\080\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\000\000\
    \000\000\000\000\000\000\030\000\000\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\000\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\000\000\000\000\000\000\000\000\077\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \000\000\000\000\000\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\031\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\000\000\
    \000\000\000\000\000\000\031\000\000\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\000\000\000\000\
    \000\000\000\000\000\000\074\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\081\000\000\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\000\000\000\000\
    \000\000\000\000\000\000\075\000\000\000\073\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\032\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\000\000\
    \000\000\000\000\000\000\032\000\000\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\000\000\
    \000\000\000\000\000\000\074\000\000\000\000\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\082\000\192\000\
    \192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\
    \192\000\000\000\000\000\075\000\000\000\073\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\000\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\033\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\000\000\
    \000\000\000\000\000\000\033\000\000\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\088\000\
    \000\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\159\000\087\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\000\000\
    \158\000\193\000\193\000\193\000\193\000\193\000\193\000\193\000\
    \193\000\193\000\193\000\000\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\000\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\000\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\005\000\056\000\
    \005\000\005\000\056\000\056\000\056\000\000\000\000\000\000\000\
    \000\000\056\000\000\000\056\000\056\000\005\000\000\000\005\000\
    \005\000\067\000\068\000\005\000\000\000\000\000\000\000\000\000\
    \056\000\000\000\056\000\056\000\056\000\056\000\056\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\056\000\000\000\
    \000\000\056\000\056\000\056\000\000\000\000\000\000\000\066\000\
    \056\000\005\000\056\000\056\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\055\000\000\000\056\000\000\000\000\000\056\000\
    \000\000\056\000\056\000\056\000\056\000\056\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\056\000\000\000\056\000\000\000\000\000\000\000\
    \000\000\055\000\000\000\056\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\033\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\000\000\000\000\000\000\
    \000\000\056\000\000\000\056\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\000\000\000\000\
    \000\000\000\000\033\000\000\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\110\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\110\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\000\000\
    \000\000\000\000\000\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\000\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\000\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\049\000\000\000\000\000\050\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\048\000\000\000\000\000\005\000\043\000\005\000\
    \000\000\000\000\000\000\045\000\046\000\000\000\046\000\043\000\
    \044\000\000\000\043\000\043\000\043\000\000\000\000\000\000\000\
    \000\000\043\000\000\000\043\000\043\000\005\000\005\000\005\000\
    \047\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \043\000\000\000\043\000\043\000\043\000\043\000\043\000\000\000\
    \000\000\000\000\000\000\000\000\044\000\000\000\000\000\044\000\
    \044\000\044\000\000\000\000\000\000\000\000\000\044\000\004\000\
    \044\000\044\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\064\000\000\000\043\000\044\000\000\000\044\000\
    \044\000\044\000\044\000\044\000\000\000\000\000\000\000\058\000\
    \000\000\000\000\058\000\058\000\058\000\005\000\000\000\047\000\
    \000\000\058\000\000\000\058\000\058\000\000\000\000\000\000\000\
    \000\000\000\000\043\000\000\000\043\000\000\000\000\000\063\000\
    \058\000\044\000\058\000\058\000\058\000\058\000\058\000\000\000\
    \000\000\000\000\046\000\000\000\000\000\046\000\046\000\046\000\
    \000\000\000\000\000\000\000\000\046\000\000\000\046\000\046\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\044\000\
    \000\000\044\000\059\000\046\000\058\000\046\000\046\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\056\000\000\000\000\000\
    \056\000\056\000\056\000\000\000\000\000\000\000\000\000\056\000\
    \000\000\056\000\056\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\058\000\000\000\058\000\057\000\056\000\046\000\
    \056\000\056\000\056\000\056\000\056\000\000\000\000\000\000\000\
    \052\000\000\000\000\000\052\000\052\000\052\000\000\000\000\000\
    \000\000\000\000\052\000\000\000\052\000\052\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\046\000\000\000\046\000\
    \055\000\052\000\056\000\052\000\053\000\052\000\052\000\052\000\
    \000\000\000\000\000\000\052\000\000\000\000\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\052\000\000\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \056\000\000\000\056\000\051\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\000\000\000\000\000\000\052\000\000\000\
    \000\000\052\000\052\000\052\000\000\000\000\000\000\000\000\000\
    \052\000\000\000\052\000\052\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\051\000\052\000\
    \052\000\052\000\054\000\052\000\052\000\052\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\052\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\052\000\000\000\
    \052\000\051\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\056\000\000\000\000\000\056\000\
    \056\000\056\000\000\000\000\000\000\000\000\000\056\000\000\000\
    \056\000\056\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\052\000\000\000\052\000\051\000\056\000\052\000\056\000\
    \056\000\056\000\056\000\056\000\000\000\000\000\000\000\058\000\
    \000\000\000\000\058\000\058\000\058\000\000\000\000\000\000\000\
    \000\000\058\000\000\000\058\000\058\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\052\000\000\000\052\000\055\000\
    \058\000\056\000\058\000\058\000\058\000\058\000\058\000\000\000\
    \000\000\000\000\060\000\000\000\000\000\060\000\060\000\060\000\
    \000\000\000\000\000\000\000\000\060\000\000\000\060\000\060\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\056\000\
    \000\000\056\000\062\000\060\000\058\000\060\000\060\000\060\000\
    \060\000\060\000\000\000\000\000\000\000\005\000\000\000\000\000\
    \005\000\005\000\005\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\005\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\058\000\000\000\058\000\061\000\005\000\060\000\
    \005\000\005\000\005\000\005\000\005\000\115\000\111\000\000\000\
    \005\000\112\000\000\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\115\000\060\000\114\000\060\000\
    \066\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\098\000\000\000\
    \098\000\000\000\000\000\000\000\000\000\098\000\000\000\000\000\
    \005\000\000\000\005\000\066\000\000\000\005\000\097\000\097\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \005\000\000\000\000\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\000\000\000\000\
    \000\000\005\000\000\000\005\000\005\000\005\000\005\000\005\000\
    \000\000\000\000\098\000\000\000\000\000\000\000\046\000\000\000\
    \098\000\046\000\046\000\046\000\000\000\000\000\000\000\000\000\
    \046\000\000\000\046\000\046\000\098\000\000\000\000\000\000\000\
    \098\000\000\000\098\000\066\000\000\000\005\000\096\000\046\000\
    \000\000\046\000\046\000\046\000\046\000\046\000\000\000\000\000\
    \000\000\046\000\000\000\000\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\000\000\046\000\000\000\046\000\046\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\000\000\000\000\
    \000\000\057\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\046\000\000\000\046\000\057\000\005\000\046\000\005\000\
    \005\000\005\000\005\000\005\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\000\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\005\000\000\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\046\000\000\000\046\000\066\000\
    \000\000\005\000\000\000\005\000\255\255\005\000\005\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\005\000\000\000\000\000\
    \005\000\005\000\005\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\005\000\000\000\000\000\000\000\255\255\005\000\
    \000\000\005\000\000\000\000\000\000\000\066\000\005\000\005\000\
    \005\000\005\000\005\000\005\000\005\000\000\000\000\000\000\000\
    \005\000\000\000\000\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \066\000\005\000\005\000\005\000\005\000\129\000\005\000\005\000\
    \000\000\000\000\000\000\005\000\000\000\000\000\005\000\005\000\
    \005\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \005\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \005\000\000\000\005\000\066\000\005\000\005\000\005\000\005\000\
    \005\000\005\000\005\000\000\000\000\000\000\000\005\000\000\000\
    \000\000\005\000\005\000\005\000\000\000\000\000\000\000\000\000\
    \005\000\000\000\005\000\005\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\066\000\005\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\052\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\066\000\052\000\005\000\052\000\142\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\052\000\000\000\000\000\052\000\
    \052\000\052\000\000\000\000\000\000\000\000\000\052\000\000\000\
    \052\000\052\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\005\000\000\000\005\000\051\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\000\000\000\000\000\000\046\000\
    \000\000\000\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \000\000\046\000\000\000\046\000\046\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\052\000\000\000\052\000\051\000\
    \046\000\052\000\046\000\046\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\046\000\000\000\000\000\046\000\046\000\046\000\
    \000\000\000\000\000\000\000\000\046\000\000\000\046\000\046\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\052\000\
    \000\000\052\000\057\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\046\000\000\000\046\000\057\000\159\000\046\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\169\000\158\000\169\000\046\000\000\000\046\000\
    \000\000\169\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\169\000\000\000\
    \000\000\000\000\000\000\000\000\169\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\000\000\000\000\000\000\000\000\000\000\
    \169\000\000\000\000\000\000\000\169\000\000\000\169\000\000\000\
    \000\000\000\000\167\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\175\000\175\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\000\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\189\000\000\000\000\000\190\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\188\000\000\000\188\000\000\000\000\000\000\000\
    \000\000\188\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\187\000\187\000\187\000\187\000\187\000\187\000\
    \187\000\187\000\187\000\187\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\194\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\194\000\194\000\194\000\188\000\000\000\
    \000\000\000\000\000\000\000\000\188\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\000\000\000\000\000\000\000\000\000\000\
    \188\000\000\000\000\000\000\000\188\000\000\000\188\000\000\000\
    \000\000\000\000\186\000\195\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\195\000\195\000\195\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\000\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\205\000\000\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\000\000\205\000\
    \204\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\000\000\000\000\204\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\041\000\000\000\000\000\041\000\050\000\
    \065\000\092\000\050\000\065\000\092\000\112\000\118\000\149\000\
    \112\000\118\000\149\000\163\000\177\000\213\000\163\000\177\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\014\000\
    \012\000\015\000\034\000\022\000\051\000\034\000\034\000\039\000\
    \055\000\051\000\039\000\057\000\059\000\055\000\061\000\062\000\
    \057\000\059\000\071\000\061\000\062\000\078\000\078\000\012\000\
    \063\000\012\000\022\000\012\000\064\000\063\000\039\000\066\000\
    \090\000\064\000\091\000\098\000\066\000\101\000\104\000\106\000\
    \122\000\130\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\113\000\131\000\135\000\113\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\154\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\145\000\155\000\003\000\003\000\003\000\145\000\
    \161\000\162\000\169\000\003\000\004\000\003\000\003\000\004\000\
    \004\000\004\000\025\000\172\000\175\000\025\000\004\000\189\000\
    \004\000\004\000\003\000\206\000\003\000\003\000\003\000\003\000\
    \003\000\114\000\207\000\116\000\114\000\004\000\116\000\004\000\
    \004\000\004\000\004\000\004\000\210\000\182\000\189\000\025\000\
    \182\000\115\000\115\000\005\000\212\000\115\000\005\000\005\000\
    \005\000\114\000\208\000\209\000\003\000\005\000\003\000\005\000\
    \005\000\190\000\191\000\196\000\190\000\191\000\196\000\004\000\
    \115\000\004\000\115\000\255\255\005\000\255\255\005\000\005\000\
    \005\000\005\000\005\000\255\255\197\000\202\000\006\000\197\000\
    \202\000\006\000\006\000\006\000\003\000\211\000\003\000\255\255\
    \006\000\255\255\006\000\006\000\025\000\203\000\255\255\004\000\
    \203\000\004\000\151\000\255\255\255\255\151\000\005\000\006\000\
    \005\000\006\000\006\000\006\000\006\000\006\000\255\255\255\255\
    \255\255\007\000\255\255\255\255\007\000\007\000\007\000\255\255\
    \255\255\255\255\255\255\007\000\255\255\007\000\007\000\151\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000\255\255\
    \005\000\006\000\007\000\006\000\007\000\007\000\007\000\007\000\
    \007\000\255\255\255\255\255\255\009\000\255\255\255\255\009\000\
    \009\000\009\000\255\255\255\255\083\000\083\000\009\000\255\255\
    \009\000\009\000\255\255\255\255\255\255\146\000\255\255\255\255\
    \146\000\006\000\113\000\006\000\007\000\009\000\007\000\009\000\
    \009\000\009\000\009\000\009\000\151\000\255\255\255\255\010\000\
    \083\000\255\255\010\000\010\000\010\000\146\000\197\000\255\255\
    \255\255\010\000\146\000\010\000\010\000\146\000\255\255\255\255\
    \255\255\255\255\146\000\083\000\007\000\255\255\007\000\009\000\
    \010\000\009\000\010\000\010\000\010\000\010\000\010\000\255\255\
    \083\000\255\255\083\000\255\255\013\000\255\255\255\255\013\000\
    \013\000\013\000\255\255\255\255\255\255\255\255\013\000\255\255\
    \013\000\013\000\255\255\255\255\255\255\255\255\255\255\009\000\
    \009\000\009\000\010\000\010\000\010\000\013\000\255\255\013\000\
    \013\000\013\000\013\000\013\000\255\255\255\255\255\255\016\000\
    \025\000\255\255\016\000\016\000\016\000\255\255\255\255\178\000\
    \255\255\016\000\178\000\016\000\016\000\255\255\255\255\114\000\
    \255\255\116\000\010\000\255\255\010\000\255\255\146\000\013\000\
    \016\000\013\000\016\000\016\000\016\000\016\000\016\000\178\000\
    \115\000\255\255\021\000\255\255\255\255\021\000\021\000\021\000\
    \208\000\209\000\255\255\255\255\021\000\255\255\021\000\021\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\013\000\
    \255\255\013\000\016\000\021\000\016\000\021\000\021\000\021\000\
    \021\000\021\000\197\000\255\255\255\255\023\000\255\255\255\255\
    \023\000\023\000\023\000\211\000\255\255\255\255\255\255\023\000\
    \255\255\023\000\023\000\023\000\255\255\255\255\255\255\255\255\
    \151\000\178\000\016\000\255\255\016\000\021\000\023\000\021\000\
    \023\000\023\000\023\000\023\000\023\000\255\255\255\255\255\255\
    \024\000\255\255\255\255\024\000\024\000\024\000\255\255\255\255\
    \255\255\024\000\024\000\255\255\024\000\024\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\021\000\255\255\021\000\
    \023\000\024\000\023\000\024\000\024\000\024\000\024\000\024\000\
    \076\000\255\255\076\000\255\255\255\255\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\255\255\
    \255\255\255\255\255\255\146\000\255\255\255\255\255\255\255\255\
    \023\000\026\000\023\000\024\000\255\255\024\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \097\000\097\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\024\000\026\000\024\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \028\000\026\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\255\255\028\000\
    \255\255\255\255\255\255\255\255\255\255\178\000\028\000\255\255\
    \255\255\029\000\255\255\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\255\255\255\255\255\255\
    \255\255\028\000\255\255\255\255\255\255\029\000\255\255\028\000\
    \029\000\255\255\255\255\084\000\255\255\255\255\028\000\029\000\
    \028\000\255\255\029\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\029\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\029\000\255\255\255\255\029\000\255\255\255\255\
    \029\000\255\255\255\255\255\255\255\255\255\255\255\255\029\000\
    \255\255\029\000\029\000\030\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\029\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\255\255\
    \255\255\255\255\255\255\030\000\255\255\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\255\255\077\000\255\255\255\255\255\255\
    \255\255\255\255\080\000\080\000\080\000\080\000\080\000\080\000\
    \100\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\255\255\255\255\255\255\255\255\077\000\255\255\
    \255\255\255\255\255\255\255\255\077\000\255\255\255\255\255\255\
    \255\255\255\255\080\000\080\000\080\000\080\000\080\000\080\000\
    \255\255\255\255\255\255\255\255\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\255\255\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\031\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\255\255\
    \255\255\255\255\255\255\031\000\255\255\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \081\000\081\000\081\000\081\000\081\000\081\000\255\255\255\255\
    \255\255\255\255\255\255\081\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\081\000\255\255\
    \081\000\081\000\081\000\081\000\081\000\081\000\255\255\255\255\
    \255\255\255\255\255\255\081\000\255\255\081\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\255\255\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\032\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\255\255\
    \255\255\255\255\255\255\032\000\255\255\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\255\255\
    \255\255\255\255\255\255\082\000\255\255\255\255\096\000\096\000\
    \096\000\096\000\096\000\096\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\082\000\187\000\
    \187\000\187\000\187\000\187\000\187\000\187\000\187\000\187\000\
    \187\000\255\255\255\255\082\000\255\255\082\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\255\255\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\033\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\255\255\
    \255\255\255\255\255\255\033\000\255\255\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\088\000\
    \255\255\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\152\000\088\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\255\255\
    \152\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\
    \192\000\192\000\192\000\255\255\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\255\255\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\255\255\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\035\000\255\255\255\255\035\000\
    \035\000\035\000\255\255\255\255\255\255\255\255\035\000\036\000\
    \035\000\035\000\036\000\036\000\036\000\255\255\255\255\255\255\
    \255\255\036\000\255\255\036\000\036\000\035\000\255\255\035\000\
    \035\000\035\000\035\000\035\000\255\255\255\255\255\255\255\255\
    \036\000\255\255\036\000\036\000\036\000\036\000\036\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\037\000\255\255\
    \255\255\037\000\037\000\037\000\255\255\255\255\255\255\035\000\
    \037\000\035\000\037\000\037\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\255\255\036\000\255\255\255\255\037\000\
    \255\255\037\000\037\000\037\000\037\000\037\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\035\000\
    \255\255\035\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\255\255\036\000\255\255\255\255\255\255\
    \255\255\037\000\255\255\037\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\038\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\255\255\255\255\255\255\
    \255\255\037\000\255\255\037\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\255\255\255\255\
    \255\255\255\255\038\000\255\255\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\110\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\110\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\103\000\
    \103\000\103\000\103\000\103\000\103\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\110\000\255\255\
    \255\255\255\255\255\255\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\255\255\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\255\255\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\042\000\255\255\255\255\042\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\042\000\255\255\255\255\042\000\042\000\042\000\
    \255\255\255\255\255\255\042\000\042\000\255\255\042\000\043\000\
    \042\000\255\255\043\000\043\000\043\000\255\255\255\255\255\255\
    \255\255\043\000\255\255\043\000\043\000\042\000\042\000\042\000\
    \042\000\042\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \043\000\255\255\043\000\043\000\043\000\043\000\043\000\255\255\
    \255\255\255\255\255\255\255\255\044\000\255\255\255\255\044\000\
    \044\000\044\000\255\255\255\255\255\255\255\255\044\000\042\000\
    \044\000\044\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\043\000\255\255\043\000\044\000\255\255\044\000\
    \044\000\044\000\044\000\044\000\255\255\255\255\255\255\045\000\
    \255\255\255\255\045\000\045\000\045\000\042\000\255\255\042\000\
    \255\255\045\000\255\255\045\000\045\000\255\255\255\255\255\255\
    \255\255\255\255\043\000\255\255\043\000\255\255\255\255\044\000\
    \045\000\044\000\045\000\045\000\045\000\045\000\045\000\255\255\
    \255\255\255\255\046\000\255\255\255\255\046\000\046\000\046\000\
    \255\255\255\255\255\255\255\255\046\000\255\255\046\000\046\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\044\000\
    \255\255\044\000\045\000\046\000\045\000\046\000\046\000\046\000\
    \046\000\046\000\255\255\255\255\255\255\047\000\255\255\255\255\
    \047\000\047\000\047\000\255\255\255\255\255\255\255\255\047\000\
    \255\255\047\000\047\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\045\000\255\255\045\000\046\000\047\000\046\000\
    \047\000\047\000\047\000\047\000\047\000\255\255\255\255\255\255\
    \048\000\255\255\255\255\048\000\048\000\048\000\255\255\255\255\
    \255\255\255\255\048\000\255\255\048\000\048\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\046\000\255\255\046\000\
    \047\000\048\000\047\000\048\000\048\000\048\000\048\000\048\000\
    \255\255\255\255\255\255\052\000\255\255\255\255\052\000\052\000\
    \052\000\255\255\255\255\255\255\255\255\052\000\255\255\052\000\
    \052\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \047\000\255\255\047\000\048\000\052\000\048\000\052\000\052\000\
    \052\000\052\000\052\000\255\255\255\255\255\255\053\000\255\255\
    \255\255\053\000\053\000\053\000\255\255\255\255\255\255\255\255\
    \053\000\255\255\053\000\053\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\048\000\255\255\048\000\052\000\053\000\
    \052\000\053\000\053\000\053\000\053\000\053\000\255\255\255\255\
    \255\255\054\000\255\255\255\255\054\000\054\000\054\000\255\255\
    \255\255\255\255\255\255\054\000\255\255\054\000\054\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\052\000\255\255\
    \052\000\053\000\054\000\053\000\054\000\054\000\054\000\054\000\
    \054\000\255\255\255\255\255\255\056\000\255\255\255\255\056\000\
    \056\000\056\000\255\255\255\255\255\255\255\255\056\000\255\255\
    \056\000\056\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\053\000\255\255\053\000\054\000\056\000\054\000\056\000\
    \056\000\056\000\056\000\056\000\255\255\255\255\255\255\058\000\
    \255\255\255\255\058\000\058\000\058\000\255\255\255\255\255\255\
    \255\255\058\000\255\255\058\000\058\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\054\000\255\255\054\000\056\000\
    \058\000\056\000\058\000\058\000\058\000\058\000\058\000\255\255\
    \255\255\255\255\060\000\255\255\255\255\060\000\060\000\060\000\
    \255\255\255\255\255\255\255\255\060\000\255\255\060\000\060\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\056\000\
    \255\255\056\000\058\000\060\000\058\000\060\000\060\000\060\000\
    \060\000\060\000\255\255\255\255\255\255\067\000\255\255\255\255\
    \067\000\067\000\067\000\255\255\255\255\255\255\255\255\067\000\
    \255\255\067\000\067\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\058\000\255\255\058\000\060\000\067\000\060\000\
    \067\000\067\000\067\000\067\000\067\000\109\000\109\000\255\255\
    \068\000\109\000\255\255\068\000\068\000\068\000\255\255\255\255\
    \255\255\255\255\068\000\255\255\068\000\068\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\109\000\060\000\109\000\060\000\
    \067\000\068\000\067\000\068\000\068\000\068\000\068\000\068\000\
    \255\255\255\255\255\255\255\255\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\089\000\255\255\
    \089\000\255\255\255\255\255\255\255\255\089\000\255\255\255\255\
    \067\000\255\255\067\000\068\000\255\255\068\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \119\000\255\255\255\255\119\000\119\000\119\000\255\255\255\255\
    \255\255\255\255\119\000\255\255\119\000\119\000\255\255\255\255\
    \255\255\255\255\255\255\068\000\255\255\068\000\255\255\255\255\
    \255\255\119\000\255\255\119\000\119\000\119\000\119\000\119\000\
    \255\255\255\255\089\000\255\255\255\255\255\255\120\000\255\255\
    \089\000\120\000\120\000\120\000\255\255\255\255\255\255\255\255\
    \120\000\255\255\120\000\120\000\089\000\255\255\255\255\255\255\
    \089\000\255\255\089\000\119\000\255\255\119\000\089\000\120\000\
    \255\255\120\000\120\000\120\000\120\000\120\000\255\255\255\255\
    \255\255\121\000\255\255\255\255\121\000\121\000\121\000\255\255\
    \255\255\255\255\255\255\121\000\255\255\121\000\121\000\255\255\
    \255\255\255\255\255\255\119\000\255\255\119\000\255\255\255\255\
    \255\255\120\000\121\000\120\000\121\000\121\000\121\000\121\000\
    \121\000\255\255\255\255\255\255\125\000\255\255\255\255\125\000\
    \125\000\125\000\255\255\255\255\255\255\255\255\125\000\255\255\
    \125\000\125\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\120\000\255\255\120\000\121\000\125\000\121\000\125\000\
    \125\000\125\000\125\000\125\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\126\000\255\255\255\255\126\000\126\000\126\000\
    \255\255\255\255\255\255\255\255\126\000\255\255\126\000\126\000\
    \255\255\255\255\255\255\255\255\121\000\255\255\121\000\125\000\
    \255\255\125\000\255\255\126\000\109\000\126\000\126\000\126\000\
    \126\000\126\000\255\255\255\255\255\255\127\000\255\255\255\255\
    \127\000\127\000\127\000\255\255\255\255\255\255\255\255\127\000\
    \255\255\127\000\127\000\255\255\255\255\255\255\089\000\125\000\
    \255\255\125\000\255\255\255\255\255\255\126\000\127\000\126\000\
    \127\000\127\000\127\000\127\000\127\000\255\255\255\255\255\255\
    \128\000\255\255\255\255\128\000\128\000\128\000\255\255\255\255\
    \255\255\255\255\128\000\255\255\128\000\128\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\126\000\255\255\126\000\
    \127\000\128\000\127\000\128\000\128\000\128\000\128\000\128\000\
    \255\255\255\255\255\255\129\000\255\255\255\255\129\000\129\000\
    \129\000\255\255\255\255\255\255\255\255\129\000\255\255\129\000\
    \129\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \127\000\255\255\127\000\128\000\129\000\128\000\129\000\129\000\
    \129\000\129\000\129\000\255\255\255\255\255\255\139\000\255\255\
    \255\255\139\000\139\000\139\000\255\255\255\255\255\255\255\255\
    \139\000\255\255\139\000\139\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\128\000\255\255\128\000\129\000\139\000\
    \129\000\139\000\139\000\139\000\139\000\139\000\255\255\255\255\
    \255\255\141\000\255\255\255\255\141\000\141\000\141\000\255\255\
    \255\255\255\255\255\255\141\000\255\255\141\000\141\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\129\000\255\255\
    \129\000\139\000\141\000\139\000\141\000\141\000\141\000\141\000\
    \141\000\255\255\255\255\255\255\142\000\255\255\255\255\142\000\
    \142\000\142\000\255\255\255\255\255\255\255\255\142\000\255\255\
    \142\000\142\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\139\000\255\255\139\000\141\000\142\000\141\000\142\000\
    \142\000\142\000\142\000\142\000\255\255\255\255\255\255\143\000\
    \255\255\255\255\143\000\143\000\143\000\255\255\255\255\255\255\
    \255\255\143\000\255\255\143\000\143\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\141\000\255\255\141\000\142\000\
    \143\000\142\000\143\000\143\000\143\000\143\000\143\000\255\255\
    \255\255\255\255\144\000\255\255\255\255\144\000\144\000\144\000\
    \255\255\255\255\255\255\255\255\144\000\255\255\144\000\144\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\142\000\
    \255\255\142\000\143\000\144\000\143\000\144\000\144\000\144\000\
    \144\000\144\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\143\000\255\255\143\000\144\000\159\000\144\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\160\000\159\000\160\000\144\000\255\255\144\000\
    \255\255\160\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\160\000\160\000\160\000\160\000\160\000\160\000\
    \160\000\160\000\160\000\160\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\160\000\255\255\
    \255\255\255\255\255\255\255\255\160\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\255\255\255\255\255\255\255\255\255\255\
    \160\000\255\255\255\255\255\255\160\000\255\255\160\000\255\255\
    \255\255\255\255\160\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\255\255\174\000\174\000\174\000\174\000\
    \174\000\174\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\183\000\255\255\255\255\183\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\174\000\174\000\174\000\174\000\
    \174\000\174\000\183\000\255\255\183\000\255\255\255\255\255\255\
    \255\255\183\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\183\000\183\000\183\000\183\000\183\000\183\000\
    \183\000\183\000\183\000\183\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\183\000\255\255\
    \255\255\255\255\255\255\255\255\183\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\255\255\255\255\255\255\255\255\255\255\
    \183\000\255\255\255\255\255\255\183\000\255\255\183\000\255\255\
    \255\255\255\255\183\000\194\000\194\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\194\000\194\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\255\255\194\000\194\000\194\000\194\000\
    \194\000\194\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\194\000\194\000\194\000\194\000\
    \194\000\194\000\199\000\255\255\199\000\199\000\199\000\199\000\
    \199\000\199\000\199\000\199\000\199\000\199\000\199\000\199\000\
    \199\000\199\000\199\000\199\000\199\000\199\000\199\000\199\000\
    \199\000\199\000\199\000\199\000\199\000\199\000\255\255\205\000\
    \199\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\205\000\205\000\205\000\205\000\205\000\
    \205\000\205\000\205\000\255\255\255\255\205\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\183\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\010\000\036\000\012\000\
    \000\000\000\000\000\000\002\000\000\000\027\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\002\000\004\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\039\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\019\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\036\000\036\000\000\000\036\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\000\000\001\000\022\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\007\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\001\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\022\000\114\000\183\000\190\000\114\000\191\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \022\000\255\255\114\000\000\000\115\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\109\000\110\000\255\255\255\255\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\110\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\110\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \114\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\004\255\255\005\255\255\007\255\006\255\255\003\255\000\004\
    \001\005\255\007\255\255\006\255\007\255\255\000\004\001\005\003\
    \006\002\007\255\001\255\255\000\001\255";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 8 (-1) ;   __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 324 "reason_lexer.mll"
                 (
      match !preprocessor with
      | None ->
        raise (Error(Illegal_character (Lexing.lexeme_char lexbuf 0),
                     Location.curr lexbuf))
      | Some _ ->
        update_loc lexbuf None 1 false 0;
        token lexbuf )
# 1403 "reason_lexer.ml"

  | 1 ->
# 333 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        match !preprocessor with
        | None -> token lexbuf
        | Some _ -> EOL
      )
# 1412 "reason_lexer.ml"

  | 2 ->
# 339 "reason_lexer.mll"
      ( token lexbuf )
# 1417 "reason_lexer.ml"

  | 3 ->
# 341 "reason_lexer.mll"
      ( UNDERSCORE )
# 1422 "reason_lexer.ml"

  | 4 ->
# 343 "reason_lexer.mll"
      ( TILDE )
# 1427 "reason_lexer.ml"

  | 5 ->
# 345 "reason_lexer.mll"
      ( QUESTION )
# 1432 "reason_lexer.ml"

  | 6 ->
# 348 "reason_lexer.mll"
      ( OPTIONAL_NO_DEFAULT )
# 1437 "reason_lexer.ml"

  | 7 ->
# 350 "reason_lexer.mll"
      ( EXPLICITLY_PASSED_OPTIONAL )
# 1442 "reason_lexer.ml"

  | 8 ->
# 352 "reason_lexer.mll"
      ( let s = Lexing.lexeme lexbuf in
        try Hashtbl.find keyword_table s
        with Not_found -> LIDENT s )
# 1449 "reason_lexer.ml"

  | 9 ->
# 356 "reason_lexer.mll"
      ( warn_latin1 lexbuf; LIDENT (Lexing.lexeme lexbuf) )
# 1454 "reason_lexer.ml"

  | 10 ->
# 358 "reason_lexer.mll"
      ( UIDENT(Lexing.lexeme lexbuf) )
# 1459 "reason_lexer.ml"

  | 11 ->
# 360 "reason_lexer.mll"
      ( warn_latin1 lexbuf; UIDENT(Lexing.lexeme lexbuf) )
# 1464 "reason_lexer.ml"

  | 12 ->
# 362 "reason_lexer.mll"
      ( try
          INT (cvt_int_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int", Location.curr lexbuf))
      )
# 1473 "reason_lexer.ml"

  | 13 ->
# 368 "reason_lexer.mll"
      ( FLOAT (remove_underscores(Lexing.lexeme lexbuf)) )
# 1478 "reason_lexer.ml"

  | 14 ->
# 370 "reason_lexer.mll"
      ( try
          INT32 (cvt_int32_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int32", Location.curr lexbuf)) )
# 1486 "reason_lexer.ml"

  | 15 ->
# 375 "reason_lexer.mll"
      ( try
          INT64 (cvt_int64_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int64", Location.curr lexbuf)) )
# 1494 "reason_lexer.ml"

  | 16 ->
# 380 "reason_lexer.mll"
      ( try
          NATIVEINT (cvt_nativeint_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "nativeint", Location.curr lexbuf)) )
# 1502 "reason_lexer.ml"

  | 17 ->
# 385 "reason_lexer.mll"
      ( reset_string_buffer();
        is_in_string := true;
        let string_start = lexbuf.lex_start_p in
        string_start_loc := Location.curr lexbuf;
        string lexbuf;
        is_in_string := false;
        lexbuf.lex_start_p <- string_start;
        STRING (get_stored_string(), None) )
# 1514 "reason_lexer.ml"

  | 18 ->
# 394 "reason_lexer.mll"
      ( reset_string_buffer();
        let delim = Lexing.lexeme lexbuf in
        let delim = String.sub delim 1 (String.length delim - 2) in
        is_in_string := true;
        let string_start = lexbuf.lex_start_p in
        string_start_loc := Location.curr lexbuf;
        quoted_string delim lexbuf;
        is_in_string := false;
        lexbuf.lex_start_p <- string_start;
        STRING (get_stored_string(), Some delim) )
# 1528 "reason_lexer.ml"

  | 19 ->
# 405 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 1;
        CHAR (Lexing.lexeme_char lexbuf 1) )
# 1534 "reason_lexer.ml"

  | 20 ->
# 408 "reason_lexer.mll"
      ( CHAR(Lexing.lexeme_char lexbuf 1) )
# 1539 "reason_lexer.ml"

  | 21 ->
# 410 "reason_lexer.mll"
      ( CHAR(char_for_backslash (Lexing.lexeme_char lexbuf 2)) )
# 1544 "reason_lexer.ml"

  | 22 ->
# 412 "reason_lexer.mll"
      ( CHAR(char_for_decimal_code lexbuf 2) )
# 1549 "reason_lexer.ml"

  | 23 ->
# 414 "reason_lexer.mll"
      ( CHAR(char_for_hexadecimal_code lexbuf 3) )
# 1554 "reason_lexer.ml"

  | 24 ->
# 416 "reason_lexer.mll"
      ( let l = Lexing.lexeme lexbuf in
        let esc = String.sub l 1 (String.length l - 1) in
        raise (Error(Illegal_escape esc, Location.curr lexbuf))
      )
# 1562 "reason_lexer.ml"

  | 25 ->
# 421 "reason_lexer.mll"
      ( let start_loc = Location.curr lexbuf  in
        comment_start_loc := [start_loc];
        reset_string_buffer ();
        let end_loc = comment lexbuf in
        let s = get_stored_string () in
        reset_string_buffer ();
        COMMENT (s, { start_loc with
                      Location.loc_end = end_loc.Location.loc_end })
      )
# 1575 "reason_lexer.ml"

  | 26 ->
# 431 "reason_lexer.mll"
      ( let loc = Location.curr lexbuf  in
        if !print_warnings then
          Location.prerr_warning loc Warnings.Comment_start;
        comment_start_loc := [loc];
        reset_string_buffer ();
        let end_loc = comment lexbuf in
        let s = get_stored_string () in
        reset_string_buffer ();
        COMMENT (s, { loc with Location.loc_end = end_loc.Location.loc_end })
      )
# 1589 "reason_lexer.ml"

  | 27 ->
# 442 "reason_lexer.mll"
      ( let loc = Location.curr lexbuf in
        Location.prerr_warning loc Warnings.Comment_not_end;
        lexbuf.Lexing.lex_curr_pos <- lexbuf.Lexing.lex_curr_pos - 1;
        let curpos = lexbuf.lex_curr_p in
        lexbuf.lex_curr_p <- { curpos with pos_cnum = curpos.pos_cnum - 1 };
        STAR
      )
# 1600 "reason_lexer.ml"

  | 28 ->
let
# 449 "reason_lexer.mll"
                                   num
# 1606 "reason_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 450 "reason_lexer.mll"
                                           name
# 1611 "reason_lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(2) in
# 452 "reason_lexer.mll"
      ( update_loc lexbuf name (int_of_string num) true 0;
        token lexbuf
      )
# 1617 "reason_lexer.ml"

  | 29 ->
# 455 "reason_lexer.mll"
         ( AMPERSAND )
# 1622 "reason_lexer.ml"

  | 30 ->
# 456 "reason_lexer.mll"
         ( AMPERAMPER )
# 1627 "reason_lexer.ml"

  | 31 ->
# 457 "reason_lexer.mll"
         ( BACKQUOTE )
# 1632 "reason_lexer.ml"

  | 32 ->
# 458 "reason_lexer.mll"
         ( QUOTE )
# 1637 "reason_lexer.ml"

  | 33 ->
# 459 "reason_lexer.mll"
         ( LPAREN )
# 1642 "reason_lexer.ml"

  | 34 ->
# 460 "reason_lexer.mll"
         ( RPAREN )
# 1647 "reason_lexer.ml"

  | 35 ->
# 461 "reason_lexer.mll"
         ( STAR )
# 1652 "reason_lexer.ml"

  | 36 ->
# 462 "reason_lexer.mll"
         ( COMMA )
# 1657 "reason_lexer.ml"

  | 37 ->
# 463 "reason_lexer.mll"
         ( MINUSGREATER )
# 1662 "reason_lexer.ml"

  | 38 ->
# 464 "reason_lexer.mll"
         ( EQUALGREATER )
# 1667 "reason_lexer.ml"

  | 39 ->
# 465 "reason_lexer.mll"
         ( SHARP )
# 1672 "reason_lexer.ml"

  | 40 ->
# 466 "reason_lexer.mll"
         ( DOT )
# 1677 "reason_lexer.ml"

  | 41 ->
# 467 "reason_lexer.mll"
         ( DOTDOT )
# 1682 "reason_lexer.ml"

  | 42 ->
# 468 "reason_lexer.mll"
         ( DOTDOTDOT )
# 1687 "reason_lexer.ml"

  | 43 ->
# 469 "reason_lexer.mll"
         ( COLON )
# 1692 "reason_lexer.ml"

  | 44 ->
# 470 "reason_lexer.mll"
         ( COLONCOLON )
# 1697 "reason_lexer.ml"

  | 45 ->
# 471 "reason_lexer.mll"
         ( COLONEQUAL )
# 1702 "reason_lexer.ml"

  | 46 ->
# 472 "reason_lexer.mll"
         ( COLONGREATER )
# 1707 "reason_lexer.ml"

  | 47 ->
# 473 "reason_lexer.mll"
         ( SEMI )
# 1712 "reason_lexer.ml"

  | 48 ->
# 474 "reason_lexer.mll"
         ( SEMISEMI )
# 1717 "reason_lexer.ml"

  | 49 ->
# 475 "reason_lexer.mll"
         ( LESS )
# 1722 "reason_lexer.ml"

  | 50 ->
# 476 "reason_lexer.mll"
         ( LESSMINUS )
# 1727 "reason_lexer.ml"

  | 51 ->
# 477 "reason_lexer.mll"
         ( EQUAL )
# 1732 "reason_lexer.ml"

  | 52 ->
# 478 "reason_lexer.mll"
         ( LBRACKET )
# 1737 "reason_lexer.ml"

  | 53 ->
# 479 "reason_lexer.mll"
         ( LBRACKETBAR )
# 1742 "reason_lexer.ml"

  | 54 ->
# 480 "reason_lexer.mll"
         ( LBRACKETLESS )
# 1747 "reason_lexer.ml"

  | 55 ->
# 481 "reason_lexer.mll"
         ( LBRACKETGREATER )
# 1752 "reason_lexer.ml"

  | 56 ->
# 482 "reason_lexer.mll"
         ( RBRACKET )
# 1757 "reason_lexer.ml"

  | 57 ->
# 483 "reason_lexer.mll"
         ( LBRACE )
# 1762 "reason_lexer.ml"

  | 58 ->
# 484 "reason_lexer.mll"
         ( LBRACELESS )
# 1767 "reason_lexer.ml"

  | 59 ->
# 485 "reason_lexer.mll"
         ( BAR )
# 1772 "reason_lexer.ml"

  | 60 ->
# 486 "reason_lexer.mll"
         ( BARBAR )
# 1777 "reason_lexer.ml"

  | 61 ->
# 487 "reason_lexer.mll"
         ( BARRBRACKET )
# 1782 "reason_lexer.ml"

  | 62 ->
# 488 "reason_lexer.mll"
         ( GREATER )
# 1787 "reason_lexer.ml"

  | 63 ->
# 494 "reason_lexer.mll"
         ( RBRACE )
# 1792 "reason_lexer.ml"

  | 64 ->
# 495 "reason_lexer.mll"
         ( GREATERRBRACE )
# 1797 "reason_lexer.ml"

  | 65 ->
# 496 "reason_lexer.mll"
         ( LBRACKETAT )
# 1802 "reason_lexer.ml"

  | 66 ->
# 497 "reason_lexer.mll"
         ( LBRACKETPERCENT )
# 1807 "reason_lexer.ml"

  | 67 ->
# 498 "reason_lexer.mll"
          ( LBRACKETPERCENTPERCENT )
# 1812 "reason_lexer.ml"

  | 68 ->
# 499 "reason_lexer.mll"
          ( LBRACKETATAT )
# 1817 "reason_lexer.ml"

  | 69 ->
# 500 "reason_lexer.mll"
           ( LBRACKETATATAT )
# 1822 "reason_lexer.ml"

  | 70 ->
# 501 "reason_lexer.mll"
         ( BANG )
# 1827 "reason_lexer.ml"

  | 71 ->
# 502 "reason_lexer.mll"
         ( INFIXOP0 "!=" )
# 1832 "reason_lexer.ml"

  | 72 ->
# 503 "reason_lexer.mll"
          ( INFIXOP0 "!==" )
# 1837 "reason_lexer.ml"

  | 73 ->
# 504 "reason_lexer.mll"
           ( INFIXOP0 "!=" )
# 1842 "reason_lexer.ml"

  | 74 ->
# 505 "reason_lexer.mll"
            ( INFIXOP0 "!==" )
# 1847 "reason_lexer.ml"

  | 75 ->
# 506 "reason_lexer.mll"
         ( PLUS )
# 1852 "reason_lexer.ml"

  | 76 ->
# 507 "reason_lexer.mll"
         ( PLUSDOT )
# 1857 "reason_lexer.ml"

  | 77 ->
# 508 "reason_lexer.mll"
         ( PLUSEQ )
# 1862 "reason_lexer.ml"

  | 78 ->
# 509 "reason_lexer.mll"
         ( MINUS )
# 1867 "reason_lexer.ml"

  | 79 ->
# 510 "reason_lexer.mll"
         ( MINUSDOT )
# 1872 "reason_lexer.ml"

  | 80 ->
# 511 "reason_lexer.mll"
         ( LESSGREATER )
# 1877 "reason_lexer.ml"

  | 81 ->
# 512 "reason_lexer.mll"
           ( LESSDOTDOTGREATER )
# 1882 "reason_lexer.ml"

  | 82 ->
# 514 "reason_lexer.mll"
            ( PREFIXOP(Lexing.lexeme lexbuf) )
# 1887 "reason_lexer.ml"

  | 83 ->
# 516 "reason_lexer.mll"
            ( PREFIXOP(Lexing.lexeme lexbuf) )
# 1892 "reason_lexer.ml"

  | 84 ->
# 518 "reason_lexer.mll"
            ( INFIXOP0(Lexing.lexeme lexbuf) )
# 1897 "reason_lexer.ml"

  | 85 ->
# 520 "reason_lexer.mll"
            ( INFIXOP1(Lexing.lexeme lexbuf) )
# 1902 "reason_lexer.ml"

  | 86 ->
# 522 "reason_lexer.mll"
            ( INFIXOP2(Lexing.lexeme lexbuf) )
# 1907 "reason_lexer.ml"

  | 87 ->
# 533 "reason_lexer.mll"
            ( INFIXOP4(Lexing.lexeme lexbuf))
# 1912 "reason_lexer.ml"

  | 88 ->
# 534 "reason_lexer.mll"
            ( PERCENT )
# 1917 "reason_lexer.ml"

  | 89 ->
# 536 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf) )
# 1922 "reason_lexer.ml"

  | 90 ->
# 538 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf))
# 1927 "reason_lexer.ml"

  | 91 ->
# 540 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf) )
# 1932 "reason_lexer.ml"

  | 92 ->
# 541 "reason_lexer.mll"
        ( EOF )
# 1937 "reason_lexer.ml"

  | 93 ->
# 543 "reason_lexer.mll"
      ( raise (Error(Illegal_character (Lexing.lexeme_char lexbuf 0),
                     Location.curr lexbuf))
      )
# 1944 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and comment lexbuf =
    __ocaml_lex_comment_rec lexbuf 146
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 549 "reason_lexer.mll"
      ( comment_start_loc := (Location.curr lexbuf) :: !comment_start_loc;
        store_lexeme lexbuf;
        comment lexbuf;
      )
# 1959 "reason_lexer.ml"

  | 1 ->
# 554 "reason_lexer.mll"
      ( match !comment_start_loc with
        | [] -> assert false
        | [_] -> comment_start_loc := []; Location.curr lexbuf
        | _ :: l -> comment_start_loc := l;
                  store_lexeme lexbuf;
                  comment lexbuf;
       )
# 1970 "reason_lexer.ml"

  | 2 ->
# 562 "reason_lexer.mll"
      (
        string_start_loc := Location.curr lexbuf;
        store_string_char '"';
        is_in_string := true;
        begin try string lexbuf
        with Error (Unterminated_string, str_start) ->
          match !comment_start_loc with
          | [] -> assert false
          | loc :: _ ->
            let start = List.hd (List.rev !comment_start_loc) in
            comment_start_loc := [];
            raise (Error (Unterminated_string_in_comment (start, str_start),
                          loc))
        end;
        is_in_string := false;
        store_string_char '"';
        comment lexbuf )
# 1991 "reason_lexer.ml"

  | 3 ->
# 580 "reason_lexer.mll"
      (
        let delim = Lexing.lexeme lexbuf in
        let delim = String.sub delim 1 (String.length delim - 2) in
        string_start_loc := Location.curr lexbuf;
        store_lexeme lexbuf;
        is_in_string := true;
        begin try quoted_string delim lexbuf
        with Error (Unterminated_string, str_start) ->
          match !comment_start_loc with
          | [] -> assert false
          | loc :: _ ->
            let start = List.hd (List.rev !comment_start_loc) in
            comment_start_loc := [];
            raise (Error (Unterminated_string_in_comment (start, str_start),
                          loc))
        end;
        is_in_string := false;
        store_string_char '|';
        store_string delim;
        store_string_char '}';
        comment lexbuf )
# 2016 "reason_lexer.ml"

  | 4 ->
# 603 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2021 "reason_lexer.ml"

  | 5 ->
# 605 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 1;
        store_lexeme lexbuf;
        comment lexbuf
      )
# 2029 "reason_lexer.ml"

  | 6 ->
# 610 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2034 "reason_lexer.ml"

  | 7 ->
# 612 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2039 "reason_lexer.ml"

  | 8 ->
# 614 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2044 "reason_lexer.ml"

  | 9 ->
# 616 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2049 "reason_lexer.ml"

  | 10 ->
# 618 "reason_lexer.mll"
      ( match !comment_start_loc with
        | [] -> assert false
        | loc :: _ ->
          let start = List.hd (List.rev !comment_start_loc) in
          comment_start_loc := [];
          raise (Error (Unterminated_comment start, loc))
      )
# 2060 "reason_lexer.ml"

  | 11 ->
# 626 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        comment lexbuf
      )
# 2068 "reason_lexer.ml"

  | 12 ->
# 631 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2073 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and string lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1) ;   __ocaml_lex_string_rec lexbuf 178
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 635 "reason_lexer.mll"
      ( () )
# 2085 "reason_lexer.ml"

  | 1 ->
let
# 636 "reason_lexer.mll"
                                  space
# 2091 "reason_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 637 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false (String.length space);
        if in_comment () then store_lexeme lexbuf;
        string lexbuf
      )
# 2098 "reason_lexer.ml"

  | 2 ->
# 642 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_backslash(Lexing.lexeme_char lexbuf 1));
        string lexbuf )
# 2105 "reason_lexer.ml"

  | 3 ->
# 646 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_decimal_code lexbuf 1);
        string lexbuf )
# 2112 "reason_lexer.ml"

  | 4 ->
# 650 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_hexadecimal_code lexbuf 2);
        string lexbuf )
# 2119 "reason_lexer.ml"

  | 5 ->
# 654 "reason_lexer.mll"
      ( if in_comment ()
        then string lexbuf
        else begin
(*  Should be an error, but we are very lax.
          raise (Error (Illegal_escape (Lexing.lexeme lexbuf),
                        Location.curr lexbuf))
*)
          let loc = Location.curr lexbuf in
          Location.prerr_warning loc Warnings.Illegal_backslash;
          store_string_char (Lexing.lexeme_char lexbuf 0);
          store_string_char (Lexing.lexeme_char lexbuf 1);
          string lexbuf
        end
      )
# 2137 "reason_lexer.ml"

  | 6 ->
# 669 "reason_lexer.mll"
      ( if not (in_comment ()) then
          Location.prerr_warning (Location.curr lexbuf) Warnings.Eol_in_string;
        update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        string lexbuf
      )
# 2147 "reason_lexer.ml"

  | 7 ->
# 676 "reason_lexer.mll"
      ( is_in_string := false;
        raise (Error (Unterminated_string, !string_start_loc)) )
# 2153 "reason_lexer.ml"

  | 8 ->
# 679 "reason_lexer.mll"
      ( store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf )
# 2159 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and quoted_string delim lexbuf =
    __ocaml_lex_quoted_string_rec delim lexbuf 197
and __ocaml_lex_quoted_string_rec delim lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 684 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        quoted_string delim lexbuf
      )
# 2174 "reason_lexer.ml"

  | 1 ->
# 689 "reason_lexer.mll"
      ( is_in_string := false;
        raise (Error (Unterminated_string, !string_start_loc)) )
# 2180 "reason_lexer.ml"

  | 2 ->
# 692 "reason_lexer.mll"
      (
        let edelim = Lexing.lexeme lexbuf in
        let edelim = String.sub edelim 1 (String.length edelim - 2) in
        if delim = edelim then ()
        else (store_lexeme lexbuf; quoted_string delim lexbuf)
      )
# 2190 "reason_lexer.ml"

  | 3 ->
# 699 "reason_lexer.mll"
      ( store_string_char(Lexing.lexeme_char lexbuf 0);
        quoted_string delim lexbuf )
# 2196 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_quoted_string_rec delim lexbuf __ocaml_lex_state

and skip_sharp_bang lexbuf =
    __ocaml_lex_skip_sharp_bang_rec lexbuf 206
and __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 704 "reason_lexer.mll"
       ( update_loc lexbuf None 3 false 0 )
# 2208 "reason_lexer.ml"

  | 1 ->
# 706 "reason_lexer.mll"
       ( update_loc lexbuf None 1 false 0 )
# 2213 "reason_lexer.ml"

  | 2 ->
# 707 "reason_lexer.mll"
       ( () )
# 2218 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state

;;

# 709 "reason_lexer.mll"
 

  let token_with_comments lexbuf =
    match !preprocessor with
    | None -> token lexbuf
    | Some (_init, preprocess) -> preprocess token lexbuf

  let last_comments = ref []
  let rec token lexbuf =
    match token_with_comments lexbuf with
        COMMENT (s, comment_loc) ->
          last_comments := (s, comment_loc) :: !last_comments;
          token lexbuf
      | tok -> tok
  let comments () = List.rev !last_comments

  let init () =
    is_in_string := false;
    last_comments := [];
    comment_start_loc := [];
    match !preprocessor with
    | None -> ()
    | Some (init, _preprocess) -> init ()

  let set_preprocessor init preprocess =
    preprocessor := Some (init, preprocess)


# 2254 "reason_lexer.ml"

end
module Reason_entry : sig 
#1 "reason_entry.mli"
(* Copyright (C) 2015-2016 Bloomberg Finance L.P.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * In addition to the permissions granted to you by the LGPL, you may combine
 * or link a "work that uses the Library" with a publicly distributed version
 * of this file to produce a combined library or application, then distribute
 * that combined work under the terms of your choosing, with no requirement
 * to comply with the obligations normally placed on you by section 4 of the
 * LGPL version 3 (or the corresponding section of a later version of the LGPL
 * should you choose to use a later version).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. *)



end = struct
#1 "reason_entry.ml"
(* Copyright (C) 2015-2016 Bloomberg Finance L.P.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * In addition to the permissions granted to you by the LGPL, you may combine
 * or link a "work that uses the Library" with a publicly distributed version
 * of this file to produce a combined library or application, then distribute
 * that combined work under the terms of your choosing, with no requirement
 * to comply with the obligations normally placed on you by section 4 of the
 * LGPL version 3 (or the corresponding section of a later version of the LGPL
 * should you choose to use a later version).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. *)

let implementation lexbuf =
  Reason_lexer.init ();
  Reason_parser.implementation Reason_lexer.token lexbuf


let () = Jsoo_main.make_compiler "reason" implementation

end
